Package: brglm2
Title: Bias Reduction in Generalized Linear Models
Version: 1.1.0
Authors@R: c(person(given = "Ioannis", family = "Kosmidis", role = c("aut", "cre"), email = "ioannis.kosmidis@warwick.ac.uk", comment = c(ORCID = "0000-0003-1556-0302")),
	     person(given = c("Euloge", "Clovis"), family = c("Kenne Pagui"), role = "aut", email = "kenne@stat.unipd.it", comment = c(ORCID = "0000-0002-8998-9251")),
 	     person(given = "Federico", family = "Boiocchi", role = "ctb", email = "federico.boiocchi@gmail.com"),
	     person(given = "Philipp", family = "Sterzinger", role = "ctb", email = "P.Sterzinger@lse.ac.uk", comment = c(ORCID = "0009-0007-7348-5810")),
             person(given = "Kjell", family = "Konis", role = "ctb", email = "kjell.konis@me.com"),
	     person(given = "Nicola", family = "Sartori", role = "ctb", email = "sartori@stat.unipd.it"))
Description: Estimation and inference from generalized linear models based on various methods for bias reduction and maximum penalized likelihood with powers of the Jeffreys prior as penalty. The 'brglmFit()' fitting method can achieve reduction of estimation bias by solving either the mean bias-reducing adjusted score equations in Firth (1993) <doi:10.1093/biomet/80.1.27> and Kosmidis and Firth (2009) <doi:10.1093/biomet/asp055>, or the median bias-reducing adjusted score equations in Kenne et al. (2017) <doi:10.1093/biomet/asx046>, or through the direct subtraction of an estimate of the bias of the maximum likelihood estimator from the maximum likelihood estimates as in Cordeiro and McCullagh (1991) <https://www.jstor.org/stable/2345592>. See Kosmidis et al (2020) <doi:10.1007/s11222-019-09860-6> for more details. Estimation in all cases takes place via a quasi Fisher scoring algorithm, and S3 methods for the construction of of confidence intervals for the reduced-bias estimates are provided. In the special case of generalized linear models for binomial and multinomial responses (both ordinal and nominal), the adjusted score approaches to mean and media bias reduction have been found to return estimates with improved frequentist properties, that are also always finite, even in cases where the maximum likelihood estimates are infinite (e.g. complete and quasi-complete separation; see Kosmidis and Firth, 2020 <doi:10.1093/biomet/asaa052>, for a proof for mean bias reduction in logistic regression). The 'mdyplFit()' fitting method fits logistic regression models using maximum Diaconis-Ylvisaker prior penalized likelihood, which also guarantees finite estimates. High-dimensionality corrections under proportional asymptotics can be applied to the resulting objects; see Sterzinger and Kosmidis (2024) <doi:10.48550/arXiv.2311.07419> for details.
URL: https://github.com/ikosmidis/brglm2
BugReports: https://github.com/ikosmidis/brglm2/issues
Depends: R (>= 3.5)
Imports: MASS, stats, Matrix, graphics, nnet, enrichwith, numDeriv,
        statmod, nleqslv
License: GPL-3
Encoding: UTF-8
LazyData: true
RoxygenNote: 7.3.3
Suggests: detectseparation, knitr, rmarkdown, covr, tinytest, VGAM,
        brglm
VignetteBuilder: knitr
NeedsCompilation: yes
Packaged: 2026-04-29 13:30:08 UTC; yiannis
Author: Ioannis Kosmidis [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-1556-0302>),
  Euloge Clovis Kenne Pagui [aut] (ORCID:
    <https://orcid.org/0000-0002-8998-9251>),
  Federico Boiocchi [ctb],
  Philipp Sterzinger [ctb] (ORCID:
    <https://orcid.org/0009-0007-7348-5810>),
  Kjell Konis [ctb],
  Nicola Sartori [ctb]
Maintainer: Ioannis Kosmidis <ioannis.kosmidis@warwick.ac.uk>
Repository: CRAN
Date/Publication: 2026-04-29 14:40:02 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-05-01 20:36:54 UTC; unix
Archs: brglm2.so.dSYM

Package: brglm2
Title: Bias Reduction in Generalized Linear Models
Version: 1.1.0
Authors@R: c(person(given = "Ioannis", family = "Kosmidis", role = c("aut", "cre"), email = "ioannis.kosmidis@warwick.ac.uk", comment = c(ORCID = "0000-0003-1556-0302")),
	     person(given = c("Euloge", "Clovis"), family = c("Kenne Pagui"), role = "aut", email = "kenne@stat.unipd.it", comment = c(ORCID = "0000-0002-8998-9251")),
 	     person(given = "Federico", family = "Boiocchi", role = "ctb", email = "federico.boiocchi@gmail.com"),
	     person(given = "Philipp", family = "Sterzinger", role = "ctb", email = "P.Sterzinger@lse.ac.uk", comment = c(ORCID = "0009-0007-7348-5810")),
             person(given = "Kjell", family = "Konis", role = "ctb", email = "kjell.konis@me.com"),
	     person(given = "Nicola", family = "Sartori", role = "ctb", email = "sartori@stat.unipd.it"))
Description: Estimation and inference from generalized linear models based on various methods for bias reduction and maximum penalized likelihood with powers of the Jeffreys prior as penalty. The 'brglmFit()' fitting method can achieve reduction of estimation bias by solving either the mean bias-reducing adjusted score equations in Firth (1993) <doi:10.1093/biomet/80.1.27> and Kosmidis and Firth (2009) <doi:10.1093/biomet/asp055>, or the median bias-reducing adjusted score equations in Kenne et al. (2017) <doi:10.1093/biomet/asx046>, or through the direct subtraction of an estimate of the bias of the maximum likelihood estimator from the maximum likelihood estimates as in Cordeiro and McCullagh (1991) <https://www.jstor.org/stable/2345592>. See Kosmidis et al (2020) <doi:10.1007/s11222-019-09860-6> for more details. Estimation in all cases takes place via a quasi Fisher scoring algorithm, and S3 methods for the construction of of confidence intervals for the reduced-bias estimates are provided. In the special case of generalized linear models for binomial and multinomial responses (both ordinal and nominal), the adjusted score approaches to mean and media bias reduction have been found to return estimates with improved frequentist properties, that are also always finite, even in cases where the maximum likelihood estimates are infinite (e.g. complete and quasi-complete separation; see Kosmidis and Firth, 2020 <doi:10.1093/biomet/asaa052>, for a proof for mean bias reduction in logistic regression). The 'mdyplFit()' fitting method fits logistic regression models using maximum Diaconis-Ylvisaker prior penalized likelihood, which also guarantees finite estimates. High-dimensionality corrections under proportional asymptotics can be applied to the resulting objects; see Sterzinger and Kosmidis (2024) <doi:10.48550/arXiv.2311.07419> for details.
URL: https://github.com/ikosmidis/brglm2
BugReports: https://github.com/ikosmidis/brglm2/issues
Depends: R (>= 3.5)
Imports: MASS, stats, Matrix, graphics, nnet, enrichwith, numDeriv,
        statmod, nleqslv
License: GPL-3
Encoding: UTF-8
LazyData: true
RoxygenNote: 7.3.3
Suggests: detectseparation, knitr, rmarkdown, covr, tinytest, VGAM,
        brglm
VignetteBuilder: knitr
NeedsCompilation: yes
Packaged: 2026-04-29 13:30:08 UTC; yiannis
Author: Ioannis Kosmidis [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-1556-0302>),
  Euloge Clovis Kenne Pagui [aut] (ORCID:
    <https://orcid.org/0000-0002-8998-9251>),
  Federico Boiocchi [ctb],
  Philipp Sterzinger [ctb] (ORCID:
    <https://orcid.org/0009-0007-7348-5810>),
  Kjell Konis [ctb],
  Nicola Sartori [ctb]
Maintainer: Ioannis Kosmidis <ioannis.kosmidis@warwick.ac.uk>
Repository: CRAN
Date/Publication: 2026-04-29 14:40:02 UTC

# Generated by roxygen2: do not edit by hand

S3method(add1,brglmFit)
S3method(add1,mdyplFit)
S3method(coef,bracl)
S3method(coef,brglmFit)
S3method(coef,brglmFit_expo)
S3method(coef,brmultinom)
S3method(coef,brnb)
S3method(confint,brglmFit)
S3method(confint,brmultinom)
S3method(confint,brnb)
S3method(confint,mdyplFit)
S3method(drop1,brglmFit)
S3method(drop1,mdyplFit)
S3method(expo,brglmFit)
S3method(expo,glm)
S3method(fitted,bracl)
S3method(fitted,brmultinom)
S3method(logLik,brmultinom)
S3method(model.matrix,brmultinom)
S3method(ordinal_superiority,bracl)
S3method(plrtest,mdyplFit)
S3method(predict,bracl)
S3method(predict,brmultinom)
S3method(print,anova_mdyplFit)
S3method(print,brglmFit_expo)
S3method(print,brmultinom)
S3method(print,brnb)
S3method(print,summary.bracl)
S3method(print,summary.brglmFit)
S3method(print,summary.brmultinom)
S3method(print,summary.brnb)
S3method(print,summary.mdyplFit)
S3method(residuals,bracl)
S3method(residuals,brmultinom)
S3method(simulate,brmultinom)
S3method(simulate,brnb)
S3method(summary,bracl)
S3method(summary,brglmFit)
S3method(summary,brmultinom)
S3method(summary,brnb)
S3method(summary,mdyplFit)
S3method(vcov,bracl)
S3method(vcov,brglmFit)
S3method(vcov,brmultinom)
S3method(vcov,brnb)
export(bracl)
export(brglmControl)
export(brglmFit)
export(brglm_control)
export(brglm_fit)
export(brmultinom)
export(brnb)
export(check_infinite_estimates)
export(detect_separation)
export(expo)
export(mdyplControl)
export(mdyplFit)
export(mdypl_control)
export(mdypl_fit)
export(mis)
export(ordinal_superiority)
export(plrtest)
export(se0)
export(se1)
export(sloe)
export(solve_se)
import(MASS)
import(enrichwith)
import(nleqslv)
import(stats)
import(Matrix)
importFrom(graphics,plot)
importFrom(nnet,class.ind)
importFrom(numDeriv,grad)
importFrom(statmod,gauss.quad)
useDynLib(brglm2)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# brglm2 <img src="man/figures/hex_brglm2.svg" width="320" align="right">

<!-- badges: start -->

[![CRAN\_Status\_Badge](https://www.r-pkg.org/badges/version/brglm2)](https://cran.r-project.org/package=brglm2)
[![R-CMD-check](https://github.com/ikosmidis/brglm2/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/ikosmidis/brglm2/actions/workflows/R-CMD-check.yaml)
[![Licence](https://img.shields.io/badge/licence-GPL--3-blue.svg)](https://www.gnu.org/licenses/gpl-3.0.html)
[![Contributor
Covenant](https://img.shields.io/badge/Contributor%20Covenant-2.1-4baaaa.svg)](https://contributor-covenant.org/version/2/1/CODE_OF_CONDUCT.html)
[![Codecov test
coverage](https://codecov.io/gh/ikosmidis/brglm2/graph/badge.svg)](https://app.codecov.io/gh/ikosmidis/brglm2)
<!-- badges: end -->

[**brglm2**](https://github.com/ikosmidis/brglm2) provides tools for the
estimation and inference from generalized linear models using various
methods for bias reduction. **brglm2** supports all generalized linear
models supported in R, and provides methods for multinomial logistic
regression (nominal responses), adjacent category models (ordinal
responses), and negative binomial regression (for potentially
overdispered count responses).

Reduction of estimation bias is achieved by solving either the mean-bias
reducing adjusted score equations in [Firth
(1993)](https://doi.org/10.1093/biomet/80.1.27) and [Kosmidis & Firth
(2009)](https://doi.org/10.1093/biomet/asp055) or the median-bias
reducing adjusted score equations in [Kenne et al
(2017)](https://doi.org/10.1093/biomet/asx046), or through the direct
subtraction of an estimate of the bias of the maximum likelihood
estimator from the maximum likelihood estimates as prescribed in
[Cordeiro and McCullagh (1991)](https://www.jstor.org/stable/2345592).
[Kosmidis et al (2020)](https://doi.org/10.1007/s11222-019-09860-6)
provides a unifying framework and algorithms for mean and median bias
reduction for the estimation of generalized linear models.

In the special case of generalized linear models for binomial and
multinomial responses (both ordinal and nominal), the adjusted score
equations return estimates with improved frequentist properties, that
are also always finite, even in cases where the maximum likelihood
estimates are infinite (e.g. complete and quasi-complete separation).
See, [Kosmidis & Firth (2021)](https://doi.org/10.1093/biomet/asaa052)
for the proof of the latter result in the case of mean bias reduction
for logistic regression (and, for more general binomial-response models
where the likelihood is penalized by a power of the Jeffreys’ invariant
prior).

For logistic regression, **brglm2** also provides methods for maximum
Diaconis-Ylvisaker prior penalized likelihood (MDYPL) estimation, and
corresponding methods for high-dimensionality corrections of the
aggregate bias of the estimator and the usual statistics used for
inference; see [Sterzinger and Kosmidis,
2024](https://arxiv.org/abs/2311.07419).

The core model fitters are implemented by the functions `brglm_fit()`
(univariate generalized linear models) and `mdyplFit()` (logistic
regression), and `brmultinom()` (baseline category logit models for
nominal multinomial responses), `bracl()` (adjacent category logit
models for ordinal multinomial responses), and `brnb()` (negative
binomial regression).

## Installation

Install the current version from CRAN:

    install.packages("brglm2")

or the development version from github:

    # install.packages("remotes")
    remotes::install_github("ikosmidis/brglm2", ref = "develop")

## Examples

### Estimation of binomial-response GLMs with separated data

Below we follow the example of [Heinze and Schemper
(2002)](https://doi.org/10.1002/sim.1047) and fit a logistic regression
model using maximum likelihood (ML) to analyze data from a study on
endometrial cancer (see `?brglm2::endometrial` for details and
references).

    library("brglm2")
    data("endometrial", package = "brglm2")
    modML <- glm(HG ~ NV + PI + EH, family = binomial("logit"), data = endometrial)
    summary(modML)
    #> 
    #> Call:
    #> glm(formula = HG ~ NV + PI + EH, family = binomial("logit"), 
    #>     data = endometrial)
    #> 
    #> Coefficients:
    #>               Estimate Std. Error z value Pr(>|z|)    
    #> (Intercept)    4.30452    1.63730   2.629 0.008563 ** 
    #> NV            18.18556 1715.75089   0.011 0.991543    
    #> PI            -0.04218    0.04433  -0.952 0.341333    
    #> EH            -2.90261    0.84555  -3.433 0.000597 ***
    #> ---
    #> Signif. codes:  0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1
    #> 
    #> (Dispersion parameter for binomial family taken to be 1)
    #> 
    #>     Null deviance: 104.903  on 78  degrees of freedom
    #> Residual deviance:  55.393  on 75  degrees of freedom
    #> AIC: 63.393
    #> 
    #> Number of Fisher Scoring iterations: 17

The ML estimate of the parameter for `NV` is actually infinite, as can
be quickly verified using the
[**detectseparation**](https://cran.r-project.org/package=detectseparation)
R package

    # install.packages("detectseparation")
    library("detectseparation")
    #> 
    #> Attaching package: 'detectseparation'
    #> The following objects are masked from 'package:brglm2':
    #> 
    #>     check_infinite_estimates, detect_separation
    update(modML, method = detect_separation)
    #> Implementation: ROI | Solver: lpsolve 
    #> Separation: TRUE 
    #> Existence of maximum likelihood estimates
    #> (Intercept)          NV          PI          EH 
    #>           0         Inf           0           0 
    #> 0: finite value, Inf: infinity, -Inf: -infinity

The reported, apparently finite estimate
`r round(coef(summary(modML))["NV", "Estimate"], 3)` for `NV` is merely
due to false convergence of the iterative estimation procedure for ML.
The same is true for the estimated standard error, and, hence the value
0.011 for the *z*-statistic cannot be trusted for inference on the
effect size for `NV`.

As mentioned earlier, many of the estimation methods implemented in
**brglm2** not only return estimates with improved frequentist
properties (e.g. asymptotically smaller mean and median bias than what
ML typically delivers), but also estimates and estimated standard errors
that are always finite in binomial (e.g. logistic, probit, and
complementary log-log regression) and multinomial regression models
(e.g. baseline category logit models for nominal responses, and adjacent
category logit models for ordinal responses). For example, the code
chunk below refits the model on the endometrial cancer study data using
mean bias reduction.

    summary(update(modML, method = "brglm_fit"))
    #> 
    #> Call:
    #> glm(formula = HG ~ NV + PI + EH, family = binomial("logit"), 
    #>     data = endometrial, method = "brglm_fit")
    #> 
    #> Deviance Residuals: 
    #>     Min       1Q   Median       3Q      Max  
    #> -1.4740  -0.6706  -0.3411   0.3252   2.6123  
    #> 
    #> Coefficients:
    #>             Estimate Std. Error z value Pr(>|z|)    
    #> (Intercept)  3.77456    1.48869   2.535 0.011229 *  
    #> NV           2.92927    1.55076   1.889 0.058902 .  
    #> PI          -0.03475    0.03958  -0.878 0.379915    
    #> EH          -2.60416    0.77602  -3.356 0.000791 ***
    #> ---
    #> Signif. codes:  0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1
    #> 
    #> (Dispersion parameter for binomial family taken to be 1)
    #> 
    #>     Null deviance: 104.903  on 78  degrees of freedom
    #> Residual deviance:  56.575  on 75  degrees of freedom
    #> AIC:  64.575
    #> 
    #> Type of estimator: AS_mixed (mixed bias-reducing adjusted score equations)
    #> Number of Fisher Scoring iterations: 6

A quick comparison of the output from mean bias reduction to that from
ML reveals a dramatic change in the *z*-statistic for `NV`, now that
estimates and estimated standard errors are finite. In particular, the
evidence against the null of `NV` not contributing to the model in the
presence of the other covariates being now stronger.

See `?brglm_fit` and `?brglm_control` for more examples and the other
estimation methods for generalized linear models, including median bias
reduction and maximum penalized likelihood with Jeffreys’ prior penalty.
Also do not forget to take a look at the vignettes
(`vignette(package = "brglm2")`) for details and more case studies.

### Improved estimation of the exponential of regression parameters

See, also `?expo` for a method to estimate the exponential of regression
parameters, such as odds ratios from logistic regression models, while
controlling for other covariate information. Estimation can be performed
using maximum likelihood or various estimators with smaller asymptotic
mean and median bias, that are also guaranteed to be finite, even if the
corresponding maximum likelihood estimates are infinite. For example,
`modML` is a logistic regression fit, so the exponential of each
coefficient is an odds ratio while controlling for other covariates. To
estimate those odds ratios using the `correction*` method for mean bias
reduction (see `?expo` for details) we do

    expoRB <- expo(modML, type = "correction*")
    expoRB
    #> 
    #> Call:
    #> expo.glm(object = modML, type = "correction*")
    #> 
    #> Odds ratios 
    #>              Estimate Std. Error     2.5 %  97.5 %
    #> (Intercept) 20.671820  33.136501  0.893141 478.451
    #> NV           8.496974   7.825239  1.397511  51.662
    #> PI           0.965089   0.036795  0.895602   1.040
    #> EH           0.056848   0.056344  0.008148   0.397
    #> 
    #> 
    #> Type of estimator: correction* (explicit mean bias correction with a multiplicative adjustment)

The odds ratio between presence of neovasculation and high histology
grade (`HG`) is estimated to be 8.497, while controlling for PI and EH.
So, for each value of `PI` and `EH`, the estimated odds of high
histology grade are about 8.5 times higher when neovasculation is
present. An approximate 95% interval for the latter odds ratio is (1.4,
51.7) providing evidence of association between `NV` and `HG` while
controlling for `PI` and `EH`. Note here that, the maximum likelihood
estimate of the odds ratio is not as useful as the `correction*`
estimate, because it is +∞ with an infinite standard error (see previous
section).

### MDYPL estimation and high-dimensionality corrections

Consider the [Multiple Features
dataset](https://doi.org/10.24432/C5HC70), which consists of digits
(0-9) extracted from a collection of maps from a Dutch public utility.
Two hundred `30 × 48` binary images per digit were available, which have
then been used to extract feature sets. The digits are shown below using
pixel averages in `2 x 3` windows.

    data("MultipleFeatures", package = "brglm2")
    par(mfrow = c(10, 20), mar = numeric(4) + 0.1)
    for (c_digit in 0:9) {
        df <- subset(MultipleFeatures, digit == c_digit)
        df <- as.matrix(df[, paste("pix", 1:240, sep = ".")])
        for (inst in 1:20) {
            m <- matrix(df[inst, ], 15, 16)[, 16:1]
            image(m, col = grey.colors(7, 1, 0), xaxt = "n", yaxt = "n")
        }
    }

<img src="man/figures/README-unnamed-chunk-6-1.png" width="100%" />

We focus on the setting of [Sterzinger and Kosmidis (2024, Section
8)](https://arxiv.org/abs/2311.07419) on explaining the character shapes
of the digit `7` in terms of 76 Fourier coefficients (`fou` features),
which are computed to be rotationally invariant, and 64 Karhunen-Loève
coefficients (`kar` features), using 1000 randomly selected digits.
Depending on the font, the level of noise introduced during
digitization, and the downscaling of the digits to binary images,
difficulties may arise in discriminating instances of the digit `7` to
instances of the digits `1` and `4`. Also, if only rotation invariant
features, like `fou`, are used difficulties may arise in discriminating
instances of the digit `7` to instances of the digit `2`.

The data is perfectly separated for both the model with only `fou`
features, and the model with `fou` and `kar` features, and the maximized
likelihood is zero for both models.

    ## Center the fou.* and kar.* features
    vars <- grep("fou|kar", names(MultipleFeatures), value = TRUE)
    train_id <- which(MultipleFeatures$training)
    MultipleFeatures[train_id, vars] <- scale(MultipleFeatures[train_id, vars], scale = FALSE)
    ## Set up module formulas
    full_fm <- formula(paste("I(digit == 7) ~", paste(vars, collapse = " + ")))
    nest_vars <- grep("fou", vars, value = TRUE)
    nest_fm <- formula(paste("I(digit == 7) ~", paste(nest_vars, collapse = " + ")))
    ## Fit the models using maximum likelihood
    full_sep <- glm(full_fm, data = MultipleFeatures, family = binomial(), subset = training,
                    method = detect_separation)
    nest_sep <- update(full_sep, nest_fm)
    full_sep$outcome
    #> [1] TRUE
    nest_sep$outcome
    #> [1] TRUE

As a result, the likelihood ratio statistic comparing the two models
will be trivially zero, regardless of any evidence against the
hypothesis that the model with only `fou` features is an as good
description of `7` as the model with both `fou` and `kar` features.

    anova(update(nest_sep, method = glm.fit),
          update(full_sep, method = glm.fit))
    #> Warning: glm.fit: algorithm did not converge
    #> Warning: glm.fit: fitted probabilities numerically 0 or 1 occurred
    #> Warning: glm.fit: algorithm did not converge
    #> Warning: glm.fit: fitted probabilities numerically 0 or 1 occurred
    #> Analysis of Deviance Table
    #> 
    #> Model 1: I(digit == 7) ~ fou.1 + fou.2 + fou.3 + fou.4 + fou.5 + fou.6 + 
    #>     fou.7 + fou.8 + fou.9 + fou.10 + fou.11 + fou.12 + fou.13 + 
    #>     fou.14 + fou.15 + fou.16 + fou.17 + fou.18 + fou.19 + fou.20 + 
    #>     fou.21 + fou.22 + fou.23 + fou.24 + fou.25 + fou.26 + fou.27 + 
    #>     fou.28 + fou.29 + fou.30 + fou.31 + fou.32 + fou.33 + fou.34 + 
    #>     fou.35 + fou.36 + fou.37 + fou.38 + fou.39 + fou.40 + fou.41 + 
    #>     fou.42 + fou.43 + fou.44 + fou.45 + fou.46 + fou.47 + fou.48 + 
    #>     fou.49 + fou.50 + fou.51 + fou.52 + fou.53 + fou.54 + fou.55 + 
    #>     fou.56 + fou.57 + fou.58 + fou.59 + fou.60 + fou.61 + fou.62 + 
    #>     fou.63 + fou.64 + fou.65 + fou.66 + fou.67 + fou.68 + fou.69 + 
    #>     fou.70 + fou.71 + fou.72 + fou.73 + fou.74 + fou.75 + fou.76
    #> Model 2: I(digit == 7) ~ fou.1 + fou.2 + fou.3 + fou.4 + fou.5 + fou.6 + 
    #>     fou.7 + fou.8 + fou.9 + fou.10 + fou.11 + fou.12 + fou.13 + 
    #>     fou.14 + fou.15 + fou.16 + fou.17 + fou.18 + fou.19 + fou.20 + 
    #>     fou.21 + fou.22 + fou.23 + fou.24 + fou.25 + fou.26 + fou.27 + 
    #>     fou.28 + fou.29 + fou.30 + fou.31 + fou.32 + fou.33 + fou.34 + 
    #>     fou.35 + fou.36 + fou.37 + fou.38 + fou.39 + fou.40 + fou.41 + 
    #>     fou.42 + fou.43 + fou.44 + fou.45 + fou.46 + fou.47 + fou.48 + 
    #>     fou.49 + fou.50 + fou.51 + fou.52 + fou.53 + fou.54 + fou.55 + 
    #>     fou.56 + fou.57 + fou.58 + fou.59 + fou.60 + fou.61 + fou.62 + 
    #>     fou.63 + fou.64 + fou.65 + fou.66 + fou.67 + fou.68 + fou.69 + 
    #>     fou.70 + fou.71 + fou.72 + fou.73 + fou.74 + fou.75 + fou.76 + 
    #>     kar.1 + kar.2 + kar.3 + kar.4 + kar.5 + kar.6 + kar.7 + kar.8 + 
    #>     kar.9 + kar.10 + kar.11 + kar.12 + kar.13 + kar.14 + kar.15 + 
    #>     kar.16 + kar.17 + kar.18 + kar.19 + kar.20 + kar.21 + kar.22 + 
    #>     kar.23 + kar.24 + kar.25 + kar.26 + kar.27 + kar.28 + kar.29 + 
    #>     kar.30 + kar.31 + kar.32 + kar.33 + kar.34 + kar.35 + kar.36 + 
    #>     kar.37 + kar.38 + kar.39 + kar.40 + kar.41 + kar.42 + kar.43 + 
    #>     kar.44 + kar.45 + kar.46 + kar.47 + kar.48 + kar.49 + kar.50 + 
    #>     kar.51 + kar.52 + kar.53 + kar.54 + kar.55 + kar.56 + kar.57 + 
    #>     kar.58 + kar.59 + kar.60 + kar.61 + kar.62 + kar.63 + kar.64
    #>   Resid. Df Resid. Dev Df   Deviance Pr(>Chi)
    #> 1       923 3.1661e-07                       
    #> 2       859 1.6140e-08 64 3.0047e-07        1

Let’s fit the models using maximum Diaconis-Ylvisaker prior penalized
likelihood (MDYPL) with a shrinkage parameter `alpha` in `(0, 1)`, which
always results in finite estimates; see `?brglm2::mdypl_fit`. The
corresponding penalized likelihood ratio test (see `?brglm2::plrtest`
and `?brglm2::summary.mdyplFit`) results again in no evidence against
the hypothesis.

    full_m <- update(full_sep, method = mdypl_fit)
    nest_m <- update(nest_sep, method = mdypl_fit, alpha = full_m$alpha)
    plrtest(nest_m, full_m)
    #> Analysis of Deviance Table
    #> 
    #> Model 1: I(digit == 7) ~ fou.1 + fou.2 + fou.3 + fou.4 + fou.5 + fou.6 + 
    #>     fou.7 + fou.8 + fou.9 + fou.10 + fou.11 + fou.12 + fou.13 + 
    #>     fou.14 + fou.15 + fou.16 + fou.17 + fou.18 + fou.19 + fou.20 + 
    #>     fou.21 + fou.22 + fou.23 + fou.24 + fou.25 + fou.26 + fou.27 + 
    #>     fou.28 + fou.29 + fou.30 + fou.31 + fou.32 + fou.33 + fou.34 + 
    #>     fou.35 + fou.36 + fou.37 + fou.38 + fou.39 + fou.40 + fou.41 + 
    #>     fou.42 + fou.43 + fou.44 + fou.45 + fou.46 + fou.47 + fou.48 + 
    #>     fou.49 + fou.50 + fou.51 + fou.52 + fou.53 + fou.54 + fou.55 + 
    #>     fou.56 + fou.57 + fou.58 + fou.59 + fou.60 + fou.61 + fou.62 + 
    #>     fou.63 + fou.64 + fou.65 + fou.66 + fou.67 + fou.68 + fou.69 + 
    #>     fou.70 + fou.71 + fou.72 + fou.73 + fou.74 + fou.75 + fou.76
    #> Model 2: I(digit == 7) ~ fou.1 + fou.2 + fou.3 + fou.4 + fou.5 + fou.6 + 
    #>     fou.7 + fou.8 + fou.9 + fou.10 + fou.11 + fou.12 + fou.13 + 
    #>     fou.14 + fou.15 + fou.16 + fou.17 + fou.18 + fou.19 + fou.20 + 
    #>     fou.21 + fou.22 + fou.23 + fou.24 + fou.25 + fou.26 + fou.27 + 
    #>     fou.28 + fou.29 + fou.30 + fou.31 + fou.32 + fou.33 + fou.34 + 
    #>     fou.35 + fou.36 + fou.37 + fou.38 + fou.39 + fou.40 + fou.41 + 
    #>     fou.42 + fou.43 + fou.44 + fou.45 + fou.46 + fou.47 + fou.48 + 
    #>     fou.49 + fou.50 + fou.51 + fou.52 + fou.53 + fou.54 + fou.55 + 
    #>     fou.56 + fou.57 + fou.58 + fou.59 + fou.60 + fou.61 + fou.62 + 
    #>     fou.63 + fou.64 + fou.65 + fou.66 + fou.67 + fou.68 + fou.69 + 
    #>     fou.70 + fou.71 + fou.72 + fou.73 + fou.74 + fou.75 + fou.76 + 
    #>     kar.1 + kar.2 + kar.3 + kar.4 + kar.5 + kar.6 + kar.7 + kar.8 + 
    #>     kar.9 + kar.10 + kar.11 + kar.12 + kar.13 + kar.14 + kar.15 + 
    #>     kar.16 + kar.17 + kar.18 + kar.19 + kar.20 + kar.21 + kar.22 + 
    #>     kar.23 + kar.24 + kar.25 + kar.26 + kar.27 + kar.28 + kar.29 + 
    #>     kar.30 + kar.31 + kar.32 + kar.33 + kar.34 + kar.35 + kar.36 + 
    #>     kar.37 + kar.38 + kar.39 + kar.40 + kar.41 + kar.42 + kar.43 + 
    #>     kar.44 + kar.45 + kar.46 + kar.47 + kar.48 + kar.49 + kar.50 + 
    #>     kar.51 + kar.52 + kar.53 + kar.54 + kar.55 + kar.56 + kar.57 + 
    #>     kar.58 + kar.59 + kar.60 + kar.61 + kar.62 + kar.63 + kar.64
    #>   Resid. Df Resid. Dev Df Deviance Pr(>Chi)
    #> 1       923     97.305                     
    #> 2       859     32.945 64   64.359   0.4639

Nevertheless, `full_m` involves 141 parameters, which is relatively
large compared to the 1000 available observations. The distribution of
the penalized likelihood ratio statistic may be far from the asymptotic
*χ*<sup>2</sup> distribution that we expect under usual asymptotics.

In stark contrast to the evidence quantified by the previous tests, the
high-dimensionality correction to the penalized likelihood ratio
statistic under proportional asymptotics proposed in [Sterzinger and
Kosmidis (2024)](https://arxiv.org/abs/2311.07419) results in strong
evidence against the model with `fou` features only.

    plrtest(nest_m, full_m, hd_correction = TRUE)
    #> Analysis of Deviance Table
    #> 
    #> Model 1: I(digit == 7) ~ fou.1 + fou.2 + fou.3 + fou.4 + fou.5 + fou.6 + 
    #>     fou.7 + fou.8 + fou.9 + fou.10 + fou.11 + fou.12 + fou.13 + 
    #>     fou.14 + fou.15 + fou.16 + fou.17 + fou.18 + fou.19 + fou.20 + 
    #>     fou.21 + fou.22 + fou.23 + fou.24 + fou.25 + fou.26 + fou.27 + 
    #>     fou.28 + fou.29 + fou.30 + fou.31 + fou.32 + fou.33 + fou.34 + 
    #>     fou.35 + fou.36 + fou.37 + fou.38 + fou.39 + fou.40 + fou.41 + 
    #>     fou.42 + fou.43 + fou.44 + fou.45 + fou.46 + fou.47 + fou.48 + 
    #>     fou.49 + fou.50 + fou.51 + fou.52 + fou.53 + fou.54 + fou.55 + 
    #>     fou.56 + fou.57 + fou.58 + fou.59 + fou.60 + fou.61 + fou.62 + 
    #>     fou.63 + fou.64 + fou.65 + fou.66 + fou.67 + fou.68 + fou.69 + 
    #>     fou.70 + fou.71 + fou.72 + fou.73 + fou.74 + fou.75 + fou.76
    #> Model 2: I(digit == 7) ~ fou.1 + fou.2 + fou.3 + fou.4 + fou.5 + fou.6 + 
    #>     fou.7 + fou.8 + fou.9 + fou.10 + fou.11 + fou.12 + fou.13 + 
    #>     fou.14 + fou.15 + fou.16 + fou.17 + fou.18 + fou.19 + fou.20 + 
    #>     fou.21 + fou.22 + fou.23 + fou.24 + fou.25 + fou.26 + fou.27 + 
    #>     fou.28 + fou.29 + fou.30 + fou.31 + fou.32 + fou.33 + fou.34 + 
    #>     fou.35 + fou.36 + fou.37 + fou.38 + fou.39 + fou.40 + fou.41 + 
    #>     fou.42 + fou.43 + fou.44 + fou.45 + fou.46 + fou.47 + fou.48 + 
    #>     fou.49 + fou.50 + fou.51 + fou.52 + fou.53 + fou.54 + fou.55 + 
    #>     fou.56 + fou.57 + fou.58 + fou.59 + fou.60 + fou.61 + fou.62 + 
    #>     fou.63 + fou.64 + fou.65 + fou.66 + fou.67 + fou.68 + fou.69 + 
    #>     fou.70 + fou.71 + fou.72 + fou.73 + fou.74 + fou.75 + fou.76 + 
    #>     kar.1 + kar.2 + kar.3 + kar.4 + kar.5 + kar.6 + kar.7 + kar.8 + 
    #>     kar.9 + kar.10 + kar.11 + kar.12 + kar.13 + kar.14 + kar.15 + 
    #>     kar.16 + kar.17 + kar.18 + kar.19 + kar.20 + kar.21 + kar.22 + 
    #>     kar.23 + kar.24 + kar.25 + kar.26 + kar.27 + kar.28 + kar.29 + 
    #>     kar.30 + kar.31 + kar.32 + kar.33 + kar.34 + kar.35 + kar.36 + 
    #>     kar.37 + kar.38 + kar.39 + kar.40 + kar.41 + kar.42 + kar.43 + 
    #>     kar.44 + kar.45 + kar.46 + kar.47 + kar.48 + kar.49 + kar.50 + 
    #>     kar.51 + kar.52 + kar.53 + kar.54 + kar.55 + kar.56 + kar.57 + 
    #>     kar.58 + kar.59 + kar.60 + kar.61 + kar.62 + kar.63 + kar.64
    #>   Resid. Df Resid. Dev Df Deviance  Pr(>Chi)    
    #> 1       923     97.305                          
    #> 2       859     32.945 64   173.34 5.095e-12 ***
    #> ---
    #> Signif. codes:  0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1
    #> 
    #> High-dimensionality correction applied with
    #> Dimentionality parameter (kappa) = 0.14
    #> Estimated signal strength (gamma) = 11.58
    #> State evolution parameters (mu, b, sigma) = (0.4, 1.84, 2.21) with max(|funcs|) = 6.300466e-09

The estimates can be corrected in terms of aggregate bias using the
`summary()` method.

    summ_full_m <- summary(full_m, hd_correction = TRUE)

The correction proceeds by estimating the constant *μ* by which the
estimates are divided in order to recover the asymptotic aggregate
unbiasedness of the estimator. The figure below illustrates that the
impact of the correction is to inflate the MDYPL estimates.

    rescaled_coefs <- coef(summ_full_m)[-1, ]
    acols <- hcl.colors(3, alpha = 0.2)
    cols <- hcl.colors(3)
    plot(coef(full_m)[-1], rescaled_coefs[, "Estimate"],
         xlim = c(-9, 9), ylim = c(-9, 9),
         xlab = "MDYPL estimates", ylab = "rescaled MDYPL estimates",
         pch = 21,
         bg = acols[grepl("kar", rownames(rescaled_coefs)) + 1],
         col = NULL)
    legend(-9, 9, legend = c("fou", "kar"), pt.bg = cols[1:2], col = NA, pch = 21,
           title = "Features")
    legend(-5.4, 9, legend = expression(1, 1/hat(mu)), lty = c(2, 1), col = "grey",
           title = "Slope")
    abline(0, 1, col = "grey", lty = 2)
    abline(0, 1/summ_full_m$se_parameters[1], col = "grey")

## Estimation methods

The workhorse function in **brglm2** is `brglm_fit()` (or equivalently
`brglmFit()` if you like camel case), which, as we did in the example
above, can be passed directly to the `method` argument of the `glm()`
function. `brglm_fit()` implements a quasi [Fisher
scoring](https://en.wikipedia.org/wiki/Scoring_algorithm) procedure,
whose special cases result in a range of explicit and implicit bias
reduction methods for generalized linear models for more details). Bias
reduction for multinomial logistic regression (nominal responses) can be
performed using the function `brmultinom()`, and for adjacent category
models (ordinal responses) using the function `bracl()`. Both
`brmultinom()` and `bracl()` rely on `brglm_fit`.

The classification of bias reduction methods into explicit and implicit
is as given in [Kosmidis (2014)](https://doi.org/10.1002/wics.1296).

For logistic regression models, in particular, the `mdypl_fit()`
function provides maximum Diaconis-Ylvisaker prior penalized likelihood
estimation, and can again be passed directly to the `method` argument of
the `glm()` function. The `summary()` method for `mdyplFit` objects,
then allows for high-dimensional corrections of aggregate bias and of
standard *z*-statistics under proportional asymptotics, and the
`plrtest()` method allows for penalized likelihood ratio tests with and
without high-dimensional corrections; see [Sterzinger and Kosmidis
(2024)](https://arxiv.org/abs/2311.07419), the example above, and the
help pages of the methods.

## References and resources

**brglm2** was presented by [Ioannis
Kosmidis](https://www.ikosmidis.com) at the useR! 2016 international
conference at University of Stanford on 16 June 2016. The presentation
was titled “Reduced-bias inference in generalized linear models”.

Motivation, details and discussion on the methods that **brglm2**
implements are provided in

> Kosmidis, I, Kenne Pagui, E C, Sartori N. (2020). Mean and median bias
> reduction in generalized linear models. [*Statistics and
> Computing*](https://doi.org/10.1007/s11222-019-09860-6) *30*, 43–59.

The [iteration
vignette](https://cran.r-project.org/package=brglm2/vignettes/iteration.html)
presents the iteration and give mathematical details for the
bias-reducing adjustments to the score functions for generalized linear
models.

Maximum Diaconis-Ylvisaker prior penalized likelihood and
high-dimensionality corrections under proportional asymptotics are
described in

> Sterzinger P, Kosmidis I (2024). Diaconis-Ylvisaker prior penalized
> likelihood for *p*/*n* → *κ* ∈ (0, 1) logistic regression.
> [*arXiv*:2311.07419v2](https://arxiv.org/abs/2311.07419).

## Code of Conduct

Please note that the **brglm2** project is released with a [Contributor
Code of
Conduct](https://contributor-covenant.org/version/2/1/CODE_OF_CONDUCT.html).
By contributing to this project, you agree to abide by its terms.

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6,
  fig.height = 6
)

## ----echo = TRUE--------------------------------------------------------------
freq <- c(15, 16, 16, 27, 33, 20,
          21, 18, 26, 41, 38, 27,
          29, 21, 33, 60, 41, 42)
dose <- rep(c(0, 10, 33, 100, 333, 1000), 3)
plate <- rep(1:3, each = 6)
(salmonella <- data.frame(freq, dose, plate))

## ----echo = TRUE--------------------------------------------------------------
ames_f <- freq ~ dose + log(dose + 10)

## ----echo = TRUE--------------------------------------------------------------
library("brglm2")
ames_ML <- brnb(ames_f, link = "log", data = salmonella,
                transformation = "identity",  type = "ML")
## Estimated regression and dispersion parameters
est <- coef(ames_ML, model = "full")
## Estimated standard errors for the regression parameters
sds <- sqrt(c(diag(ames_ML$vcov.mean), ames_ML$vcov.dispersion))
round(cbind(est, sds), 4)

## ----echo = TRUE--------------------------------------------------------------
ames_BC <- update(ames_ML, type = "correction")
## Estimated regression and dispersion parameters
est <- coef(ames_BC, model = "full")
## Estimated standard errors for the regression parameters
sds <- sqrt(c(diag(ames_BC$vcov.mean), ames_BC$vcov.dispersion))
round(cbind(est, sds), 4)

## ----echo = TRUE--------------------------------------------------------------
ames_BRmean <- update(ames_ML, type = "AS_mean")
## Estimated regression and dispersion parameters
est <- coef(ames_BRmean, model = "full")
## Estimated standard errors for the regression parameters
sds <- sqrt(c(diag(ames_BRmean$vcov.mean), ames_BRmean$vcov.dispersion))
round(cbind(est, sds), 4)

## ----echo = TRUE--------------------------------------------------------------
ames_BRmedian <- update(ames_ML, type = "AS_median")
## Estimated regression and dispersion parameters
est <- coef(ames_BRmedian, model = "full")
## Estimated standard errors for the regression parameters
sds <- sqrt(c(diag(ames_BRmedian$vcov.mean), ames_BRmedian$vcov.dispersion))
round(cbind(est, sds), 4)

## ----echo = TRUE--------------------------------------------------------------
ames_BRmixed <- update(ames_ML, type = "AS_mixed")
## Estimated regression and dispersion parameters
est <- coef(ames_BRmixed, model = "full")
## Estimated standard errors for the regression parameters
sds <- sqrt(c(diag(ames_BRmixed$vcov.mean), ames_BRmixed$vcov.dispersion))
round(cbind(est, sds), 4)

## Bias-reduced Poisson and multinomial log-linear models
## using maximum penalized likelihood as in Firth (1993) Biometrika.
##
## Author: David Firth, d.firth@warwick.ac.uk, 2 Sep 2010
##
## NOT A POLISHED PIECE OF PUBLIC-USE SOFTWARE!  Provided "as is",
## licensed under GPL2.  NO WARRANTY OF FITNESS FOR ANY PURPOSE!
##
## This code still has lots of rough edges, some of which are
## indicated by the embedded comments.
##
## The intention is to include a more polished implementation as part
## of the CRAN package "brglm", in due course.

## Arguments are as for a Poisson log-linear glm,
## with one new argument:
## -- "fixed.totals", a factor indexing groups of observations
##    whose response totals are fixed (for example, this might be
##    the rows of a table).
## If fixed.totals is NULL (the default), pure Poisson sampling (with
## no totals fixed) is assumed.

brpr <- function(formula,
                  data,
                  subset,
                  na.action,
                  offset = NULL,
                  control = glm.control(...),
                  contrasts = NULL,
                  fixed.totals = NULL,
                  weights = NULL,
                  ...) {
    ## The "weights" and "offset" arguments here can only be NULL -- anything
    ## else is an error (for now, at least)
    if (!is.null(weights)) stop("The weights argument here can only be NULL")
    if (!is.null(offset)) stop("The offset argument here can only be NULL")
    ##  work needed here, to allow the use of offset as an argument?

    ## Initial setup as in glm (except that `prior weights' are not used here,
    ## and an offset (if any) must be specified through the formula rather
    ## than (as is also allowed by glm) as a separate argument:
    call <- match.call()
    theFormula <- formula
    if (missing(data))
        data <- environment(formula)
    if (is.null(call$epsilon)) control$epsilon <- 1e-8
    ##  Is 1e-8 too stringent?  Work needed here!
    fixed.totals <- eval(substitute(fixed.totals), envir = data)
    mf <- match.call(expand.dots = FALSE)
    m <- match(c("formula", "data", "subset", "na.action",
                 "offset", "fixed.totals"), names(mf), 0)
    mf <- mf[c(1, m)]
    mf$drop.unused.levels <- TRUE
    mf[[1]] <- as.name("model.frame")
    mf <- eval(mf, parent.frame())
    mt <- attr(mf, "terms")
    Y <- model.response(mf, "numeric")

    ## Correct the fixed.totals variable for subsetting, NA treatment, etc.
    if (!is.null(fixed.totals)) {
        fixed.totals <-  mf$"(fixed.totals)"
        if (!is.factor(fixed.totals)) stop("fixed.totals must be a factor")
        rowTotals <-  as.vector(tapply(Y, fixed.totals, sum))[fixed.totals]
    }

    ## Do an initial fit of the model, with constant-adjusted counts
    adj <- 0.5
    X <- model.matrix(formula, data = mf)
    offset <- model.offset(mf)
    nFixed <- if (is.null(fixed.totals)) 0 else nlevels(fixed.totals)
    mf$y.adj <- Y + adj * (ncol(X) - nFixed)/nrow(X)
    formula <- update(formula, y.adj ~ .)
    op <- options(warn = -1)
    fit <- glm.fit(X, mf$y.adj, family = poisson(), offset = offset,
                   control = glm.control(maxit = 1))
    options(op)
    ## Update the model iteratively, refining the adjustment at each iteration
    criterion <- 1
    iter <- 1
    coefs <- coef(fit)
    Xwork <- X[ , !is.na(coefs), drop = FALSE]
    muAdj <- fitted(fit)
    coefs <- na.omit(coefs)
    while (criterion > control$epsilon && iter < control$maxit) {
        iter <- iter + 1
        if (!is.null(fixed.totals)){
            muTotals <-  as.vector(tapply(muAdj, fixed.totals,
                                          sum))[fixed.totals]
            mu <- muAdj * rowTotals / muTotals
        } else { ## case fixed.totals is NULL
            mu <- muAdj
        }
        W.X <- sqrt(mu) * Xwork
        XWXinv <- chol2inv(chol(crossprod(W.X)))
        coef.se <- sqrt(diag(XWXinv))
        h <- diag(Xwork %*% XWXinv %*% t(mu * Xwork))
        mf$y.adj <- Y + h * adj
        z <- log(muAdj) + (mf$y.adj - muAdj)/muAdj
        lsfit <- lm.wfit(Xwork, z, muAdj, offset = offset)
        criterion <- max((abs(lsfit$coefficients - coefs))/coef.se)
        if (control$trace) cat("Iteration ", iter,
                               ": largest (scaled) coefficient change is ",
                               criterion, "\n", sep = "")
        coefs <- lsfit$coefficients
        muAdj <- exp(lsfit$fitted.values)
    }

    ## The object `lsfit' uses *adjusted* counts, fitted values, etc. --
    ## so we need to correct various things before returning (so that, for example,
    ## deviance and standard errors are correct).
    fit$coefficients[!is.na(fit$coefficients)] <- coefs
    fit$y <- Y
    fit$fitted.values <- fit$weights <- mu
    fit$linear.predictors <- log(mu)
    dev.resids <- poisson()$dev.resids
    wt <- fit$prior.weights
    fit$deviance <- sum(dev.resids(Y, mu, wt))
    fit$null.deviance <- sum(dev.resids(Y, mean(Y), wt))
    fit$aic <- NA # sum((poisson()$aic)(Y, 1, mu, wt, 0)) ??  work needed!
    fit$residuals <- Y/mu - 1

    ## Next bit is straight from glm -- not sure it applies here (work needed!)
    if (any(offset) && attr(mt, "intercept") > 0) {
        fit$null.deviance <- glm.fit(x = X[, "(Intercept)", drop = FALSE],
                                     y = Y, weights = rep(1, nrow(mf)),
                                     offset = offset, family = poisson(),
                                     intercept = TRUE)$deviance
    }

    fit$iter <- iter
    fit$model <- mf
    fit$na.action <- attr(mf, "na.action")
    fit$x <- X
    fit$qr <- qr(sqrt(mu) * X)
    fit$R <- qr.R(fit$qr, complete = TRUE)
    ## The "effects" component of the fit object is almost certainly not
    ## correct as is -- but where does it actually get used?  (work needed!)
    ##  rownames(fit$R) <- colnames(fit$R)  ##  FIX THIS!
    fit <- c(fit, list(call = call, formula = theFormula, terms = mt,
                       data = data, offset = offset, method = "brpr",
                       contrasts = attr(X, "contrasts"),
                       xlevels = .getXlevels(mt, mf)))
    class(fit) <- c("glm", "lm")
    return(fit)
}

## source(system.file("brpr", "brpr.R", package = "brglm2"))
source("brpr.R")

## Dobson (1990) Page 93: Randomized Controlled Trial :
counts <- c(18,17,15,20,10,20,25,13,12)
outcome <- gl(3,1,9)
treatment <- gl(3,3)
dobson <- data.frame(counts, outcome, treatment)

tol <- 1e-06
## BR estimates and std. errors from brglmFit and from brpr are the same for poisson
expect_warning({
    br1 <- summary(glm(counts ~ outcome + treatment, family = poisson(), method = "brglmFit"))
    br2 <- summary(brpr(counts ~ outcome + treatment, data = dobson))
})
c1 <- coef(br1)
c2 <- coef(br2)
c2 <- c2[rownames(c1), ]
expect_equal(c1,c2, tolerance = tol)


## ## mbest::firthglm.fit crashes for this example
## ## A Gamma example, from McCullagh & Nelder (1989, pp. 300-2)
## clotting <- data.frame(
##     u = c(5,10,15,20,30,40,60,80,100, 5,10,15,20,30,40,60,80,100),
##     conc = c(118,58,42,35,27,25,21,19,18,69,35,26,21,18,16,13,12,12),
##     lot = factor(c(rep(1, 9), rep(2, 9))))
## mod <- glm(conc ~ lot*log(u), data = clotting, family = Gamma())

## tol <- sqrt(.Machine$double.eps)
## test_that("BR estimates from brglmFit and from mbest::firthglm.fit are the same for gamma", {
##     br1 <- update(mod, method = "firthglm.fit")
##     br2 <- update(mod, method = "brglmFit")
##     c1 <- coef(br1)
##     c2 <- coef(br2)
##     c2 <- c2[names(c1)]
##     expect_equal(c1,c2, tolerance = tol)
## })

#############################################################
### Median bias reduction in poisson regression  ############
#############################################################
mod1 <- function(X,mu,InfoInv) {
    X<-as.matrix(X)
    n <- nrow(X)
    p <- ncol(X)
    nu_r_s_t  <- array(0,c(p,p,p))
    k1 <- k2 <- k3 <- rep(NA,p)
    for(r in 1:p) {
        nu_r_s_t[r,,] = t(X)%*%(X[,r]*mu*X)
    }
    k2 <- 1/diag(InfoInv)
    barA <- InfoInv*k2
    for (r in 1:p) {
        sum_s1 <- rep(0,p)
        sum_s3 <- rep(0,p)
        nu.tu <- InfoInv-outer(InfoInv[,r]*k2[r],InfoInv[,r])
        for (s in 1:p){
            sum_s1[s] <- sum(diag(nu.tu%*%(nu_r_s_t[s,,])))
            sum_s3[s] <- sum((barA[r,]%*%nu_r_s_t[s,,])*barA[r,])
        }
        k1[r] <- -0.5*sum(sum_s1*barA[r,])
        k3[r] <- sum(sum_s3*barA[r,])
    }
    return(-k1/k2+k3/(6*k2^2))
}


mbrpr <- function(par,y,X,eps=1e-06,maxit=500) {
    step <- .Machine$integer.max
    nIter <- 0
    test <- TRUE
    while ( test & (nIter < maxit)) {
        nIter <- nIter + 1
        eta = drop(X%*%par)
        mu <- exp(eta)
        info <- t(X)%*%(W<-diag(mu))%*%X
        score <- t(X)%*%(y-mu)
        InfoInv <- try(chol2inv(chol(info)),TRUE)
        if(failedInv <- inherits(InfoInv, "try-error"))  {
            warning("failed to invert the information matrix: iteration stopped prematurely")
            break
        }
        mod <- mod1(X, mu, InfoInv)
        modscore <- score + info%*%mod
        y.adj <- X%*%(par+mod)+ solve(diag(mu))%*%(y-mu)
        par <- InfoInv%*%t(X)%*%(W%*%y.adj)
        test <- sqrt(crossprod(drop(modscore))) > eps
    }
    converged <- nIter < maxit
    list(coefficients=drop(par),converged=converged,nIter=nIter,InfoInv=InfoInv)
}

## mbrpr is supplied under the permission of M. Zambelli, and is part
## of his unpublished undergraduate thesis submitted in 2016 at the
## Department of Statisitcal Science, University of Padova

## source(system.file("mbrpr", "mbrpr.R", package = "brglm2"))
source("mbrpr.R")

## Dobson (1990) Page 93: Randomized Controlled Trial :
counts <- c(18, 17, 15, 20, 10, 20, 25, 13, 12)
outcome <- gl(3, 1,9)
treatment <- gl(3, 3)
dobson <- data.frame(counts, outcome, treatment)

tol <- 1e-06

## MBR estimates and std. errors from brglmFit and from mbrpr are the same for poisson
expect_warning(
    br1 <- summary(fit<-glm(counts ~ outcome + treatment, family = poisson(), method = "brglmFit",
                            type="AS_median"))
)

X <- model.matrix(fit)
y <- fit$y
br2 <- mbrpr(par = rep(0, ncol(X)), y = y, X = X, eps = 1e-10, maxit = 500)
c1 <- coef(br1)[, c(1,2)]
dimnames(c1) <- NULL
c2 <- cbind(br2$coefficients,sqrt(diag(br2$InfoInv)))
expect_equal(c1,c2, tolerance = tol)

## mbrpr is supplied under the permission of M. Zambelli, and is part
## of his unpublished undergraduate thesis submitted in 2016 at the
## Department of Statisitcal Science, University of Padova

## source(system.file("mbrpr", "mbrpr.R", package = "brglm2"))
source("mbrpr.R")

## Dobson (1990) Page 93: Randomized Controlled Trial :
counts <- c(18, 17, 15, 20, 10, 20, 25, 13, 12)
outcome <- gl(3, 1,9)
treatment <- gl(3, 3)
dobson <- data.frame(counts, outcome, treatment)

tol <- 1e-06

## MBR estimates and std. errors from brglmFit and from mbrpr are the same for poisson
br1 <- summary(fit <- glm(counts ~ outcome + treatment, family = poisson(), method = "brglmFit",
                          type = "AS_median"))

X <- model.matrix(fit)
y <- fit$y
br2 <- mbrpr(par = rep(0, ncol(X)), y = y, X = X, eps = 1e-10, maxit = 500)
c1 <- coef(br1)[, c(1,2)]
dimnames(c1) <- NULL
c2 <- cbind(br2$coefficients,sqrt(diag(br2$InfoInv)))
expect_equal(c1,c2, tolerance = tol)

acl_probs <- function(theta, X, ncat, po = TRUE) {
    X <- as.matrix(X)
    int <- (ncat - 1):1
    if (po) {
        nslopes <- ncol(X) - 1L
        if (length(theta) != (ncat - 1) + nslopes) {
            stop("for `po = TRUE`, `length(theta)` must be `ncat - 1 + ncol(X) - 1`")
        }
        coefs <- cbind(rev(cumsum(theta[int])),
                       int * matrix(theta[-int], nrow = ncat - 1,
                                    ncol = nslopes, byrow = TRUE))
    } else {
        if (length(theta) != (ncat - 1) * ncol(X)) {
            stop("for `po = FALSE`, `length(theta)` must be `(ncat - 1) * ncol(X)`")
        }
        coefs <- matrix(theta, nrow = ncat - 1)
        coefs <- apply(coefs, 2, function(x) rev(cumsum(x[int])))
        if (is.null(dim(coefs))) {
            coefs <- matrix(coefs, ncol = 1L)
        }
    }
    eta <- X %*% t(coefs)
    fits <- cbind(eta, 0)
    exp_fits <- exp(fits - apply(fits, 1, max))
    exp_fits / rowSums(exp_fits)
}


loglik <- function(theta, Y, X, po = TRUE) {
    X <- as.matrix(X)
    if (is.factor(Y)) {
        y <- as.integer(Y)
        ncat <- nlevels(Y)
    } else {
        ylev <- sort(unique(Y))
        y <- match(Y, ylev)
        ncat <- length(ylev)
    }
    probs <- acl_probs(theta, X, ncat = ncat, po = po)
    sum(log(pmax(probs[cbind(seq_along(y), y)], .Machine$double.xmin)))
}

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6,
  fig.height = 6
)

## ----echo = TRUE--------------------------------------------------------------
library("brglm2")
data("alligators", package = "brglm2")

## ----echo = TRUE--------------------------------------------------------------
agresti_contrasts <- list(lake = contr.treatment(levels(alligators$lake), base = 4),
                          size = contr.treatment(levels(alligators$size), base = 2))
all_ml <- brmultinom(foodchoice ~ size + lake , weights = freq,
                     data = alligators,
                     contrasts = agresti_contrasts,
                     ref = 1,
                     type = "ML")
all_ml_summary <- summary(all_ml)
## Estimated regression parameters
round(all_ml_summary$coefficients, 2)
## Estimated standard errors
round(all_ml_summary$standard.errors, 2)

## ----echo = TRUE--------------------------------------------------------------
all_mean <- update(all_ml, type = "AS_mean")
summary(all_mean)

## ----echo = TRUE--------------------------------------------------------------
all_median <- update(all_ml, type = "AS_median")
summary(all_median)

## ----echo = TRUE, error = TRUE------------------------------------------------
try({
all_ml_sparse <- update(all_ml, weights = round(freq/3), slowit = 0.1)
summary(all_ml_sparse)
})

## ----echo = TRUE--------------------------------------------------------------
library("detectseparation")
se_ratios <- check_infinite_estimates(all_ml_sparse)
plot(se_ratios)

## ----echo = TRUE--------------------------------------------------------------
all_mean_sparse <- update(all_ml_sparse, type = "AS_mean")
summary(all_mean_sparse)

all_median_sparse <- update(all_ml_sparse, type = "AS_median")
summary(all_median_sparse)

# Copyright (C) 2021- Ioannis Kosmidis

#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 or 3 of the License
#  (at your option).
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  A copy of the GNU General Public License is available at
#  http://www.r-project.org/Licenses/

#' Ordinal superiority scores of Agresti and Kateri (2017)
#'
#' [ordinal_superiority()] is a method for the estimation and
#' inference about model-based ordinal superiority scores introduced
#' in Agresti and Kateri (2017, Section 5) from fitted objects. The
#' mean bias of the estimates of the ordinal superiority scores can be
#' corrected.
#'
#' @aliases ordinal_superiority
#'
#' @param object a fitted object from an ordinal regression
#'     model. Currently only models from class [`"bracl"`][bracl] are supported.
#' @param formula a RHS formula indicating the group variable to use.
#' @param data an optional data frame in which to look for variables
#'     with which to compute ordinal superiority measures.  If
#'     omitted, an attempt is made to use the data that produced
#'     `object`.
#' @param measure either `"gamma"` (default) or `"Delta"`, specifying
#'     the ordinal superiority measure to be returned.
#' @param level the confidence level required when computing
#'     confidence intervals for the ordinal superiority measures.
#' @param bc logical. If `FALSE` (default) then the ordinal
#'     superiority measures are computed using the estimates in
#'     `object`. If `TRUE` then the ordinal superiority measure
#'     estimates are corrected for mean bias.
#'
#' @references
#'
#' Agresti, A., Kateri, M. (2017). Ordinal probability effect measures
#' for group comparisons in multinomial cumulative link models.
#' *Biometrics*, **73** 214-219. \doi{10.1111/biom.12565}.
#'
#' @examples
#'
#' data("stemcell", package = "brglm2")
#'
#' # Adjacent category logit (proportional odds)
#' stem <- within(stemcell, {nreligion = as.numeric(religion)})
#' fit_bracl_p <- bracl(research ~ nreligion + gender, weights = frequency,
#'                      data = stem, type = "ML", parallel = TRUE)
#'
#' # Estimates and 95% confidence intervals for the probabilities that the response
#' # category for gender "female" is higher than the response category for gender "male",
#' # while adjusting for religion.
#' ordinal_superiority(fit_bracl_p, ~ gender)
#'
#' \donttest{
#' # And their (very-similar in value here) bias corrected versions
#' # with 99% CIs
#' ordinal_superiority(fit_bracl_p, ~ gender, bc = TRUE, level = 0.99)
#' # Note that the object is refitted with type = "AS_mean"
#'
#' }
#'
#' @export
ordinal_superiority.bracl <- function(object, formula, data,
                                      measure = c("gamma", "Delta"),
                                      level = 0.95,
                                      bc = FALSE) {
    measure <- match.arg(measure)
    if (!inherits(object, "bracl")) {
        stop("ordinal superiority measures are not available for objects of class ", class(object)[1])
    }
    ## If bc is TRUE and object is not a reduced mean-bias fit then
    ## compute reduced mean-bias estimators.
    if (isTRUE(bc) & !(object$type %in% c("AS_mean", "AS_mixed"))) {
        object <- update(object, type = "AS_mean")
        warning('`object` was refitted using `type = "AS_mean"`.')
    }
    source_data <- if (missing(data)) {
        if (!is.null(object$call$data)) {
            eval(object$call$data, environment(formula(object)), parent.frame())
        } else {
            environment(formula(object))
        }
    } else {
        data
    }
    source_data <- get_all_vars(delete.response(object$terms), source_data)
    row_ids <- if (missing(data)) {
        row.names(model.frame(object))
    } else {
        row.names(model.frame(delete.response(object$terms), source_data,
                              na.action = na.omit, xlev = object$xlevels))
    }
    source_data <- source_data[row_ids, , drop = FALSE]
    mf <- model.frame(formula, source_data)
    Terms <- attr(mf, "terms")
    z <- model.matrix(Terms, mf, object$contrasts[all.vars(formula)])
    znames <- colnames(z)[-match("(Intercept)", colnames(z), nomatch = 0L)]
    if (isTRUE(length(znames) > 1) | !isTRUE(all(sort(unique(z[, znames])) %in% c(0, 1)))) {
        stop("`formula` can have only one grouping explanatory variable with two levels")
    }
    group_name <- all.vars(formula)
    if (length(group_name) != 1L) {
        stop("`formula` can have only one grouping explanatory variable with two levels")
    }
    predictor_data <- source_data[, !(names(source_data) %in% c(as.character(formula(object)[[2L]]), "(weights)")), drop = FALSE]
    if (!(group_name %in% names(predictor_data))) {
        stop("the grouping explanatory variable must be one of the explanatory variables in `object`")
    }
    group_values <- if (is.factor(mf[[1L]])) levels(droplevels(mf[[1L]])) else sort(unique(mf[[1L]]))
    base_data <- unique(predictor_data[, setdiff(names(predictor_data), group_name), drop = FALSE])
    X0data <- X1data <- base_data
    X0data[[group_name]] <- rep(group_values[1L], nrow(base_data))
    X1data[[group_name]] <- rep(group_values[2L], nrow(base_data))
    X0 <- model.matrix(object, data = X0data)
    X1 <- model.matrix(object, data = X1data)
    same_cols <- colSums(abs(X0 - X1)) == 0
    Xnoz <- X0[, same_cols, drop = FALSE]
    nx <- nrow(X0)
    gammas <- .ordsup(coef(object),
                      X0 = X0, X1 = X1, ncat = object$ncat,
                      ref = object$ref, lev = object$lev, measure = "gamma",
                      po = object$parallel)
    coef_vcov <- vcov(object)
    ## mean bias reduction of gammas
    bias_gammas <- numeric(length(gammas))
    if (bc) {
        for (j in seq.int(nx)) {
            hess <- numDeriv::hessian(function(theta, ...) .ordsup(theta, ...)[j],
                                      x = coef(object),
                                      X0 = X0, X1 = X1, ncat = object$ncat,
                                      ref = object$ref, lev = object$lev, measure = "gamma",
                                      po = object$parallel)
            bias_gammas[j] <- sum(diag(coef_vcov %*% hess)) / 2
        }
    }
    gammas <- gammas - bias_gammas
    ## mean_gammas <- mean(gammas)
    ## compute standard error for gamma
    grads <- numDeriv::jacobian(.ordsup, x = coef(object),
                                X0 = X0, X1 = X1, ncat = object$ncat,
                                ref = object$ref, lev = object$lev, measure = "gamma",
                                po = object$parallel)
    ## grad_mean <- apply(grads, 2, mean)
    se <- apply(grads, 1, function(x) sqrt(crossprod(x, (coef_vcov %*% x))))
    ## se_mean <- sqrt(crossprod(grad_mean, (coef_vcov %*% grad_mean)))
    ## Confidence intervals as in Agresti and Kateri
    a <- 1/2 + level/2
    pct <- paste(format(100 * c(1 - a, a), trim = TRUE, scientific = FALSE, digits = 3), "%")
    lsd <- drop(qnorm(a) * se / (gammas * (1 - gammas)))
    ci <- qlogis(gammas) + cbind(rep(-1, nx),  1) * lsd
    ## lsd_mean <- drop(qnorm(a) * se_mean / (mean_gammas * (1 - mean_gammas)))
    ## ci_mean <- qlogis(mean_gammas) + c(-1, 1) * lsd_mean
    if (isTRUE(measure == "Delta")) {
        out <- cbind(Xnoz, 2 * gammas - 1, 2 * se, 2 * plogis(ci) - 1)
        ## out_mean <- c(2 * mean_gammas - 1, 2 * se_mean, 2 * plogis(ci_mean) - 1)
        colnames(out)[ncol(Xnoz) + 1:4] <- c("Delta", "se", pct)
        ## names(out_mean) <- c("Delta*", "se", pct)
    } else {
        out <- cbind(Xnoz, gammas, se, plogis(ci))
        ## out_mean <- c(mean_gammas, se_mean, plogis(ci_mean))
        colnames(out)[ncol(Xnoz) + 1:4] <- c("gamma", "se", pct)
        ## names(out_mean) <- c("gamma*", "se", pct)
    }
    ## list(individual = out, mean = out_mean)
    out
}

.ordsup <- function(coef, X0, X1, ncat, ref, lev, measure = "gamma", po = TRUE) {
    nams <- names(coef)
    int <- (ncat - 1):1
    sl <- nams[-int]
    if (po) {
        coefs <- cbind(rev(cumsum(coef[int])),
                       int * matrix(coef[sl], nrow = ncat - 1, ncol = length(sl), byrow = TRUE))
    } else {
        coefs <- matrix(coef, nrow = ncat - 1)
        coefs <- apply(coefs, 2, function(x) rev(cumsum(x[int])))
    }
    rownames(coefs) <- lev[-ref]
    probs0 <- .bracl_probs(coefs, X0, ncat, lev)
    probs1 <- .bracl_probs(coefs, X1, ncat, lev)
    gamma_fun <- function(p0, p1) {
        out <- outer(p0, p1, "*")
        sum(out[upper.tri(out)]) + sum(diag(out)) / 2
    }
    out <- numeric(nrow(X0))
    for (i in seq_len(nrow(X0))) {
        out[i] <- gamma_fun(probs0[i, ], probs1[i, ])
    }
    if (measure == "gamma") out else 2 * out - 1
}

.bracl_probs <- function(coefs, X, ncat, lev) {
    fits <- matrix(0, nrow = nrow(X), ncol = ncat, dimnames = list(rownames(X), lev))
    fits1 <- apply(coefs, 1, function(b) X %*% b)
    fits[, rownames(coefs)] <- fits1
    t(apply(fits, 1, function(x) exp(x) / sum(exp(x))))
}

data("stemcell", package = "brglm2")

# Adjacent category logit (non-proportional odds)
fit_bracl <- bracl(research ~ religion + gender, weights = frequency,
                   data = stemcell, type = "ML")
# Adjacent category logit (proportional odds)
fit_bracl_p <- bracl(research ~ religion + gender, weights = frequency,
                    data = stemcell, type = "ML", parallel = TRUE)

tol <- 1e-08
## ordinal_superiority agrees with fitted probabilities for main-effects models
gamma_fun <- function(p0, p1) {
    out <- outer(p0, p1, "*")
    sum(out[upper.tri(out)]) + sum(diag(out)) / 2
}
expected_ordsup <- function(fit) {
    nd0 <- data.frame(religion = factor(levels(stemcell$religion),
                                        levels = levels(stemcell$religion),
                                        ordered = is.ordered(stemcell$religion)),
                      gender = factor("male", levels = levels(stemcell$gender)))
    nd1 <- nd0
    nd1$gender <- factor("female", levels = levels(stemcell$gender))
    p0 <- predict(fit, nd0, type = "probs")
    p1 <- predict(fit, nd1, type = "probs")
    vapply(seq_len(nrow(p0)), function(i) gamma_fun(p0[i, ], p1[i, ]), numeric(1))
}
os_np <- ordinal_superiority(fit_bracl, ~ gender)
os_p <- ordinal_superiority(fit_bracl_p, ~ gender)

expect_equal(unname(os_np[, "gamma"]), expected_ordsup(fit_bracl), tolerance = tol)
expect_equal(unname(os_p[, "gamma"]), expected_ordsup(fit_bracl_p), tolerance = tol)
expect_equal(unname(ordinal_superiority(fit_bracl, ~ gender, measure = "Delta")[, "Delta"]),
             2 * unname(os_np[, "gamma"]) - 1, tolerance = tol)
expect_equal(unname(ordinal_superiority(fit_bracl_p, ~ gender, measure = "Delta")[, "Delta"]),
             2 * unname(os_p[, "gamma"]) - 1, tolerance = tol)

## data argument works with predictor-only data
pred_data <- stemcell[c("religion", "gender")]
os_np_data <- ordinal_superiority(fit_bracl, ~ gender, data = pred_data)
os_p_data <- ordinal_superiority(fit_bracl_p, ~ gender, data = pred_data)
expect_equal(os_np_data, os_np, tolerance = tol, check.attributes = FALSE)
expect_equal(os_p_data, os_p, tolerance = tol, check.attributes = FALSE)

# Adjacent category logit models with interaction
fit_bracl_int <- bracl(research ~ religion * gender, weights = frequency,
                       data = stemcell, type = "ML")
fit_bracl_p_int <- bracl(research ~ religion * gender, weights = frequency,
                         data = stemcell, type = "ML", parallel = TRUE)

expected_ordsup_int <- function(fit) {
    nd0 <- data.frame(religion = factor(levels(stemcell$religion),
                                        levels = levels(stemcell$religion),
                                        ordered = is.ordered(stemcell$religion)),
                      gender = factor("male", levels = levels(stemcell$gender)))
    nd1 <- nd0
    nd1$gender <- factor("female", levels = levels(stemcell$gender))
    p0 <- predict(fit, nd0, type = "probs")
    p1 <- predict(fit, nd1, type = "probs")
    vapply(seq_len(nrow(p0)), function(i) gamma_fun(p0[i, ], p1[i, ]), numeric(1))
}

os_np_int <- ordinal_superiority(fit_bracl_int, ~ gender)
os_p_int <- ordinal_superiority(fit_bracl_p_int, ~ gender)

expect_identical(nrow(os_np_int), 3L)
expect_identical(nrow(os_p_int), 3L)
expect_equal(unname(os_np_int[, "gamma"]), expected_ordsup_int(fit_bracl_int), tolerance = tol)
expect_equal(unname(os_p_int[, "gamma"]), expected_ordsup_int(fit_bracl_p_int), tolerance = tol)

pred_data_int <- stemcell[c("religion", "gender")]
os_np_int_data <- ordinal_superiority(fit_bracl_int, ~ gender, data = pred_data_int)
os_p_int_data <- ordinal_superiority(fit_bracl_p_int, ~ gender, data = pred_data_int)
expect_equal(os_np_int_data, os_np_int, tolerance = tol, check.attributes = FALSE)
expect_equal(os_p_int_data, os_p_int, tolerance = tol, check.attributes = FALSE)

# Adjacent category logit models with transformed covariates in formula
fit_bracl_tf <- bracl(research ~ as.numeric(religion) + gender, weights = frequency,
                      data = stemcell, type = "ML")
fit_bracl_p_tf <- bracl(research ~ as.numeric(religion) + gender, weights = frequency,
                        data = stemcell, type = "ML", parallel = TRUE)

os_np_tf <- ordinal_superiority(fit_bracl_tf, ~ gender)
os_p_tf <- ordinal_superiority(fit_bracl_p_tf, ~ gender)

expect_equal(unname(os_np_tf[, "gamma"]), expected_ordsup(fit_bracl_tf), tolerance = tol)
expect_equal(unname(os_p_tf[, "gamma"]), expected_ordsup(fit_bracl_p_tf), tolerance = tol)

pred_data_tf <- stemcell[c("religion", "gender")]
os_np_tf_data <- ordinal_superiority(fit_bracl_tf, ~ gender, data = pred_data_tf)
os_p_tf_data <- ordinal_superiority(fit_bracl_p_tf, ~ gender, data = pred_data_tf)
expect_equal(os_np_tf_data, os_np_tf, tolerance = tol, check.attributes = FALSE)
expect_equal(os_p_tf_data, os_p_tf, tolerance = tol, check.attributes = FALSE)

# Something odd
fit_bracl_tf <- bracl(research ~ sqrt(as.numeric(religion)) + gender, weights = frequency,
                      data = stemcell, type = "ML")
fit_bracl_p_tf <- bracl(research ~ sqrt(as.numeric(religion)) + gender, weights = frequency,
                        data = stemcell, type = "ML", parallel = TRUE)

os_np_tf <- ordinal_superiority(fit_bracl_tf, ~ gender)
os_p_tf <- ordinal_superiority(fit_bracl_p_tf, ~ gender)

expect_equal(unname(os_np_tf[, "gamma"]), expected_ordsup(fit_bracl_tf), tolerance = tol)
expect_equal(unname(os_p_tf[, "gamma"]), expected_ordsup(fit_bracl_p_tf), tolerance = tol)

pred_data_tf <- stemcell[c("religion", "gender")]
os_np_tf_data <- ordinal_superiority(fit_bracl_tf, ~ gender, data = pred_data_tf)
os_p_tf_data <- ordinal_superiority(fit_bracl_p_tf, ~ gender, data = pred_data_tf)
expect_equal(os_np_tf_data, os_np_tf, tolerance = tol, check.attributes = FALSE)
expect_equal(os_p_tf_data, os_p_tf, tolerance = tol, check.attributes = FALSE)

# Copyright (C) 2016- Ioannis Kosmidis

#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 or 3 of the License
#  (at your option).
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  A copy of the GNU General Public License is available at
#  http://www.r-project.org/Licenses/


#' Bias reduction for adjacent category logit models for ordinal
#' responses using the Poisson trick.
#'
#' [bracl()] is a wrapper of [brglmFit()] that fits adjacent category
#' logit models with or without proportional odds using implicit and
#' explicit bias reduction methods. See Kosmidis & Firth (2011) for
#' details.
#'
#' @inheritParams MASS::polr
#' @param control a list of parameters for controlling the fitting
#'     process. See [brglmControl()] for details.
#' @param parallel if `FALSE` (default), then a non-proportional odds
#'     adjacent category model is fit, assuming different effects per
#'     category; if `TRUE` then a proportional odds adjacent category
#'     model is fit. See Details.
#' @param x should the model matrix be included with in the result
#'     (default is `TRUE`).
#' @param ... arguments to be used to form the default `control`
#'     argument if it is not supplied directly.
#'
#' @details
#'
#' The [bracl()] function fits adjacent category models, which assume
#' multinomial observations with probabilities with proportional odds
#' of the form
#'
#' \deqn{\log\frac{\pi_{ij}}{\pi_{ij + 1}} = \alpha_j + \beta^T x_i}{log(pi[i, j]/pi[i, j+1]) = alpha[j] + sum(beta * x[i, ])}
#'
#' or with non-proportional odds of the form
#'
#' \deqn{\log\frac{\pi_{ij}}{\pi_{ij + 1}} = \alpha_j + \beta_j^T x_i}{log(pi[i, j]/pi[i, j+1]) = alpha[j] + sum(beta[j, ] * x[i, ])}
#'
#' where \eqn{x_i}{x[i, ]} is a vector of covariates and \eqn{\pi_{ij}}{pi[i, j]} is the
#' probability that category \eqn{j} is observed at the covariate setting \eqn{i}.
#'
#' @author Ioannis Kosmidis `[aut, cre]` \email{ioannis.kosmidis@warwick.ac.uk}
#'
#' @seealso [nnet::multinom()], [brmultinom()]
#'
#' @references
#'
#' Kosmidis I, Kenne Pagui E C, Sartori N (2020). Mean and median bias
#' reduction in generalized linear models. *Statistics and Computing*,
#' **30**, 43-59. \doi{10.1007/s11222-019-09860-6}.
#'
#' Agresti, A (2010). *Analysis of Ordinal Categorical Data* (2nd
#' edition).  Wiley Series in Probability and Statistics. Wiley.
#'
#' Albert A, Anderson J A (1984). On the Existence of Maximum
#' Likelihood Estimates in Logistic Regression Models. *Biometrika*,
#' **71**, 1-10. \doi{10.2307/2336390}.
#'
#' Kosmidis I, Firth D (2011). Multinomial logit bias reduction
#' via the Poisson log-linear model. *Biometrika*, **98**,
#' 755-759. \doi{10.1093/biomet/asr026}.
#'
#' Palmgren J (1981). The Fisher Information Matrix for Log Linear
#' Models Arguing Conditionally on Observed Explanatory
#' Variables. *Biometrika*, **68**,
#' 563-566. \doi{10.1093/biomet/68.2.563}.
#'
#' @examples
#'
#' data("stemcell", package = "brglm2")
#'
#' # Adjacent category logit (non-proportional odds)
#' fit_bracl <- bracl(research ~ as.numeric(religion) + gender, weights = frequency,
#'                    data = stemcell, type = "ML")
#' # Adjacent category logit (proportional odds)
#' fit_bracl_p <- bracl(research ~ as.numeric(religion) + gender, weights = frequency,
#'                     data = stemcell, type = "ML", parallel = TRUE)
#'
#'
#' @export
bracl <- function(formula, data, weights, subset, na.action,
                  parallel = FALSE,
                  contrasts = NULL,
                  model = TRUE, x= TRUE,
                  control = list(...), ...) {
    call <- match.call()
    if (missing(data)) {
        data <- environment(formula)
    }
    mf <- match.call(expand.dots = FALSE)
    m <- match(c("formula", "data", "subset", "weights", "na.action", "offset"), names(mf), 0L)
    mf <- mf[c(1L, m)]
    mf$drop.unused.levels <- TRUE
    mf[[1L]] <- quote(model.frame)
    mf <- eval.parent(mf)
    Terms <- attr(mf, "terms")
    ## Model matrix
    X <- model.matrix(Terms, mf, contrasts)
    xint <- match("(Intercept)", colnames(X), nomatch = 0L)
    ## If there is no intercept include it and issue warning
    if (xint == 0L) {
        cn <- colnames(X)
        X <- cbind(1L, X)
        xint <- 1
        colnames(X) <- c("(Intercept)", cn)
        warning("an intercept is needed and assumed")
    }
    Xcontrasts <- attr(X, "contrasts")
    ## Response
    Y <- model.response(mf, "any")
    n <- length(Y)
    if (!is.factor(Y)) {
        stop("response must be a factor")
    }
    lev <- levels(Y)
    ncat <- length(lev)
    if (ncat <= 2L) {
        stop("response must have 3 or more levels")
    }
    Y <- nnet::class.ind(Y)
    cons <- attr(X, "contrasts")
    ## Weights
    w <- model.weights(mf)
    if (!length(w)) {
        w <- rep(1, n)
    }

    ## Offset
    offset <- model.offset(mf)
    if (length(offset) <= 1L) {
        offset <- rep(0, n)
    }

    nvar <- ncol(X)
    keep <- w > 0
    nkeep <- sum(keep)

    ## cats0 <- apply(Y, 1, function(x) which(x == 1))
    cats <- rep(seq.int(ncat), each = nrow(X))[keep]

    fixed_totals <- rep(seq.int(nkeep), ncat)

    ## Set up the model matrix for the poisson fit
    Xnuisance <- Matrix::Diagonal(nkeep)
    int <- seq.int(nkeep)
    nd <- paste0("%0", nchar(max(int)), "d")
    if (parallel) {
        Xextended <- cbind(Matrix::kronecker(rep(1, ncat), Xnuisance),
                           Matrix::kronecker(Matrix::Diagonal(ncat)[, -ncat, drop = FALSE], X[keep, xint]),
        (ncat - cats) * Matrix::kronecker(c(rep(1, ncat - 1), 0), X[keep, -xint, drop = FALSE]))
        ofInterest <- c(paste(lev[-ncat], rep("(Intercept)", ncat - 1), sep = ":"), colnames(X)[-xint])
    } else {
        Xextended <- cbind(Matrix::kronecker(rep(1, ncat), Xnuisance),
                           Matrix::kronecker(Matrix::Diagonal(ncat)[, -ncat, drop = FALSE], X[keep, , drop = FALSE]))
        ofInterest <- paste(rep(lev[-ncat], each = nvar),
                            rep(colnames(X), ncat - 1), sep = ":")
    }
    colnames(Xextended) <- c(paste0(".nuisance", sprintf(nd, int)),
                             ofInterest)

    ## Set up the extended response
    Yextended <- c(Y[keep] * w[keep])

    fit <- brglmFit(x = Xextended, y = Yextended,
                    start = NULL,
                    family = poisson("log"), control = control, intercept = TRUE, fixed_totals = fixed_totals)

    ## Fitted values
    fitted <- do.call("rbind", tapply(fit$fitted, fixed_totals, function(x) x/sum(x)))
    rownames(fitted) <- rownames(X)[keep]
    colnames(fitted) <- lev
    fit$fitted.values <- fitted
    fit$parallel  <- parallel
    fit$call <- call
    class(fit) <- c("bracl", "brmultinom", fit$class, "glm")
    fit$ofInterest <- ofInterest
    fit$ncat <- ncat
    fit$lev <- lev
    fit$ref <- ncat
    if (model) {
        fit$model  <- mf
    }
    if (x) {
        fit$x  <- X
    }
    fit$contrasts <- attr(X, "contrasts")
    fit$xlevels = .getXlevels(Terms, mf)
    fit$terms <- Terms
    fit$coefNames <- colnames(X)
    fit$null.deviance <- NULL
    fit
}

#' @method fitted bracl
#' @export
fitted.bracl <- function(object, ...) {
    object$fitted.values
}

#' @method coef bracl
#' @export
coef.bracl <- function(object, ...) {
    if (length(object$ofInterest)) {
        if (object$parallel) {
            with(object, {
                coefs <- coefficients[ofInterest]
                intercept_names <- paste0(lev[-ref], ":", "(Intercept)")
                coefs[intercept_names] <- -diff(c(coefs[intercept_names], 0))
                names(coefs[intercept_names]) <- intercept_names
                coefs
            })
        } else {
            with(object, {
                coefs <- matrix(coefficients[ofInterest], ncol = ncat - 1)
                coefs <- -apply(cbind(coefs, 0), 1, diff)
                coefs <- c(coefs)
                names(coefs) <- c(sapply(coefNames, function(x) paste(object$lev[-ref], x, sep = ":")))
                coefs
            })
        }
    } else {
        NULL
    }
}

#' @method vcov bracl
#' @export
vcov.bracl <- function(object, ...) {
    vc <- vcov.brglmFit(object, ...)
    coefNames <- names(coefficients(object))
    ofInterest <- object$ofInterest
    vc <- vc[ofInterest, ofInterest]
    levs <- object$lev[-object$ref]
    intercept_names <- paste0(levs, ":", "(Intercept)")
    ddiff <- function(mat) {
        mat <- diff(rbind(mat, 0))
        diff(rbind(t(mat), 0))
    }
    if (object$parallel) {
        beta_names <- ofInterest
        beta_names <- beta_names[!(beta_names %in% intercept_names)]
        vbeta <- vc[beta_names, beta_names]
        vint <- ddiff(vc[intercept_names, intercept_names])
        vintslo <- -diff(rbind(vc[intercept_names, beta_names], 0))
        if (nrow(vintslo) == 1) {
            vintslo <- drop(vintslo)
        }
        par_names <- c(intercept_names, beta_names)
        vc[par_names, par_names] <- rbind(cbind(vint, vintslo),
                                          cbind(t(vintslo), vbeta))
    } else {
        betas <- object$coefNames
        npar <- length(betas)
        for (j in 1:npar) {
            for (k in 1:npar) {
                par_names1 <- paste(levs, betas[j], sep = ":")
                par_names2 <- paste(levs, betas[k], sep = ":")
                mat <- ddiff(vc[par_names1, par_names2])
                vc[par_names1, par_names2] <- mat
            }
        }
        ## re-order
        vc <- vc[coefNames, coefNames]
    }
    vc
}

#' @method summary bracl
#' @export
summary.bracl <- function(object, correlation = FALSE, digits = 3, ...) {
    object$digits <- digits
    object$logLik <- logLik(object)
    object$AIC <- AIC(object)
    coefs <- coefficients(object)
    aliased <- is.na(coefs)
    vc <- vcov(object)
    var_coef <- diag(vc)
    s.err <- sqrt(var_coef)
    tvalue <- coefs/s.err
    dn <- c("Estimate", "Std. Error")
    pvalue <- 2 * pnorm(-abs(tvalue))
    coef_table <- cbind(coefs, s.err, tvalue, pvalue)
    dimnames(coef_table) <- list(names(coefs), c(dn, "z value", "Pr(>|z|)"))
    object$coefficients <- coef_table
    if (correlation) {
        object$correlation <- vc/outer(s.err, s.err)
    }
    class(object) <- "summary.bracl"
    return(object)
}

#' @method print summary.bracl
#' @export
print.summary.bracl <- function(x, digits = x$digits, ...) {
    if (!is.null(cl <- x$call)) {
        cat("Call:\n")
        dput(cl, control = NULL)
    }
    cat("\nCoefficients:\n")
    printCoefmat(x$coefficients, digits = digits)
    cat("\nResidual Deviance:", format(x$deviance), "\n")
    cat("Log-likelihood:", format(x$logLik), "\n")
    cat("AIC:", format(x$AIC), "\n")
    cat("\n\nType of estimator:", x$type, get_type_description(x$type))
    cat("\n", "Number of Fisher Scoring iterations: ", x$iter, "\n", sep = "")
    if (!is.null(correl <- x$correlation)) {
        p <- dim(correl)[2L]
        if (p > 1) {
            cat("\nCorrelation of Coefficients:\n")
            ll <- lower.tri(correl)
            correl[ll] <- format(round(correl[ll], digits))
            correl[!ll] <- ""
            print(correl[-1L, -p], quote = FALSE, ...)
        }
    }
    invisible(x)
}

#' Predict method for [bracl] fits
#'
#' Obtain class and probability predictions from a fitted adjacent
#' category logits model.
#'
#' @param object a fitted object of class inheriting from [`"bracl"`][bracl].
#' @param newdata optionally, a data frame in which to look for
#'     variables with which to predict.  If omitted, the fitted linear
#'     predictors are used.
#' @param type the type of prediction required. The default is
#'     `"class"`, which produces predictions of the response category
#'     at the covariate values supplied in `"newdata"`, selecting the
#'     category with the largest probability; the alternative
#'     `"probs"` returns all category probabilities at the covariate
#'     values supplied in `newdata`.
#' @param ... further arguments passed to or from other methods.
#'
#'
#' @details
#'
#' If `newdata` is omitted the predictions are based on the data
#' used for the fit.
#'
#' @return
#'
#' If `type = "class"` a vector with the predicted response
#' categories; if `type = "probs"` a matrix of probabilities for all
#' response categories at `newdata`.
#'
#' @examples
#'
#' data("stemcell", package = "brglm2")
#'
#' # Adjacent category logit (non-proportional odds)
#' fit_bracl <- bracl(research ~ as.numeric(religion) + gender, weights = frequency,
#'                    data = stemcell, type = "ML")
#' # Adjacent category logit (proportional odds)
#' fit_bracl_p <- bracl(research ~ as.numeric(religion) + gender, weights = frequency,
#'                     data = stemcell, type = "ML", parallel = TRUE)
#'
#' # New data
#' newdata <- expand.grid(gender = c("male", "female"),
#'                        religion = c("liberal", "moderate", "fundamentalist"))
#'
#' # Predictions
#' sapply(c("class", "probs"), function(what) predict(fit_bracl, newdata, what))
#' sapply(c("class", "probs"), function(what) predict(fit_bracl_p, newdata, what))
#'
#' @method predict bracl
#' @export
predict.bracl <- function(object, newdata, type = c("class", "probs"), ...) {
    ## Adapted from nnet:::predict.multinom
    if (!inherits(object, "bracl"))
        stop("not a \"bracl\" fit")
    type <- match.arg(type)
    X <- if (missing(newdata)) model.matrix(object) else model.matrix(object, data = newdata)
    rn <- attr(X, "rn_data")
    keep <- attr(X, "rn_kept")
    cc <- coef(object)
    ## Ignore unidentifiable parameters
    cc[is.na(cc)] <- 0
    nams <- names(cc)
    if (object$parallel) {
        int <- (object$ncat - 1):1
        sl <- nams[-int]
        coefs <- cbind(rev(cumsum(cc[int])),
                       int * matrix(cc[sl], nrow = object$ncat - 1, ncol = length(sl), byrow = TRUE))
        rownames(coefs) <- object$lev[-object$ref]
    } else {
        coefs <- matrix(cc, nrow = object$ncat - 1)
        rownames(coefs) <- object$lev[-object$ref]
        coefs <- apply(coefs, 2, function(x) cumsum(rev(x)))
    }

    fits <- matrix(0, nrow = nrow(X), ncol = object$ncat, dimnames = list(rn[keep], object$lev))
    fits1 <- apply(coefs, 1, function(b) X %*% b)
    fits[, rownames(coefs)] <- fits1
    Y1 <- t(apply(fits, 1, function(x) exp(x) / sum(exp(x))))
    Y <- matrix(NA, length(rn), ncol(Y1), dimnames = list(rn, colnames(Y1)))
    Y[keep, ] <- Y1
    switch(type, class = {
        if (length(object$lev) > 2L) Y <- factor(max.col(Y),
            levels = seq_along(object$lev), labels = object$lev)
        if (length(object$lev) == 2L) Y <- factor(1 + (Y > 0.5),
            levels = 1L:2L, labels = object$lev)
        if (length(object$lev) == 0L) Y <- factor(max.col(Y),
            levels = seq_along(object$lab), labels = object$lab)
    }, probs = {
    })
    drop(Y)
}

library("VGAM")

### tests for bracl and its agreement with other methods

## Stem cell research data from Agresti (2017, Analysis of ordinal data, Table 4.1)
freq <- c(34, 41, 58, 21, 30, 64,
          67, 83, 63, 52, 52, 50,
          30, 23, 15, 24, 18, 16,
          25, 14, 12, 15, 11, 11)
fund_research <- factor(rep(c("definitely", "probably", "probably not", "definitely not"), each = 6),
                        levels = c("definitely", "probably", "probably not", "definitely not"),
                        ordered = TRUE)
gender <- factor(rep(rep(c("female", "male"), each = 3), 4), levels = c("male", "female"))
## gender <- rep(rep(c(1, 0), each = 3), 4)
religion <- factor(rep(c("fundamentalist", "moderate", "liberal"), 8),
                   levels = c("fundamentalist", "moderate", "liberal"),
                   ordered = TRUE)
stemcell <- data.frame(frequency = freq, research = fund_research, gender = gender, religion = religion)

ff <- matrix(freq, nrow = 4, byrow = TRUE)
re <- c(1, 2, 3, 1, 2, 3)
ge <- c(1, 1, 1, 0, 0, 0)

## With proportional odds
suppressWarnings(fit_vgam_p <- vglm(cbind(ff[1, ], ff[2, ], ff[3, ], ff[4, ]) ~ re + ge, family = acat(reverse = TRUE, parallel = TRUE)))


fit_bracl_p <- bracl(research ~ as.numeric(religion) + gender, weights = frequency, data = stemcell, type = "ML", parallel = TRUE)


fit_vgam <- vglm(cbind(ff[1, ], ff[2, ], ff[3, ], ff[4, ]) ~ re + ge, family = acat(reverse = TRUE, parallel = FALSE))
## Without proportional odds
fit_bracl <- bracl(research ~ as.numeric(religion) + gender, weights = frequency, data = stemcell, type = "ML")


tol <- 1e-06
## "VGAM::vglm and bracl return the same coefficients
expect_equal(unname(coef(fit_vgam)), unname(coef(fit_bracl)), tolerance = tol)
expect_equal(unname(coef(fit_vgam_p)), unname(coef(fit_bracl_p)), tolerance = tol)

## Difference in deviance is the same with VGAM::vglm and bracl
expect_equal(logLik(fit_vgam) - logLik(fit_vgam_p),
             unclass(logLik(fit_bracl) - logLik(fit_bracl_p)), tolerance = tol,
             check.attributes = FALSE)


## logLik returns the correct df for cracl
expect_identical(attr(logLik(fit_bracl), "df"), as.integer(9))
expect_identical(attr(logLik(fit_bracl_p), "df"), as.integer(5))

## bracl returns the correct fitted values
expect_equal(fitted(fit_bracl)[1:6, ], fit_vgam@fitted.values, tolerance = tol,
             check.attributes = FALSE)
expect_equal(fitted(fit_bracl_p)[1:6, ], fit_vgam_p@fitted.values, tolerance = tol,
             check.attributes = FALSE)

shu <- function(dat)  dat[sample(seq.int(nrow(dat)), nrow(dat)), ]

## bracl results is invariance to shuffling of the data
for (j in 1:10) {
    fit_bracl_p_r <- bracl(research ~ as.numeric(religion) + gender, weights = frequency, data = shu(stemcell), type = "ML", parallel = TRUE)

     fit_bracl_r <- bracl(research ~ as.numeric(religion) + gender, weights = frequency, data = shu(stemcell), type = "ML", parallel = FALSE)

    expect_equal(coef(fit_bracl), coef(fit_bracl_r), tolerance = tol)
    expect_equal(coef(fit_bracl_p), coef(fit_bracl_p_r), tolerance = tol)
}

tol  <-  1e-02
## vcov method for bracl returns the correct vcov matrix
expect_equal(vcov(fit_vgam_p), vcov(fit_bracl_p), tolerance = tol, check.attributes = FALSE)
expect_equal(vcov(fit_vgam), vcov(fit_bracl), tolerance = tol, check.attributes = FALSE)

s1 <- summary(fit_bracl)
s1p <- summary(fit_bracl_p)
s2 <- summary(fit_vgam)
s2p <- summary(fit_vgam_p)

tol <- 1e-06
## summary method for bracl returns the correct coef mat
expect_equal(coef(s1), s2@coef3, tolerance = tol, check.attributes = FALSE)
expect_equal(coef(s1p), s2p@coef3, tolerance = tol, check.attributes = FALSE)

newdata <- expand.grid(gender = c("male", "female"),  religion = c("moderate", "fundamentalist"))
## predict.bracl works as expected
pp <- predict(fit_bracl_p, newdata = stemcell, type = "probs")
p <- predict(fit_bracl, newdata = stemcell, type = "probs")
expect_equal(predict(fit_vgam_p, type = "response"),
             pp[19:24, ],
             tolerance = 1e-06, check.attributes = FALSE)
expect_equal(predict(fit_vgam, type = "response"),
             p[19:24, ],
             tolerance = 1e-06, check.attributes = FALSE)

## no intercept returns warning
expect_warning(
    fit_bracl_p_r <- bracl(research ~ -1 + as.numeric(religion) + gender, weights = frequency, data = shu(stemcell), type = "ML", parallel = FALSE)
)

## prediction with NAs works
newd <- newdata
newd[3, 2] <- NA
expect_true(is.na(predict(fit_bracl_p, newd, "class")[3]))
expect_true(all(is.na(predict(fit_bracl_p, newd, "probs")[3, ])))



## simulate from bracl objects returns a data frame with expected characteristics
simu_df <- simulate(fit_bracl_p)
nam_mf <- names(model.frame(fit_bracl_p))
nam_simu <- names(simu_df)
expect_identical(nrow(simu_df),
                 nrow(stemcell) * nlevels(stemcell$research))
expect_identical(levels(simu_df$research),
                 levels(stemcell$research))
expect_identical(is.ordered(simu_df$research),
                 is.ordered(stemcell$research))
expect_identical(nam_mf[!(nam_mf %in% nam_simu)],
                 "(weights)")
expect_identical(nam_simu[!(nam_simu %in% nam_mf)],
                 as.character(fit_bracl_p$call$weights))
expect_identical(nam_simu[(nam_simu %in% nam_mf)],
                 nam_mf[(nam_mf %in% nam_simu)])

## simulate from bracl objects returns a data frame with expected characteristics
simu_df <- simulate(fit_bracl)
nam_mf <- names(model.frame(fit_bracl))
nam_simu <- names(simu_df)
expect_identical(nrow(simu_df),
                 nrow(stemcell) * nlevels(stemcell$research))
expect_identical(levels(simu_df$research),
                 levels(stemcell$research))
expect_identical(is.ordered(simu_df$research),
                 is.ordered(stemcell$research))
expect_identical(nam_mf[!(nam_mf %in% nam_simu)],
                 "(weights)")
expect_identical(nam_simu[!(nam_simu %in% nam_mf)],
                 as.character(fit_bracl$call$weights))
expect_identical(nam_simu[(nam_simu %in% nam_mf)],
                 nam_mf[(nam_mf %in% nam_simu)])

## source(system.file("inst", "brglm0/brglm0.R", package = "brglm2"))
data("lizards", package = "brglm2")


types <- list("ML" = "(maximum likelihood)",
               "correction" = "(bias correction)",
               "AS_mean" = "(mean bias-reducing adjusted score equations)",
               "AS_median" = "(median bias-reducing adjusted score equations)",
               "AS_mixed" = "(mixed bias-reducing adjusted score equations)",
               "MPL_Jeffreys" = "(maximum Jeffreys' prior penalized likelihood)")


for (type in names(types)) {
    liz <- glm(cbind(grahami, opalinus) ~ height + diameter + light + time, family = binomial(),
               data = lizards,
               method = "brglmFit", type = type)
    expect_stdout(print(liz), "Coefficients")
    expect_stdout(print(liz), "Degrees of Freedom")
    expect_stdout(print(liz), "Null Deviance")
    expect_equal(liz$type, type)
    summ <- summary(liz)
    expect_true(all(class(summ) %in% c("summary.brglmFit", "summary.glm")))
    expect_stdout(print(summ), "Type of estimator:")
    expect_stdout(print(summ), type)
    expect_stdout(print(summ), types[[type]])
}


## brnb
salmonella <- data.frame(freq = c(15, 16, 16, 27, 33, 20,
                                   21, 18, 26, 41, 38, 27,
                                   29, 21, 33, 60, 41, 42),
                         dose = rep(c(0, 10, 33, 100, 333, 1000), 3),
                         observation = rep(1:3, each = 6))
salmonella_fm <- freq ~ dose + log(dose + 10)
fit_brnb <- brnb(salmonella_fm, data = salmonella,
                 link = "log", transformation = "inverse", type = "ML")
summ <- summary(fit_brnb)
expect_stdout(print(summ), "Type of estimator:")
expect_stdout(print(summ), "ML")
expect_stdout(print(summ), "(maximum likelihood)")

## brmultinom
data("housing", package = "MASS")
fit_brmultinom <- brmultinom(Sat ~ Infl + Type + Cont, weights = Freq,
                             data = housing, type = "ML", ref = 1)
summ <- summary(fit_brmultinom)
expect_stdout(print(summ), "Type of estimator:")
expect_stdout(print(summ), "ML")
expect_stdout(print(summ), "(maximum likelihood)")


## bracl
data("stemcell", package = "brglm2")
fit_bracl <- bracl(research ~ as.numeric(religion) + gender, weights = frequency,
                    data = stemcell, type = "ML")
summ <- summary(fit_bracl)
expect_stdout(print(summ), "Type of estimator:")
expect_stdout(print(summ), "ML")
expect_stdout(print(summ), "(maximum likelihood)")

# Copyright (C) 2016- Ioannis Kosmidis

#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 or 3 of the License
#  (at your option).
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  A copy of the GNU General Public License is available at
#  http://www.r-project.org/Licenses/


#' Auxiliary function for [glm()] fitting using the [brglmFit()]
#' method.
#'
#' Typically only used internally by [brglmFit()], but may be used to
#' construct a `control` argument.
#'
#' @aliases brglm_control
#' @param epsilon positive convergence tolerance epsilon. Default is
#'     `1e-06`.
#' @param check_aliasing logical indicating where a QR decomposition
#'     of the model matrix should be used to check for
#'     aliasing. Default is `TRUE`. See Details.
#' @param maxit integer giving the maximal number of iterations
#'     allowed. Default is `100`.
#' @param trace logical indicating if output should be produced for
#'     each iteration. Default is `FALSE`.
#' @param type the type of fitting method to be used. The options are
#'     `"AS_mean"` (mean-bias reducing adjusted scores), `"AS_median"`
#'     (median-bias reducing adjusted scores), `"AS_mixed"` (bias
#'     reduction using mixed score adjustments; default),
#'     `"correction"` (asymptotic bias correction), `"MPL_Jeffreys"`
#'     (maximum penalized likelihood with powers of the Jeffreys prior
#'     as penalty) and `"ML"` (maximum likelihood).
#' @param transformation the transformation of the dispersion to be
#'     estimated. Default is `"identity"`. See Details.
#' @param slowit a positive real used as a multiplier for the
#'     stepsize. The smaller it is the smaller the steps are. Default
#'     is `1`.
#' @param max_step_factor the maximum number of step halving steps to
#'     consider. Default is `12`.
#' @param response_adjustment a (small) positive constant or a vector
#'     of such. Default is `NULL`. See Details.
#' @param a power of the Jeffreys prior penalty. See Details.
#' @param ... further arguments passed to [brglmControl()]. Currently
#'     ignored in the output.
#'
#' @details
#'
#' [brglmControl()] provides default values and sanity checking for
#' the various constants that control the iteration and generally the
#' behaviour of [brglmFit()].
#'
#' When `trace = TRUE`, calls to [cat()] produce the output for each
#' iteration.  Hence, `options(digits = *)` can be used to increase
#' the precision.
#'
#' When `check_aliasing = TRUE` (default), a QR decomposition of the
#' model matrix is computed to check for aliasing. If the model matrix
#' is known to be of full rank, then `check_aliasing = FALSE` avoids
#' the extra computational overhead of an additional QR decomposition,
#' which can be substantial for large model matrices. However, setting
#' `check_aliasing = FALSE` tells [brglmFit()] that the model matrix
#' is full rank, and hard to trace back errors will result if it is
#' rank deficient.
#'
#' `transformation` sets the transformation of the dispersion
#' parameter for which the bias reduced estimates are computed. Can be
#' one of `"identity"`, `"sqrt"`, `"inverse"`, `"log"` and
#' `"inverseSqrt"`. Custom transformations are accommodated by
#' supplying a list of two expressions (transformation and inverse
#' transformation). See the examples for more details.
#'
#' The value of `response_adjustment` is only relevant if [brglmFit()]
#' is called with `start = NULL`, and `family` is [binomial()] or
#' [poisson()]. For those models, an initial maximum likelihood fit is
#' obtained on adjusted data to provide starting values for the
#' iteration in [brglmFit()]. The value of `response_adjustment`
#' governs how the data is adjusted. Specifically, if `family` is
#' [binomial()], then the responses and totals are adjusted by
#' `response_adjustment` and `2 * response_adjustment`, respectively;
#' if `family` is [poisson()], then the responses are adjusted by and
#' `response_adjustment`. `response_adjustment = NULL` (default)
#' is equivalent to setting it to
#' `"number of parameters" / "number of observations"`.
#'
#' When `type = "AS_mixed"` (default), mean bias reduction is used for
#' the regression parameters, and median bias reduction for the
#' dispersion parameter, if that is not fixed. This adjustment has
#' been developed based on equivariance arguments (see, Kosmidis et
#' al, 2020, Section 4) in order to produce regression parameter
#' estimates that are invariant to arbitrary contrasts, and estimates
#' for the dispersion parameter that are invariant to arbitrary
#' non-linear transformations. `type = "AS_mixed"` and `type =
#' "AS_mean"` return the same results if [brglmFit()] is called with
#' `family` [binomial()] or [poisson()] (i.e. families with fixed
#' dispersion).
#'
#' When `type = "MPL_Jeffreys"`, [brglmFit()] will maximize the
#' penalized log-likelihood \deqn{l(\beta, \phi) + a\log \det i(\beta,
#' \phi)}{l(beta, phi) + a log det i(beta, phi)} where \eqn{i(\beta,
#' \phi)}{i(beta, phi)} is the expected information matrix about the
#' regression parameters \eqn{\beta} and the dispersion parameter
#' \eqn{\phi}. See, `vignette("iteration", "brglm2")` for more
#' information. The argument `a` controls the amount of penalization
#' and its default value is `a = 1/2`, corresponding to maximum
#' penalized likelihood using a Jeffreys-prior penalty. See, Kosmidis
#' & Firth (2021) for proofs and discussion about the finiteness and
#' shrinkage properties of the maximum penalized likelihood estimators
#' for binomial-response generalized linear models.
#'
#' The estimates from `type = "AS_mean"` and `type =
#' "MPL_Jeffreys"` with `a = 1/2` (default) are identical for
#' Poisson log-linear models and logistic regression models, i.e. for
#' binomial and Poisson regression models with canonical links. See,
#' Firth (1993) for details.
#'
#' [brglm_control()] is an alias to [brglmControl()].
#'
#' @return a list with components named as the arguments, including
#'     symbolic expressions for the dispersion transformation
#'     (`Trans`) and its inverse (`inverseTrans`)
#'
#' @author Ioannis Kosmidis `[aut, cre]` \email{ioannis.kosmidis@warwick.ac.uk}
#'
#' @seealso [brglm_fit()] and [glm.fit()]
#'
#' @references
#'
#' Kosmidis I, Firth D (2021). Jeffreys-prior penalty, finiteness
#' and shrinkage in binomial-response generalized linear
#' models. *Biometrika*, **108**, 71-82. \doi{10.1093/biomet/asaa052}.
#'
#' Kosmidis I, Kenne Pagui E C, Sartori N (2020). Mean and median bias
#' reduction in generalized linear models. *Statistics and Computing*,
#' **30**, 43-59. \doi{10.1007/s11222-019-09860-6}.
#'
#' Firth D (1993). Bias reduction of maximum likelihood estimates.
#' Biometrika, **80**, 27-38. \doi{10.2307/2336755}.
#'
#' @examples
#'
#' data("coalition", package = "brglm2")
#' ## The maximum likelihood fit with log link
#' coalitionML <- glm(duration ~ fract + numst2, family = Gamma, data = coalition)
#'
#' ## Bias reduced estimation of the dispersion parameter
#' coalitionBRi <- glm(duration ~ fract + numst2, family = Gamma, data = coalition,
#'                     method = "brglmFit")
#' coef(coalitionBRi, model = "dispersion")
#'
#' ## Bias reduced estimation of log(dispersion)
#' coalitionBRl <- glm(duration ~ fract + numst2, family = Gamma, data = coalition,
#'                     method = "brglmFit", transformation = "log")
#' coef(coalitionBRl, model = "dispersion")
#'
#' ## Just for illustration: Bias reduced estimation of dispersion^0.25
#' my_transformation <- list(expression(dispersion^0.25), expression(transformed_dispersion^4))
#' coalitionBRc <- update(coalitionBRi, transformation = my_transformation)
#' coef(coalitionBRc, model = "dispersion")
#'
#' @export
brglmControl <- function(epsilon = 1e-06, maxit = 100,
                         check_aliasing = TRUE,
                         trace = FALSE,
                         type = c("AS_mixed", "AS_mean", "AS_median", "correction", "MPL_Jeffreys", "ML"),
                         transformation = "identity",
                         slowit = 1,
                         response_adjustment = NULL,
                         max_step_factor = 12,
                         a = 1/2, ...) {
    type <- match.arg(type)

    if (is.character(transformation)) {
        Trans <- switch(transformation,
                        identity = expression(dispersion),
                        sqrt = expression(dispersion^0.5),
                        inverse = expression(1/dispersion),
                        log = expression(log(dispersion)),
                        inverseSqrt = expression(1/sqrt(dispersion)),
                        stop(transformation, " is not one of the implemented dispersion transformations"))
        inverseTrans <- switch(transformation,
                               identity = expression(transformed_dispersion),
                               sqrt = expression(transformed_dispersion^2),
                               inverse = expression(1/transformed_dispersion),
                               log = expression(exp(transformed_dispersion)),
                               inverseSqrt = expression(1/transformed_dispersion^2))
    } else {
        if (is.list(transformation) && (length(transformation) == 2)) {
            Trans <- transformation[[1]]
            inverseTrans <- transformation[[2]]
            transformation <- "custom_transformation"
        } else {
            stop("transformation can be either one of 'identity', 'sqrt', 'inverse', 'log' and 'inverseSqrt', or a list of two expressions")
        }
    }
    if (!is.numeric(epsilon) || epsilon <= 0)
        stop("value of 'epsilon' must be > 0")

    if (!is.numeric(max_step_factor) || max_step_factor < 1) {
        warning("`max_step_factor = ", deparse(max_step_factor), "` is not a permissible value. Defaulting to 12")
        max_step_factor <- 12
    }
    list(epsilon = epsilon, maxit = maxit, trace = trace,
         check_aliasing = check_aliasing,
         response_adjustment = response_adjustment,
         type = type,
         Trans = Trans,
         inverseTrans = inverseTrans,
         transformation = transformation,
         slowit = slowit,
         max_step_factor = max_step_factor,
         a = a)
}

b_control <- brglmControl()

## the object brglmControl() returns is as expected
expect_identical(b_control$epsilon, 1e-06)
expect_identical(b_control$maxit, 100)
expect_false(b_control$trace)
expect_null(b_control$response_adjustment)
expect_equal(b_control$Trans, expression(dispersion))
expect_equal(b_control$inverseTrans, expression(transformed_dispersion))
expect_equal(b_control$transformation, "identity")
expect_equal(b_control$slowit, 1)
expect_equal(b_control$max_step_factor, 12)

b_control <- brglmControl(epsilon = 1e-02, trace = TRUE, response_adjustment = c(0.3, 0.2))
## the object brglmControl returns with defaults is as expected
expect_identical(b_control$epsilon, 1e-02)
expect_identical(b_control$maxit, 100)
expect_true(b_control$trace)
expect_identical(b_control$response_adjustment, c(0.3, 0.2))
expect_equal(b_control$Trans, expression(dispersion))
expect_equal(b_control$inverseTrans, expression(transformed_dispersion))
expect_equal(b_control$transformation, "identity")
expect_equal(b_control$slowit, 1)
expect_equal(b_control$max_step_factor, 12)

data("coalition", package = "brglm2")
mod <- glm(duration ~ fract + numst2, family = Gamma, data = coalition, method = "brglmFit")
expect_warning(mod <- update(mod, maxit = 7, transformation = "sqrt"), "brglmFit: algorithm did not converge")
expect_equal(mod$iter, 7)
expect_equal(mod$transformation, "sqrt")

b_control <- brglmControl(epsilon = 1e-02, ABCDEFG123 = 1, response_adjustment = c(0.3, 0.2), trace = TRUE, )
## the object brglmControl returns with defaults is as expected
expect_identical(b_control$epsilon, 1e-02)
expect_identical(b_control$maxit, 100)
expect_true(b_control$trace)
expect_identical(b_control$response_adjustment, c(0.3, 0.2))
expect_equal(b_control$Trans, expression(dispersion))
expect_equal(b_control$inverseTrans, expression(transformed_dispersion))
expect_equal(b_control$transformation, "identity")
expect_equal(b_control$slowit, 1)
expect_equal(b_control$max_step_factor, 12)

## A Gamma example, from McCullagh & Nelder (1989, pp. 300-2)
clotting <- data.frame(
    u = c(5,10,15,20,30,40,60,80,100, 5,10,15,20,30,40,60,80,100),
    conc = c(118,58,42,35,27,25,21,19,18,69,35,26,21,18,16,13,12,12),
    lot = factor(c(rep(1, 9), rep(2, 9))))

clot_formula <- conc ~ lot*log(u) + I(2*log(u))

## check_aliasing option works as expected
expect_error(glm(clot_formula, data = clotting, family = Gamma(), method = brglm_fit, check_aliasing = FALSE),
             pattern = "NA/NaN/Inf in foreign")
mod <- glm(clot_formula, data = clotting, family = Gamma(), method = brglm_fit, check_aliasing = TRUE)
expect_true(is.na(coef(mod)["I(2 * log(u))"]))
## check defaults
mod <- glm(clot_formula, data = clotting, family = Gamma(), method = brglm_fit)
expect_true(is.na(coef(mod)["I(2 * log(u))"]))

## data("sex2", package = "logistf")
## ## brglmControl arguments can be passed directly from the brglmFit call
## expect_warning(glm(case ~ dia, data = sex2, family = binomial(), method = brglm_fit))
## ## Set response_adjustment to 2 just to avoid non-integer successes warning
## m1 <- glm(case ~ dia, data = sex2, family = binomial(), method = brglm_fit, response_adjustment = 2)
## expect_equal(m1$iter, 7, tolerance = .Machine$double.eps/2)

## Essentially stats:::add1.glm() with `brglmFit()` calls in place of `glm.fit()` calls
#' @export
add1.brglmFit <- function(object, scope, scale = 0,
                          test = c("none", "Rao", "LRT", "Chisq", "F"), x = NULL, k = 2, ...)  {

    Fstat <- function(table, rdf) {
        dev <- table$Deviance
        df <- table$Df
        diff <- pmax(0, (dev[1L] - dev)/df)
        Fs <- diff/(dev/(rdf - df))
        Fs[df < .Machine$double.eps] <- NA
        P <- Fs
        nnas <- !is.na(Fs)
        P[nnas] <- safe_pf(Fs[nnas], df[nnas], rdf - df[nnas],
                           lower.tail = FALSE)
        list(Fs = Fs, P = P)
    }
    test <- match.arg(test)
    if (test == "Chisq")
        test <- "LRT"
    if (!is.character(scope))
        scope <- add.scope(object, update.formula(object, scope))
    if (!length(scope))
        stop("no terms in scope for adding to object")
    oTerms <- attr(object$terms, "term.labels")
    int <- attr(object$terms, "intercept")
    ns <- length(scope)
    dfs <- numeric(ns + 1L)
    names(dfs) <- c("<none>", scope)
    dev <- score <- dfs
    add.rhs <- eval(str2lang(paste("~ . +", paste(scope, collapse = "+"))))
    new.form <- update.formula(object, add.rhs)
    Terms <- terms(new.form)
    y <- object$y
    if (is.null(x)) {
        fc <- object$call
        fc$formula <- Terms
        fob <- list(call = fc, terms = Terms)
        class(fob) <- oldClass(object)
        m <- model.frame(fob, xlev = object$xlevels)
        offset <- model.offset(m)
        wt <- model.weights(m)
        x <- model.matrix(Terms, m, contrasts.arg = object$contrasts)
        oldn <- length(y)
        y <- model.response(m)
        if (!is.factor(y))
            storage.mode(y) <- "double"
        if (NCOL(y) == 2) {
            n <- y[, 1] + y[, 2]
            y <- ifelse(n == 0, 0, y[, 1]/n)
            wt <- (wt %||% rep.int(1, length(y))) * n
        }
        newn <- length(y)
        if (newn < oldn)
            warning(sprintf(ngettext(newn, "using the %d/%d row from a combined fit",
                                     "using the %d/%d rows from a combined fit"),
                            newn, oldn), domain = NA)
    }
    else {
        wt <- object$prior.weights
        offset <- object$offset
    }
    n <- nrow(x)
    if (is.null(wt))
        wt <- rep.int(1, n)
    Terms <- attr(Terms, "term.labels")
    asgn <- attr(x, "assign")
    ousex <- match(asgn, match(oTerms, Terms), 0L) > 0L
    if (int)
        ousex[1L] <- TRUE
    X <- x[, ousex, drop = FALSE]
    z <- brglmFit(X, y, wt, offset = offset, family = object$family,
                  control = object$control)
    dfs[1L] <- z$rank
    dev[1L] <- z$deviance
    r <- z$residuals
    w <- z$weights
    sTerms <- sapply(strsplit(Terms, ":", fixed = TRUE), function(x) paste(sort(x),
                                                                           collapse = ":"))
    for (tt in scope) {
        stt <- paste(sort(strsplit(tt, ":")[[1L]]), collapse = ":")
        usex <- match(asgn, match(stt, sTerms), 0L) > 0L
        X <- x[, usex | ousex, drop = FALSE]
        z <- brglmFit(X, y, wt, offset = offset, family = object$family,
                      control = object$control)
        dfs[tt] <- z$rank
        dev[tt] <- z$deviance
        if (test == "Rao") {
            zz <- glm.fit(X, r, w, offset = offset)
            score[tt] <- zz$null.deviance - zz$deviance
        }
    }
    dispersion <- if (scale == 0)
                      summary(object, dispersion = NULL)$dispersion
    else scale
    fam <- object$family$family
    loglik <- if (fam == "gaussian") {
                  if (scale > 0)
                      dev/scale - n
                  else n * log(dev/n)
              }
              else dev/dispersion
    aic <- loglik + k * dfs
    aic <- aic + (extractAIC(object, k = k)[2L] - aic[1L])
    dfs <- dfs - dfs[1L]
    dfs[1L] <- NA
    aod <- data.frame(Df = dfs, Deviance = dev, AIC = aic, row.names = names(dfs),
                      check.names = FALSE)
    if (all(is.na(aic)))
        aod <- aod[, -3]
    test <- match.arg(test)
    if (test == "LRT") {
        dev <- pmax(0, loglik[1L] - loglik)
        dev[1L] <- NA
        LRT <- if (dispersion == 1)
                   "LRT"
               else "scaled dev."
        aod[, LRT] <- dev
        nas <- !is.na(dev)
        dev[nas] <- safe_pchisq(dev[nas], aod$Df[nas], lower.tail = FALSE)
        aod[, "Pr(>Chi)"] <- dev
    }
    else if (test == "Rao") {
        dev <- pmax(0, score)
        dev[1L] <- NA
        nas <- !is.na(dev)
        SC <- if (dispersion == 1)
                  "Rao score"
              else "scaled Rao sc."
        dev <- dev/dispersion
        aod[, SC] <- dev
        dev[nas] <- safe_pchisq(dev[nas], aod$Df[nas], lower.tail = FALSE)
        aod[, "Pr(>Chi)"] <- dev
    }
    else if (test == "F") {
        if (fam == "binomial" || fam == "poisson")
            warning(gettextf("F test assumes quasi%s family",
                             fam), domain = NA)
        rdf <- object$df.residual
        aod[, c("F value", "Pr(>F)")] <- Fstat(aod, rdf)
    }
    head <- c("Single term additions", "\nModel:", deparse(formula(object)),
              if (scale > 0) paste("\nscale: ", format(scale), "\n"))
    class(aod) <- c("anova", "data.frame")
    attr(aod, "heading") <- head
    aod
}

#' @export
add1.mdyplFit <- function(object, scope, scale = 0,
                          test = c("none", "Rao", "LRT", "Chisq", "F"), x = NULL, k = 2, ...)  {

    Fstat <- function(table, rdf) {
        dev <- table$Deviance
        df <- table$Df
        diff <- pmax(0, (dev[1L] - dev)/df)
        Fs <- diff/(dev/(rdf - df))
        Fs[df < .Machine$double.eps] <- NA
        P <- Fs
        nnas <- !is.na(Fs)
        P[nnas] <- safe_pf(Fs[nnas], df[nnas], rdf - df[nnas],
                           lower.tail = FALSE)
        list(Fs = Fs, P = P)
    }
    test <- match.arg(test)
    if (test == "Chisq")
        test <- "LRT"
    if (!is.character(scope))
        scope <- add.scope(object, update.formula(object, scope))
    if (!length(scope))
        stop("no terms in scope for adding to object")
    oTerms <- attr(object$terms, "term.labels")
    int <- attr(object$terms, "intercept")
    ns <- length(scope)
    dfs <- numeric(ns + 1L)
    names(dfs) <- c("<none>", scope)
    dev <- score <- dfs
    add.rhs <- eval(str2lang(paste("~ . +", paste(scope, collapse = "+"))))
    new.form <- update.formula(object, add.rhs)
    Terms <- terms(new.form)
    y <- object$y
    if (is.null(x)) {
        fc <- object$call
        fc$formula <- Terms
        fob <- list(call = fc, terms = Terms)
        class(fob) <- oldClass(object)
        m <- model.frame(fob, xlev = object$xlevels)
        offset <- model.offset(m)
        wt <- model.weights(m)
        x <- model.matrix(Terms, m, contrasts.arg = object$contrasts)
        oldn <- length(y)
        y <- model.response(m)
        if (!is.factor(y))
            storage.mode(y) <- "double"
        if (NCOL(y) == 2) {
            n <- y[, 1] + y[, 2]
            y <- ifelse(n == 0, 0, y[, 1]/n)
            wt <- (wt %||% rep.int(1, length(y))) * n
        }
        newn <- length(y)
        if (newn < oldn)
            warning(sprintf(ngettext(newn, "using the %d/%d row from a combined fit",
                                     "using the %d/%d rows from a combined fit"),
                            newn, oldn), domain = NA)
    }
    else {
        wt <- object$prior.weights
        offset <- object$offset
    }
    n <- nrow(x)
    if (is.null(wt))
        wt <- rep.int(1, n)
    Terms <- attr(Terms, "term.labels")
    asgn <- attr(x, "assign")
    ousex <- match(asgn, match(oTerms, Terms), 0L) > 0L
    if (int)
        ousex[1L] <- TRUE
    X <- x[, ousex, drop = FALSE]
    z <- mdyplFit(X, y, wt, offset = offset, family = object$family,
                  control = object$control)
    dfs[1L] <- z$rank
    dev[1L] <- z$deviance
    r <- z$residuals
    w <- z$weights
    sTerms <- sapply(strsplit(Terms, ":", fixed = TRUE), function(x) paste(sort(x),
                                                                           collapse = ":"))
    for (tt in scope) {
        stt <- paste(sort(strsplit(tt, ":")[[1L]]), collapse = ":")
        usex <- match(asgn, match(stt, sTerms), 0L) > 0L
        X <- x[, usex | ousex, drop = FALSE]
        z <- mdyplFit(X, y, wt, offset = offset, family = object$family,
                      control = object$control)
        dfs[tt] <- z$rank
        dev[tt] <- z$deviance
        if (test == "Rao") {
            zz <- glm.fit(X, r, w, offset = offset)
            score[tt] <- zz$null.deviance - zz$deviance
        }
    }
    dispersion <- if (scale == 0)
                      summary(object, dispersion = NULL)$dispersion
    else scale
    fam <- object$family$family
    loglik <- if (fam == "gaussian") {
                  if (scale > 0)
                      dev/scale - n
                  else n * log(dev/n)
              }
              else dev/dispersion
    aic <- loglik + k * dfs
    aic <- aic + (extractAIC(object, k = k)[2L] - aic[1L])
    dfs <- dfs - dfs[1L]
    dfs[1L] <- NA
    aod <- data.frame(Df = dfs, Deviance = dev, AIC = aic, row.names = names(dfs),
                      check.names = FALSE)
    if (all(is.na(aic)))
        aod <- aod[, -3]
    test <- match.arg(test)
    if (test == "LRT") {
        dev <- pmax(0, loglik[1L] - loglik)
        dev[1L] <- NA
        LRT <- if (dispersion == 1)
                   "LRT"
               else "scaled dev."
        aod[, LRT] <- dev
        nas <- !is.na(dev)
        dev[nas] <- safe_pchisq(dev[nas], aod$Df[nas], lower.tail = FALSE)
        aod[, "Pr(>Chi)"] <- dev
    }
    else if (test == "Rao") {
        dev <- pmax(0, score)
        dev[1L] <- NA
        nas <- !is.na(dev)
        SC <- if (dispersion == 1)
                  "Rao score"
              else "scaled Rao sc."
        dev <- dev/dispersion
        aod[, SC] <- dev
        dev[nas] <- safe_pchisq(dev[nas], aod$Df[nas], lower.tail = FALSE)
        aod[, "Pr(>Chi)"] <- dev
    }
    else if (test == "F") {
        if (fam == "binomial" || fam == "poisson")
            warning(gettextf("F test assumes quasi%s family",
                             fam), domain = NA)
        rdf <- object$df.residual
        aod[, c("F value", "Pr(>F)")] <- Fstat(aod, rdf)
    }
    head <- c("Single term additions", "\nModel:", deparse(formula(object)),
              if (scale > 0) paste("\nscale: ", format(scale), "\n"))
    class(aod) <- c("anova", "data.frame")
    attr(aod, "heading") <- head
    aod
}


safe_pf <- utils::getFromNamespace("safe_pf", "stats")
safe_pchisq <- utils::getFromNamespace("safe_pchisq", "stats")

# Copyright (C) 2020- Ioannis Kosmidis

#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 or 3 of the License
#  (at your option).
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  A copy of the GNU General Public License is available at
#  http://www.r-project.org/Licenses/


function_moves_to_new_package <- function(function_name,
                                          removal_version,
                                          current_pkg,
                                          new_pkg,
                                          extra_message = NULL) {
    function_name <- paste0("'", function_name, "'")
    current_pkg <- paste0("'", current_pkg, "'")
    new_pkg <- paste0("'", new_pkg, "'")
    msg <- paste(function_name,
                 "will be removed from",
                 current_pkg,
                 "at version",
                 paste0(removal_version, "."),
                 "A new version of",
                 function_name,
                 "is now maintained in the",
                 new_pkg,
                 "package.",
                 extra_message)
    .Deprecated(msg = msg, package = current_pkg)
}

function_moved_to_new_package <- function(function_name,
                                          removal_version,
                                          current_pkg,
                                          new_pkg,
                                          extra_message = NULL) {
    function_name <- paste0("'", function_name, "'")
    current_pkg <- paste0("'", current_pkg, "'")
    new_pkg <- paste0("'", new_pkg, "'")
    msg <- paste(function_name,
                 "has been removed from",
                 current_pkg,
                 "at version",
                 paste0(removal_version, "."),
                 "A new version of",
                 function_name,
                 "is now maintained in the",
                 new_pkg,
                 "package.",
                 extra_message)
    .Defunct(msg = msg, package = current_pkg)
}

# Copyright (C) 2016- Ioannis Kosmidis

#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 or 3 of the License
#  (at your option).
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  A copy of the GNU General Public License is available at
#  http://www.r-project.org/Licenses/


#' Fitting function for [glm()] for reduced-bias estimation and
#' inference
#'
#' [brglmFit()] is a fitting method for [glm()] that fits generalized
#' linear models using implicit and explicit bias reduction methods
#' (Kosmidis, 2014), and other penalized maximum likelihood
#' methods. Currently supported methods include the mean bias-reducing
#' adjusted scores approach in Firth (1993) and Kosmidis & Firth
#' (2009), the median bias-reduction adjusted scores approach in Kenne
#' Pagui et al. (2017), the correction of the asymptotic bias in
#' Cordeiro & McCullagh (1991), the mixed bias-reduction adjusted
#' scores approach in Kosmidis et al (2020), maximum penalized
#' likelihood with powers of the Jeffreys prior as penalty, and
#' maximum likelihood. Estimation is performed using a quasi Fisher
#' scoring iteration (see `vignette("iteration", "brglm2")`, which, in
#' the case of mean-bias reduction, resembles an iterative correction
#' of the asymptotic bias of the Fisher scoring iterates.
#'
#' @inheritParams stats::glm.fit
#' @aliases brglm_fit
#' @param x a design matrix of dimension `n * p`.
#' @param y a vector of observations of length `n`.
#' @param control a list of parameters controlling the fitting
#'     process. See [brglmControl()] for details.
#' @param start starting values for the parameters in the linear
#'     predictor. If `NULL` (default) then the maximum likelihood
#'     estimates are calculated and used as starting values.
#' @param mustart applied only when start is not `NULL`. Starting
#'     values for the vector of means to be passed to
#'     [glm.fit()] when computing starting values using maximum
#'     likelihood.
#' @param etastart applied only when start is not `NULL`. Starting
#'     values for the linear predictor to be passed to
#'     [glm.fit()] when computing starting values using maximum
#'     likelihood.
#' @param fixed_totals effective only when `family` is
#'     [poisson()]. Either `NULL` (no effect) or a vector that
#'     indicates which counts must be treated as a group. See Details
#'     for more information and [brmultinom()].
#' @param singular.ok logical. If `FALSE`, a singular model is an
#'     error.
#'
#' @details
#'
#' A detailed description of the supported adjustments and the quasi
#' Fisher scoring iteration is given in the iteration vignette (see,
#' `vignette("iteration", "brglm2")` or Kosmidis et al, 2020).  A
#' shorter description of the quasi Fisher scoring iteration is also
#' given in one of the vignettes of the *enrichwith* R package (see,
#' \url{https://cran.r-project.org/package=enrichwith/vignettes/bias.html}).
#' Kosmidis and Firth (2010) describe a parallel quasi Newton-Raphson
#' iteration with the same stationary point.
#'
#' In the special case of generalized linear models for binomial,
#' Poisson and multinomial responses, the adjusted score equation
#' approaches for `type = "AS_mixed"`, `type = "AS_mean"`, and `type =
#' "AS_median"` (see below for what methods each `type` corresponds)
#' return estimates with improved frequentist properties, that are
#' also always finite, even in cases where the maximum likelihood
#' estimates are infinite (e.g. complete and quasi-complete separation
#' in multinomial regression). See, Kosmidis and Firth (2021) for a
#' proof for binomial-response GLMs with Jeffreys-prior penalties to
#' the log-likelihood, which is equivalent to mean bias reduction for
#' logistic regression. See, also,
#' [detectseparation::detect_separation()] and
#' [detectseparation::check_infinite_estimates()] for pre-fit and
#' post-fit methods for the detection of infinite estimates in
#' binomial response generalized linear models.
#'
#' The type of score adjustment to be used is specified through the
#' `type` argument (see [brglmControl()] for details). The available
#' options are
#'
#' * `type = "AS_mixed"`: the mixed bias-reducing score adjustments in
#' Kosmidis et al (2020) that result in mean bias reduction for the
#' regression parameters and median bias reduction for the dispersion
#' parameter, if any; default.
#'
#' * `type = "AS_mean"`: the mean bias-reducing score adjustments in
#' Firth, 1993 and Kosmidis & Firth, 2009. `type = "AS_mixed"` and
#' `type = "AS_mean"` will return the same results when `family` is
#' [binomial()] or [poisson()], i.e. when the dispersion is fixed
#'
#' * `type = "AS_median"`: the median bias-reducing score
#' adjustments in Kenne Pagui et al. (2017)
#'
#' * `type = "MPL_Jeffreys"`: maximum penalized likelihood
#' with powers of the Jeffreys prior as penalty.
#'
#' * `type = "ML"`: maximum likelihood.
#'
#' * `type = "correction"`: asymptotic bias correction, as in
#' Cordeiro & McCullagh (1991).
#'
#' The null deviance is evaluated based on the fitted values using the
#' method specified by the `type` argument (see [brglmControl()]).
#'
#' The `family` argument of the current version of [brglmFit()] can
#' accept any combination of [`"family"`][family] objects and link functions,
#' including families with user-specified link functions, [mis()]
#' links, and [power()] links, but excluding [quasi()],
#' [quasipoisson()] and [quasibinomial()] families.
#'
#' The description of `method` argument and the `Fitting functions`
#' section in [glm()] gives information on supplying fitting
#' methods to [glm()].
#'
#' `fixed_totals` specifies groups of observations for which the sum
#' of the means of a Poisson model will be held fixed to the observed
#' count for each group. This argument is used internally in
#' [brmultinom()] and [bracl()] for baseline-category logit models and
#' adjacent category logit models, respectively.
#'
#' [brglm_fit()] is an alias to [brglmFit()].
#'
#' @return
#'
#' An object inheriting from [`"brglmFit"`][brglmFit()] object, which
#' is a list having the same elements to the list that
#' [stats::glm.fit()] returns, with a few extra arguments.
#'
#' @author Ioannis Kosmidis `[aut, cre]` \email{ioannis.kosmidis@warwick.ac.uk}, Euloge Clovis Kenne Pagui `[ctb]` \email{kenne@stat.unipd.it}
#'
#' @seealso [brglmControl()], [glm.fit()], [glm()]
#'
#' @references
#'
#' Kosmidis I, Firth D (2021). Jeffreys-prior penalty, finiteness
#' and shrinkage in binomial-response generalized linear
#' models. *Biometrika*, **108**, 71-82. \doi{10.1093/biomet/asaa052}.
#'
#' Kosmidis I, Kenne Pagui E C, Sartori N (2020). Mean and median bias
#' reduction in generalized linear models. *Statistics and Computing*,
#' **30**, 43-59. \doi{10.1007/s11222-019-09860-6}.
#'
#' Cordeiro G M, McCullagh P (1991). Bias correction in generalized
#' linear models. *Journal of the Royal Statistical Society. Series B
#' (Methodological)*, **53**, 629-643. \doi{10.1111/j.2517-6161.1991.tb01852.x}.
#'
#' Firth D (1993). Bias reduction of maximum likelihood estimates.
#' *Biometrika*. **80**, 27-38. \doi{10.2307/2336755}.
#'
#' Kenne Pagui E C, Salvan A, Sartori N (2017). Median bias
#' reduction of maximum likelihood estimates. *Biometrika*, **104**,
#' 923–938. \doi{10.1093/biomet/asx046}.
#'
#' Kosmidis I, Firth D (2009). Bias reduction in exponential family
#' nonlinear models. *Biometrika*, **96**, 793-804. \doi{10.1093/biomet/asp055}.
#'
#' Kosmidis I, Firth D (2010). A generic algorithm for reducing
#' bias in parametric estimation. *Electronic Journal of Statistics*,
#' **4**, 1097-1112. \doi{10.1214/10-EJS579}.
#'
#' Kosmidis I (2014). Bias in parametric estimation: reduction and
#' useful side-effects. *WIRE Computational Statistics*, **6**,
#' 185-196. \doi{10.1002/wics.1296}.
#'
#' @examples
#' ## The lizards example from ?brglm::brglm
#' data("lizards", package = "brglm2")
#' # Fit the model using maximum likelihood
#' lizardsML <- glm(cbind(grahami, opalinus) ~ height + diameter +
#'                  light + time, family = binomial(logit), data = lizards,
#'                  method = "glm.fit")
#' # Mean bias-reduced fit:
#' lizardsBR_mean <- glm(cbind(grahami, opalinus) ~ height + diameter +
#'                       light + time, family = binomial(logit), data = lizards,
#'                       method = "brglmFit")
#' # Median bias-reduced fit:
#' lizardsBR_median <- glm(cbind(grahami, opalinus) ~ height + diameter +
#'                         light + time, family = binomial(logit), data = lizards,
#'                         method = "brglmFit", type = "AS_median")
#' summary(lizardsML)
#' summary(lizardsBR_median)
#' summary(lizardsBR_mean)
#'
#' # Maximum penalized likelihood with Jeffreys prior penatly
#' lizards_Jeffreys <- glm(cbind(grahami, opalinus) ~ height + diameter +
#'                         light + time, family = binomial(logit), data = lizards,
#'                         method = "brglmFit", type = "MPL_Jeffreys")
#' # lizards_Jeffreys is the same fit as lizardsBR_mean (see Firth, 1993)
#' all.equal(coef(lizardsBR_mean), coef(lizards_Jeffreys))
#'
#' # Maximum penalized likelihood with powers of the Jeffreys prior as
#' # penalty. See Kosmidis & Firth (2021) for the finiteness and
#' # shrinkage properties of the maximum penalized likelihood
#' # estimators in binomial response models
#' \donttest{
#' a <- seq(0, 20, 0.5)
#' coefs <- sapply(a, function(a) {
#'       out <- glm(cbind(grahami, opalinus) ~ height + diameter +
#'              light + time, family = binomial(logit), data = lizards,
#'              method = "brglmFit", type = "MPL_Jeffreys", a = a)
#'       coef(out)
#' })
#' # Illustration of shrinkage as a grows
#' matplot(a, t(coefs), type = "l", col = 1, lty = 1)
#' abline(0, 0, col = "grey")
#'}
#'
#' \donttest{
#' ## Another example from
#' ## King, Gary, James E. Alt, Nancy Elizabeth Burns and Michael Laver
#' ## (1990).  "A Unified Model of Cabinet Dissolution in Parliamentary
#' ## Democracies", _American Journal of Political Science_, **34**, 846-870
#'
#' data("coalition", package = "brglm2")
#' # The maximum likelihood fit with log link
#' coalitionML <- glm(duration ~ fract + numst2, family = Gamma, data = coalition)
#' # The mean bias-reduced fit
#' coalitionBR_mean <- update(coalitionML, method = "brglmFit")
#' # The bias-corrected fit
#' coalitionBC <- update(coalitionML, method = "brglmFit", type = "correction")
#' # The median bias-corrected fit
#' coalitionBR_median <- update(coalitionML, method = "brglmFit", type = "AS_median")
#' }
#'
#' \donttest{
#' ## An example with offsets from Venables & Ripley (2002, p.189)
#' data("anorexia", package = "MASS")
#'
#' anorexML <- glm(Postwt ~ Prewt + Treat + offset(Prewt),
#'                 family = gaussian, data = anorexia)
#' anorexBC <- update(anorexML, method = "brglmFit", type = "correction")
#' anorexBR_mean <- update(anorexML, method = "brglmFit")
#' anorexBR_median <- update(anorexML, method = "brglmFit", type = "AS_median")
#'
#' # All methods return the same estimates for the regression
#' # parameters because the maximum likelihood estimator is normally
#' # distributed around the `true` value under the model (hence, both
#' # mean and component-wise median unbiased). The Wald tests for
#' # anorexBC and anorexBR_mean differ from anorexML because the
#' # bias-reduced estimator of the dispersion is the unbiased, by
#' # degree of freedom adjustment (divide by n - p), estimator of the
#' # residual variance. The Wald tests from anorexBR_median are based
#' # on the median bias-reduced estimator of the dispersion that
#' # results from a different adjustment of the degrees of freedom
#' # (divide by n - p - 2/3)
#' summary(anorexML)
#' summary(anorexBC)
#' summary(anorexBR_mean)
#' summary(anorexBR_median)
#' }
#'
#' ## endometrial data from Heinze & Schemper (2002) (see ?endometrial)
#' data("endometrial", package = "brglm2")
#' endometrialML <- glm(HG ~ NV + PI + EH, data = endometrial,
#'                      family = binomial("probit"))
#' endometrialBR_mean <- update(endometrialML, method = "brglmFit",
#'                              type = "AS_mean")
#' endometrialBC <- update(endometrialML, method = "brglmFit",
#'                         type = "correction")
#' endometrialBR_median <- update(endometrialML, method = "brglmFit",
#'                                type = "AS_median")
#' summary(endometrialML)
#' summary(endometrialBC)
#' summary(endometrialBR_mean)
#' summary(endometrialBR_median)
#'
#' @export
brglmFit <- function(x, y, weights = rep(1, nobs), start = NULL, etastart = NULL,
                     mustart = NULL, offset = rep(0, nobs), family = gaussian(),
                     control = list(), intercept = TRUE,
                     ## Arguments that glm will not use in its call to brglmFit (be wise with defaults!)
                     fixed_totals = NULL, singular.ok = TRUE) {
    trace_iteration <- function() {
        if (iter %% control$trace == 0) {
            st <-  max(abs(step_beta), na.rm = TRUE)
            gr <- max(abs(adjusted_grad_beta), na.rm = TRUE)
            cat("Coefficients update:\t")
            cat("Outer/Inner iteration:\t", sprintf("%03d", iter), "/", sprintf("%03d", step_factor), "\n", sep = "")
            if (!no_dispersion) {
                st <- abs(step_zeta)
                gr <- abs(adjusted_grad_zeta)
                cat("Dispersion update:\t")
                cat("Outer iteration:\t", sprintf("%03d", iter), "\n")
            }
            cat("max |step|:", format(round(st, 6), nsmall = 6, scientific = FALSE), "\t",
                "max |gradient|:", format(round(gr, 6), nsmall = 6, scientific = FALSE), "\n")
        }
    }

    ## key_quantities, grad, info and bias are ALWAYS in beta, dispersion parameterization
    key_quantities <- function(pars, y, level = 0, scale_totals = FALSE, qr = TRUE) {
        betas <- pars[seq.int(nvars)]
        dispersion <- pars[nvars + 1]
        prec <- 1/dispersion
        etas <- drop(x %*% betas + offset)
        mus <- linkinv(etas)
        if (scale_totals) {
            ## Rescale mus
            mus_totals <-  as.vector(tapply(mus, fixed_totals, sum))[fixed_totals]
            mus <- mus * row_totals / mus_totals
            etas <- linkfun(mus)
        }
        out <- list(precision = prec,
                    betas = betas,
                    dispersion = dispersion,
                    etas = etas,
                    mus = mus,
                    scale_totals = scale_totals)
        mean_quantities <- function(out) {
            d1mus <- mu.eta(etas)
            d2mus <- d2mu.deta(etas)
            varmus <- variance(mus)
            working_weights <- weights * d1mus^2 / varmus
            wx <- sqrt(working_weights) * x
            out$d1mus <- d1mus
            out$d2mus <- d2mus
            out$varmus <- varmus
            out$d1varmus <- d1variance(mus)
            out$working_weights <- working_weights
            if (qr) out$qr_decomposition <- qr(wx)
            out
        }
        dispersion_quantities <- function(out) {
            zetas <- -weights * prec
            out$zetas <- zetas
            ## Evaluate the derivatives of the a function only for
            ## objervations with non-zero weight
            d1afuns <- d2afuns <- d3afuns <- rep(NA_real_, nobs)
            d1afuns[keep] <- d1afun(zetas[keep])
            ## because of the way dev.resids is implemented, this is
            ## d1afun is the expectation of dev.resids + 2 for gamma
            ## families, so subtract 2
            if (family$family == "Gamma") d1afuns <- d1afuns - 2
            d2afuns[keep] <- d2afun(zetas[keep])
            d3afuns[keep] <- d3afun(zetas[keep])
            out$d2afuns <- d2afuns
            out$d3afuns <- d3afuns
            out$deviance_residuals <- dev.resids(y, mus, weights)
            out$Edeviance_residuals <- weights * d1afuns
            out
        }
        if (level == 0) {
            out <- mean_quantities(out)
        }
        if (level == 1) {
            out <- dispersion_quantities(out)
        }
        if (level > 1) {
            out <- mean_quantities(out)
            out <- dispersion_quantities(out)
        }
        out
    }

    gradient <- function(pars, level = 0, fit = NULL) {
        if (is.null(fit)) {
            fit <- key_quantities(pars, y = y, level = level, qr = FALSE)
        }
        with(fit, {
            if (level == 0) {
                score_components <- weights * d1mus  * (y - mus) / varmus * x
                return(precision * .colSums(score_components, nobs, nvars, TRUE))
            }
            if (level == 1) {
                return(1/2 * precision^2 * sum(deviance_residuals - Edeviance_residuals, na.rm = TRUE))
            }
        })
    }

    information <- function(pars, level = 0, fit = NULL, inverse = FALSE) {
        if (is.null(fit)) {
            fit <- key_quantities(pars, y = y, level = level, qr = TRUE)
        }
        with(fit, {
            if (level == 0) {
                R_matrix <- qr.R(qr_decomposition)
                if (inverse) {
                    ## return(dispersion * tcrossprod(solve(R_matrix)))
                    return(dispersion * chol2inv(R_matrix))
                } else {
                    return(precision * crossprod(R_matrix))
                }
            }
            if (level == 1) {
                info <- 0.5 * sum(weights^2 * d2afuns, na.rm = TRUE)/dispersion^4
                if (inverse) {
                    return(1/info)
                } else {
                    return(info)
                }
            }
        })
    }

    hat_values <- function(pars, fit = NULL) {
        if (is.null(fit)) {
            fit <- key_quantities(pars, y = y, level = 0, qr = TRUE)
        }
        with(fit, {
            Qmat <- qr.Q(qr_decomposition)
            .rowSums(Qmat * Qmat, nobs, nvars, TRUE)
        })
    }

    ## FIXME: Redundant function for now
    refit <- function(y, betas_start = NULL) {
        ## Estimate Beta
        betas <- coef(glm.fit(x = x, y = y, weights = weights,
                              start = betas_start,
                              offset = offset,
                              family = family,
                              control = list(epsilon = control$epsilon,
                                             maxit = 2, trace = FALSE),
                              intercept = intercept))
        betas
    }

    ## Estimate the ML of the dispersion parameter for gaussian, gamma and inverse Gaussian
    ## Set the dispersion to 1 if Poisson or binomial
    ## betas is only the regression parameters
    estimate_dispersion <- function(betas, y) {
        if (no_dispersion) {
            disp <- 1
            dispML <- 1
        } else {
            if (df_residual > 0) {
                dispFit <- try(uniroot(f = function(phi) {
                    theta <- c(betas, phi)
                    cfit <- key_quantities(theta, y = y, level = 1, qr = FALSE)
                    gradient(theta, level = 1, fit = cfit)
                }, lower = .Machine$double.eps, upper = 10000, tol = control$epsilon), silent = FALSE)
                if (inherits(dispFit, "try-error")) {
                    warning("the ML estimate of the dispersion could not be calculated. An alternative estimate had been used as starting value.")
                    dispML <- NA_real_
                    disp <- NA_real_
                } else {
                    disp <- dispML <- dispFit$root
                }
            } else { ## if the model is saturated dispML is NA_real_
                disp <- 1 ## A convenient value
                dispML <- NA_real_
            }
        }
        list(dispersion = disp, dispersion_ML = dispML)
    }

    AS_mean_adjustment <- function(pars, level = 0, fit = NULL) {
        if (is.null(fit)) {
            fit <- key_quantities(pars, y = y, level = level, qr = TRUE)
        }
        with(fit, {
            if (level == 0) {
                hatvalues <- hat_values(pars, fit = fit)
                ## Use only observations with keep = TRUE to ensure that no division with zero takes place
                return(.colSums(0.5 * hatvalues * d2mus/d1mus * x, nobs, nvars, TRUE))
            }
            if (level == 1) {
                s1 <- sum(weights^3 * d3afuns, na.rm = TRUE)
                s2 <- sum(weights^2 * d2afuns, na.rm = TRUE)
                return((nvars - 2)/(2 * dispersion) + s1/(2 * dispersion^2 * s2))
            }
        })
    }

    AS_Jeffreys_adjustment <- function(pars, level = 0, fit = NULL) {
        if (is.null(fit)) {
            fit <- key_quantities(pars, y = y, level = level, qr = TRUE)
        }
        with(fit, {
            if (level == 0) {
                hatvalues <- hat_values(pars, fit = fit)
                ## Use only observations with keep = TRUE to ensure that no division with zero takes place
                return(2 * control$a * .colSums(0.5 * hatvalues * (2 * d2mus/d1mus - d1varmus * d1mus / varmus) * x, nobs, nvars, TRUE))
            }
            if (level == 1) {
                s1 <- sum(weights^3 * d3afuns, na.rm = TRUE)
                s2 <- sum(weights^2 * d2afuns, na.rm = TRUE)
                return(2 * control$a * (-(nvars + 4)/(2 * dispersion) + s1/(2 * dispersion^2 * s2)))
            }
        })
    }

    AS_median_adjustment <- function(pars, level = 0, fit = NULL) {
        if (is.null(fit)) {
            fit <- key_quantities(pars, y = y, level = level, qr = TRUE)
        }
        with(fit, {
            if (level == 0) {
                hatvalues <- hat_values(pars, fit = fit)
                R_matrix <- qr.R(qr_decomposition)
                info_unscaled <- crossprod(R_matrix)
                inverse_info_unscaled <- chol2inv(R_matrix)
                ## FIXME: There is 1) definitely a better way to do this, 2) no time...
                b_vector <- numeric(nvars)
                for (j in seq.int(nvars)) {
                    inverse_info_unscaled_j <- inverse_info_unscaled[j, ]
                    vcov_j <- tcrossprod(inverse_info_unscaled_j) / inverse_info_unscaled_j[j]
                    hats_j <- .rowSums((x %*% vcov_j) * x, nobs, nvars, TRUE) * working_weights
                    b_vector[j] <- inverse_info_unscaled_j %*% .colSums(x * (hats_j * (d1mus * d1varmus / (6 * varmus) - 0.5 * d2mus/d1mus)), nobs, nvars, TRUE)
                }
                return(.colSums(0.5 * hatvalues * d2mus / d1mus * x, nobs, nvars, TRUE) +
                       info_unscaled %*% b_vector)
            }
            if (level == 1) {
                s1 <- sum(weights^3 * d3afuns, na.rm = TRUE)
                s2 <- sum(weights^2 * d2afuns, na.rm = TRUE)
                return(nvars/(2 * dispersion) + s1/(6 * dispersion^2 * s2))
            }
        })
    }

    AS_mixed_adjustment <- function(pars, level = 0, fit = NULL) {
        if (is.null(fit)) {
            fit <- key_quantities(pars, y = y, level = level, qr = TRUE)
        }
        with(fit, {
            if (level == 0) {
                hatvalues <- hat_values(pars, fit = fit)
                ## Use only observations with keep = TRUE to ensure that no division with zero takes place
                return(.colSums(0.5 * hatvalues * d2mus/d1mus * x, nobs, nvars, TRUE))
            }
            if (level == 1) {
                s1 <- sum(weights^3 * d3afuns, na.rm = TRUE)
                s2 <- sum(weights^2 * d2afuns, na.rm = TRUE)
                return(nvars/(2 * dispersion) + s1/(6 * dispersion^2 * s2))
            }
        })
    }


    ## compute_step_components does everything on the scale of the /transformed/ dispersion
    compute_step_components <- function(pars, level = 0, fit = NULL) {
        if (is.null(fit)) {
            fit <- key_quantities(pars, y = y, level = level, qr = TRUE)
        }
        if (level == 0) {
            grad <-  gradient(pars, fit = if (has_fixed_totals) NULL else fit, level = 0)
            inverse_info <- try(information(pars, inverse = TRUE, fit = fit, level = 0))
            failed_inversion <- inherits(inverse_info, "try-error")
            adjustment <- adjustment_function(pars, fit = fit, level = 0)
            failed_adjustment <- any(is.na(adjustment))
        }
        if (level == 1) {
            if (no_dispersion | df_residual < 1) {
                grad <- adjustment <- inverse_info <- NA_real_
                failed_adjustment <- failed_inversion <- FALSE
            } else {
                d1zeta <- eval(d1_transformed_dispersion)
                d2zeta <- eval(d2_transformed_dispersion)
                grad <-  gradient(theta, fit = fit, level = 1)/d1zeta
                inverse_info <- 1/information(theta, inverse = FALSE, fit = fit, level = 1) * d1zeta^2
                failed_inversion <- !is.finite(inverse_info)
                ## adjustment <- adjustment_function(theta, fit = fit, level = 1)/d1zeta - if (is_ML | is_AS_median) 0 else 0.5 * d2zeta / d1zeta^2
                adjustment <- adjustment_function(theta, fit = fit, level = 1)/d1zeta - 0.5 * d2zeta / d1zeta^2
                failed_adjustment <- is.na(adjustment)
            }
        }
        out <- list(grad = grad,
                    inverse_info = inverse_info,
                    adjustment = adjustment,
                    failed_adjustment = failed_adjustment,
                    failed_inversion = failed_inversion)
        out
    }

    customTransformation <- is.list(control$transformation) & length(control$transformation) == 2
    if (customTransformation) {
        transformation0 <- control$transformation
    }

    control <- do.call("brglmControl", control)

    adjustment_function <- switch(control$type,
                                  "correction" = AS_mean_adjustment,
                                  "AS_mean" = AS_mean_adjustment,
                                  "AS_median" = AS_median_adjustment,
                                  "AS_mixed" = AS_mixed_adjustment,
                                  "MPL_Jeffreys" = AS_Jeffreys_adjustment,
                                  "ML" = function(pars, ...) 0)

    ## Some useful quantities
    is_ML <- control$type == "ML"
    is_AS_median <- control$type == "AS_median"
    is_AS_mixed <- control$type == "AS_mixed"
    is_correction <- control$type == "correction"
    no_dispersion <- family$family %in% c("poisson", "binomial")


    if (is_ML | is_AS_median | is_AS_mixed) {
        transformation1 <- control$transformation
        Trans1 <- control$Trans
        inverseTrans1 <- control$inverseTrans
        ## Set the transformation to identity
        control$transformation <- "identity"
        control$Trans <- expression(dispersion)
        control$inverseTrans <- expression(transformed_dispersion)
    }


    ## If fixed_totals is specified the compute row_totals
    if (is.null(fixed_totals)) {
        has_fixed_totals <- FALSE
    } else {
        if (family$family == "poisson") {
            row_totals <-  as.vector(tapply(y, fixed_totals, sum))[fixed_totals]
            has_fixed_totals <- TRUE
        } else {
            has_fixed_totals <- FALSE
        }
    }

    ## Ensure x is a matrix, extract variable names, observation
    ## names, nobs, nvars, and initialize weights and offsets if
    ## needed

    x <- as.matrix(x)
    betas_names <- dimnames(x)[[2L]]
    nvars <- ncol(x)
    EMPTY <- nvars == 0
    if (is.null(betas_names) & !EMPTY) {
        betas_names <- colnames(x) <- paste0("x", seq.int(nvars))
    }
    ynames <- if (is.matrix(y)) rownames(y) else names(y)
    converged <- FALSE
    nobs <- NROW(y)
    if (is.null(weights)) {
        weights <- rep.int(1, nobs)
    }
    if (missing_offset <- is.null(offset)) {
        offset <- rep.int(0, nobs)
    }

    ok_links <- c("logit", "probit", "cauchit",
                  "cloglog", "identity", "log",
                  "sqrt", "inverse")


    if (isTRUE(family$family %in% c("quasi", "quasibinomial", "quasipoisson"))) {
        stop("`brglmFit` does not currently support the `quasi`, `quasipoisson` and `quasibinomial` families.")
    }

    ## Enrich family
    family <- enrichwith::enrich(family, with = c("d1afun", "d2afun", "d3afun", "d1variance"))
    if ((family$link %in% ok_links) | (grepl("mu\\^", family$link))) {
        ## Enrich the link object with d2mu.deta and update family object
        linkglm <- make.link(family$link)
        linkglm <- enrichwith::enrich(linkglm, with = "d2mu.deta")
        ## Put everything into the family object
        family[names(linkglm)] <- linkglm
    }
    ## Annoying thing is that link-glm components other than the
    ## standard ones disappear when extra arguments are passed to a
    ## family functions... Anyway, we only require d2mu.deta here.

    ## Extract functions from the enriched family object
    variance <- family$variance
    d1variance <- family$d1variance
    linkinv <- family$linkinv
    linkfun <- family$linkfun
    if (!is.function(variance) || !is.function(linkinv))
        stop("'family' argument seems not to be a valid family object",
             call. = FALSE)
    dev.resids <- family$dev.resids
    aic <- family$aic
    mu.eta <- family$mu.eta
    ## If the family is custom then d2mu.deta cannot survive when
    ## passing throguh current family functions. But mu.eta does; so
    ## we compute d2mu.deta numerically; this allows also generality,
    ## as the users can then keep their custom link implementations
    ## unaltered. Issue is scalability, due to the need of evaluating
    ## n numerical derivatives
    if (is.null(family$d2mu.deta)) {
        d2mu.deta <- function(eta) {
            numDeriv::grad(mu.eta, eta)
        }
    } else {
        d2mu.deta <- family$d2mu.deta
    }
    d1afun <- family$d1afun
    d2afun <- family$d2afun
    d3afun <- family$d3afun
    simulate <- family$simulate
    d1_transformed_dispersion <- DD(control$Trans, "dispersion", order = 1)
    d2_transformed_dispersion <- DD(control$Trans, "dispersion", order = 2)


    ## Check for invalid etas and mus
    valid_eta <- unless_null(family$valideta, function(eta) TRUE)
    valid_mu <- unless_null(family$validmu, function(mu) TRUE)

    ## FIXME: mustart and etastart set to NULL by default
    mustart <- NULL
    etastart <- NULL

    ## Initialize as prescribed in family
    eval(family$initialize)

    ## If there are no covariates in the model then evaluate only the offset
    if (EMPTY) {
        etas <- rep.int(0, nobs) + offset
        if (!valid_eta(etas))
            stop("invalid linear predictor values in empty model", call. = FALSE)
        mus <- linkinv(etas)
        if (!valid_mu(mus))
            stop("invalid fitted means in empty model", call. = FALSE)
        ## deviance <- sum(dev.resids(y, mus, weights))
        working_weights <- ((weights * mu.eta(etas)^2)/variance(mus))^0.5
        residuals <- (y - mus)/mu.eta(etas)
        keep <- rep(TRUE, length(residuals))
        boundary <- converged <- TRUE
        betas_all <- numeric()
        rank <- 0
        iter <- 0L
        keep <- weights > 0
        nkeep <- sum(keep)
        df_residual <- nkeep
    } else {
        boundary <- converged <- FALSE
        ## Detect aliasing
        if (!isTRUE(control$check_aliasing)) {
            is_full_rank <- TRUE ## Assumption
            rank <- nvars_all <- nvars
            betas_names_all <- betas_names
        } else {
            qrx <- qr(x)
            rank <- qrx$rank
            is_full_rank <- rank == nvars
            if (!isTRUE(singular.ok) && !isTRUE(is_full_rank)) {
                stop("singular fit encountered")
            }
            if (!isTRUE(is_full_rank)) {
                aliased <- qrx$pivot[seq.int(qrx$rank + 1, nvars)]
                X_all <- x
                x <- x[, -aliased]
                nvars_all <- nvars
                nvars <- ncol(x)
                betas_names_all <- betas_names
                betas_names <- betas_names[-aliased]
            } else {
                nvars_all <- nvars
                betas_names_all <- betas_names
            }
        }
        betas_all <- structure(rep(NA_real_, nvars_all), .Names = betas_names_all)
        keep <- weights > 0
        ## Check for zero weights
        ## if (any(!keep)) {
        ##     warning("Observations with non-positive weights have been omited from the computations")
        ## }
        nkeep <- sum(keep)
        df_residual <- nkeep - rank
        ## Handle starting values
        ## If start is NULL then start at the ML estimator else use start
        if (is.null(start)) {
            ## Adjust counts if binomial or Poisson in order to avoid infinite estimates
            adj <- control$response_adjustment
            if (is.null(adj)) {
                adj <- nvars/nobs
            }
            if (family$family == "binomial") {
                weights.adj <- weights + (!(is_correction)) * adj
                y.adj <- (weights * y + (!(is_correction)) * 0.5 * adj)/weights.adj
            } else {
                weights.adj <- weights
                y.adj <- y + if (family$family == "poisson") (!(is_correction)) * 0.5 * adj else 0
            }
            ## ML fit to get starting values
            ## Get startng values and kill warnings whilst doing that
            suppressWarnings(
                tempFit <- glm.fit(x = x, y = y.adj, weights = weights.adj,
                                   etastart = etastart, mustart = mustart,
                                   offset = offset, family = family,
                                   control = list(epsilon = control$epsilon,
                                                  maxit = 10000, trace = FALSE),
                                   intercept = intercept)
            )
            betas <- coef(tempFit)
            names(betas) <- betas_names
            dispList <- estimate_dispersion(betas, y = y)
            dispersion <- dispList$dispersion
            if (is.na(dispersion)) dispersion <- var(y)/variance(sum(weights * y)/sum(weights))
            dispersion_ML <- dispList$dispersion_ML
            transformed_dispersion <- eval(control$Trans)
        } else {
            if ((length(start) == nvars_all) & is.numeric(start)) {
                betas_all <- start
                names(betas_all) <- betas_names_all
                if (!isTRUE(is_full_rank)) {
                    betas_all[aliased] <- NA_real_
                    betas <- betas_all[-aliased]
                } else {
                    betas <- betas_all
                }
                ## Estimate dispersion based on current value for betas
                dispList <- estimate_dispersion(betas, y = y)
                dispersion <- dispList$dispersion
                if (is.na(dispersion)) dispersion <- var(y)/variance(sum(weights * y)/sum(weights))
                dispersion_ML <- dispList$dispersion_ML
                transformed_dispersion <- eval(control$Trans)
            }
            if ((length(start) == nvars_all + 1) & is.numeric(start)) {
                betas_all <- start[seq.int(nvars_all)]
                names(betas_all) <- betas_names_all
                if (!isTRUE(is_full_rank)) {
                    betas_all[aliased] <- NA_real_
                    betas <- betas_all[-aliased]
                } else {
                    betas <- betas_all
                }
                transformed_dispersion <- start[nvars_all + 1]
                dispersion_ML <- NA_real_
                dispersion <- eval(control$inverseTrans)
            }
            if (length(start) > nvars_all + 1 | length(start) < nvars_all) {
                stop(paste(paste(gettextf("length of 'start' should be equal to %d and correspond to initial betas for %s", nvars_all, paste(deparse(betas_names_all), collapse = ", "), "or", gettextf("to %d and also include a starting value for the transformed dispersion", nvars_all + 1)))), domain = NA_real_)
            }
        }

        adjusted_grad_all <- rep(NA_real_, nvars_all + 1)
        names(adjusted_grad_all) <- c(betas_names_all, "Transformed dispersion")
        if (is_correction) {
            if (control$maxit > 0) control$maxit <- 1
            control$slowit <- 1
            control$max_step_factor <- 1
        }

        ## Evaluate at the starting values
        theta <- c(betas, dispersion)
        transformed_dispersion <- eval(control$Trans)
        ## Mean quantities
        ## If fixed_totals is provided (i.e. multinomial regression
        ## via the Poisson trick) then evaluate everything expect
        ## the score function at the scaled fitted means
        quantities <- key_quantities(theta, y = y, level = 2 * !no_dispersion, scale_totals = has_fixed_totals, qr = TRUE)
        step_components_beta <- compute_step_components(theta, level = 0, fit = quantities)
        step_components_zeta <- compute_step_components(theta, level = 1, fit = quantities)
        if (step_components_beta$failed_inversion) {
            warning("failed to invert the information matrix")
        }
        if (step_components_beta$failed_adjustment) {
            warning("failed to calculate score adjustment")
        }
        adjusted_grad_beta <- with(step_components_beta, {
            grad + adjustment
        })
        step_beta <- drop(step_components_beta$inverse_info %*% adjusted_grad_beta)
        ## Dispersion quantities
        if (no_dispersion) {
            adjusted_grad_zeta <- step_zeta <- NA_real_
        } else {
            if (step_components_zeta$failed_inversion) {
                warning("failed to invert the information matrix")
            }
            if (step_components_zeta$failed_adjustment) {
                warning("failed to calculate score adjustment")
            }
            adjusted_grad_zeta <- with(step_components_zeta, {
                grad + adjustment
            })
            step_zeta <- as.vector(adjusted_grad_zeta * step_components_zeta$inverse_info)
        }

        ## Main iterations
        slowit <- control$slowit
        if (control$maxit == 0) {
            iter <- 0
            failed <- FALSE
        } else {
            ## Outer iteration
            for (iter in seq.int(control$maxit)) {
                step_factor <- 0
                testhalf <- TRUE
                ## Inner iteration
                while (testhalf & step_factor < control$max_step_factor) {
                    ## store previous values
                    betas0 <- betas
                    dispersion0 <- dispersion
                    step_beta_previous <- step_beta
                    step_zeta_previous <- step_zeta

                    ## Update betas
                    betas <- betas + slowit * 2^(-step_factor) * step_beta

                    ## Update zetas
                    if (!no_dispersion & df_residual > 0) {
                        transformed_dispersion <- eval(control$Trans)
                        transformed_dispersion <- transformed_dispersion + 2^(-step_factor) * step_zeta
                        dispersion <- eval(control$inverseTrans)
                    }

                    ## Compute key quantities
                    theta <- c(betas, dispersion)
                    transformed_dispersion <- eval(control$Trans)

                    ## Mean quantities
                    quantities <- try(key_quantities(theta, y = y, level = 2 * !no_dispersion, scale_totals = has_fixed_totals, qr = TRUE), silent = TRUE)
                    ## This is to capture qr failing and revering to previous estimates
                    if (failed_adjustment_beta <- inherits(quantities, "try-error")) {
                        betas <- betas0
                        dispersion <- dispersion0
                        warning("failed to calculate score adjustment")
                        break
                    }
                    step_components_beta <- compute_step_components(theta, level = 0, fit = quantities)
                    step_components_zeta <- compute_step_components(theta, level = 1, fit = quantities)
                    if (failed_inversion_beta <- step_components_beta$failed_inversion) {
                        betas <- betas0
                        dispersion <- dispersion0
                        warning("failed to invert the information matrix")
                        break
                    }
                    if (failed_adjustment_beta <- step_components_beta$failed_adjustment) {
                        betas <- betas0
                        dispersion <- dispersion0
                        warning("failed to calculate score adjustment")
                        break
                    }
                    adjusted_grad_beta <- with(step_components_beta, grad + adjustment)
                    step_beta <- drop(step_components_beta$inverse_info %*% adjusted_grad_beta)

                    ## Dispersion quantities
                    if (no_dispersion) {
                        adjusted_grad_zeta <- step_zeta <- NA_real_
                        failed_inversion_zeta <- failed_adjustment_zeta <- FALSE
                    } else {
                        if (failed_inversion_zeta <- step_components_zeta$failed_inversion) {
                            warning("failed to invert the information matrix")
                            break
                        }
                        if (failed_adjustment_zeta <- step_components_zeta$failed_adjustment) {
                            warning("failed to calculate score adjustment")
                            break
                        }
                        adjusted_grad_zeta <- with(step_components_zeta, grad + adjustment)
                        step_zeta <- as.vector(adjusted_grad_zeta * step_components_zeta$inverse_info)
                    }

                    ## Convergence criteria
                    linf_current <- max(abs(c(step_beta, step_zeta)), na.rm = TRUE)
                    linf_previous <- max(abs(c(step_beta_previous, step_zeta_previous)), na.rm = TRUE)
                    testhalf <- linf_current > linf_previous

                    ## Continue inner loop
                    ## if (step_factor == 0 & iter == 1)  {
                    ##     testhalf <- TRUE
                    ## }
                    step_factor <- step_factor + 1

                    ##  Trace here
                    if (control$trace) {
                        trace_iteration()
                    }
                }
                failed <- failed_adjustment_beta || failed_inversion_beta || failed_adjustment_zeta || failed_inversion_zeta
                if (failed || linf_current < control$epsilon) {
                    break
                }
            }
        }

        adjusted_grad_all[betas_names] <- adjusted_grad_beta
        adjusted_grad_all["Transformed dispersion"] <- adjusted_grad_zeta
        betas_all[betas_names] <- betas

        ## Convergence analysis
        if ((failed | iter >= control$maxit) & !(is_correction)) {
            warning("brglmFit: algorithm did not converge. Try changing the optimization algorithm defaults, e.g. the defaults for one or more of `maxit`, `epsilon`, `slowit`, and `response_adjustment`; see `?brglm_control` for default values and available options", call. = FALSE)
            converged <- FALSE
        } else {
            converged <- TRUE
        }

        if (boundary) {
            warning("brglmFit: algorithm stopped at boundary value", call. = FALSE)
        }

        ## QR decomposition and fitted values are at the final value
        ## for the coefficients
        ## QR decomposition for cov.unscaled
        if (!isTRUE(is_full_rank)) {
            x <- X_all
            betas <- betas_all
            betas[is.na(betas)] <- 0
            nvars <- nvars_all
        }

        ## If has_fixed_totals = TRUE, then scale fitted values before
        ## calculating QR decompositions, fitted values, etas,
        ## residuals and working_weights

        quantities <- key_quantities(c(betas, dispersion), y = y, level = 2 * !no_dispersion, scale_totals = has_fixed_totals, qr = TRUE)

        qr.Wx <- quantities$qr_decomposition

        mus <- quantities$mus
        etas <- quantities$etas
        ## Residuals
        residuals <- with(quantities, (y - mus)/d1mus)
        working_weights <- quantities$working_weights

        ## info_transformed_dispersion will be NA if is_ML | is_AS_median | is_AS_mixed
        info_transformed_dispersion <- 1/step_components_zeta$inverse_info
        if (is_ML | is_AS_median | is_AS_mixed) {
            transformed_dispersion <- eval(Trans1)
            d1zeta <- eval(DD(Trans1, "dispersion", order = 1))
            adjusted_grad_all["Transformed dispersion"] <- adjusted_grad_all["Transformed dispersion"] / d1zeta
            info_transformed_dispersion <- info_transformed_dispersion / d1zeta^2
            control$transformation <- transformation1
            control$trans <- Trans1
            control$inverseTrans <- inverseTrans1
        }

        eps <- 10 * .Machine$double.eps
        if (family$family == "binomial") {
            if (any(mus > 1 - eps) || any(mus < eps)) {
                warning("brglmFit: fitted probabilities numerically 0 or 1 occurred", call. = FALSE)
                boundary <- TRUE
            }
        }
        if (family$family == "poisson") {
            if (any(mus < eps)) {
                warning("brglmFit: fitted rates numerically 0 occurred", call. = FALSE)
                boundary <- TRUE
            }
        }
        if (df_residual == 0 & !no_dispersion) {
            dispersion <- NA_real_
        }

        ## ## Estimate of first-order bias from the last iteration (so
        ## ## not at the final value for the coefficients)
        ## if (is_ML) {
        ##     ## For now... To be set to calculate biases at a later version
        ##     bias_betas <- bias_zeta <- NULL
        ## }
        ## else {
        ##     bias_betas <- with(step_components_beta, -drop(inverse_info %*% adjustment))
        ##     bias_zeta <- with(step_components_zeta, -drop(inverse_info %*% adjustment))
        ##     bias_betas_all <- betas_all
        ##     bias_betas_all[betas_names] <- bias_betas
        ##     ## If correction has been requested then add estimated biases an attribute to the coefficients
        ##     if (is_correction) {
        ##         attr(betas_all, "biases") <- bias_betas_all
        ##         attr(transformed_dispersion, "biases") <- bias_zeta
        ##     }
        ## }
    }

    ## Working weights
    wt <- rep.int(0, nobs)
    wt[keep] <- working_weights[keep]
    names(wt) <- names(residuals) <- names(mus) <- names(etas) <- names(weights) <- names(y) <- ynames
    ## For the null deviance:
    ##
    ## If there is an intercept but not an offset then the ML fitted
    ## value is the weighted average and is calculated easily below if
    ## ML is used
    ##
    control0 <- control
    control0$maxit <- 1000
    if (customTransformation) {
        control0$transformation <- transformation0
    }
    if (intercept & missing_offset) {
        nullFit <- brglmFit(x = x[, "(Intercept)", drop = FALSE], y = y, weights = weights,
                            offset = rep(0, nobs), family = family, intercept = TRUE,
                            control = control0[c("epsilon", "maxit", "type", "transformation", "slowit")],
                            start = if (no_dispersion) linkfun(mean(y)) else c(linkfun(mean(y)), 1))
        ## FIX: Starting values above are hard-coded. Change in future versions
        nullmus <- nullFit$fitted.values
    }
    ## If there is an offset but not an intercept then the fitted
    ## value is the inverse link evaluated at the offset
    ##
    ## If there is neither an offset nor an intercept then the fitted
    ## values is the inverse link at zero (and hence covered by
    ## linkinv(offset) because offset is zero
    if (!intercept) {
        nullmus <- linkinv(offset)
    }
    ## If there is an intercept and an offset then, for calculating
    ## the null deviance glm will make a call to the fitter to fit the
    ## glm with intercept and the offset
    if (intercept & !missing_offset) {
        nullmus <- mus
        ## doen't really matter what nullmus is set to. glm will make
        ## a new call to brglmFit and use the deviance from that call
        ## as null
    }
    nulldev <- sum(dev.resids(y, nullmus, weights))
    nulldf <- nkeep - as.integer(intercept)
    deviance <- sum(dev.resids(y, mus, weights))
    aic.model <- aic(y, n, mus, weights, deviance) + 2 * rank

    list(coefficients = betas_all,
         residuals = residuals,
         fitted.values = mus,
         ## TODO: see effects?
         ## effects = if (!EMPTY) effects,
         R = if (!EMPTY) qr.R(qr.Wx),
         rank = rank,
         qr = if (!EMPTY) structure(qr.Wx[c("qr", "rank", "qraux", "pivot", "tol")], class = "qr"),
         family = family,
         linear.predictors = etas,
         deviance = deviance,
         aic = aic.model,
         null.deviance = nulldev,
         iter = iter,
         weights = wt,
         prior.weights = weights,
         df.residual = df_residual,
         df.null = nulldf,
         y = y,
         converged = converged,
         boundary = boundary,
         dispersion = dispersion,
         dispersion_ML = dispersion_ML,
         transformed_dispersion = transformed_dispersion,
         info_transformed_dispersion = if (no_dispersion) NA_real_ else info_transformed_dispersion,
         grad = adjusted_grad_all,
         transformation = control$transformation,
         ## cov.unscaled = tcrossprod(R_matrix),
         type = control$type,
         control = control,
         class = "brglmFit")
}

#' Extract model coefficients from [`"brglmFit"`][brglmFit] objects
#'
#' @inheritParams stats::coef
#' @param model one of `"mean"` (default), `"dispersion"`, `"full",
#'     to return the estimates of the parameters in the linear
#'     prediction only, the estimate of the dispersion parameter only,
#'     or both, respectively.
#'
#' @details
#'
#' See [coef()] for more details.
#'
#' @seealso
#'
#' [coef()]
#'
#' @export
coef.brglmFit <- function(object, model = c("mean", "full", "dispersion"), ...) {
    model <- match.arg(model)
    switch(model,
           "mean" = {
        object$coefficients
    },
    "dispersion" = {
        transDisp <- object$transformed_dispersion
        names(transDisp) <- paste0(object$transformation, "(dispersion)")
        transDisp
        ## This will ALWAYS be on the scale of the TRANSFORMED dispersion
    },
    "full" = {
        transDisp <- object$transformed_dispersion
        ntd <- paste0(object$transformation, "(dispersion)")
        names(transDisp) <- ntd
        betas <- object$coefficients
        thetaTrans <- c(betas, transDisp)
        ## if (object$type == "correction") {
        ##     bcf <- attr(betas, "biases")
        ##     btd <- attr(transDisp, "biases")
        ##     names(btd) <- ntd
        ##     attr(thetaTrans, "biases") <- c(bcf, btd)
        ## }
        thetaTrans
    })
}

#' [summary()] method for [`"brglmFit"`][brglmFit()] objects
#'
#' @inheritParams stats::summary.glm
#'
#' @details The interface of the summary method for [`"brglmFit"`][brglmFit]
#'     objects is identical to that of [`"glm"`][glm] objects. The summary
#'     method for [`"brglmFit"`][brglmFit] objects computes the p-values of the
#'     individual Wald statistics based on the standard normal
#'     distribution, unless the family is Gaussian, in which case a t
#'     distribution with appropriate degrees of freedom is used.
#'
#' @seealso [summary.glm()] and [glm()]
#'
#' @examples
#' ## For examples see `examples(brglmFit)`
#'
#' @method summary brglmFit
#' @export
summary.brglmFit <- function(object, dispersion = NULL,
                             correlation = FALSE, symbolic.cor = FALSE,
                             ...) {
    if (is.null(dispersion)) {
        if (object$family$family == "Gaussian") {
            dispersion <- NULL
        } else {
            dispersion <- object$dispersion
        }
    }
    out <- summary.glm(object, dispersion = dispersion,
                       correlation = correlation,
                       symbolic.cor = symbolic.cor, ...)
    out$type <- object$type
    class(out) <- c("summary.brglmFit", class(out))
    out
}

#' Method for computing confidence intervals for one or more
#' regression parameters in a [`"brglmFit"`][brglmFit()] object
#'
#' @inheritParams stats::confint
#'
#' @method confint brglmFit
#' @export
confint.brglmFit <- function(object, parm, level = 0.95, ...) {
    confint.default(object, parm, level, ...)
}

#' Return the variance-covariance matrix for the regression parameters
#' in a [brglmFit()] object
#'
#' @inheritParams stats::vcov.glm
#' @param model character specifying for which component of the model coefficients should be extracted.
#'
#' @details
#'
#' The options for `model` are `"mean"` for mean regression parameters
#' only (default), `"dispersion"` for the dispersion parameter (or the
#' transformed dispersion; see [brglm_control()]), and `"full"` for
#' both the mean regression and the (transformed) dispersion
#' parameters.
#'
#' @method vcov brglmFit
#' @export
vcov.brglmFit <- function(object, model = c("mean", "full", "dispersion"), complete = TRUE, ...) {
    model <- match.arg(model)
    switch(model,
           mean = {
        vcov(summary.brglmFit(object, ...), complete = complete)
    },
    dispersion = {
        vtd <- 1/object$info_transformed_dispersion
        ntd <- paste0(object$transformation, "(dispersion)")
        names(vtd) <- ntd
        vtd
    },
    full = {
        vbetas <- vcov(summary.brglmFit(object, ...), complete = complete)
        vtd <- 1/object$info_transformed_dispersion
        nBetasAll <- c(rownames(vbetas), paste0(object$transformation, "(dispersion)"))
        vBetasAll <- cbind(rbind(vbetas, 0),
                           c(numeric(nrow(vbetas)), vtd))
        dimnames(vBetasAll) <- list(nBetasAll, nBetasAll)
        vBetasAll
    })
}


DD <- function(expr,name, order = 1) {
    if(order < 1) stop("'order' must be >= 1")
    if(order == 1) D(expr,name)
    else DD(D(expr, name), name, order - 1)
}



## Almost all code in print.summary.brglmFit is from
## stats:::print.summary.glm apart from minor modifications
#' @rdname summary.brglmFit
#' @method print summary.brglmFit
#' @export
print.summary.brglmFit <- function (x, digits = max(3L, getOption("digits") - 3L),
                                    symbolic.cor = x$symbolic.cor,
                                    signif.stars = getOption("show.signif.stars"), ...) {
    cat("\nCall:\n", paste(deparse(x$call), sep = "\n", collapse = "\n"),
        "\n\n", sep = "")
    cat("Deviance Residuals: \n")
    if (x$df.residual > 5) {
        x$deviance.resid <- setNames(quantile(x$deviance.resid,
                                              na.rm = TRUE), c("Min", "1Q", "Median", "3Q", "Max"))
    }
    xx <- zapsmall(x$deviance.resid, digits + 1L)
    print.default(xx, digits = digits, na.print = "", print.gap = 2L)
    if (length(x$aliased) == 0L) {
        cat("\nNo Coefficients\n")
    } else {
        df <- if ("df" %in% names(x))
                  x[["df"]]
              else NULL
        if (!is.null(df) && (nsingular <- df[3L] - df[1L]))
            cat("\nCoefficients: (", nsingular, " not defined because of singularities)\n",
                sep = "")
        else cat("\nCoefficients:\n")
        coefs <- x$coefficients
        if (!is.null(aliased <- x$aliased) && any(aliased)) {
            cn <- names(aliased)
            coefs <- matrix(NA, length(aliased), 4L, dimnames = list(cn,
                                                                     colnames(coefs)))
            coefs[!aliased, ] <- x$coefficients
        }
        printCoefmat(coefs, digits = digits, signif.stars = signif.stars,
                     na.print = "NA", ...)
    }
    cat("\n(Dispersion parameter for ", x$family$family, " family taken to be ",
        format(x$dispersion), ")\n\n", apply(cbind(paste(format(c("Null",
                                                                  "Residual"), justify = "right"), "deviance:"), format(unlist(x[c("null.deviance",
                                                                                                                                   "deviance")]), digits = max(5L, digits + 1L)), " on",
                                                   format(unlist(x[c("df.null", "df.residual")])), " degrees of freedom\n"),
                                             1L, paste, collapse = " "), sep = "")
    if (nzchar(mess <- naprint(x$na.action)))
        cat("  (", mess, ")\n", sep = "")
    cat("AIC: ", format(x$aic, digits = max(4L, digits + 1L)))
    cat("\n\nType of estimator:", x$type, get_type_description(x$type))
    cat("\n", "Number of Fisher Scoring iterations: ", x$iter, "\n", sep = "")
    correl <- x$correlation
    if (!is.null(correl)) {
        p <- NCOL(correl)
        if (p > 1) {
            cat("\nCorrelation of Coefficients:\n")
            if (is.logical(symbolic.cor) && symbolic.cor) {
                print(symnum(correl, abbr.colnames = NULL))
            } else {
                correl <- format(round(correl, 2L), nsmall = 2L,
                                 digits = digits)
                correl[!lower.tri(correl)] <- ""
                print(correl[-1, -p, drop = FALSE], quote = FALSE)
            }
        }
    }
    invisible(x)
}

# Copyright (C) 2022- Ioannis Kosmidis

#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 or 3 of the License
#  (at your option).
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  A copy of the GNU General Public License is available at
#  http://www.r-project.org/Licenses/

#' Estimate the exponential of parameters of generalized linear models
#' using various methods
#'
#' The [expo()] method uses the supplied [`"brglmFit"`][brglmFit] or
#' [`"glm"`][glm] object to estimate the exponential of parameters of
#' generalized linear models with maximum likelihood or various mean
#' and median bias reduction methods. [expo()] is useful for computing
#' (corrected) estimates of the multiplicative impact of a unit
#' increase on a covariate on the mean of a Poisson log-linear model
#' (`family = poisson("log")` in [glm()]) while adjusting for other
#' covariates, the odds ratio associated with a unit increase on a
#' covariate in a logistic regression model (`family =
#' binomial("logit")` in [glm()]) while adjusting for other
#' covariates, the relative risk associated with a unit increase on a
#' covariate in a relative risk regression model (`family =
#' binomial("log")` in [glm()]) while adjusting for other covariates,
#' among others.
#'
#' @aliases brglmFit_expo
#' @aliases expo
#' @param object an object of class [`"brglmFit"`][brglmFit] or
#'     [`"glm"`][glm].
#' @param type the type of correction to be used. The available
#'     options are `"correction*"` (explicit mean bias correction with
#'     a multiplicative adjustment), `"correction*"` (explicit mean
#'     bias correction with an additive adjustment), `"Lylesetal2012"`
#'     (explicit median bias correction using the multiplicative
#'     adjustment in Lyles et al., 2012), `"AS_median"` (median bias
#'     reduction), and `"ML"` (maximum likelihood). See Details.
#' @param level the confidence level required. Default is `0.95`.
#'
#' @details
#'
#' The supported methods through the `type` argument are:
#'
#' * `"ML"`: the estimates of the exponentiated parameters are
#' \eqn{\exp(\hat\theta_j)}, where \eqn{\theta_j} is the maximum
#' likelihood estimates for the \eqn{j}th regression parameter.
#'
#' * `"correction*"`: the estimates of the exponentiated parameters
#' are \eqn{\exp(\hat\theta_j) / (1 + \hat{v}_j / 2)}, where
#' \eqn{\hat\theta_j} is the estimate of the \eqn{j}th regression
#' parameter using `type = "AS_mixed"` in [brglmFit()].
#'
#' * `"correction+"`: the estimates of the exponentiated parameters
#' are \eqn{\exp(\hat\theta_j) (1 - \hat{v}_j / 2)}, where
#' \eqn{\hat\theta_j} is the estimate of the \eqn{j}th regression
#' parameter using `type = "AS_mixed"` in [brglmFit()].
#'
#' * `"Lylesetal2012"`: the estimates of the exponentiated parameters
#' are \eqn{\exp(\hat\theta_j) exp(- \hat{v}_j / 2)}, where
#' \eqn{\hat\theta_j} is the estimate of the \eqn{j}th regression
#' parameter using `type = "AS_mixed"` in [brglmFit()]. This estimator
#' has been proposed in Lyles et al. (2012).
#'
#' * `"AS_median"`: the estimates of the exponentiated parameters are
#' \eqn{\exp(\hat\theta_j)}, where \eqn{\hat\theta_j} is the estimate
#' of the \eqn{j}th regression parameter using `type = "AS_median"` in
#' [brglmFit()].
#'
#' `"correction*"` and `"correction+"` are based on multiplicative and
#' additive adjustments, respectively, of the exponential of a
#' reduced-bias estimator (like the ones coming from [brglmFit()] with
#' `type = "AS_mixed"`, `type = "AS_mean"`, and `type =
#' "correction"`). The form of those adjustments results from the
#' expression of the first-term in the mean bias expansion of the
#' exponential of a reduced-bias estimator. See, for example, Di
#' Caterina & Kosmidis (2019, expression 12) for the general form of
#' the first-term of the mean bias of a smooth transformation of a
#' reduced-bias estimator.
#'
#' The estimators from `"correction+"`, `"correction*"`,
#' `"Lylesetal2012"` have asymptotic mean bias of order smaller than
#' than of the maximum likelihood estimator. The estimators from
#' `"AS_median"` are asymptotically closed to being median unbiased
#' than the maximum likelihood estimator is.
#'
#' Estimated standard errors are computed using the delta method,
#' where both the Jacobin and the information matrix are evaluated at
#' the logarithm of the estimates of the exponentiated parameters.
#'
#' Confidence intervals results by taking the exponential of the
#' limits of standard Wald-type intervals computed at the logarithm of
#' the estimates of the exponentiated parameters.
#'
#'
#' @return a list inheriting from class [`"brglmFit_expo"`][brglmFit_expo] with
#'     components `coef` (the estimates of the exponentiated
#'     regression parameters), `se` (the corresponding estimated
#'     standard errors for the exponentiated parameters), `ci`
#'     (confidence intervals of level `level` for the exponentiated
#'     parameters), and `type` for the `type` of correction that has
#'     been requested.
#'
#' @author Ioannis Kosmidis `[aut, cre]` \email{ioannis.kosmidis@warwick.ac.uk}
#'
#' @seealso [brglm_fit()] and and [brglm_control()]
#'
#' @references
#'
#' Di Caterina C, Kosmidis I (2019). Location-Adjusted Wald Statistics for Scalar
#' Parameters. *Computational Statistics & Data Analysis*, **138**,
#' 126-142. \doi{10.1016/j.csda.2019.04.004}.
#'
#' Kosmidis I, Kenne Pagui E C, Sartori N (2020). Mean and median bias
#' reduction in generalized linear models. *Statistics and Computing*,
#' **30**, 43-59. \doi{10.1007/s11222-019-09860-6}.
#'
#' Cordeiro G M, McCullagh P (1991). Bias correction in generalized
#' linear models. *Journal of the Royal Statistical Society. Series B
#' (Methodological)*, **53**, 629-643. \doi{10.1111/j.2517-6161.1991.tb01852.x}.
#'
#' Lyles R H, Guo Y, Greenland S (2012). Reducing bias and mean
#' squared error associated with regression-based odds ratio
#' estimators. *Journal of Statistical Planning and Inference*,
#' **142** 3235–3241. \doi{10.1016/j.jspi.2012.05.005}.
#'
#' @examples
#'
#' ## The lizards example from ?brglm::brglm
#' lizardsML <- glm(cbind(grahami, opalinus) ~ height + diameter +
#'                  light + time, family = binomial(logit), data = lizards,
#'                  method = "glm.fit")
#' # Get estimates, standard errors, and confidence intervals of odds
#' # ratios with various methods
#' expo(lizardsML, type = "ML")
#' expo(lizardsML, type = "correction*")
#' expo(lizardsML, type = "Lylesetal2012")
#' expo(lizardsML, type = "correction+")
#' expo(lizardsML, type = "AS_median")
#'
#' ## Example from ?glm
#' ## Dobson (1990) Page 93: Randomized Controlled Trial :
#' df_D93 <- data.frame(counts = c(18,17,15,20,10,20,25,13,12),
#'                      outcome = gl(3,1,9),
#'                      treatment = gl(3,3))
#' glm.D93 <- glm(counts ~ outcome + treatment, family = poisson(), data = df_D93)
#' expo(glm.D93, type = "correction*")
#'
#' @export
expo.brglmFit <- function(object, type = c("correction*", "correction+", "Lylesetal2012", "AS_median", "ML"), level = 0.95) {
    type <- match.arg(type)
    to_correct <- type %in% c("correction*", "correction+", "Lylesetal2012")
    if (to_correct) {
        if (isTRUE(object$type != "AS_mixed")) {
            object <- update(object, type = "AS_mixed", data = object$data)
        }
    } else {
        ## AS_mean and ML are invariant to transformation
        if (isTRUE(object$type != type)) {
            object <- update(object, type = type, data = object$data)
        }
    }

    info_fun <- get_information_function(object)
    dispersion <- object$dispersion
    coefs <- coef(object, model = "mean")
    ses <- sqrt(diag(solve(info_fun(coefs, dispersion))))[seq_along(coefs)]
    trans_coefs <- exp(coefs) * switch(type,
                                       "ML" = 1,
                                       "AS_median" = 1,
                                       "Lylesetal2012" = exp(- ses^2/2),
                                       "correction*" = 1 / (1 + ses^2/2),
                                       "correction+" = (1 - ses^2/2))

    if (to_correct) {
        ## Recompute coefs and ses on the original scale from trans_coefs if method is not invariant
        coefs <- log(trans_coefs)
        ses <- sqrt(diag(solve(info_fun(coefs, dispersion))))[seq_along(coefs)]
    }

    trans_ses <- trans_coefs * ses
    a <- (1 - level)/2
    ci <- coefs + qnorm(a) * cbind(ses, -ses)
    colnames(ci) <- paste(format(100 * c(a, 1 - a), trim = TRUE, scientific = FALSE, digits = 3), "%")
    out <- list(coef = trans_coefs, se = trans_ses, ci = exp(ci), type = type)
    out$call <- match.call()
    out$family <- object$family
    class(out) <- c("brglmFit_expo", class(out))
    out
}

#' @rdname expo.brglmFit
#' @export
expo.glm <- function(object, type = c("correction*", "correction+", "Lylesetal2012", "AS_median", "ML"), level = 0.95) {
    type <- match.arg(type)
    fit_type <- switch(type,
                       "correction*" = "AS_mixed",
                       "correction+" = "AS_mixed",
                       "Lylesetal2012" = "AS_mixed",
                       "AS_median" = "AS_median",
                       "ML" = "ML")
    object <- update(object, method = brglmFit, type = fit_type)
    out <- expo(object, type, level)
    out$call <- match.call()
    out
}


#' @method print brglmFit_expo
#' @export
print.brglmFit_expo <- function(x, digits = max(3L, getOption("digits") - 3L), ...) {
    cmat <- cbind(x$coef, x$se, x$ci)
    fam_link <- paste0(x$family$family, x$family$link)
    interpr <- switch(fam_link,
                      "binomiallogit" = "Odds ratios",
                      "poissonlog" = "Multiplicative effects to the mean",
                      "Gammalog" = "Multiplicative effects to the mean",
                      "binomiallog" = "Relative risks",
                      "inverse.gaussianlog"= "Multiplicative effects to the mean",
                      NA)
    colnames(cmat) <- c("Estimate", "Std. Error", colnames(x$ci))
    cat("\nCall:\n", paste(deparse(x$call), sep = "\n", collapse = "\n"),
        "\n\n", sep = "")
    if (!is.na(interpr)) {
        cat(interpr, "\n")
    }
    printCoefmat(cmat, digits = digits, ...)
    cat("\n\nType of estimator:", x$type, get_type_description(x$type), "\n")
}

#' Extract estimates from [`"brglmFit_expo"`][brglmFit_expo] objects
#'
#' @inheritParams stats::coef
#'
#' @method coef brglmFit_expo
#' @export
coef.brglmFit_expo <- function(object, ...) {
    object$coef
}

# Copyright (C) 2018- Ioannis Kosmidis

#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 or 3 of the License
#  (at your option).
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  A copy of the GNU General Public License is available at
#  http://www.r-project.org/Licenses/

#' A [`"link-glm"`][make.link] object for misclassified responses in binomial regression models
#'
#' [mis()] is a [`"link-glm"`][make.link] object that specifies the link function in Neuhaus (1999, expression (8)) for handling misclassified responses in binomial regression models using maximum likelihood. A prior specification of the sensitivity and specificity is required.
#'
#' @param link the baseline link to be used.
#' @param sensitivity the probability of observing a success given that a success actually took place given any covariate values.
#' @param specificity the probability of observing a failure given that a failure actually took place given any covariate values.
#'
#' @details
#'
#' `sensitivity + specificity` should be greater or equal to 1,
#' otherwise it is implied that the procedure producing the responses
#' performs worse than chance in terms of misclassification.
#'
#' @references
#'
#' Neuhaus J M (1999). Bias and efficiency loss due to misclassified
#' responses in binary regression. Biometrika, **86**, 843-855.
#' \url{https://www.jstor.org/stable/2673589}.
#'
#' @seealso [glm()], [brglm_fit()]
#'
#' @examples
#'
#' ## Define a few links with some misclassification
#' logit_mis <- mis(link = "logit", sensitivity = 0.9, specificity = 0.9)
#'
#' lizards_f <- cbind(grahami, opalinus) ~ height + diameter + light + time
#'
#' lizardsML <- glm(lizards_f, family = binomial(logit), data = lizards)
#'
#' lizardsML_mis <- update(lizardsML, family = binomial(logit_mis),
#'                         start = coef(lizardsML))
#'
#' ## A notable change is coefficients is noted here compared to when
#' ## specificity and sensitity are 1
#' coef(lizardsML)
#' coef(lizardsML_mis)
#'
#' ## Bias reduction is also possible
#' update(lizardsML_mis, method = "brglmFit", type = "AS_mean",
#'        start = coef(lizardsML))
#'
#' update(lizardsML_mis, method = "brglmFit", type = "AS_median",
#'        start = coef(lizardsML))
#'
#' @export
mis <- function(link = "logit", sensitivity = 1, specificity = 1) {
    link <- make.link(link)
    linkfun <- function(mu) {
        link$linkfun((mu -1 + specificity) / (sensitivity + specificity - 1))
    }
    linkinv <- function(eta) {
        (sensitivity + specificity - 1) * link$linkinv(eta) + 1 - specificity
    }
    mu.eta <- function(eta) {
        (sensitivity + specificity - 1) * link$mu.eta(eta)
    }
    valideta <- function(eta) {
        TRUE
    }
    structure(list(linkfun = linkfun,
                   linkinv = linkinv,
                   mu.eta = mu.eta,
                   valideta = valideta,
                   name = "miss"),
              class = "link-glm")
}

## Dobson (1990) Page 93: Randomized Controlled Trial :

logit_mis <- mis(link = "logit", sensitivity = 1, specificity = 1)
probit_mis <- mis(link = "probit", sensitivity = 1, specificity = 1)

expect_identical(class(logit_mis), "link-glm")
expect_identical(class(probit_mis), "link-glm")

lizards_f <- cbind(grahami, opalinus) ~ height + diameter + light + time
lizards_logit <- glm(lizards_f, family = binomial(logit), data = lizards)
lizards_probit <- glm(lizards_f, family = binomial(probit), data = lizards)

lizards_logit_mis <- update(lizards_logit, family = binomial(logit_mis),
                            start = coef(lizards_logit))
lizards_probit_mis <- update(lizards_probit, family = binomial(probit_mis),
                             start = coef(lizards_probit))

## mis link with sensitivity and specificity 1 are the same as original links
expect_equal(coef(lizards_logit), coef(lizards_logit_mis), tolerance = 1e-06)
expect_equal(coef(lizards_probit), coef(lizards_probit_mis), tolerance = 1e-06)

# Copyright (C) 2025- Ioannis Kosmidis

#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 or 3 of the License
#  (at your option).
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  A copy of the GNU General Public License is available at
#  http://www.r-project.org/Licenses/

#' Fitting function for [glm()] for maximum Diaconis-Ylvisaker prior
#' penalized likelihood estimation of logistic regression models
#'
#' [mdyplFit()] is a fitting method for [glm()] that fits logistic
#' regression models using maximum Diaconis-Ylvisaker prior penalized
#' likelihood estimation.
#'
#' @inheritParams stats::glm.fit
#' @aliases mdypl_fit
#' @param x a design matrix of dimension `n * p`.
#' @param y a vector of observations of length `n`.
#' @param control a list of parameters controlling the fitting
#'     process. See [mdyplControl()] for details.
#'
#' @details
#'
#' [mdyplFit()] uses [stats::glm.fit()] to fit a logistic regression
#' model on responses `alpha * y + (1 - alpha) / 2`, where `y` are the
#' original binomial responses scaled by the binomial totals. This is
#' equivalent to penalizing the likelihood by the Diaconis-Ylvisaker
#' prior with shrinkage parameter \eqn{\alpha} and regression parameters
#' set to zero. See Rigon & Aliverti (2023) and Sterzinger & Kosmidis
#' (2024).
#'
#' By default, `alpha = n / (p + n)` is used, where `n` is the sum of
#' the binomial totals. Alternative values of `alpha` can be passed to
#' the `control` argument; see [mdyplControl()] for setting up the
#' list passed to `control`. If `alpha = 1` then [mdyplFit()] will
#' simply do maximum likelihood estimation.
#'
#' Note that `null.deviance`, `deviance` and `aic` in the resulting
#' object are computed at the adjusted responses. Hence, methods such
#' as [logLik()][stats::logLik()] and [AIC()][stats::AIC()] use the
#' penalized log-likelihood. With the default `alpha`, the inferential
#' procedures based on penalized likelihood are asymptotically
#' equivalent to the ones that use the unpenalized likelihood when
#' `p/n` is vanishing asymptotically.
#'
#' For high-dimensionality corrected estimates, standard errors and z
#' statistics, use the [`summary`][summary.mdyplFit()] method for
#' [`"mdyplFit"`][mdyplFit()] objects with `hd_correction = TRUE`.
#'
#' [mdypl_fit()] is an alias to [mdyplFit()].
#'
#' @return
#'
#' An object inheriting from [`"mdyplFit"`][mdyplFit()] object, which
#' is a list having the same elements to the list that
#' [stats::glm.fit()] returns, with a few extra arguments.
#'
#' @author Ioannis Kosmidis `[aut, cre]` \email{ioannis.kosmidis@warwick.ac.uk}
#'
#' @seealso [mdyplControl()], [summary.mdyplFit()], [plrtest.mdyplFit()], [glm()]
#'
#' @references
#'
#' Sterzinger P, Kosmidis I (2024). Diaconis-Ylvisaker prior
#' penalized likelihood for \eqn{p/n \to \kappa \in (0,1)} logistic
#' regression. *arXiv*:2311.07419v2, \url{https://arxiv.org/abs/2311.07419}.
#'
#' Rigon T, Aliverti E (2023). Conjugate priors and bias reduction for
#' logistic regression models. *Statistics & Probability Letters*,
#' **202**, 109901. \doi{10.1016/j.spl.2023.109901}.
#'
#' @examples
#'
#' data("lizards", package = "brglm2")
#' liz_fm <- cbind(grahami, opalinus) ~ height + diameter + light + time
#' ## ML fit = MDYPL fit with `alpha = 1`
#' liz_ml <- glm(liz_fm, family = binomial(), data = lizards,
#'               method = "mdyplFit", alpha = 1)
#' liz_ml0 <- glm(liz_fm, family = binomial(), data = lizards)
#'
#' ## liz_ml is the same fit as liz_ml0
#' summ_liz_ml <- summary(liz_ml)
#' summ_liz_ml0 <- summary(liz_ml0)
#' all.equal(coef(summ_liz_ml), coef(summ_liz_ml0))
#'
#' ## MDYPL fit with default `alpha` (see `?mdyplControl`)
#' liz_fm <- cbind(grahami, opalinus) ~ height + diameter + light + time
#' liz_mdypl <- glm(liz_ml, family = binomial(), data = lizards,
#'                  method = "mdyplFit")
#'
#' ## Comparing outputs from ML and MDYPL, with and without
#' ## high-dimensionality corrections.
#' summary(liz_mdypl)
#' summary(liz_mdypl, hd_correction = TRUE)
#' summ_liz_ml
#' summary(liz_ml, hd_correction = TRUE)
#' ## Not much difference in fits here as this is a low dimensional
#' ## problem with dimensionality constant
#' (liz_ml$rank - 1) / sum(weights(liz_ml))
#'
#'
#'
#' ## The case study in Section 8 of Sterzinger and
#' ## Kosmidis (2024)
#' data("MultipleFeatures", package = "brglm2")
#'
#' ## Center the fou.* and kar.* features
#' vars <- grep("fou|kar", names(MultipleFeatures), value = TRUE)
#' train_id <- which(MultipleFeatures$training)
#' MultipleFeatures[train_id, vars] <- scale(MultipleFeatures[train_id, vars], scale = FALSE)
#' ## Compute the MDYPL fits
#' kappa <- length(vars) / sum(MultipleFeatures$training)
#' full_fm <- formula(paste("I(digit == 7) ~", paste(vars, collapse = " + ")))
#' nest_vars <- grep("fou", vars, value = TRUE)
#' nest_fm <- formula(paste("I(digit == 7) ~", paste(nest_vars, collapse = " + ")))
#' full_m <- glm(full_fm, data = MultipleFeatures, family = binomial(),
#'               method = mdyplFit, alpha = 1 / (1 + kappa), subset = training)
#' nest_m <- update(full_m, nest_fm)
#'
#' ## With a naive penalized likelihood ratio test we get no evidence
#' ## against the hypothesis that the model with only `fou` features
#' ## is an as good descrition of `7` as the model with both `fou` and
#' ## `kar` features.
#' plrtest(nest_m, full_m)
#'
#' ## With a high-dimensionality correction theres is strong evidence
#' ## against the model with only `fou` features
#' plrtest(nest_m, full_m, hd_correction = TRUE)
#'
#'
#' \donttest{
#' ## A simulated data set as in Rigon & Aliverti (2023, Section 4.3)
#'
#' set.seed(123)
#' n <- 1000
#' p <- 500
#' gamma <- sqrt(5)
#' X <- matrix(rnorm(n * p, 0, 1), nrow = n, ncol = p)
#' betas0 <- rep(c(-1, -1/2, 0, 2, 3), each = p / 5)
#' betas <- gamma * betas0 / sqrt(sum(betas0^2))
#' probs <- plogis(drop(X %*% betas))
#' y <- rbinom(n, 1, probs)
#' fit_mdypl <- glm(y ~ -1 + X, family = binomial(), method = "mdyplFit")
#'
#' ## The default value of `alpha` is `n / (n + p)` here
#' identical(n / (n + p), fit_mdypl$alpha)
#'
#' ## Aggregate bias of MDYPL and rescaled MDYPL estimators
#' ag_bias <- function(estimates, beta) mean(estimates - beta)
#' ag_bias(coef(summary(fit_mdypl))[, "Estimate"], betas)
#' ag_bias(coef(summary(fit_mdypl, hd_correction = TRUE))[, "Estimate"], betas)
#'
#' }
#' @export
mdyplFit <- function(x, y, weights = rep(1, nobs), start = NULL, etastart = NULL,
                     mustart = NULL, offset = rep(0, nobs), family = binomial(),
                     control = list(), intercept = TRUE,
                     singular.ok = TRUE) {

    nobs <- NROW(y)
    if (!isTRUE(family$family == "binomial" && family$link == "logit")) {
        stop('`mdyplFit` currently supports only `binomial` family with `"logit"` link')
    }

    control <- do.call("mdyplControl", control)

    missing_offset <- is.null(offset)

    if (is.null(weights)) {
        weights <- rep.int(1, nobs)
    }

    if (is.null(mustart)) {
        eval(family$initialize)
    }
    else {
        mukeep <- mustart
        eval(family$initialize)
        mustart <- mukeep
    }

    if (missing_offset) {
        offset <- rep.int(0, nobs)
    }

    alpha <- unless_null(control$alpha, sum(weights) / (sum(weights) + ncol(x) - intercept))

    ## adjust responses as per MDYPL with beta_P = 0
    y_adj <- alpha * y + (1 - alpha) / 2
    control_glm <- glm.control(epsilon = control$epsilon,
                               maxit = control$maxit, trace = control$trace)
    out <- glm.fit(x = x, y = y_adj, weights = weights,
                   etastart = etastart, mustart = mustart,
                   offset = offset, family = quasibinomial(),
                   control = control_glm,
                   intercept = intercept, singular.ok = singular.ok)

    mus <- out$fitted.values
    if (intercept & missing_offset) {
        nullmus <- mdyplFit(x = x[, "(Intercept)", drop = FALSE], y = y, weights = weights,
                            offset = rep(0, nobs), family = family, intercept = TRUE,
                            control = control,
                            start = family$linkfun(mean(y)))$fitted.values
    }

    if (!intercept) {
        nullmus <- family$linkinv(offset)
    }
    ## If there is an intercept and an offset then, for calculating
    ## the null deviance glm will make a call to the fitter to fit the
    ## glm with intercept and the offset
    if (intercept & !missing_offset) {
        nullmus <- mus
        ## doen't really matter what nullmus is set to. glm will make
        ## a new call to mdyplFit and use the deviance from that call
        ## as null
    }

    out$family <- family

    ## Reset quantities in terms of original responses wherever needed
    dev.resids <- family$dev.resids
    out$null.deviance <- sum(dev.resids(y_adj, nullmus, weights))
    out$deviance <- sum(dev.resids(y_adj, mus, weights))
    out$aic <- logist_aic(y_adj, n, mus, weights, deviance) + 2 * out$rank
    out$residuals <- (y - mus) / (mus * (1 - mus))
    out$y_adj <- y_adj
    out$y <- y
    out$alpha <- alpha
    out$type <- "MPL_DY"
    out$control <- control
    out$class <- c("mdyplFit")
    out$n_init <- n ## needed when `hd_correction = TRUE` in summary where aic is recomputed
    out
}

## Similar to binomial()$aic but works with y in (0, 1)
dbinom2 <- function(x, size, prob, log = FALSE) {
    su <- x
    fa <- size - x
    db <- prob^su * (1 - prob)^fa / beta(su + 1, fa + 1) / (size + 1)
    db[size < su] <- 0
    if (isTRUE(log)) log(db) else db
}

logist_aic <- function(y, n, mu, wt, dev) {
    m <- if (any(n > 1)) n else wt
    -2 * sum(ifelse(m > 0, (wt/m), 0) * dbinom2(m * y, m, mu, log = TRUE))
}

#' Auxiliary function for [glm()] fitting using the [brglmFit()]
#' method.
#'
#' Typically only used internally by [brglmFit()], but may be used to
#' construct a `control` argument.
#'
#' @aliases mdypl_control
#' @param alpha the shrinkage parameter (in `[0, 1]`) in the
#'     Diaconis-Ylvisaker prior penalty. Default is \code{NULL}, which
#'     results in `alpha = n / (n + p)`, where `n` is the sum of the
#'     binomial totals and `p` is the number of model
#'     parameters. Setting `alpha = 1` corresponds to using maximum
#'     likelihood, i.e. no penalization. See Details.
#' @param epsilon positive convergence tolerance epsilon. Default is
#'     `1e-08`.
#' @param maxit integer giving the maximal number of iterations
#'     allowed. Default is `25`.
#' @param trace logical indicating if output should be produced for
#'     each iteration. Default is `FALSE`.
#'
#' @details
#'
#' Internally, [mdyplFit()] uses [stats::glm.fit()] to fit a logistic
#' regression model on responses `alpha * y + (1 - alpha) / 2`, where
#' `y` are the original binomial responses scaled by the binomial
#' totals. `epsilon`, `maxit` and `trace` control the
#' [stats::glm.fit()] call; see [stats::glm.control()].
#'
#' @return
#'
#' A list with components named as the arguments.
#'
#' @author Ioannis Kosmidis `[aut, cre]` \email{ioannis.kosmidis@warwick.ac.uk}
#'
#' @seealso [mdyplFit()], [glm.control()]
#'
#' @export
mdyplControl <- function(alpha = NULL, epsilon = 1e-08, maxit = 25, trace = FALSE) {
    out <- glm.control(epsilon, maxit, trace)
    if (!is.null(alpha)) {
        if (!(is.numeric(alpha)) || isTRUE(alpha < 0) || isTRUE(alpha > 1))
            stop("`alpha` should be in [0, 1]")
    }
    out$alpha <- alpha
    out
}


#' Method for computing confidence intervals for one or more
#' regression parameters in a [`"mdyplFit"`][mdyplFit()] object
#'
#' @inheritParams stats::confint
#' @inheritParams summary.mdyplFit
#'
#' @author Ioannis Kosmidis `[aut, cre]` \email{ioannis.kosmidis@warwick.ac.uk}
#'
#' @seealso [mdyplFit()], [summary.mdyplFit()]
#'
#' @examples
#'
#' \donttest{
#'
#' set.seed(123)
#' n <- 2000
#' p <- 800
#' set.seed(123)
#' betas <- c(rnorm(p / 4, mean = 7, sd = 1), rep(0, 3 * p / 4))
#' X <- matrix(rnorm(n * p, 0, 1/sqrt(n)), nrow = n, ncol = p)
#' probs <- plogis(drop(X %*% betas))
#' y <- rbinom(n, 1, probs)
#' fit_mdypl <- glm(y ~ -1 + X, family = binomial(), method = "mdyplFit")
#'
#' wald_ci <- confint(fit_mdypl)
#' adj_wald_ci <- confint(fit_mdypl, hd_correction = TRUE)
#' ag_coverage <- function(cis, beta) mean((cis[, 1] < beta) & (cis[, 2] > beta))
#' ag_coverage(wald_ci, betas)
#' ag_coverage(adj_wald_ci, betas)
#'
#' }
#'
#' @method confint mdyplFit
#' @export
confint.mdyplFit <- function(object, parm, level = 0.95, hd_correction = FALSE, ...) {
    ## A modification of confint.default to use summary objects
    summ <- summary(object, hd_correction = hd_correction, ...)
    coefs <- coef(summ)
    cf <- coefs[, "Estimate"]
    pnames <- rownames(coefs)
    if (missing(parm))
        parm <- pnames
    else if (is.numeric(parm))
        parm <- pnames[parm]
    a <- (1 - level)/2
    a <- c(a, 1 - a)
    ## Directly from stats:::.format_perc
    pct <- paste(format(100 * a, trim = TRUE, scientific = FALSE, digits = 3), "%")
    fac <- qnorm(a)
    ci <- array(NA, dim = c(length(parm), 2L), dimnames = list(parm, pct))
    ses <- coefs[, "Std. Error"]
    ci[] <- cf[parm] + ses %o% fac
    ci
}

library("brglm")
data("lizards", package = "brglm2")

a <- 0.3

liz_fm <- cbind(grahami, opalinus) ~ height + diameter + light + time
expect_error(
    glm(liz_fm, family = binomial("probit"), data = lizards,
        method = "mdyplFit", alpha = a)
)

expect_error(
    glm(liz_fm, family = gaussian(), data = lizards,
        method = "mdyplFit", alpha = a)
)

liz_fit_DY <- glm(liz_fm, family = binomial(), data = lizards,
                  method = "mdyplFit", alpha = a)

adj_lizards <- lizards |>
    transform(totals = grahami + opalinus) |>
    transform(g_y = grahami / totals) |>
    transform(grahami = a * grahami + totals * (1 - a) / 2) |>
    transform(opalinus = totals - grahami)
expect_warning(
    temp_fit <- glm(cbind(grahami, opalinus) ~ height + diameter +
                        light + time, family = binomial(),
                    data = adj_lizards)
)

## Correct alpha
expect_equal(a, liz_fit_DY$alpha)

## Correct coefficients
expect_equal(coef(liz_fit_DY), coef(temp_fit))

## Correct deviance
yy <- adj_lizards$g_y
mm <- fitted(temp_fit)
tt <- adj_lizards$totals
expect_equal(deviance(liz_fit_DY),
             sum(binomial()$dev.resids(a * yy + (1- a)/2, mm, tt)))

## Correct estimated standard errors
v <- tt * mm * (1 - mm)
xx <- model.matrix(liz_fit_DY)
vv <- solve(crossprod(xx * sqrt(v)))
expect_equal(sqrt(diag(vv)), coef(summary(liz_fit_DY))[, "Std. Error"])

## Naming conventions
liz_fit_DY1 <- update(liz_fit_DY, method = "mdypl_fit")
objs <- names(liz_fit_DY)
objs <- objs[!(objs %in% c("call", "method"))]
for (what in objs) {
    expect_equal(liz_fit_DY1[[what]],
                 liz_fit_DY[[what]])
}


## taus
xx <- model.matrix(liz_fit_DY)
expect_equal(
    sapply(2:ncol(xx), function(j) sum( lm.fit(xx[, -c(1, j)], xx[, j])$residuals^2 ) / (nrow(xx) - ncol(xx) + 2)) |> sqrt(),
    brglm2:::taus(xx))


## logist_aic
set.seed(111)
tots <- c(11, 15, 5, 5, 15, 111, 8)
succ <- c(0, 11, 3, 5, 2, 5, 7)
fail <- tots - succ
probs <- c(0, runif(length(fail) - 2), 1)
expect_equal(brglm2:::logist_aic(succ / tots, tots, probs, tots),
             binomial()$aic(succ / tots, tots, probs, tots))
expect_equal(-brglm2:::logist_aic(succ / tots, tots, probs, tots) / 2,
             sum(dbinom(succ, tots, probs, log = TRUE)))


expect_equal(brglm2:::dbinom2(succ, tots, probs, log = FALSE),
             dbinom(succ, tots, probs, log = FALSE))

expect_equal(brglm2:::dbinom2(succ, tots, probs, log = TRUE),
             dbinom(succ, tots, probs, log = TRUE))

expect_equal(brglm2:::dbinom2(1, 1, 0), dbinom(1, 1, 0))
expect_equal(brglm2:::dbinom2(1, 1, 1), dbinom(1, 1, 1))
expect_equal(brglm2:::dbinom2(0, 1, 0), dbinom(0, 1, 0))
expect_equal(brglm2:::dbinom2(0, 1, 1), dbinom(0, 1, 1))
expect_equal(brglm2:::dbinom2(1, 0, 0), dbinom(1, 0, 0))
expect_equal(brglm2:::dbinom2(1, 0, 1), dbinom(1, 0, 1))
expect_equal(brglm2:::dbinom2(0, 0, 1), dbinom(0, 0, 1))
expect_equal(brglm2:::dbinom2(0, 0, 0), dbinom(0, 0, 0))
expect_equal(brglm2:::dbinom2(1, 1, 0), dbinom(1, 1, 0), log = TRUE)
expect_equal(brglm2:::dbinom2(1, 1, 1), dbinom(1, 1, 1), log = TRUE)
expect_equal(brglm2:::dbinom2(0, 1, 0), dbinom(0, 1, 0), log = TRUE)
expect_equal(brglm2:::dbinom2(0, 1, 1), dbinom(0, 1, 1), log = TRUE)
expect_equal(brglm2:::dbinom2(1, 0, 0), dbinom(1, 0, 0), log = TRUE)
expect_equal(brglm2:::dbinom2(1, 0, 1), dbinom(1, 0, 1), log = TRUE)
expect_equal(brglm2:::dbinom2(0, 0, 1), dbinom(0, 0, 1), log = TRUE)
expect_equal(brglm2:::dbinom2(0, 0, 0), dbinom(0, 0, 0), log = TRUE)

#' Estimate the corrupted signal strength in a model with
#' (sub-)Gaussian covariates
#'
#' @param object an [`"mdyplFit"`][mdyplFit()] object.
#'
#'
#' @details
#'
#' The Signal Strength Leave-One-Out Estimator (SLOE) is defined in
#' Yadlowsky et al. (2021) when the model is estimated using maximum
#' likelihood (i.e. when `object$alpha = 1`; see [mdyplControl()] for
#' what `alpha` is). The SLOE adaptation when estimation is through
#' maximum Diaconis-Ylvisaker prior penalized likelihood
#' ([mdypl_fit()]) has been put forward in Sterzinger & Kosmidis
#' (2025).
#'
#' In particular, [sloe()] computes an estimate of the corrupted
#' signal strength which is the limit \deqn{\nu^2} of \eqn{var(X
#' \hat\beta(\alpha))}, where \eqn{\hat\beta(\alpha)} is the maximum
#' Diaconis-Ylvisaker prior penalized likelihood (MDYPL) estimator as
#' computed by [mdyplFit()] with shrinkage parameter \eqn{alpha}.
#'
#' @return
#'
#' A scalar.
#'
#' @author Ioannis Kosmidis `[aut, cre]` \email{ioannis.kosmidis@warwick.ac.uk}
#'
#' @seealso [summary.mdyplFit()]
#'
#' @references
#'
#' Sterzinger P, Kosmidis I (2024). Diaconis-Ylvisaker prior
#' penalized likelihood for \eqn{p/n \to \kappa \in (0,1)} logistic
#' regression. *arXiv*:2311.07419v2, \url{https://arxiv.org/abs/2311.07419}.
#'
#' Yadlowsky S, Yun T, McLean C Y, D' Amour A (2021). SLOE: A Faster
#' Method for Statistical Inference in High-Dimensional Logistic
#' Regression. In M Ranzato, A Beygelzimer, Y Dauphin, P Liang, J W
#' Vaughan (eds.), *Advances in Neural Information Processing
#' Systems*, **34**, 29517–29528. Curran Associates,
#' Inc. \url{https://proceedings.neurips.cc/paper_files/paper/2021/file/f6c2a0c4b566bc99d596e58638e342b0-Paper.pdf}.
#'
#' @export
sloe <- function(object) {
    mu <- fitted(object)
    v <- mu * (1 - mu)
    h <- hatvalues(object)
    S <- object$linear.predictors - (object$y_adj - mu) / v * h / (1 - h)
    inds <- is.infinite(S) | is.na(S)
    sd(S[!inds])
}

taus <- function(X) {
    X <- X[, colnames(X) != "(Intercept)"]
    R <- qr.R(qr(X))
    rss <- 1 / colSums(backsolve(R, diag(ncol(X)), transpose = TRUE)^2)
    sqrt(rss / (nrow(X) - ncol(X) + 1))
}

#' Summary method for [`"mdyplFit"`][mdyplFit()] objects
#'
#' @inheritParams stats::summary.glm
#' @inheritParams solve_se
#' @param hd_correction if `FALSE` (default), then the corresponding
#'     quantities are computed according to standard asymptotics. If
#'     `TRUE` then the high-dimensionality corrections in Sterzinger &
#'     Kosmidis (2024) are employed to updates estimates, estimated
#'     standard errors, z-statistics, etc. See Details.
#' @param solve_se_control a list of further arguments to be passed to
#'     [solve_se()]. Even if explicitly specified, the arguments
#'     `kappa`, `ss`, `alpha`, and `intercept` are always set
#'     according to `object`, and `corrupted` is set to `TRUE`.
#' @param ... further arguments to be passed to [summary.glm()].
#'
#' @details
#'
#' If `hd_correction = TRUE`, the [sloe()] estimator of the square
#' root of the corrupted signal strength is estimated from `object`,
#' as are the conditional variances of each covariate given the others
#' (excluding the intercept). The latter are estimated using residual
#' sums of squares from the linear regression of each covariate on all
#' the others, as proposed in Zhao et al (2021, Section 5.1). Then the
#' appropriate state evolution equations are solved using [solve_se()]
#' with `corrupted = TRUE`, and the obtained constants are used to
#' rescale the estimates, and adjust estimated standard errors and
#' z-statistics as in Sterzinger & Kosmidis (2024).
#'
#' The key assumptions under which the rescaled estimates and corrected
#' standard errors and z-statistics are asymptotically valid are that
#' the covariates have sub-Gaussian distributions, and that the signal
#' strength, which is the limit \eqn{\gamma^2} of \eqn{var(X \beta)}
#' is finite as \eqn{p / n \to \kappa \in (0, 1)}, with \eqn{\kappa \in
#' (0, 1)}. See Sterzinger & Kosmidis (2024).
#'
#' If `hd_correction = TRUE`, and the model has an intercept, then the
#' result provides only a corrected estimate of the intercept with no
#' accompanying standard error, z-statistic, and p-value. Also,
#' `vcov(summary(object, hd_correction = TRUE))` is always
#' `NULL`. Populating those objects with appropriate estimates is the
#' subject of current work.
#'
#' @return
#'
#' A list with objects as in the result of [stats::summary.glm()],
#' with extra component `se_parameters`, which is the vector of the
#' solution to the state evolution equations with extra attributes
#' (see [solve_se()]).
#'
#' @author Ioannis Kosmidis `[aut, cre]` \email{ioannis.kosmidis@warwick.ac.uk}
#'
#' @seealso [mdyplFit()], [solve_se()]
#'
#' @references
#'
#' Zhao Q, Sur P, Cand\`es E J (2022). The asymptotic distribution of
#' the MLE in high-dimensional logistic models: Arbitrary
#' covariance. *Bernoulli*, **28**,
#' 1835–1861. \doi{10.3150/21-BEJ1401}.
#'
#' Sterzinger P, Kosmidis I (2024). Diaconis-Ylvisaker prior
#' penalized likelihood for \eqn{p/n \to \kappa \in (0,1)} logistic
#' regression. *arXiv*:2311.07419v2, \url{https://arxiv.org/abs/2311.07419}.
#'
#' @examples
#'
#' \donttest{
#'
#' set.seed(123)
#' n <- 2000
#' p <- 400
#' set.seed(123)
#' betas <- c(rnorm(p / 2, mean = 7, sd = 1), rep(0, p / 2))
#' X <- matrix(rnorm(n * p, 0, 1/sqrt(n)), nrow = n, ncol = p)
#' probs <- plogis(drop(X %*% betas))
#' y <- rbinom(n, 1, probs)
#' fit_mdypl <- glm(y ~ -1 + X, family = binomial(), method = "mdyplFit")
#'
#' st_summary <- summary(fit_mdypl)
#' hd_summary <- summary(fit_mdypl, hd_correction = TRUE)
#'
#' cols <- hcl.colors(3, alpha = 0.2)
#' par(mfrow = c(1, 2))
#' plot(betas, type = "l", ylim = c(-3, 14),
#'      main = "MDYPL estimates",
#'      xlab = "Parameter index", ylab = NA)
#' points(coef(st_summary)[, "Estimate"], col = NA, bg = cols[1], pch = 21)
#'
#' plot(betas, type = "l", ylim = c(-3, 14),
#'      main = "rescaled MDYPL estimates",
#'      xlab = "Parameter index", ylab = NA)
#' points(coef(hd_summary)[, "Estimate"], col = NA, bg = cols[2], pch = 21)
#'
#' ## z-statistics
#' z_mdypl <- coef(st_summary)[betas == 0, "z value"]
#' qqnorm(z_mdypl, col = NA, bg = cols[1], pch = 21, main = "z value")
#' abline(0, 1, lty = 2)
#' z_c_mdypl <- coef(hd_summary)[betas == 0, "z value"]
#' qqnorm(z_c_mdypl, col = NA, bg = cols[2], pch = 21, main = "corrected z value")
#' abline(0, 1, lty = 2)
#'
#'}
#'
#' @method summary mdyplFit
#' @export
summary.mdyplFit <- function(object, hd_correction = FALSE,
                             solve_se_control = list(), ...) {
    ## Get summary object
    summ <- summary.glm(object, ...)
    if (isTRUE(hd_correction)) {
        coefs <- coef(summ)
        nobs <- sum(pw <- weights(object))
        has_intercept <- attr(terms(object), "intercept")
        p <- nrow(coefs) - has_intercept
        nu_sloe <- sloe(object)
        theta <- if (has_intercept) coefs["(Intercept)", "Estimate"] else NULL
        solve_se_control$kappa <- ka <- p / nobs
        solve_se_control$ss <- nu_sloe
        solve_se_control$alpha <- object$alpha
        solve_se_control$intercept <- theta
        if (is.null(solve_se_control$start)) {
            solve_se_control$start <- c(0.5, 1, 1, theta)
        }
        solve_se_control$corrupted <- TRUE
        se_pars <- try(do.call("solve_se", solve_se_control), silent = TRUE)
        if (inherits(se_pars, "try-error")) {
            msg <- paste(se_pars, "Unable to solve the state evolution equations. See `?solve_se` and use `solve_se_control` to control the optimizer, including supplying a vector of", 3 + has_intercept, " starting values for `mu` (in (0, 1)), `b` (> 0), `sigma` (> 0)", if (has_intercept) ", `intercept.`" else ".")
            stop(msg)
        }
        xx <- model.matrix(object)[, !summ$aliased]
        tt <- taus(xx)
        no_int <- !(rownames(coefs) %in% "(Intercept)")
        coefs[no_int, "Estimate"] <- coefs[no_int, "Estimate"] / se_pars[1]
        coefs[no_int, "Std. Error"] <- se_pars[3] / (sqrt(nobs) * tt * se_pars[1])
        coefs[, "z value"] <- coefs[, "Estimate"] / coefs[, "Std. Error"]
        coefs[, "Pr(>|z|)"] <- 2 * pnorm(-abs(coefs[, "z value"]))
        if (has_intercept) {
            coefs["(Intercept)", ] <- c(se_pars[4], NA, NA, NA)
        }
        summ$coefficients <- coefs

        family <- object$family
        dev.resids <- family$dev.resids
        mus <- family$linkinv(drop(xx %*% coefs[, "Estimate"]))
        y <- object$y
        ## Null deviance is not updated
        d_res <- sqrt(pmax(family$dev.resids(y, mus, pw), 0))
        summ$deviance.resid <- ifelse(y > mus, d_res, -d_res)
        summ$deviance <- sum(summ$deviance.resid^2)
        summ$aic <- logist_aic(object$y_adj, object$n_init, mus, pw, summ$deviance) + 2 * object$rank
        summ$cov.scaled <- summ$cov.unscaled <- NULL
        summ$se_parameters <- se_pars
        if (!isTRUE(all(abs(attr(se_pars, "funcs")) < 1e-04))) {
            msg <- paste("Unable to solve the state evolution equations. See `?solve_se` and use `solve_se_control` to control the optimizer, including supplying a vector of", 3 + has_intercept, " starting values for `mu` (in (0, 1)), `b` (> 0), `sigma` (> 0)", if (has_intercept) ", `intercept.`" else ".")
            warning(msg)
        }
        summ$signal_strength <- (nu_sloe^2 - ka * se_pars[3]^2) / se_pars[1]^2
        summ$kappa <- ka
    }
    summ$hd_correction <- hd_correction
    summ$alpha <- object$alpha
    summ$type <- object$type
    class(summ) <- c("summary.mdyplFit", class(summ))
    summ
}

## Almost all code in print.summary.mdyplFit is from
## stats:::print.summary.glm apart from minor modifications
#' @rdname summary.mdyplFit
#' @method print summary.mdyplFit
#' @export
print.summary.mdyplFit <- function (x, digits = max(3L, getOption("digits") - 3L),
                                    symbolic.cor = x$symbolic.cor,
                                    signif.stars = getOption("show.signif.stars"), ...) {
    cat("\nCall:\n", paste(deparse(x$call), sep = "\n", collapse = "\n"),
        "\n\n", sep = "")
    cat("Deviance Residuals: \n")
    if (x$df.residual > 5) {
        x$deviance.resid <- setNames(quantile(x$deviance.resid,
                                              na.rm = TRUE), c("Min", "1Q", "Median", "3Q", "Max"))
    }
    xx <- zapsmall(x$deviance.resid, digits + 1L)
    print.default(xx, digits = digits, na.print = "", print.gap = 2L)
    if (length(x$aliased) == 0L) {
        cat("\nNo Coefficients\n")
    } else {
        df <- if ("df" %in% names(x))
                  x[["df"]]
              else NULL
        if (!is.null(df) && (nsingular <- df[3L] - df[1L]))
            cat("\nCoefficients: (", nsingular, " not defined because of singularities)\n",
                sep = "")
        else cat("\nCoefficients:\n")
        coefs <- x$coefficients
        if (!is.null(aliased <- x$aliased) && any(aliased)) {
            cn <- names(aliased)
            coefs <- matrix(NA, length(aliased), 4L, dimnames = list(cn,
                                                                     colnames(coefs)))
            coefs[!aliased, ] <- x$coefficients
        }
        printCoefmat(coefs, digits = digits, signif.stars = signif.stars,
                     na.print = "NA", ...)
    }
    cat("\n(Dispersion parameter for ", x$family$family, " family taken to be ",
        format(x$dispersion), ")\n\n", apply(cbind(paste(format(c("Null",
                                                                  "Residual"), justify = "right"), "deviance:"), format(unlist(x[c("null.deviance",
                                                                                                                                   "deviance")]), digits = max(5L, digits + 1L)), " on",
                                                   format(unlist(x[c("df.null", "df.residual")])), " degrees of freedom\n"),
                                             1L, paste, collapse = " "), sep = "")
    if (nzchar(mess <- naprint(x$na.action)))
        cat("  (", mess, ")\n", sep = "")
    cat("AIC: ", format(x$aic, digits = max(4L, digits + 1L)))
    cat("\n\nType of estimator:", x$type, get_type_description(x$type), "with alpha =", round(x$alpha, 2))
    cat("\n", "Number of Fisher Scoring iterations: ", x$iter, "\n", sep = "")
    correl <- x$correlation
    if (!is.null(correl)) {
        p <- NCOL(correl)
        if (p > 1) {
            cat("\nCorrelation of Coefficients:\n")
            if (is.logical(symbolic.cor) && symbolic.cor) {
                print(symnum(correl, abbr.colnames = NULL))
            } else {
                correl <- format(round(correl, 2L), nsmall = 2L,
                                 digits = digits)
                correl[!lower.tri(correl)] <- ""
                print(correl[-1, -p, drop = FALSE], quote = FALSE)
            }
        }
    }
    if (x$hd_correction) {
        cat("\nHigh-dimensionality correction applied with")
        cat("\nDimentionality parameter (kappa) =", round(x$kappa, 2))
        cat("\nEstimated signal strength (gamma^2) =", round(x$signal_strength, 2))
        cat("\nState evolution parameters (mu, b, sigma) =", paste0("(", paste(round(x$se_parameters[1:3], 2), collapse = ", "), ")"), "with max(|funcs|) =", max(abs(attr(x$se_parameters, "funcs"))), "\n")
    }
    invisible(x)
}

library("brglm2")

data("lizards", package = "brglm2")
liz_fm <- cbind(grahami, opalinus) ~ height + diameter + light + time
## ML fit = MDYPL fit with `alpha = 1`
liz_ml <- glm(liz_fm, family = binomial(), data = lizards,
              method = "mdyplFit", alpha = 1)
liz_ml0 <- glm(liz_fm, family = binomial(), data = lizards)
## liz_ml is the same fit as liz_ml0
summ_liz_ml <- summary(liz_ml)
summ_liz_ml0 <- summary(liz_ml0)
all.equal(coef(summ_liz_ml), coef(summ_liz_ml0))

to_check <- names(summ_liz_ml)
to_check <- to_check[!(to_check %in% c("hd_correction", "alpha", "type", "call"))]
for (nam in to_check) {
    expect_equal(summ_liz_ml[nam], summ_liz_ml0[nam])
}




liz_mdypl <- glm(liz_fm, family = binomial(), data = lizards,
                 method = "mdyplFit", alpha = 0.8761)
summ_mdypl <- summary(liz_mdypl)
summ_mdypl_c <- summary(liz_mdypl, hd_correction = TRUE)

eta_sloe <- sloe(liz_mdypl)
nn <- sum(weights(liz_mdypl))
p <- length(coef(liz_mdypl)) - 1
se_pars <- solve_se(p / nn, eta_sloe, liz_mdypl$alpha, intercept = coef(liz_mdypl)["(Intercept)"], start = c(0.5, 1, 1, unname(coef(liz_mdypl)["(Intercept)"])), corrupted = TRUE)
expect_equal(se_pars, summ_mdypl_c$se_parameters)

cc <- coef(liz_mdypl)
cc["(Intercept)"] <- se_pars[4]
cc[-1] <- cc[-1] / se_pars[1]
expect_equal(cc, coef(summ_mdypl_c)[, "Estimate"])

ttt <- brglm2:::taus(model.matrix(liz_mdypl))
ses_c <- c(NA, se_pars[3] / (sqrt(nn) * ttt * se_pars[1]))
expect_equal(unname(coef(summ_mdypl_c)[, "Std. Error"]), ses_c)
expect_equal(coef(summ_mdypl_c)[, "z value"], cc / ses_c)

tots <- lizards$grahami + lizards$opalinus
lizX <- model.matrix(liz_mdypl)
expect_equal(summ_mdypl_c$aic,
             brglm2:::logist_aic(liz_mdypl$y_adj, weights(liz_ml), drop(plogis(lizX %*% cc)), weights(liz_ml)) + 2 * liz_mdypl$rank)


## aliasing

X <- model.matrix(liz_ml)
glm(I(grahami / (grahami +  opalinus)) ~ -1 + X, weights = grahami + opalinus, data = lizards, family = binomial())

# Copyright (C) 2017- Ioannis Kosmidis

#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 or 3 of the License
#  (at your option).
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  A copy of the GNU General Public License is available at
#  http://www.r-project.org/Licenses/

## Some convenience names

#' @rdname brglmControl
#' @export
brglm_control <- brglmControl

#' @rdname brglmFit
#' @export
brglm_fit <- brglmFit

#' @rdname mdyplFit
#' @export
mdypl_fit <- mdyplFit

#' @rdname mdyplControl
#' @export
mdypl_control <- mdyplControl



#### Method conventions

#' @rdname residuals.brmultinom
#' @method residuals bracl
#' @export
residuals.bracl <- residuals.brmultinom

## Essentially stats:::drop1.glm() with `brglmFit()` calls in place of `glm.fit()` calls
#' @export
drop1.brglmFit <- function(object, scope, scale = 0, test = c("none", "Rao", "LRT",  "Chisq", "F"), k = 2, ...) {
    test <- match.arg(test)
    if (test == "Chisq")
        test <- "LRT"
    x <- model.matrix(object)
    n <- nrow(x)
    asgn <- attr(x, "assign")
    tl <- attr(object$terms, "term.labels")
    if (missing(scope))
        scope <- drop.scope(object)
    else {
        if (!is.character(scope))
            scope <- attr(terms(update.formula(object, scope)),
                          "term.labels")
        if (!all(match(scope, tl, 0L) > 0L))
            stop("scope is not a subset of term labels")
    }
    ndrop <- match(scope, tl)
    ns <- length(scope)
    rdf <- object$df.residual
    chisq <- object$deviance
    dfs <- numeric(ns)
    dev <- numeric(ns)
    score <- numeric(ns)
    y <- object$y
    if (is.null(y)) {
        y <- model.response(model.frame(object))
        if (!is.factor(y))
            storage.mode(y) <- "double"
    }
    wt <- object$prior.weights %||% rep.int(1, n)
    for (i in seq_len(ns)) {
        ii <- seq_along(asgn)[asgn == ndrop[i]]
        jj <- setdiff(seq(ncol(x)), ii)
        z <- brglmFit(x[, jj, drop = FALSE], y, wt, offset = object$offset,
                      family = object$family, control = object$control)
        dfs[i] <- z$rank
        dev[i] <- z$deviance
        if (test == "Rao") {
            r <- z$residuals
            w <- z$weights
            zz <- glm.fit(x, r, w)
            score[i] <- zz$null.deviance - zz$deviance
        }
    }
    scope <- c("<none>", scope)
    dfs <- c(object$rank, dfs)
    dev <- c(chisq, dev)
    if (test == "Rao") {
        score <- c(NA, score)
    }
    dispersion <- if (is.null(scale) || scale == 0)
                      summary(object, dispersion = NULL)$dispersion
    else scale
    fam <- object$family$family
    loglik <- if (fam == "gaussian") {
                  if (scale > 0)
                      dev/scale - n
                  else n * log(dev/n)
              }
              else dev/dispersion
    aic <- loglik + k * dfs
    dfs <- dfs[1L] - dfs
    dfs[1L] <- NA
    aic <- aic + (extractAIC(object, k = k)[2L] - aic[1L])
    aod <- data.frame(Df = dfs, Deviance = dev, AIC = aic, row.names = scope,
                      check.names = FALSE)
    if (all(is.na(aic)))
        aod <- aod[, -3]
    if (test == "LRT") {
        dev <- pmax(0, loglik - loglik[1L])
        dev[1L] <- NA
        nas <- !is.na(dev)
        LRT <- if (dispersion == 1)
                   "LRT"
               else "scaled dev."
        aod[, LRT] <- dev
        dev[nas] <- safe_pchisq(dev[nas], aod$Df[nas], lower.tail = FALSE)
        aod[, "Pr(>Chi)"] <- dev
    }
    else if (test == "Rao") {
        dev <- pmax(0, score)
        nas <- !is.na(dev)
        SC <- if (dispersion == 1)
                  "Rao score"
              else "scaled Rao sc."
        dev <- dev/dispersion
        aod[, SC] <- dev
        dev[nas] <- safe_pchisq(dev[nas], aod$Df[nas], lower.tail = FALSE)
        aod[, "Pr(>Chi)"] <- dev
    }
    else if (test == "F") {
        if (fam == "binomial" || fam == "poisson")
            warning(gettextf("F test assumes 'quasi%s' family",
                             fam), domain = NA)
        dev <- aod$Deviance
        rms <- dev[1L]/rdf
        dev <- pmax(0, dev - dev[1L])
        dfs <- aod$Df
        rdf <- object$df.residual
        Fs <- (dev/dfs)/rms
        Fs[dfs < 1e-04] <- NA
        P <- Fs
        nas <- !is.na(Fs)
        P[nas] <- safe_pf(Fs[nas], dfs[nas], rdf, lower.tail = FALSE)
        aod[, c("F value", "Pr(>F)")] <- list(Fs, P)
    }
    head <- c("Single term deletions", "\nModel:", deparse(formula(object)),
              if (!is.null(scale) && scale > 0) paste("\nscale: ",
                                                      format(scale), "\n"))
    class(aod) <- c("anova", "data.frame")
    attr(aod, "heading") <- head
    aod
}

#' @export
drop1.mdyplFit <- function(object, scope, scale = 0, test = c("none", "Rao", "LRT",  "Chisq", "F"), k = 2, ...) {
    test <- match.arg(test)
    if (test == "Chisq")
        test <- "LRT"
    x <- model.matrix(object)
    n <- nrow(x)
    asgn <- attr(x, "assign")
    tl <- attr(object$terms, "term.labels")
    if (missing(scope))
        scope <- drop.scope(object)
    else {
        if (!is.character(scope))
            scope <- attr(terms(update.formula(object, scope)),
                          "term.labels")
        if (!all(match(scope, tl, 0L) > 0L))
            stop("scope is not a subset of term labels")
    }
    ndrop <- match(scope, tl)
    ns <- length(scope)
    rdf <- object$df.residual
    chisq <- object$deviance
    dfs <- numeric(ns)
    dev <- numeric(ns)
    score <- numeric(ns)
    y <- object$y
    if (is.null(y)) {
        y <- model.response(model.frame(object))
        if (!is.factor(y))
            storage.mode(y) <- "double"
    }
    wt <- object$prior.weights %||% rep.int(1, n)
    for (i in seq_len(ns)) {
        ii <- seq_along(asgn)[asgn == ndrop[i]]
        jj <- setdiff(seq(ncol(x)), ii)
        z <- mdyplFit(x[, jj, drop = FALSE], y, wt, offset = object$offset,
                      family = object$family, control = object$control)
        dfs[i] <- z$rank
        dev[i] <- z$deviance
        if (test == "Rao") {
            r <- z$residuals
            w <- z$weights
            zz <- glm.fit(x, r, w)
            score[i] <- zz$null.deviance - zz$deviance
        }
    }
    scope <- c("<none>", scope)
    dfs <- c(object$rank, dfs)
    dev <- c(chisq, dev)
    if (test == "Rao") {
        score <- c(NA, score)
    }
    dispersion <- if (is.null(scale) || scale == 0)
                      summary(object, dispersion = NULL)$dispersion
    else scale
    fam <- object$family$family
    loglik <- if (fam == "gaussian") {
                  if (scale > 0)
                      dev/scale - n
                  else n * log(dev/n)
              }
              else dev/dispersion
    aic <- loglik + k * dfs
    dfs <- dfs[1L] - dfs
    dfs[1L] <- NA
    aic <- aic + (extractAIC(object, k = k)[2L] - aic[1L])
    aod <- data.frame(Df = dfs, Deviance = dev, AIC = aic, row.names = scope,
                      check.names = FALSE)
    if (all(is.na(aic)))
        aod <- aod[, -3]
    if (test == "LRT") {
        dev <- pmax(0, loglik - loglik[1L])
        dev[1L] <- NA
        nas <- !is.na(dev)
        LRT <- if (dispersion == 1)
                   "LRT"
               else "scaled dev."
        aod[, LRT] <- dev
        dev[nas] <- safe_pchisq(dev[nas], aod$Df[nas], lower.tail = FALSE)
        aod[, "Pr(>Chi)"] <- dev
    }
    else if (test == "Rao") {
        dev <- pmax(0, score)
        nas <- !is.na(dev)
        SC <- if (dispersion == 1)
                  "Rao score"
              else "scaled Rao sc."
        dev <- dev/dispersion
        aod[, SC] <- dev
        dev[nas] <- safe_pchisq(dev[nas], aod$Df[nas], lower.tail = FALSE)
        aod[, "Pr(>Chi)"] <- dev
    }
    else if (test == "F") {
        if (fam == "binomial" || fam == "poisson")
            warning(gettextf("F test assumes 'quasi%s' family",
                             fam), domain = NA)
        dev <- aod$Deviance
        rms <- dev[1L]/rdf
        dev <- pmax(0, dev - dev[1L])
        dfs <- aod$Df
        rdf <- object$df.residual
        Fs <- (dev/dfs)/rms
        Fs[dfs < 1e-04] <- NA
        P <- Fs
        nas <- !is.na(Fs)
        P[nas] <- safe_pf(Fs[nas], dfs[nas], rdf, lower.tail = FALSE)
        aod[, c("F value", "Pr(>F)")] <- list(Fs, P)
    }
    head <- c("Single term deletions", "\nModel:", deparse(formula(object)),
              if (!is.null(scale) && scale > 0) paste("\nscale: ",
                                                      format(scale), "\n"))
    class(aod) <- c("anova", "data.frame")
    attr(aod, "heading") <- head
    aod
}

# Copyright (C) 2020- Ioannis Kosmidis

#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 or 3 of the License
#  (at your option).
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  A copy of the GNU General Public License is available at
#  http://www.r-project.org/Licenses/

#'
#' brglm2: Bias Reduction in Generalized Linear Models
#'
#' Estimation and inference from generalized linear models using
#' implicit and explicit bias reduction methods (Kosmidis, 2014), and
#' other penalized maximum likelihood methods. Currently supported
#' methods include the mean bias-reducing adjusted scores approach in
#' Firth (1993) and Kosmidis & Firth (2009), the median bias-reduction
#' adjusted scores approach in Kenne Pagui et al. (2017), the
#' correction of the asymptotic bias in Cordeiro & McCullagh (1991),
#' the mixed bias-reduction adjusted scores approach in Kosmidis et al
#' (2020), maximum penalized likelihood with powers of the Jeffreys
#' prior as penalty, and maximum likelihood.
#'
#'
#' In the special case of generalized linear models for binomial,
#' Poisson and multinomial responses (both nominal and ordinal), mean
#' and median bias reduction and maximum penalized likelihood return
#' estimates with improved frequentist properties, that are also
#' always finite, even in cases where the maximum likelihood estimates
#' are infinite (e.g. complete and quasi-complete separation in
#' multinomial regression). Estimation in all cases takes place via a
#' modified Fisher scoring algorithm, and S3 methods for the
#' construction of confidence intervals for the reduced-bias estimates
#' are provided.
#'
#' The core model fitters are implemented by the functions
#' [brglm_fit()] (univariate generalized linear models),
#' [brmultinom()] (baseline category logit models for nominal
#' multinomial responses), [bracl()] (adjacent category logit models
#' for ordinal multinomial responses), and [brnb()] for negative
#' binomial regression.
#'
#' @details
#'
#'
#' The similarly named **brglm** R package can only handle generalized
#' linear models with binomial responses. Special care has been taken
#' when developing **brglm2** in order not to have conflicts when the
#' user loads **brglm2** and **brglm** simultaneously. The development
#' and maintenance of the two packages will continue in parallel,
#' until **brglm2** incorporates all **brglm** functionality and
#' provides an appropriate wrapper to the [brglm::brglm()] function.
#'
#' @author Ioannis Kosmidis `[aut, cre]` \email{ioannis.kosmidis@warwick.ac.uk}
#'
#' @seealso
#'
#' [brglm_fit()], [brmultinom()], [bracl()]
#'
#' @references
#'
#' Kosmidis I, Firth D (2021). Jeffreys-prior penalty, finiteness
#' and shrinkage in binomial-response generalized linear
#' models. *Biometrika*, **108**, 71-82. \doi{10.1093/biomet/asaa052}.
#'
#' Cordeiro G M, McCullagh P (1991). Bias correction in generalized
#' linear models. *Journal of the Royal Statistical Society. Series B
#' (Methodological)*, **53**, 629-643. \doi{10.1111/j.2517-6161.1991.tb01852.x}.
#'
#' Firth D (1993). Bias reduction of maximum likelihood estimates,
#' Biometrika, **80**, 27-38. \doi{10.2307/2336755}.
#'
#' Kenne Pagui E C, Salvan A, Sartori N (2017). Median bias
#' reduction of maximum likelihood estimates. *Biometrika*, **104**,
#' 923–938. \doi{10.1093/biomet/asx046}.
#'
#' Kosmidis I, Kenne Pagui E C, Sartori N (2020). Mean and median bias
#' reduction in generalized linear models. *Statistics and Computing*,
#' **30**, 43-59. \doi{10.1007/s11222-019-09860-6}.
#'
#' Kosmidis I, Firth D (2009). Bias reduction in exponential family
#' nonlinear models. *Biometrika*, **96**, 793-804. \doi{10.1093/biomet/asp055}.
#'
#' Kosmidis I, Firth D (2010). A generic algorithm for reducing
#' bias in parametric estimation. *Electronic Journal of Statistics*,
#' **4**, 1097-1112. \doi{10.1214/10-EJS579}.
#'
#' Kosmidis I (2014). Bias in parametric estimation: reduction and
#' useful side-effects. *WIRE Computational Statistics*, **6**,
#' 185-196. \doi{10.1002/wics.1296}.
#'
#' @docType package
#' @aliases brglm2-package
#' @name brglm2
#' @import stats
#' @import enrichwith
#' @import Matrix
#' @import MASS
#' @import nleqslv
#' @importFrom statmod gauss.quad
#' @importFrom graphics plot
#' @importFrom nnet class.ind
#' @importFrom numDeriv grad
#' @useDynLib brglm2
#'
"_PACKAGE"

## NAMESPACE should have import(stats), import(Matrix)


## Suggestion by Kurt Hornik to avoid a warning related to the binding
## of n which is evaluated by family$initialize
if (getRversion() >= "2.15.1") globalVariables(c("n", "lambda"))

#' @export
ordinal_superiority <- function(object, formula, data,
                                measure = c("gamma", "Delta"),
                                level = 0.95,
                                bc = FALSE) {
    UseMethod("ordinal_superiority")
}

#' @export
expo <- function(object, type = c("ML", "correction", "AS_median", "Lylesetal2012"), level = 0.95) {
    UseMethod("expo")
}

#' @export
plrtest <- function(object1, object2, ...) {
    UseMethod("plrtest")
}

## source(system.file("inst", "brglm0/brglm0.R", package = "brglm2"))
data("lizards", package = "brglm2")

lfit <- glm(cbind(grahami, opalinus) ~ height + diameter +
                       light + time, family = binomial(), data=lizards,
                   method = "brglmFit", type = "ML", epsilon = 1e-10, maxit = 1000)

info <- get_information_function(lfit)

start_methods <- c("AS_mixed", "AS_mean", "AS_median", "correction", "MPL_Jeffreys", "ML")
keep <- c("coef", "se", "ci")

## ML
or_est_ml <- exp(coef(lfit))
or_ses_ml <- sqrt(diag(diag(or_est_ml) %*% vcov(lfit) %*% diag(or_est_ml)))
ml <- expo(update(lfit, type = "ML"), type = "ML")
for (met in start_methods) {
    obj <- expo(update(lfit, type = met), type = "ML")
    expect_identical(ml[keep], obj[keep])
}
expect_equal(coef(ml), or_est_ml)
expect_equal(ml$se, or_ses_ml, check.attributes = FALSE)
expect_equal(ml$ci, exp(confint(lfit)))

## Lyles et al (2012)
lfit_mix <- update(lfit, type = "AS_mixed")
or_est_bc1 <- exp(coef(lfit_mix)) * exp(-diag(vcov(lfit_mix))/2)
or_ses_bc1 <- sqrt(diag(diag(or_est_bc1) %*% solve(info(log(or_est_bc1))) %*% diag(or_est_bc1)))
for (met in start_methods) {
    bc1 <- expo(update(lfit_mix, type = met), type = "Lylesetal2012")
    expect_equal(coef(bc1), or_est_bc1)
    expect_equal(bc1$se, or_ses_bc1, check.attributes = FALSE)
}

## correction*
or_est_bc2 <- exp(coef(lfit_mix)) / (1 + diag(vcov(lfit_mix))/2)
or_ses_bc2 <- sqrt(diag(diag(or_est_bc2) %*% solve(info(log(or_est_bc2))) %*% diag(or_est_bc2)))
for (met in start_methods) {
    bc2 <- expo(update(lfit_mix, type = met), type = "correction*")
    expect_equal(coef(bc2), or_est_bc2)
    expect_equal(bc2$se, or_ses_bc2, check.attributes = FALSE)
}

## correction+
or_est_bc3 <- exp(coef(lfit_mix)) * (1 - diag(vcov(lfit_mix))/2)
or_ses_bc3 <- sqrt(diag(diag(or_est_bc3) %*% solve(info(log(or_est_bc3))) %*% diag(or_est_bc3)))
for (met in start_methods) {
    bc3 <- expo(update(lfit, type = met), type = "correction+")
    expect_equal(coef(bc3), or_est_bc3)
    expect_equal(bc3$se, or_ses_bc3, check.attributes = FALSE)
}

## AS median
lfit_med <- update(lfit, type = "AS_median")
or_est_med <- exp(coef(lfit_med))
or_ses_med <- sqrt(diag(diag(or_est_med) %*% solve(info(log(or_est_med))) %*% diag(or_est_med)))
for (met in start_methods) {
    med <- expo(update(lfit_med, type = met), type = "AS_median")
    expect_equal(coef(med), or_est_med)
    expect_equal(med$se, or_ses_med, check.attributes = FALSE)
}

## starting from a glm object
expo_methods <- c("ML", "correction*", "correction+", "Lylesetal2012", "AS_median")
lfit_glm <- glm(cbind(grahami, opalinus) ~ height + diameter +
                    light + time, family = binomial(), data=lizards)
for (met in expo_methods) {
    out_brglmFit <- expo(lfit_med, type = met)[keep]
    out_glm <- expo(lfit_glm, type = met)[keep]
    expect_equal(out_brglmFit, out_glm, tolerance = 1e-06)
}

expect_stdout(print(expo(lfit_glm)), "Odds ratios")

## ## Interpretation
## set.seed(123)
## dat <- data.frame(y = rexp(10), x = rnorm(10))
## mod <- glm(y ~ x, family = Gamma("log"), data = dat)
## expo_mod <- expo(mod, type = "ML")
## expect_stdout(print(expo_mod), "Multiplicative effects to the mean")

## set.seed(111)
## dat <- data.frame(y = exp(rnorm(10)), x = rnorm(10))
## mod <- glm(y ~ x, family = inverse.gaussian("log"), data = dat)
## expo_mod <- expo(mod, type = "ML")
## expect_stdout(print(expo_mod), "Multiplicative effects to the mean")

## set.seed(111)
## dat <- data.frame(x = rexp(10), y = rpois(10, 10))
## mod <- glm(y ~ x, family = poisson("log"), data = dat)
## expo_mod <- expo(mod, type = "ML")
## expect_stdout(print(expo_mod), "Multiplicative effects to the mean")

## set.seed(111)
## dat <- data.frame(x = rexp(100, 0.5))
## dat$y <- rbinom(100, 1, exp(-1 -0.2 * dat$x))
## mod <- glm(y ~ x, family = binomial("log"), data = dat)
## expo_mod <- expo(mod, type = "ML")
## expect_stdout(print(expo_mod), "Relative risks")


## library(parallel)
## nsimu <- 1000
## set.seed(123)
## Y <- simulate(lfit, nsimu)
## true_psi <- exp(coef(lfit))
## X <- model.matrix(lfit)

## methods <- c("ML", "correction+", "correction*", "Lylesetal2012", "AS_median")
## results <- as.list(numeric(length(methods)))
## names(results) <- methods
## for (met in methods) {
##     results[[met]] <- mclapply(1:nsimu, function(k) {
##         temp_data <- lizards
##         temp_data[c("grahami", "opalinus")] <- Y[[k]]
##         mod <- update(lfit, data = temp_data)
##         expo(mod, met)
##     }, mc.cores = 8)
## }

## get_bias <- function(res, truth) rowMeans(sapply(res, function(x) x$coef - truth))
## get_pu <- function(res, truth) rowMeans(sapply(res, function(x) x$coef < truth))
## get_mse <- function(res, truth) rowMeans(sapply(res, function(x) (x$coef - truth)^2))
## get_avgv <- function(res, truth) rowMeans(sapply(res, function(x) x$se^2))
## get_pu <- function(res, truth) rowMeans(sapply(res, function(x) x$coef < truth))
## get_coverage <- function(res, truth) {
##     rowMeans(sapply(res, function(x) {
##         (x$ci[, 1] < truth) & (x$ci[, 2] > truth)
##     }))
## }

## round(sapply(results, get_bias, truth = true_psi), 2)
## round(sapply(results, get_mse, truth = true_psi), 2)
## round(sapply(results, get_pu, truth = true_psi), 2)
## round(sapply(results, get_coverage, truth = true_psi), 2)
## vars <- sapply(results, get_mse, truth = true_psi) - sapply(results, get_bias, truth = true_psi)^2
## avgvars <- sapply(results, get_avgv, truth = true_psi)

# Copyright (C) 2016- Ioannis Kosmidis

#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 or 3 of the License
#  (at your option).
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  A copy of the GNU General Public License is available at
#  http://www.r-project.org/Licenses/

unless_null <- function(x, if_null) {
    if (is.null(x)) {
        if_null
    } else {
        x
    }
}



get_type_description <- function(type, parenthesized = TRUE) {
    pp <- function(txt) {
        ifelse(parenthesized, paste0("(", txt, ")"), txt)
    }
    switch(type,
           "ML" = pp("maximum likelihood"),
           "correction" = pp("bias correction"),
           "AS_mean" = pp("mean bias-reducing adjusted score equations"),
           "AS_median" = pp("median bias-reducing adjusted score equations"),
           "AS_mixed" = pp("mixed bias-reducing adjusted score equations"),
           "MPL_Jeffreys" = pp("maximum Jeffreys' prior penalized likelihood"),
           "MPL_DY" = pp("maximum Diaconis-Ylvisaker prior penalized likelihood"),
           "Lylesetal2012" = pp("Lyles et al., 2012; doi: 10.1016/j.jspi.2012.05.005"),
           "correction*" = pp("explicit mean bias correction with a multiplicative adjustment"),
           "correction+" = pp("explicit mean bias correction with an additive adjustment")
           )
}


## Just the inverse logistic link aoviding conventions in `plogis()`
plogis2 <- function(x) 1 / (1 + exp(-x))

## Vectorized version (in x and b) of the proximal operator arg min_u
## {b * log(1 + e^u) + (x - u)^2 / 2}
prox <- function(x, b, tol) {
    u <- 0
    g0 <- x - b / 2
    while (!isTRUE(all(abs(g0) < tol))) {
        pr <- plogis2(u)
        g0 <- x - u - b * pr
        u <- u + g0 / (b * pr * (1 - pr) + 1)
    }
    u
}

#' Defunct Functions in package \pkg{brglm2}
#'
#'  The functions or variables listed here are no longer part of
#'  \pkg{brglm2}.
#'
#'
#' @param ... arguments to be passed to functions and methods.
#'
#'
#' @details
#'
#' * [detect_separation()]: This function is defunct from \pkg{brglm2}
#' since version 0.8.0. A new version of [detect_separation()] is now
#' maintained in the
#' [\pkg{detectseparation}](https://cran.r-project.org/package=detectseparation)
#' R package.
#'
#' * [check_infinite_estimates()] is defunct from \pkg{brglm2} since
#' version 0.8.0. An new version of [check_infinite_estimates()] is
#' now maintained in the
#' [\pkg{detectseparation}](https://cran.r-project.org/package=detectseparation)
#' R package.
#'
#' @name brglm2-defunct
NULL

#' @rdname brglm2-defunct
#' @export
check_infinite_estimates <- function(...) {
    function_moved_to_new_package(gsub("\\(|\\)", "", deparse(match.call()[1])),
                                  "0.8.0",
                                  "brglm2",
                                  "detectseparation")
}

#' @rdname brglm2-defunct
#' @export
detect_separation <- function(...) {
    function_moved_to_new_package(gsub("\\(|\\)", "", deparse(match.call()[1])),
                                  "0.8.0",
                                  "brglm2",
                                  "detectseparation")
}

## endometrial data from Heinze \& Schemper (2002) (see ?endometrial)
data("endometrial", package = "brglm2")
expect_warning({
    endometrial_ml <- glm(HG ~ I(-NV) + PI + EH, data = endometrial,
                          family = binomial("cloglog"))
})

expect_error(cie <- check_infinite_estimates(endometrial_ml))

# Copyright (C) 2025- Ioannis Kosmidis, Federico Boiocchi, Philipp Sterzinger

#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 or 3 of the License
#  (at your option).
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  A copy of the GNU General Public License is available at
#  http://www.r-project.org/Licenses/

#' Solve the MDYPL state evolution equations with or without
#' intercept, with signal strength or contaminated signal strength
#'
#' @param kappa asymptotic ratio of columns/rows of the design
#'     matrix. `kappa` should be in `(0, 1)`.
#' @param ss square root of signal strength or of corrupted signal
#'     strength, depending on whether `corrupted = TRUE` or not. See
#'     Details.
#' @param alpha the shrinkage parameter of the MDYPL
#'     estimator. `alpha` should be in `(0, 1)`.
#' @param intercept if `NULL` (default) then the MDYPL state evolution
#'     equations for the model with no intercept parameter are
#'     solved. If a real then the equations for the models with
#'     intercept parameter equal to `intercept` are solved. See
#'     Details.
#' @param start a vector with starting values for `mu`, `b`,`sigma`
#'     (and `iota` if `intercept` is numeric).
#' @param corrupted if `FALSE` (default) then `ss` is the square root
#'     of the signal strength and `intercept`, if numeric, is the
#'     oracle intercept value. If `TRUE`, then `ss` is the square root
#'     of the corrupted signal strength, and `intercept`, if numeric,
#'     is the limit of the estimator computed by [mdyplFit()] with
#'     shrinkage parameter `alpha`. See Details.
#' @param gh A list with the Gauss-Hermite quadrature nodes and
#'     weights, as returned from `statmod::gauss.quad()` with `kind =
#'     "hermite"`. Default is `NULL`, in which case `gh` is set to
#'     `statmod::gauss.quad(200, kind = "hermite")`.
#' @param prox_tol tolerance for the computation of the proximal
#'     operator; default is `1e-10`.
#' @param transform if `TRUE` (default), the optimization is with
#'     respect to `log(mu)`, `log(b)`,`log(sigma)`, (and `iota` if
#'     `intercept` is numeric). If `FALSE`, then it is over `mu`, `b`,
#'     `sigma` (and `iota` if `intercept` is numeric). The solution is
#'     returned in terms of the latter set, regardless of how
#'     optimization took place.
#' @param init_iter how many iterations of [optim()] should we make to
#'     get starting values for [nleqslv::nleqslv()]? Default is `50`,
#'     but can also be `0` in which case `start` is directly passed to
#'     `nleqslv:nleqslv()`. `init_iter = "only"` results in only
#'     [optim()] being used. See Details.
#' @param init_method The method to be passed to [optim()]. Default is
#'     `"Nelder-Mead"`.
#' @param ... further arguments to be passed to [nleqslv::nleqslv()],
#'     unless `init_iter = "only"`, in which case `...` is further
#'     arguments to be passed to [optim()].
#'
#' @details
#'
#' `init_iter` iterations of [optim()] with `method = init_method` are
#' used towards minimizing `sum(se)^2`, where `se` is a vector of the
#' state evolution functions. The solution is then passed to
#' `nleqslv::nleqslv()` for a more aggressive iteration. The state
#' evolution equations are given in expressions (8) (model without
#' intercept) and expression (15) (model with intercept) in Sterzinger
#' & Kosmidis (2024).
#'
#' If `corrupted = FALSE` (default), then `ss` is the square root of
#' the signal strength, which is the limit \eqn{\gamma^2} of
#' \eqn{var(X \beta)}. If `corrupted = TRUE`, then `ss` is the square
#' root of the corrupted signal strength which is the limit
#' \eqn{\nu^2} of \eqn{var(X hat(beta)(\alpha))}, where
#' \eqn{hat(\beta)(\alpha)} is the maximum Diaconis-Ylvisaker prior
#' penalized likelihood (MDYPL) estimator as computed by [mdyplFit()]
#' with shrinkage parameter \eqn{alpha}.
#'
#' If `intercept = NULL`, then the state evolution equations are
#' solved for the model without intercept. If `intercept` is a real
#' number, then the state evolution equations for the model with
#' intercept are solved (i.e. with predictor \eqn{\eta_i = \theta +
#' x_i^T \beta}). In that case, what `intercept` represents depends on
#' the value of `corrupted`. If `corrupted = FALSE`, `intercept`
#' represents the oracle value of \eqn{\theta}, otherwise it represents
#' the limit `iota` of the MDYPL estimator of \eqn{\theta} as computed by
#' [mdyplFit()] with shrinkage parameter `alpha`.
#'
#' Note that `start` is always for `mu`, `b`,`sigma`, as is the
#' result, regardless whether `transform = TRUE` or
#' not. Transformations during optimization are done internally.
#'
#' @return
#'
#' If `intercept = NULL`, a vector with the values of `mu`,
#' `b`,`sigma`. Otherwise, a vector with the values of `mu`,
#' `b`,`sigma`, and `iota`, if `corrupted = FALSE`, or the value of
#' the intercept otherwise. The vector has attributes the state
#' evolution functions at the solution (`"funcs"`), the number of
#' iterations used by the last optimization method (`"iter"`), any
#' messages from the last optimization method (`"message"`), and
#' information on the optimization methods used
#' (`"optimization-chain"`).
#'
#' @author Ioannis Kosmidis `[aut, cre]` \email{ioannis.kosmidis@warwick.ac.uk}, Federico Boiocchi `[ctb]` \email{federico.boiocchi@gmail.com}, Philipp Sterzinger `[ctb, earlier Julia code by]` \email{P.Sterzinger@lse.ac.uk}
#'
#'
#' @references
#'
#' Zhao Q, Sur P, Cand\`es E J (2022). The asymptotic distribution of
#' the MLE in high-dimensional logistic models: Arbitrary
#' covariance. *Bernoulli*, **28**, 1835–1861. \doi{10.3150/21-BEJ1401}.
#'
#' Sterzinger P, Kosmidis I (2024). Diaconis-Ylvisaker prior
#' penalized likelihood for \eqn{p/n \to \kappa \in (0,1)} logistic
#' regression. *arXiv*:2311.07419v2, \url{https://arxiv.org/abs/2311.07419}.
#'
#'
#' @examples
#'
#' ## Reproducing Table 13 of Zhao et al. (2022, DOI: 10.3150/21-BEJ1401)
#' \donttest{
#'
#' thetas <- c(0, 0.5, 1, 2, 2.5)
#' gamma0 <- 5
#' pars3 <- matrix(NA, length(thetas), 3)
#' pars4 <- matrix(NA, length(thetas), 4)
#' colnames(pars4) <- c("I_mu", "I_b", "I_sigma", "I_iota")
#' colnames(pars3) <- c("II_mu", "II_b", "II_sigma")
#' for (i in seq_along(thetas)) {
#'     start3 <- c(0.5, 1, 1)
#'     pars3[i, ] <- solve_se(kappa = 0.2, ss = sqrt(5 + thetas[i]^2),
#'                            alpha = 1, start = start3, init_iter = 0)
#'     start4 <- c(pars3[i, ], thetas[i])
#'     pars4[i, ] <- solve_se(kappa = 0.2, ss = sqrt(5), intercept = thetas[i],
#'                            alpha = 1, start = start4, init_iter = 0)
#' }
#'
#' cbind(pars3, pars4)
#'
#' }
#' @export
solve_se <- function(kappa, ss, alpha, intercept = NULL, start, corrupted = FALSE, gh = NULL, prox_tol = 1e-10, transform = TRUE, init_method = "Nelder-Mead", init_iter = 50, ...) {
    no_intercept <- is.null(intercept)
    npar <- 3 + !no_intercept
    stopifnot(length(start) == npar)
    is_corrupted <- isTRUE(corrupted)
    init_solver <- if (is_corrupted) optim_se_corrupted else optim_se
    main_solver <- if (is_corrupted) nleqslv_se_corrupted else nleqslv_se
    if (is.null(gh))
        gh <- gauss.quad(200, kind = "hermite")
    if (isTRUE(init_iter == "only")) {
        start <- init_solver(kappa, ss, alpha, intercept, start, gh, prox_tol, method = init_method, ...)
        opt_chain <- paste0("optim(method = ", init_method, ")")
    } else {
        if (init_iter > 0) {
            start <- init_solver(kappa, ss, alpha, intercept, start, gh, prox_tol, method = init_method, control = list(maxit = init_iter));
            opt_chain <- paste0("optim(method = ", init_method, ", maxit = ", init_iter, ") -> ")
        } else {
            opt_chain <- ""
        }
        start <- main_solver(kappa, ss, alpha, intercept, start, gh, prox_tol, transform, ...)
        opt_chain <- paste0(opt_chain, "nleqslv()")
    }
    attr(start, "optimization-chain") <- opt_chain
    start
}

se_funcs <- function(kappa, ss, alpha, intercept = NULL, iota = NULL,
                     gh, prox_tol, corrupted = FALSE, transform = TRUE) {
    ## corrupted signal strength
    if (isTRUE(corrupted)) {
        if (is.null(iota)) {
            if (transform) {
                g <- function(pars) {
                    pars <- exp(pars)
                    mu <- pars[1]
                    b <- pars[2]
                    sigma <- pars[3]
                    gamma <- sqrt(ss^2 - kappa * sigma^2) / mu
                    if (is.na(gamma)) return(rep(NA, 3))
                    se0(mu = mu, b = b, sigma = sigma, kappa = kappa, gamma = gamma, alpha = alpha, gh = gh, prox_tol = prox_tol)
                }
            } else {
                g <- function(pars) {
                    mu <- pars[1]
                    b <- pars[2]
                    sigma <- pars[3]
                    gamma <- sqrt(ss^2 - kappa * sigma^2) / mu
                    if (is.na(gamma)) return(rep(NA, 3))
                    se0(mu = mu, b = b, sigma = sigma, kappa = kappa, gamma = gamma, alpha = alpha, gh = gh, prox_tol = prox_tol)
                }
            }
        } else {
            no_int <- 1:3
            if (transform) {
                g <- function(pars) {
                    pars[no_int] <- exp(pars[no_int])
                    mu <- pars[1]
                    b <- pars[2]
                    sigma <- pars[3]
                    gamma <- sqrt(ss^2 - kappa * sigma^2) / mu
                    if (is.na(gamma)) return(rep(NA, 4))
                    se1(mu = mu, b = b, sigma = sigma, iota = iota, kappa = kappa, gamma = gamma, alpha = alpha, intercept = pars[4], gh = gh, prox_tol = prox_tol)
                }
            } else {
                g <- function(pars) {
                    mu <- pars[1]
                    b <- pars[2]
                    sigma <- pars[3]
                    gamma <- sqrt(ss^2 - kappa * sigma^2) / mu
                    if (is.na(gamma)) return(rep(NA, 4))
                    se1(mu = mu, b = b, sigma = sigma, iota = iota, kappa = kappa, gamma = gamma, alpha = alpha, intercept = pars[4], gh = gh, prox_tol = prox_tol)
                }
            }
        }
    } else {
        ## signal strength
        if (is.null(intercept)) {
            if (transform) {
                g <- function(pars) {
                    pars <- exp(pars)
                    se0(mu = pars[1], b = pars[2], sigma = pars[3], kappa = kappa, gamma = ss, alpha = alpha, gh = gh, prox_tol = prox_tol)
                }
            } else {
                g <- function(pars) {
                    se0(mu = pars[1], b = pars[2], sigma = pars[3], kappa = kappa, gamma = ss, alpha = alpha, gh = gh, prox_tol = prox_tol)
                }
            }
        } else {
            no_int <- 1:3
            if (transform) {
                g <- function(pars) {
                    pars[no_int] <- exp(pars[no_int])
                    se1(mu = pars[1], b = pars[2], sigma = pars[3], iota = pars[4], kappa = kappa, gamma = ss, alpha = alpha, intercept = intercept, gh = gh, prox_tol = prox_tol)
                }
            } else {
                g <- function(pars) {
                    se1(mu = pars[1], b = pars[2], sigma = pars[3], iota = pars[4], kappa = kappa, gamma = ss, alpha = alpha, intercept = intercept, gh = gh, prox_tol = prox_tol)
                }
            }
        }
    }
    g
}

nleqslv_se <- function(kappa, gamma, alpha, intercept = NULL, start, gh = NULL, prox_tol = 1e-10, transform = TRUE, ...) {
    no_intercept <- is.null(intercept)
    g <- se_funcs(kappa, gamma, alpha, intercept, iota = NULL, gh, prox_tol, corrupted = FALSE, transform)
    start <- c(if (transform) log(start[1:3]) else start[1:3],
               if (no_intercept) NULL else start[4])
    res <- nleqslv(start, g, ...)
    if (transform) {
        if (no_intercept) {
            soln <- exp(res$x)
        } else {
            soln <- c(exp(res$x[1:3]), res$x[4])
        }
    } else {
        soln <- res$x
    }
    attr(soln, "funcs") <- res$fvec
    attr(soln, "iter") <- res$iter
    attr(soln, "message") <- res$message
    attr(soln, "nleqslv_termination_code") <- res$termcd
    soln
}

nleqslv_se_corrupted <- function(kappa, nu, alpha, iota = NULL, start, gh = NULL, prox_tol = 1e-10, transform = TRUE, ...) {
    no_intercept <- is.null(iota)
    g <- se_funcs(kappa, nu, alpha, intercept = NULL, iota, gh, prox_tol, corrupted = TRUE, transform)
    start <- c(if (transform) log(start[1:3]) else start[1:3],
               if (no_intercept) NULL else start[4])
    suppressWarnings(res <- nleqslv(start, g, ...))
    if (transform) {
        if (no_intercept) {
            soln <- exp(res$x)
        } else {
            soln <- c(exp(res$x[1:3]), res$x[4])
        }
    } else {
        soln <- res$x
    }
    attr(soln, "funcs") <- res$fvec
    attr(soln, "iter") <- res$iter
    attr(soln, "message") <- res$message
    attr(soln, "nleqslv_termination_code") <- res$termcd
    soln
}


optim_se <- function(kappa, gamma, alpha, intercept = NULL, start, gh = NULL, prox_tol = 1e-10, transform = TRUE, ...) {
    no_intercept <- is.null(intercept)
    g <- se_funcs(kappa, gamma, alpha, intercept, iota = NULL, gh, prox_tol, corrupted = FALSE, transform = TRUE)
    npar <- 3 + !no_intercept
    obj <- function(pars) {
        sum(g(pars)^2)
    }
    res <- optim(start, obj, ...)
    if (no_intercept) {
        soln <- exp(res$par)
    } else {
        soln <- c(exp(res$par[1:3]), res$par[4])
    }
    attr(soln, "funcs") <- g(res$par)
    attr(soln, "iter") <- res$counts
    attr(soln, "convergence") <- res$convergence
    attr(soln, "message") <- res$message
    soln
}

optim_se_corrupted <- function(kappa, nu, alpha, iota = NULL, start, gh = NULL, prox_tol = 1e-10, transform = TRUE, ...) {
    no_intercept <- is.null(iota)
    g <- se_funcs(kappa, nu, alpha, intercept = NULL, iota, gh, prox_tol, corrupted = TRUE, transform = TRUE)
    start <- c(log(start[1:3]), if (no_intercept) NULL else start[4])
    obj <- function(pars) {
        sum(g(pars)^2)
    }
    suppressWarnings(res <- optim(start, obj, ...))
    if (no_intercept) {
        soln <- exp(res$par)
    } else {
        soln <- c(exp(res$par[1:3]), res$par[4])
    }
    suppressWarnings(attr(soln, "funcs") <- g(res$par))
    attr(soln, "iter") <- res$counts
    attr(soln, "convergence") <- res$convergence
    attr(soln, "message") <- res$message
    soln
}

## Logistic ridge regression; currently not exported
solve_se_ridge <- function(kappa, ss, lambda, start, gh = NULL, prox_tol = 1e-10, transform = TRUE, init_method = "Nelder-Mead", init_iter = 50, ...) {
    init_solver <- optim_se_ridge
    main_solver <- nleqslv_se_ridge
    stopifnot(length(start) == 3)
    if (is.null(gh))
        gh <- gauss.quad(200, kind = "hermite")
    if (isTRUE(init_iter == "only")) {
        start <- init_solver(kappa, ss, lambda, start, gh, prox_tol, method = init_method, ...)
        opt_chain <- paste0("optim(method = ", init_method, ")")
    } else {
        if (init_iter > 0) {
            start <- init_solver(kappa, ss, lambda, start, gh, prox_tol, method = init_method, control = list(maxit = init_iter));
            opt_chain <- paste0("optim(method = ", init_method, ", maxit = ", init_iter, ") -> ")
        } else {
            opt_chain <- ""
        }
        start <- main_solver(kappa, ss, lambda, start, gh, prox_tol, transform, ...)
        opt_chain <- paste0(opt_chain, "nleqslv()")
    }
    attr(start, "optimization-chain") <- opt_chain
    start
}

nleqslv_se_ridge <- function(kappa, ss, lambda, start, gh = NULL, prox_tol = 1e-10, transform = TRUE, ...) {
    if (transform) {
        start <- log(start)
        g <- function(pars) {
            pars <- exp(pars)
            se0_ridge(pars[1], pars[2], pars[3], kappa, ss, lambda, gh, prox_tol)
        }
    } else {
        g <- function(pars) {
            se0_ridge(pars[1], pars[2], pars[3], kappa, ss, lambda, gh, prox_tol)
        }
    }
    res <- nleqslv(start, g, ...)
    if (transform) {
        soln <- exp(res$x)
    } else {
        soln <- res$x
    }
    attr(soln, "funcs") <- res$fvec
    attr(soln, "iter") <- res$iter
    attr(soln, "message") <- res$message
    attr(soln, "nleqslv_termination_code") <- res$termcd
    soln
}

optim_se_ridge <- function(kappa, ss, lambda, start, gh = NULL, prox_tol = 1e-10, transform = TRUE, ...) {
    if (transform) {
        start <- log(start)
        g <- function(pars) {
            pars <- exp(pars)
            se0_ridge(pars[1], pars[2], pars[3], kappa, ss, lambda, gh, prox_tol)
        }
    } else {
        g <- function(pars) {
            se0_ridge(pars[1], pars[2], pars[3], kappa, ss, lambda, gh, prox_tol)
        }
    }
    obj <- function(pars) {
        sum(g(pars)^2)
    }
    res <- optim(start, obj, ...)
    if (transform) {
        soln <- exp(res$par)
    } else {
        soln <- res$par
    }
    attr(soln, "funcs") <- g(res$par)
    attr(soln, "iter") <- res$counts
    attr(soln, "convergence") <- res$convergence
    attr(soln, "message") <- res$message
    soln
}

# Copyright (C) 2016- Ioannis Kosmidis

#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 or 3 of the License
#  (at your option).
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  A copy of the GNU General Public License is available at
#  http://www.r-project.org/Licenses/


#' Bias reduction for multinomial response models using the
#' Poisson trick.
#'
#' [brmultinom()] is a wrapper of [brglmFit()] that fits multinomial
#' regression models using implicit and explicit bias reduction
#' methods. See Kosmidis & Firth (2011) for details.
#'
#' @inheritParams nnet::multinom
#' @param control a list of parameters for controlling the fitting
#'     process. See [brglmControl()] for details.
#' @param ref the reference category to use for multinomial
#'     regression. Either an integer, in which case
#'     `levels(response)[ref]` is used as a baseline, or a character
#'     string. Default is 1.
#' @param x should the model matrix be included with in the result
#'     (default is `TRUE`).
#' @param ... arguments to be used to form the default `control`
#'     argument if it is not supplied directly.
#'
#' @details
#'
#' The models [brmultinom()] handles are also known as
#' baseline-category logit models (see, Agresti, 2002, Section 7.1),
#' because they model the log-odds of every category against a
#' baseline category. The user can control which baseline (or
#' reference) category is used via the `ref`. By default
#' [brmultinom()] uses the first category as reference.
#'
#' The maximum likelihood estimates for the parameters of
#' baseline-category logit models have infinite components with
#' positive probability, which can result in problems in their
#' estimation and the use of inferential procedures (e.g. Wad
#' tests). Albert and Andreson (1984) have categorized the possible
#' data patterns for such models into the exclusive and exhaustive
#' categories of complete separation, quasi-complete separation and
#' overlap, and showed that infinite maximum likelihood estimates
#' result when complete or quasi-complete separation occurs.
#'
#' The adjusted score approaches to bias reduction that [brmultinom()]
#' implements for `type = "AS_mean"` and `type = "AS_median"` are
#' alternatives to maximum likelihood that result in estimates with
#' smaller asymptotic mean and median bias, respectively, that are
#' also *always* finite, even in cases of complete or quasi-complete
#' separation.
#'
#' [brmultinom()] is a wrapper of [brglmFit()] that fits multinomial
#' logit regression models through the 'Poisson trick' (see, for
#' example, Palmgren, 1981; Kosmidis & Firth, 2011).
#'
#' The implementation relies on the construction of an extended model
#' matrix for the log-linear model and constraints on the sums of the
#' Poisson means. Specifically, a log-linear model is fitted on a
#' [Kronecker
#' product](https://en.wikipedia.org/wiki/Kronecker_product) of the
#' original model matrix `X` implied by the formula, augmented by
#' `nrow(X)` dummy variables.
#'
#' The extended model matrix is sparse, and the
#' [\pkg{Matrix}](https://cran.r-project.org/package=Matrix) package is
#' used for its effective storage.
#'
#' While [brmultinom()] can be used for analyses using multinomial
#' regression models, the current implementation is more of a proof of
#' concept and is not expected to scale well with either of `nrow(X)`,
#' `ncol(X)` or the number of levels in the categorical response.
#'
#' @author Ioannis Kosmidis `[aut, cre]` \email{ioannis.kosmidis@warwick.ac.uk}
#'
#' @seealso [nnet::multinom()], [bracl()] for adjacent category logit models for ordinal responses
#'
#' @references
#'
#' Kosmidis I, Kenne Pagui E C, Sartori N (2020). Mean and median bias
#' reduction in generalized linear models. *Statistics and Computing*,
#' **30**, 43-59. \doi{10.1007/s11222-019-09860-6}.
#'
#' Agresti A (2002). *Categorical data analysis* (2nd edition). Wiley
#' Series in Probability and Statistics. Wiley.
#'
#' Albert A, Anderson J A (1984). On the Existence of Maximum
#' Likelihood Estimates in Logistic Regression Models. *Biometrika*,
#' **71** 1--10. \doi{10.2307/2336390}.
#'
#' Kosmidis I, Firth D (2011). Multinomial logit bias reduction
#' via the Poisson log-linear model. *Biometrika*, **98**, 755-759.
#' \doi{10.1093/biomet/asr026}.
#'
#' Palmgren, J (1981). The Fisher Information Matrix for Log Linear
#' Models Arguing Conditionally on Observed Explanatory
#' Variables. *Biometrika*, **68**, 563-566.
#' \doi{10.1093/biomet/68.2.563}.
#'
#' @examples
#' ## The housing data analysis from ?MASS::housing
#'
#' data("housing", package = "MASS")
#'
#' # Maximum likelihood using nnet::multinom
#' houseML1nnet <- nnet::multinom(Sat ~ Infl + Type + Cont, weights = Freq,
#'                                data = housing)
#' # Maximum likelihood using brmultinom with baseline category 'Low'
#' houseML1 <- brmultinom(Sat ~ Infl + Type + Cont, weights = Freq,
#'                        data = housing, type = "ML", ref = 1)
#' # The estimates are numerically the same as houseML0
#' all.equal(coef(houseML1nnet), coef(houseML1), tolerance = 1e-04)
#'
#' # Maximum likelihood using brmultinom with 'High' as baseline
#' houseML3 <- brmultinom(Sat ~ Infl + Type + Cont, weights = Freq,
#'                       data = housing, type = "ML", ref = 3)
#' # The fitted values are the same as houseML1
#' all.equal(fitted(houseML3), fitted(houseML1), tolerance = 1e-10)
#'
#' # Bias reduction
#' houseBR3 <- update(houseML3, type = "AS_mean")
#' # Bias correction
#' houseBC3 <- update(houseML3, type = "correction")
#'
#' ## Reproducing Bull et al. (2002, Table 3)
#' data("hepatitis", package = "brglm2")
#'
#' # Construct a variable with the multinomial categories according to
#' # the HCV and nonABC columns
#' hepat <- hepatitis
#' hepat$type <- with(hepat, factor(1 - HCV * nonABC + HCV + 2 * nonABC))
#' hepat$type <- factor(hepat$type, labels = c("noDisease", "C", "nonABC"))
#' contrasts(hepat$type) <- contr.treatment(3, base = 1)
#'
#' # Maximum likelihood estimation fails to converge because some estimates are infinite
#' hepML <- brmultinom(type ~ group * time, data = hepat, weights = counts, type = "ML", slowit = 0.1)
#'
#' # Mean bias reduction returns finite estimates
#' hep_meanBR <- brmultinom(type ~ group * time, data = hepat, weights = counts, type = "AS_mean")
#' # The estimates in Bull et al. (2002, Table 3, DOI: 10.1016/S0167-9473(01)00048-2)
#' coef(hep_meanBR)
#'
#' # Median bias reduction also returns finite estimates, which are a bit larger in absolute value
#' hep_medianBR <- brmultinom(type ~ group * time, data = hepat, weights = counts, type = "AS_median")
#' coef(hep_medianBR)
#'
#' @export
brmultinom <- function(formula, data, weights, subset, na.action,
                       contrasts = NULL, ref = 1,
                       model = TRUE, x = TRUE,
                       control = list(...), ...) {
    call <- match.call()
    if (missing(data)) {
        data <- environment(formula)
    }
    mf <- match.call(expand.dots = FALSE)
    m <- match(c("formula", "data", "subset", "weights", "na.action", "offset"), names(mf), 0L)
    mf <- mf[c(1L, m)]
    mf$drop.unused.levels <- TRUE
    mf[[1L]] <- quote(model.frame)
    mf <- eval.parent(mf)
    Terms <- attr(mf, "terms")
    X <- model.matrix(Terms, mf, contrasts)
    Xcontrasts <- attr(X, "contrasts")
    Y <- model.response(mf, "any")
    ## The chunk of code between +BEGIN and +END has been adopted from
    ## nnet::multinom
    ##+BEGIN
    if (!is.matrix(Y)) {
        Y <- as.factor(Y)
        lev <- levels(Y)
    } else {
        lev <- colnames(Y)
    }
    w <- model.weights(mf)
    if (length(w) == 0L)
        if (is.matrix(Y))
            w <- rep(1, dim(Y)[1L])
        else w <- rep(1, length(Y))
    if (is.factor(Y)) {
        counts <- table(Y)
        if (any(counts == 0L)) {
            empty <- lev[counts == 0L]
            warning(sprintf(ngettext(length(empty), "group %s is empty",
                                     "groups %s are empty"), paste(sQuote(empty),
                                                                   collapse = " ")), domain = NA)
            Y <- factor(Y, levels = lev[counts > 0L])
            lev <- lev[counts > 0L]
        }
        if (length(lev) < 2L)
            stop("need two or more classes to fit a multinomial logit model")
        ## if (length(lev) == 2L)
        ##     Y <- as.integer(Y) - 1
        else Y <- nnet::class.ind(Y)
    }
    ##+END

    ncat <- if (is.matrix(Y)) ncol(Y) else length(lev)
    nvar <- ncol(X)

    if (is.character(ref)) {
        refc <- ref
        ref <- match(ref, lev, nomatch = NA)
        if (is.na(ref)) stop("reference category ", refc, " does not exist")
    }

    keep <- w > 0
    nkeep <- sum(keep)
    fixed_totals <- rep(seq.int(nkeep), ncat)

    ## Set up the model matrix for the poisson fit
    Xnuisance <- Matrix::Diagonal(nkeep)
    Xextended <- cbind(Matrix::kronecker(rep(1, ncat), Xnuisance),
                       Matrix::kronecker(Matrix::Diagonal(ncat)[, -ref, drop = FALSE], X[keep, ]))
    int <- seq.int(nkeep)
    ## Set up the extended response
    Yextended <- c(Y[keep] * w[keep])

    nd <- paste0("%0", nchar(max(int)), "d")
    colnames(Xextended) <- c(paste0(".nuisance", sprintf(nd, int)),
                             ## CHECK: lev[-1] contrasts?
                             ofInterest <- paste(rep(lev[-ref], each = nvar),
                                                 rep(colnames(X), ncat - 1), sep = ":"))

    fit <- brglmFit(x = Xextended, y = Yextended,
                    start = NULL,
                    family = poisson("log"), control = control, intercept = TRUE, fixed_totals = fixed_totals)

    ## TODO:
    ## + starting values
    ## + subset
    ## + na.action
    ## + control

    ## Fitted values
    fitted <- do.call("rbind", tapply(fit$fitted, fixed_totals, function(x) x/sum(x)))
    rownames(fitted) <- rownames(X)[keep]
    colnames(fitted) <- lev
    fit$fitted.values <- fitted
    fit$call <- call
    ## fit$fitted.values <- matrix(fit$fitted.values, ncol = ncat)/w[keep]
    ## rownames(fit$fitted.values) <- rownames(X)[keep]
    ## colnames(fit$fitted.values) <- lev

    class(fit) <- c("brmultinom", fit$class, "glm")
    fit$ofInterest <- ofInterest
    fit$ncat <- ncat
    fit$lev <- lev
    fit$ref <- ref
    if (model) {
        fit$model  <- mf
    }
    if (x) {
        fit$x  <- X
    }
    fit$contrasts <- attr(X, "contrasts")
    fit$xlevels = .getXlevels(Terms, mf)
    fit$terms <- Terms
    fit$coefNames <- colnames(X)
    fit$null.deviance <- NULL
    fit
}

#' @method fitted brmultinom
#' @export
fitted.brmultinom <- function(object, ...) {
    object$fitted.values
}

#' Residuals for multinomial logistic regression and adjacent category logit models
#'
#' @param object the object coming out of [bracl()] and
#'     [brmultinom()].
#' @param type the type of residuals which should be returned.  The
#'     options are: `"pearson"` (default), `"response"`, `"deviance"`,
#'     `"working"`. See Details.
#' @param ... Currently not used.
#'
#' @details
#'
#' The residuals computed are the residuals from the equivalent
#' Poisson log-linear model fit, organized in a form that matches the
#' output of `fitted(object, type = "probs")`. As a result, the output
#' is residuals defined in terms of the object and expected
#' multinomial counts.
#'
#' @seealso brmultinom bracl
#'
#' @method residuals brmultinom
#' @export
residuals.brmultinom <- function(object, type = c("pearson", "response", "deviance", "working"), ...) {
    type <- match.arg(type)
    ## This is a Poisson log-linear models, so the working weights are
    ## the fitted counts
    fitted <- weights(object, type = "working")
    ## The poisson responses
    y <- object$y
    out <- switch(type,
                  "pearson" = (y - fitted)/sqrt(fitted),
                  "response" = (y - fitted),
                  "working" = object$residuals,
                  "deviance" = object$family$dev.resids(y, fitted, 1))
    matrix(out, ncol = object$ncat, dimnames = dimnames(fitted(object)))
}

#' @method coef brmultinom
#' @export
coef.brmultinom <- function(object, ...) {
    if (length(object$ofInterest)) {
        with(object, {
            coefs <- matrix(coefficients[ofInterest], nrow = ncat - 1, byrow = TRUE)
            dimnames(coefs) <- list(lev[-object$ref], coefNames)
            coefs
        })
    } else {
        NULL
    }
}

#' @method print brmultinom
#' @export
print.brmultinom <- function(x, digits = max(5L, getOption("digits") - 3L), ...) {
     if (!is.null(cl <- x$call)) {
        cat("Call:\n")
        dput(cl, control = NULL)
     }
     cat("\nCoefficients:\n")
     if (is.null(coef(x))) {
         print("No coefficients")
     } else {
         print(format(coef(x), digits = digits), print.gap = 2, quote = FALSE)
     }
     cat("\nResidual Deviance:", format(x$deviance), "\n")
}

#' @method logLik brmultinom
#' @export
logLik.brmultinom <- function(object, ...) {
    structure(-object$deviance/2,
              df = sum(!is.na(coef(object))),
              nobs = sum(object$weights),
              class = "logLik")
}

#' @method summary brmultinom
#' @export
summary.brmultinom <- function(object, correlation = FALSE, digits = options()$digits, Wald.ratios = FALSE, ...) {
    ncat <- object$ncat
    coefficients <- coef.brmultinom(object)
    object$digits <- digits
    object$AIC <- AIC(object)
    object$logLik <- logLik(object)
    if (is.null(coefficients)) {
        object$coefficients <- NULL
        object$standard.errors <- NULL
        if (Wald.ratios)
            object$Wald.ratios <- NULL
        if (correlation)
            object$correlation <- NULL
    } else {
        vc <- vcov.brglmFit(object)
        vc <- vc[object$ofInterest, object$ofInterest]
        se <- sqrt(diag(vc))
        ses <- matrix(se, nrow = ncat - 1, byrow = TRUE, dimnames = dimnames(coefficients))
        object$coefficients <- coefficients
        object$standard.errors <- ses
        ## object$AIC <- AIC(object)
        if (Wald.ratios) {
            object$Wald.ratios <- coefficients/ses
            object$Wald.pvalues <-  2 * pnorm(-abs(object$Wald.ratios))
        }
        if (correlation)
            object$correlation <- vc/outer(se, se)
    }
    class(object) <- "summary.brmultinom"
    object
}

#' @method vcov brmultinom
#' @export
vcov.brmultinom <- function(object, ...) {
    vc <- vcov.brglmFit(object, ...)
    vc <- vc[object$ofInterest, object$ofInterest]
    vc
}


#' @method print summary.brmultinom
#' @export
print.summary.brmultinom <- function(x, digits = x$digits, ...)
{
    if (!is.null(cl <- x$call)) {
        cat("Call:\n")
        dput(cl, control = NULL)
    }
    cat("\nCoefficients:\n")
    print(x$coefficients, digits = digits)
    cat("\nStd. Errors:\n")
    print(x$standard.errors, digits = digits)
    if (!is.null(x$Wald.ratios)) {
        cat("\nValue/SE (Wald statistics):\n")
        print(x$coefficients/x$standard.errors, digits = digits)
    }
    cat("\nResidual Deviance:", format(x$deviance), "\n")
    cat("Log-likelihood:", format(x$logLik), "\n")
    cat("AIC:", format(x$AIC))
    cat("\n\nType of estimator:", x$type, get_type_description(x$type))
    cat("\n", "Number of Fisher Scoring iterations: ", x$iter, "\n", sep = "")
    if (!is.null(correl <- x$correlation)) {
        p <- dim(correl)[2L]
        if (p > 1) {
            cat("\nCorrelation of Coefficients:\n")
            ll <- lower.tri(correl)
            correl[ll] <- format(round(correl[ll], digits))
            correl[!ll] <- ""
            print(correl[-1L, -p], quote = FALSE, ...)
        }
    }
    invisible(x)
}

#' Predict method for [brmultinom] fits
#'
#' Obtain class and probability predictions from a fitted baseline
#' category logits model.
#'
#' @param object a fitted object of class inheriting from
#'     [`"brmultinom"`][brmultinom].
#' @param newdata optionally, a data frame in which to look for
#'     variables with which to predict.  If omitted, the fitted linear
#'     predictors are used.
#' @param type the type of prediction required. The default is
#'     `"class"`, which produces predictions of the response category
#'     at the covariate values supplied in `"newdata"`, selecting the
#'     category with the largest probability; the alternative
#'     `"probs"` returns all category probabilities at the covariate
#'     values supplied in `newdata`.
#' @param ... further arguments passed to or from other methods.
#'
#'
#' @details
#'
#' If `newdata` is omitted the predictions are based on the data used
#' for the fit.
#'
#' @return
#'
#' If `type = "class"` a vector with the predicted response
#' categories; if `type = "probs"` a matrix of probabilities for all
#' response categories at `newdata`.
#'
#' @examples
#'
#' data("housing", package = "MASS")
#'
#' # Maximum likelihood using brmultinom with baseline category 'Low'
#' houseML1 <- brmultinom(Sat ~ Infl + Type + Cont, weights = Freq,
#'                        data = housing, type = "ML", ref = 1)
#'
#' # New data
#' newdata <- expand.grid(Infl = c("Low", "Medium"),
#'                        Type = c("Tower", "Atrium", "Terrace"),
#'                        Cont = c("Low", NA, "High"))
#'
#' ## Predictions
#' sapply(c("class", "probs"), function(what) predict(houseML1, newdata, what))
#'
#' @method predict brmultinom
#' @export
predict.brmultinom <- function(object, newdata, type = c("class", "probs"), ...) {
    ## Adapted from nnet:::predict.multinom
    if (!inherits(object, "brmultinom"))
        stop("not a \"brmultinom\" fit")
    type <- match.arg(type)
    X <- if (missing(newdata)) model.matrix(object) else model.matrix(object, data = newdata)
    rn <- attr(X, "rn_data")
    keep <- attr(X, "rn_kept")
    coefs <- coef(object)
    fits <- matrix(0, nrow = nrow(X), ncol = object$ncat, dimnames = list(rn[keep], object$lev))
    fits1 <- apply(coefs, 1, function(b) X %*% b)
    fits[, rownames(coefs)] <- fits1
    Y1 <- t(apply(fits, 1, function(x) exp(x) / sum(exp(x))))
    Y <- matrix(NA, length(rn), ncol(Y1), dimnames = list(rn, colnames(Y1)))
    Y[keep, ] <- Y1
    switch(type, class = {
        if (length(object$lev) > 2L) Y <- factor(max.col(Y),
            levels = seq_along(object$lev), labels = object$lev)
        if (length(object$lev) == 2L) Y <- factor(1 + (Y > 0.5),
            levels = 1L:2L, labels = object$lev)
        if (length(object$lev) == 0L) Y <- factor(max.col(Y),
            levels = seq_along(object$lab), labels = object$lab)
    }, probs = {
    })
    drop(Y)
}

#' @method model.matrix brmultinom
#' @export
model.matrix.brmultinom <- function(object, data, ...) {
    if (!inherits(object, "brmultinom"))
        stop("not a \"brmultinom\" fit")
    if (missing(data)) {
        data <- model.frame(object)
    } else {
        data <- as.data.frame(data)
    }
    Terms <- delete.response(object$terms)
    m <- model.frame(Terms, data, na.action = na.omit,
                     xlev = object$xlevels)
    if (!is.null(cl <- attr(Terms, "dataClasses")))
        .checkMFClasses(cl, m)
    X <- model.matrix(Terms, m, contrasts = object$contrasts)
    rn <- row.names(data)
    attr(X, "rn_data") <- rn
    attr(X, "rn_kept") <-  match(row.names(m), rn)
    X
}


#' Method for computing confidence intervals for one or more
#' regression parameters in a [`"brmultinom"`][brmultinom] object
#'
#' @inheritParams stats::confint
#'
#' @export
confint.brmultinom <- function (object, parm, level = 0.95, ...)  {
    ## Apart from formatting changes this function is identical to
    ## nnet:::confint.multinom
    cf <- coef(object)
    pnames <- if (is.matrix(cf)) colnames(cf) else names(cf)
    if (missing(parm)) {
        parm <- seq_along(pnames)
    } else {
        if (is.character(parm))  {
            parm <- match(parm, pnames, nomatch = 0L)
        }
    }
    a <- (1 - level) / 2
    a <- c(a, 1 - a)
    pct <- paste(round(100 * a, 1), "%")
    fac <- qnorm(a)
    if (is.matrix(cf)) {
        ses <- matrix(sqrt(diag(vcov(object))), ncol = ncol(cf),
            byrow = TRUE)[, parm, drop = FALSE]
        cf <- cf[, parm, drop = FALSE]
        ci <- array(NA, dim = c(dim(cf), 2L), dimnames = c(dimnames(cf),
            list(pct)))
        ci[, , 1L] <- cf + ses * fac[1L]
        ci[, , 2L] <- cf + ses * fac[2L]
        aperm(ci, c(2L, 3L, 1L))
    } else {
        ci <- array(NA, dim = c(length(parm), 2L), dimnames = list(pnames[parm],
            pct))
        ses <- sqrt(diag(vcov(object)))[parm]
        ci[] <- cf[parm] + ses %o% fac
        ci
    }
}


#' Method for simulating a data set from [`"brmultinom"`][brmultinom] and [`"bracl"`][bracl]
#' objects
#'
#' @param object an object of class [`"brmultinom"`][brmultinom] or [`"bracl"`][bracl].
#' @param ... currently not used.
#'
#' @return
#'
#' A [`"data.frame"`][data.frame] with `object$ncat` times the rows that
#' `model.frame(object)` have and the same variables. If `weights` has
#' been specified in the call that generated `object`, then the
#' simulate frequencies will populate the weights variable. Otherwise,
#' the resulting [data.frame] will have a `".weights"` variable with
#' the simulated multinomial counts.
#'
#' @examples
#'
#' ## Multinomial logistic regression
#' data("housing", package = "MASS")
#' houseML1 <- brmultinom(Sat ~ Infl + Type + Cont, weights = Freq,
#'                        data = housing, type = "ML", ref = 1)
#' simulate(houseML1)
#'
#' ## Adjacent-category logits
#' data("stemcell", package = "brglm2")
#' stemML1 <- bracl(research ~ religion + gender, weights = frequency,
#'                 data = stemcell, type = "ML")
#'
#' simulate(stemML1)
#'
#' @export
simulate.brmultinom <- function(object, ...) {
    mf <- model.frame(object)
    probs <- predict(object, type = "probs")
    categories <- colnames(probs)
    ncat <- object$ncat
    weights <- model.weights(mf)
    if (is.null(weights)) {
        weights <- rep.int(1L, nrow(mf))
    }
    samples <- sapply(1:nrow(probs), function(j) rmultinom(1, weights[j], probs[j, ]))
    mf <- mf[rep(1:nrow(mf), each = ncat), ]
    mf[, 1] <- factor(colnames(probs),
                      levels = levels(mf[, 1]),
                      ordered = is.ordered(mf[, 1]))
    weights_ind <- grep("(weights)", names(mf))
    if (length(weights_ind)) {
        weights_nam <- as.character(object$call$weights)
        names(mf)[weights_ind] <- weights_nam
    } else {
        weights_nam <- ".weights"
    }
    mf[[weights_nam]] <- c(samples)
    mf
}

## Tolerance for comparisons
tolerance <- 1e-05

data("lizards")

lizards_grahami <- lizards[, c("grahami", "height", "diameter", "light", "time")]
lizards_grahami <- lizards_grahami[rep(seq.int(nrow(lizards_grahami)), lizards_grahami$grahami), ]
lizards_grahami$species <- "grahami"
lizards_grahami$grahami <- NULL
lizards_opalinus <- lizards[, c("opalinus", "height", "diameter", "light", "time")]
lizards_opalinus <- lizards_opalinus[rep(seq.int(nrow(lizards_opalinus)), lizards_opalinus$opalinus), ]
lizards_opalinus$species <- "opalinus"
lizards_opalinus$opalinus <- NULL
lizards1 <- rbind(lizards_grahami, lizards_opalinus)
lizards1$species <- factor(lizards1$species, levels = c("opalinus", "grahami"))

model1 <- glm(formula = species ~ height + diameter + light + time, family = binomial(logit),
              data = lizards1,
              method = "brglmFit")
model2 <- brmultinom(species ~ height + diameter + light + time, data = lizards1)

tolerance <- 1e-06
## coefficients from the binomial fit match those from the multinomial model with the poisson trick
expect_equal(coef(model1), coef(model2)[1,], tol = tolerance)

library("nnet")

#####################################################################
data("hepatitis", package = "brglm2")
## Construct a variable with the multinomial categories according to
## the HCV and nonABC columns
hepat <- hepatitis
hepat$type <- with(hepat, factor(1 - HCV * nonABC + HCV + 2 * nonABC))
hepat$type <- factor(hepat$type, labels = c("noDisease", "C", "nonABC"))
contrasts(hepat$type) <- contr.treatment(3, base = 1)

hepbr <- brmultinom(type ~ group * time,
                    data = hepat, weights = counts)

## ML fails when there is separation
expect_warning(brmultinom(type ~ group * time, data = hepat, weights = counts, type = "ML"),
               pattern = "algorithm did not converge|failed to calculate score adjustment")

## brmultinom returns the same estimates as in Table 3 of Bull et al. (2002)
bulletall2002table3 <- matrix(c(-2.43, -1.57, 1.96, -0.36, -0.38, 0.26), ncol = 2)
expect_equal(coef(hepbr)[, -1], t(bulletall2002table3), tolerance = 1e-02, check.attributes = FALSE)

##############
hepbr_mat <- brmultinom(counts * nnet::class.ind(type) ~ group * time,
                        data = hepat)

tol <- 1e-05
## brmultinom returns the same estimates if counts are supplied as a matrix
expect_equal(coef(hepbr), coef(hepbr_mat), tolerance = tol)

#####################################################################
## Analysis of the enzymes data set in ?pmlr
#####################################################################
data("enzymes", package = "brglm2")
## Exclude patients in Group 4 (post-necrotic cirrhosis)
enzymes <- subset(enzymes, Group != 4)
## Center and scale covariates
AST <- scale(log(enzymes$AST))
ALT <- scale(log(enzymes$ALT))
GLDH <- scale(log(enzymes$GLDH))
OCT <- scale(log(enzymes$OCT))
enzymes <- data.frame(Patient = enzymes$Patient,
                      Group = enzymes$Group, AST, ALT, GLDH, OCT)
## Multinomial: acute viral hepatitis and aggressive chronic hepatits
## vs. persistent chronic hepatitis
## Assign Group 2 (persistent chronic hepatitis) as baseline category
enzymes$Group <- factor(enzymes$Group, levels=c("2","1","3"))
enzymes$counts <- rep(1, nrow(enzymes))
##
## this expect_warning here is preventive for the non-integer counts
## warnings that are generated internally but are not visible to the
## user
enzbrmultinom_ml <- brmultinom(Group ~ AST + GLDH, weights = counts, data = enzymes, type = "ML")

enzmultinom <- multinom(Group ~ AST + GLDH, weights = counts, data = enzymes, trace = FALSE)



## confint methods works as expected
cnnet <- confint(enzmultinom, level = 0.9123)
ccbrm <- confint(enzbrmultinom_ml, level = 0.9123)
expect_equal(cnnet, ccbrm, tolerance = 1e-04)

c1 <- drop(confint(enzbrmultinom_ml, level = 0.99, parm = 3))
c2 <- with(summary(enzbrmultinom_ml), {
    rbind(coefficients[, "GLDH"] - qnorm(1 - 0.01/2) * standard.errors[, "GLDH"],
          coefficients[, "GLDH"] + qnorm(1 - 0.01/2) * standard.errors[, "GLDH"])
})
expect_equal(c1, c2, tolerance = 1e-10, check.attributes = FALSE)

aic1 <- AIC(enzbrmultinom_ml)
aic2 <- AIC(enzmultinom)
aic3 <- -2 * logLik(enzbrmultinom_ml) + 2 * length(coef(enzbrmultinom_ml))
aic4 <- summary(enzbrmultinom_ml)$AIC

###
## AIC with brmultinon
expect_equal(aic1, aic2, tolerance = 1e-06)
expect_equal(aic1, unclass(aic3), tolerance = 1e-06, check.attributes = FALSE)
expect_equal(aic1, aic4, tolerance = 1e-06)


###############
## brmultinom returns the same estimates as nnet::multinom if type = 'ML'
expect_equal(coef(enzbrmultinom_ml), coef(enzmultinom), tolerance = 1e-04)

## brmultinom returns the same fitted values as nnet::multinom if type = 'ML'", {
expect_equal(fitted(enzbrmultinom_ml), fitted(enzmultinom), tolerance = 1e-04)


## brmultinom returns the same model matrix as nnet::multinom", {
expect_equal(model.matrix(enzbrmultinom_ml), model.matrix(enzmultinom), check.attributes = FALSE)


## brmultinom returns the same maximized loglikelihood as nnet::multinom if type = 'ML'", {
expect_equal(logLik(enzbrmultinom_ml), logLik(enzmultinom), tolerance = tol)


## brmultinom returns the same standard errors as nnet::multinom if type = 'ML'", {
expect_equal(summary(enzbrmultinom_ml)$standard.errors, summary(enzmultinom)$standard.errors, tolerance = 1e-04)


hepbr1 <- brmultinom(type ~ group * time,
                     data = hepat, weights = counts, ref = 1)
hepbr2 <- brmultinom(type ~ group * time,
                     data = hepat, weights = counts, ref = 2)
hepbr3 <- brmultinom(type ~ group * time,
                     data = hepat, weights = counts, ref = 3)

## brmultinom fits are invariant to the value of ref (ref1 vs ref2)'", {
expect_equal(hepbr1$fitted.values, hepbr2$fitted.values, tolerance = 1e-04)


## brmultinom fits are invariant to the value of ref  (ref1 vs ref3)'", {
expect_equal(hepbr1$fitted.values, hepbr3$fitted.values, tolerance = 1e-04)


newdata <- data.frame(group = c("no-withhold", "withhold", "no-withhold", "withhold"),
                      time = c("pre", "pre", "post", "post"))

hepnnet <- multinom(type ~ group * time, data = hepat, weights = counts, trace = FALSE)
expect_warning({
    hepml1 <- brmultinom(type ~ group * time,
                         data = hepat, weights = counts, ref = 1, type = "ML", maxit = 5)
    hepml2 <- brmultinom(type ~ group * time,
                         data = hepat, weights = counts, ref = 2, type = "ML", maxit = 5)
    hepml3 <- brmultinom(type ~ group * time,
                         data = hepat, weights = counts, ref = 3, type = "ML", maxit = 5)
})

## predict.brmultinom returns the right result", {
expect_equal(predict(hepnnet), predict(hepml1), tolerance = 1e-04)
expect_equal(predict(hepnnet), predict(hepml2), tolerance = 1e-04)
expect_equal(predict(hepnnet), predict(hepml3), tolerance = 1e-04)
expect_equal(predict(hepnnet, newdata = newdata), predict(hepml1, newdata = newdata), tolerance = 1e-04)
expect_equal(predict(hepml1, newdata = newdata), predict(hepml2, newdata = newdata), tolerance = 1e-04)
expect_equal(predict(hepml1, newdata = newdata), predict(hepml3, newdata = newdata), tolerance = 1e-04)


## simulate method returns a data frame with expected characteristics", {
simu_df <- simulate(hepbr1)
nam_mf <- names(model.frame(hepbr1))
nam_simu <- names(simu_df)
expect_identical(nrow(simu_df),
                 nrow(hepat) * nlevels(hepat$type))
expect_identical(levels(simu_df$type),
                 levels(hepat$type))
expect_identical(is.ordered(simu_df$type),
                 is.ordered(hepat$type))
expect_identical(nam_mf[!(nam_mf %in% nam_simu)],
                 "(weights)")
expect_identical(nam_simu[!(nam_simu %in% nam_mf)],
                 as.character(hepbr1$call$weights))
expect_identical(nam_simu[(nam_simu %in% nam_mf)],
                 nam_mf[(nam_mf %in% nam_simu)])




## Aligator data
## data("alligators", package = "brglm2")
## all_ml <- brmultinom(foodchoice ~ size + lake , weights = freq,
##                      data = alligators, type = "ML", ref = 1)
## out <- sapply(1:1000, function(j) coef(update(all_ml, data = simulate(all_ml))))

## all_mean <- brmultinom(foodchoice ~ size + lake , weights = round(freq/k),
##                        data = alligators, type = "AS_mean", ref = 1)

## all_median <- brmultinom(foodchoice ~ size + lake , weights = round(freq/k),
##                          data = alligators, type = "AS_median", ref = 1)

## library("ggplot2")

## ## Collect probabilities
## pml <- data.frame(stack(data.frame(fitted(all_ml))), method = "ML")
## pmean <- data.frame(stack(data.frame(fitted(all_mean))), method = "AS_mean")
## pmedian <- data.frame(stack(data.frame(fitted(all_median))), method = "AS_median")
## probabilities <- rbind(pml, pmean, pmedian)
## names(probabilities) <- c("probability", "category", "method")
## probabilities1 <- NULL
## for (j in levels(probabilities$category)) {
##     cdat <- subset(probabilities, category == j)
##     cdat$id <- rep(seq.int(nrow(pml)/5), 3)
##     cdat <- reshape(cdat, timevar = "method", v.names = "probability", direction = "wide")
##     probabilities1 <- rbind(probabilities1, cdat)
## }

## ## Shrinkage plots
## ggplot(probabilities1) +
##     geom_point(aes(x = probability.ML, y = probability.AS_mean)) +
##     geom_abline(aes(intercept = 0, slope = 1), alpha = 0.5) +
##     lims(x = c(0,1), y = c(0,1)) +
##     facet_grid(~ category) +
##     theme_bw()

## ggplot(probabilities1) +
##     geom_point(aes(x = probability.ML, y = probability.AS_median)) +
##     geom_abline(aes(intercept = 0, slope = 1), alpha = 0.5) +
##     lims(x = c(0,1), y = c(0,1)) +
##     facet_grid(~ category) +
##     theme_bw()

#' MDYPL state evolution functions with no intercept
#'
#' @param mu aggregate bias parameter.
#' @param b parameter `b` in the state evolution functions.
#' @param sigma square root of the aggregate variance of the MDYPL
#'     estimator.
#' @param kappa asymptotic ratio of columns/rows of the design
#'     matrix. `kappa` should be in `(0, 1)`.
#' @param gamma the square root of the limit of the variance of the
#'     linear predictor.
#' @param alpha the shrinkage parameter of the MDYPL
#'     estimator. `alpha` should be in `(0, 1)`.
#' @param gh A list with the Gauss-Hermite quadrature nodes and
#'     weights, as returned from `statmod::gauss.quad()` with `kind =
#'     "hermite"`. Default is `NULL`, in which case `gh` is set to
#'     `statmod::gauss.quad(200, kind = "hermite")` is used.
#' @param prox_tol tolerance for the computation of the proximal
#'     operator; default is `1e-10`.
#'
#' @export
se0 <- function(mu, b, sigma, kappa, gamma, alpha, gh = NULL, prox_tol = 1e-10) {
    if (is.null(gh))
        gh <- gauss.quad(200, kind = "hermite")
    a_frac <- 0.5 * (1 + alpha)
    xi <- gh$nodes
    wi <- gh$weights
    n_quad <- length(xi)
    q1 <- rep(sqrt(2) * gamma * xi, times = n_quad)
    q2 <- mu * q1 + rep(sqrt(2 * kappa) * sigma * xi, each = n_quad)
    w2 <- rep(wi, times = n_quad) * rep(wi, each = n_quad)

    w2p <- 2 * w2 * plogis2(q1)  / pi
    p_prox <- plogis2(prox(q2 + a_frac * b, b, prox_tol))
    prox_resid <- a_frac - p_prox
    c(sum(w2p * q1 * prox_resid),
      1 - kappa - sum(w2p / (1 + b * p_prox * (1 - p_prox))),
      kappa^2 * sigma^2 - b^2 * sum(w2p * prox_resid^2))
}

#' MDYPL state evolution functions with intercept
#'
#' @param mu aggregate bias parameter.
#' @param b parameter `b` in the state evolution functions.
#' @param sigma square root of the aggregate variance of the MDYPL
#'     estimator.
#' @param iota limits of the MDYPL estimate for the intercept as the sample size goes to +Inf
#' @param kappa asymptotic ratio of columns/rows of the design
#'     matrix. `kappa` should be in `(0, 1)`.
#' @param gamma the square root of the limit of the variance of the
#'     linear predictor.
#' @param alpha the shrinkage parameter of the MDYPL
#'     estimator. `alpha` should be in `(0, 1)`.
#' @param intercept intercept of the logistic regression model
#' @param gh A list with the Gauss-Hermite quadrature nodes and
#'     weights, as returned from `statmod::gauss.quad()` with `kind =
#'     "hermite"`. Default is `NULL`, in which case `gh` is set to
#'     `statmod::gauss.quad(200, kind = "hermite")` is used.
#' @param prox_tol tolerance for the computation of the proximal
#'     operator; default is `1e-10`. fixed point problem solved via Newton-Raphson
#' @export
se1 <- function(mu, b, sigma, iota, kappa, gamma, alpha, intercept, gh = NULL, prox_tol = 1e-10) {
    if (is.null(gh))
        gh <- gauss.quad(200, kind = "hermite")
    a_frac <- 0.5 * (1 + alpha)
    xi <- gh$nodes
    wi <- gh$weights
    n_quad <- length(xi)
    q1n <- sqrt(2) * gamma * xi
    q1 <- rep(q1n + intercept, times = n_quad)
    q2 <- mu * q1n + rep(sqrt(2 * kappa) * sigma * xi + iota, each = n_quad)
    w2 <- rep(wi, times = n_quad) * rep(wi, each = n_quad)

    p_q1_pos <- plogis2(q1)
    w2pi <- w2 / pi
    w2p_pos <- w2pi * p_q1_pos
    w2p_neg <- w2pi - w2p_pos

    p_prox_pos <- plogis2(prox(a_frac * b + q2, b, prox_tol))
    p_prox_neg <- plogis2(prox(a_frac * b - q2, b, prox_tol))

    prox_pos_resid <- a_frac - p_prox_pos
    prox_neg_resid <- a_frac - p_prox_neg

    c(sum(q1 * (w2p_pos * prox_pos_resid - w2p_neg * prox_neg_resid)),
      1 - kappa - sum(w2p_pos / (1 + b * p_prox_pos * (1 - p_prox_pos)) + w2p_neg / (1 + b * p_prox_neg * (1 - p_prox_neg))),
      kappa^2 * sigma^2 - b^2 * sum(w2p_pos * prox_pos_resid^2 + w2p_neg * prox_neg_resid^2),
      sum(w2p_pos * prox_pos_resid - w2p_neg * prox_neg_resid))
}


#' Logistic ridge regression state evolution functions with no intercept
#'
#' @param mu aggregate bias parameter.
#' @param b parameter `b` in the state evolution functions.
#' @param sigma square root of the aggregate variance of the MDYPL
#'     estimator.
#' @param kappa asymptotic ratio of columns/rows of the design
#'     matrix. `kappa` should be in `(0, 1)`.
#' @param gamma the square root of the limit of the variance of the
#'     linear predictor.
#' @param lambda the shrinkage parameter of the logistic regression penalty
#'     estimator. `lambda` should be in greater than zero.
#' @param gh A list with the Gauss-Hermite quadrature nodes and
#'     weights, as returned from `statmod::gauss.quad()` with `kind =
#'     "hermite"`. Default is `NULL`, in which case `gh` is set to
#'     `statmod::gauss.quad(200, kind = "hermite")` is used.
#' @param prox_tol tolerance for the computation of the proximal
#'     operator; default is `1e-10`.
#'
#' @details
#'
#' It is assumed that the ridge penalty to the logistic regression
#' log-likelihood is `n * lambda * sum(beta^2) / (2 * length(beta))`,
#' where `n` is the sum of the binomial totals.
#'
#' @export
se0_ridge <- function(mu, b, sigma, kappa, gamma, lambda, gh = NULL, prox_tol = 1e-10) {
    if (is.null(gh))
        gh <- gauss.quad(200, kind = "hermite")
    xi <- gh$nodes
    wi <- gh$weights
    n_quad <- length(xi)
    q1 <- rep((sqrt(2) * gamma) * xi, times = n_quad)
    q2 <- rep((sqrt(2) * sqrt(kappa) * sigma) * xi, each = n_quad)
    w2pi <- rep((2 / pi) * wi, times = n_quad) * rep(wi, each = n_quad)
    g <- mu * q1 + q2
    prox <- prox(g, b, prox_tol)
    p_prox <- plogis2(prox)
    p_q1 <- plogis2(-q1)
    c(sum(w2pi * p_q1 * (1 -  p_q1) * prox) + mu * kappa,
      1 - kappa + b * lambda - sum(w2pi * p_q1 / (1 + b * p_prox * (1 - p_prox))),
      (sigma * kappa)^2 - sum(w2pi * p_q1 * (g - prox)^2))
}

library("brglm2")

kappa0 <- 0.2
gamma0 <- 5
alpha0 <- 0.88
theta0 <- 1

iota0 <- 2
mu0 <- 0.7
b0 <- 1.2
sigma0 <- 2.3

expect_equal(se0(mu0, b0, sigma0, kappa0, gamma0, alpha0),
             se1(mu0, b0, sigma0, 0, kappa0, gamma0, alpha0, 0)[1:3])

## 3 par
soln0 <- solve_se(kappa0, gamma0, alpha0, start = c(0.5, 1, 1),
                  init_iter = 23, init_method = "BFGS")
expect_equal(attr(soln0, "funcs"), rep(0, 3))

opt_str <- c("optim", "23", "BFGS", "nleqslv")
for (st in opt_str) {
    expect_true(grepl(st, attr(soln0, "optimization-chain")))
}

nu0 <- sqrt(soln0[1]^2 * gamma0^2 + kappa0 * soln0[3]^2)
soln0_c <- solve_se(kappa0, nu0, alpha0, start = soln0, corrupted = TRUE,
                    init_iter = "only", init_method = "Nelder-Mead")
expect_equal(soln0_c, soln0, check.attributes = FALSE, tolerance = 1e-07)

## 4 par
soln1 <- solve_se(kappa0, gamma0, alpha0, theta0, start = c(0.5, 1, 1, 1),
                  init_iter = 44, init_method = "Nelder-Mead")
expect_equal(attr(soln1, "funcs"), rep(0, 4))

soln1t <- solve_se(kappa0, gamma0, alpha0, theta0, start = c(0.5, 1, 1, 1),
                   init_iter = 44, init_method = "Nelder-Mead", transform = FALSE)
expect_equal(attr(soln1t, "funcs"), rep(0, 4))
expect_equal(soln1t, soln1, tolerance = 1e-07, check.attributes = FALSE)

opt_str <- c("optim", "44", "Nelder-Mead", "nleqslv")
for (st in opt_str) {
    expect_true(grepl(st, attr(soln1, "optimization-chain")))
}

nu0 <- sqrt(soln1[1]^2 * gamma0^2 + kappa0 * soln1[3]^2)
soln1_c <- solve_se(kappa0, nu0, alpha0, soln1[4], start = soln1, corrupted = TRUE,
                    init_iter = 44, init_method = "Nelder-Mead")
expect_equal(soln1_c[1:3], soln1[1:3], check.attributes = FALSE, tolerance = 1e-07)
expect_equal(soln1_c[4], theta0, check.attributes = FALSE, tolerance = 1e-07)


## Reproducing Table 13 of Zhao et al. (2022)
## at https://doi.org/10.3150/21-BEJ1401

thetas <- c(0, 0.5, 1, 2, 2.5)
gamma0 <- 5
pars3 <- matrix(NA, length(thetas), 3)
pars4 <- matrix(NA, length(thetas), 4)
colnames(pars4) <- c("I_mu", "I_b", "I_sigma", "I_iota")
colnames(pars3) <- c("I_mu", "I_b", "I_sigma")
for (i in seq_along(thetas)) {
    pars3[i, ] <- solve_se(kappa = 0.2, ss = sqrt(5 + thetas[i]^2), alpha = 1, start = c(0.5, 1, 1), init_iter = 0)
    pars4[i, ] <- solve_se(kappa = 0.2, ss = sqrt(5), alpha = 1, intercept = thetas[i], start = c(pars3[i, ], thetas[i]), init_iter = 0)
}

cbind(pars4, pars3)
Zhaoetal_t13_I <- matrix(c(1.5, 1.51, 1.56, 1.83, 2.31,
                           3.03, 3.13, 3.45, 5.47, 8.96,
                           4.75, 4.84, 5.16, 7.01, 10.00,
                           0, 0.76, 1.559, 3.68, 5.8), ncol = 4)
Zhaoetal_t13_II <- matrix(c(1.5, 1.51, 1.55, 1.75, 1.95,
                           3.03, 3.12, 3.42, 4.83, 6.26,
                           4.75, 4.84, 5.13, 6.45, 7.73), ncol = 3)
expect_equal(Zhaoetal_t13_II, pars3, tolerance = 1e-02, check.attributes = FALSE)
expect_equal(Zhaoetal_t13_I, pars4, tolerance = 1e-01, check.attributes = FALSE)

#' Habitat preferences of lizards
#'
#' @format A data frame with 23 rows and 6 columns:
#'
#' * `grahami`. count of grahami lizards
#' * `opalinus`. count of opalinus lizards
#' * `height`. a factor with levels `<5ft`, `>=5ft`
#' * `diameter`. a factor with levels `<=2in`, `>2in`
#' * `light`. a factor with levels `sunny`, `shady`
#' * `time`. a factor with levels `early`, `midday`, `late`
#'
#' The variables `grahami` and `opalinus` are counts of two lizard
#' species at two different perch heights, two different perch
#' diameters, in sun and in shade, at three times of day.
#'
#' @seealso
#'
#' [brglm_fit()]

#'
#' @source
#'
#'   McCullagh P, Nelder J A (1989) _Generalized Linear
#'   Models_ (2nd Edition).  London: Chapman and Hall.
#'
#' Originally from
#'
#'     Schoener T W (1970) Nonsynchronous spatial overlap of lizards
#'     in patchy habitats.  _Ecology_ *51*, 408-418.
#'
"lizards"

#' Histology grade and risk factors for 79 cases of endometrial cancer
#'
#' @format A data frame with 79 rows and 4 variables:
#'
#' * `NV`: neovasculization with coding 0 for absent and 1 for present
#' * `PI`: pulsality index of arteria uterina
#' * `EH`: endometrium height
#' * `HG` histology grade with coding 0 for low grade and 1 for high grade
#'
#' @source The packaged data set was downloaded in `.dat` format from
#'     \url{https://users.stat.ufl.edu/~aa/glm/data/}. The latter link
#'     provides the data sets used in Agresti (2015).
#'
#'     The endometrial data set was first analyzed in Heinze and
#'     Schemper (2002), and was originally provided by Dr
#'     E. Asseryanis from the Medical University of Vienna.
#'
#' @seealso
#'
#' [brglm_fit()]
#'
#'
#' @references
#'
#' Agresti A (2015). *Foundations of Linear and Generalized Linear
#' Models*.  Wiley Series in Probability and Statistics. Wiley.
#'
#' Heinze G, Schemper M (2002). A Solution to the Problem of
#' Separation in Logistic Regression. *Statistics in Medicine*,
#' **21**, 2409–2419. \doi{10.1002/sim.1047}.
#'
#' Kosmidis I, Firth D (2021). Jeffreys-prior penalty, finiteness
#' and shrinkage in binomial-response generalized linear
#' models. *Biometrika*, **108**, 71-82. \doi{10.1093/biomet/asaa052}.
#'
#'
"endometrial"


#' Coalition data
#'
#' This data set contains survival data on government coalitions in
#' parliamentary democracies (Belgium, Canada, Denmark, Finland,
#' France, Iceland, Ireland, Israel, Italy, Netherlands, Norway,
#' Portugal, Spain, Sweden, and the United Kingdom) for the period
#' 1945-1987.  For parsimony, country indicator variables are omitted
#' in the sample data.
#'
#' @format
#'
#' A data frame with 314 rows and the 7 variables "duration",
#' "ciep12", "invest", "fract", "polar", "numst2", and "crisis".  For
#' variable descriptions, please refer to King et al (1990).
#'
#' @note
#'
#' Data is as it is provided by the
#' [\pkg{Zeilig}](https://cran.r-project.org/package=Zelig) R package.
#'
#' @seealso
#'
#' [brglm_fit()]
#'
#' @references
#'
#'  King G, Alt J E, Burns N E, Laver M. (1990). A Unified
#'  Model of Cabinet Dissolution in Parliamentary
#'  Democracies. *American Journal of Political Science*, **34**,
#'  846-870. \doi{10.2307/2111401}.
#'
#'  King G, Alt J E, Burns N E, Laver M. ICPSR
#'  Publication Related Archive, 1115.
#'
"coalition"

#' Effects of AZT in slowing the development of AIDS symptoms
#'
#' The data is from a 3-year study on the effects of AZT in slowing the
#' development of AIDS symptoms. 338 veterans whose immune systems
#' were beginning to falter after infection with the AIDS virus were
#' randomly assigned either to receive AZT immediately or to wait
#' until their T cells showed severe immune weakness.
#'
#' @format A data frame with 4 rows and 4 variables:
#'
#' * `symptomatic`: counts of veterans showing AIDS symptoms during the 3-year study
#'
#' * `asymptomatic`: counts of veterans not showing AIDS symptoms during the 3-year study
#'
#' * `race`: the race of the veterans with levels `"White"` and `"Black"`
#'
#' * `AZT_use`: whether the veterans received AZT immediately (`"Yes"`)
#' or waited until their T cells showed severe immune weakness (`"No"`)
#'
#' @source
#'
#' The data set is analyzed in Agresti (2002, Subsection 5.4.2). Its
#' original source is New York Times, Feb. 15, 1991.
#'
#' @seealso
#'
#' [brmultinom()]
#'
#' @references
#'
#' Agresti A (2002). *Categorical Data Analysis*.  Wiley Series in
#' Probability and Statistics. Wiley.
"aids"

#' Alligator food choice data
#'
#' @format A data frame with 80 rows and 5 variables:
#'
#' * `foodchoice`: primary food type, in volume, found in an alligator’s stomach, with levels `fish`, `invertebrate`,`reptile`, `bird`, `other`
#' * `lake`: lake of capture with levels `Hancock`, `Oklawaha`, `Trafford`, `George`.
#' * `gender`: gender of the alligator with levels `Male` and `Female`
#' * `size`: size of the alligator with levels `<=2.3` meters long and `>2.3` meters long
#' * `freq`: number of alligators for each foodchoice, lake, gender and size combination
#'
#'
#' @source
#'
#' The alligators data set is analyzed in Agresti (2002, Subsection 7.1.2).
#'
#' @seealso
#'
#' [brmultinom()]
#'
#' @references
#'
#' Agresti A (2002). *Categorical Data Analysis*.  Wiley Series in
#' Probability and Statistics. Wiley.
#'
"alligators"



#' Opinion on stem cell research and religious fundamentalism
#'
#' A data set from the 2006 General Social Survey that shows the
#' relationship in the United States between opinion about funding
#' stem cell research and the fundamentalism/liberalism of one’s
#' religious beliefs, stratified by gender.
#'
#' @format A data frame with 24 rows and 4 variables:
#'
#' * `research`: opinion about funding stem cell research with levels `definitely`, `probably`, `probably not`, `definitely not`
#' * `gender`: the gender of the respondent with levels `female` and `male`
#' * `religion`: the fundamentalism/liberalism of one’s religious beliefs with levels `fundamentalist`, `moderate`,
#' `liberal`
#' `frequency`: the number of times a respondent fell in each of the combinations of levels for `research`, `religion` and `gender`
#'
#'
#' @seealso
#'
#' [bracl()]
#'
#' @source
#'
#' The `stemcell` data set is analyzed in Agresti (2010, Subsection 4.1.5).
#'
#' @references
#'
#' Agresti A (2010). *Analysis of Ordinal Categorical Data* (2nd edition).  Wiley Series in
#' Probability and Statistics. Wiley.
#'
"stemcell"


#' Post-transfusion hepatitis: impact of non-A, non-B hepatitis
#' surrogate tests
#'
#' Data from a randomized double-blind trial to assess whether
#' withholding donor blood positive for the non-A, non-B (`"NANB"`)
#' surrogate markers would reduce the frequency of post-transfusion
#' hepatitis.  The dataset contains `4588` subjects enrolled from 1988
#' to 1992 into two study groups that received allogenic blood from
#' which units positive for NANB surrogate markers were withheld (n =
#' `2311`) or not withheld (n = `2277`).  Subjects were followed up
#' for 6 months and assessed for the presence of post-transfusion
#' hepatitis.
#'
#' @format A data frame with 28 rows and the following 6 columns:
#'
#' * `city`: Subjects were recruited from 3 Canadian Red Cross Society
#' Blood Centres and 13 university-affiliated hospitals in 3 cities:
#' Toronto, Hamilton and Winnipeg.
#'
#' * `group`: Eligible subjects were assigned to one of two allogenic
#' blood recipient groups.  One group received products that had only
#' routine Canadian transfusion-transmissible disease marker screening
#' (no-withhold).  The other group received only products that were
#' not positive for NANB surrogate markers (withhold).
#'
#' * `time`: Hepatitis C (HCV) screening was introduced in Canada in
#' May, 1990.  Subjects were recruited into the study before (pre) and
#' after (post) the introduction of anti-HCV testing.
#'
#' * `HCV`: Post-transfusion HCV hepatitis present (1) or absent (0).
#'
#' * `nonABC`: Post-transfusion non-A, non-B, non-C hepatitis present (1) or absent (0)
#'
#' * `counts`: Number of subjects
#'
#' @source
#'
#' Data is from Blajchman et al. (1995), also analyzed in Bull et
#' al. (2002), and is also provided by the
#' [\pkg{pmlr}](https://cran.r-project.org/package=pmlr) R package.
#'
#' @references
#'
#' Bull S B, Mak C, Greenwood C M T (2002). A modified score function
#' estimator for multinomial logistic regression in small
#' samples. *Computational Statistics & Data Analysis*, **39**,
#' 57-74. \doi{10.1016/S0167-9473(01)00048-2}
#'
#' Blajchman M A, Bull S B and Feinman S V (1995). Post-transfusion
#' hepatitis: impact of non-A, non-B hepatitis surrogate tests. *The
#' Lancet*, **345**, 21--25. \doi{10.1016/S0140-6736(95)91153-7}
#'
"hepatitis"

#' Liver enzyme data
#'
#' Liver enzyme data collected from 218 patients with liver disease
#' (Plomteux, 1980). The laboratory profile consists of enzymatic
#' activity measured for four liver enzymes: aspartate
#' aminotransferase (`AST`), alanine aminotransferase (`ALT`),
#' glutamate dehydrogenase (`GLDH`) and ornithine carbamyltransferase
#' (`OCT`).
#'
#' @format A data frame with 218 rows and the following 6 columns:
#'
#' * `Patient`: Patient ID
#'
#' * `Group`: Four diagnostic groups were considered: acute viral
#' hepatitis (1), persistent chronic hepatitis (2), aggressive chronic
#' hepatitis (3) and post-necrotic cirrhosis (4).
#'
#' * `AST`: Aspartate aminotransferase (in U/L)
#'
#' * `ALT`: Alanine aminotransferase (in U/L)
#'
#' * `GLDH`: Glutamate dehydrogenase (in U/L)
#'
#' * `OCT`: Ornithine carbamyltransferase (in U/L)
#'
#' @source
#'
#' Data from Albert and Harris (1984, Chapter 5, Appendix I), and is
#' also provided by the
#' [\pkg{pmlr}](https://cran.r-project.org/package=pmlr) R package.
#'
#' @references
#'
#' Albert A, Harris E K (1984). *Multivariate Interpretation of
#' Clinical Laboratory Data*. Dekker: New York.
#'
#' Plomteux G (1980). Multivariate analysis of an enzyme profile for
#' the differential diagnosis of viral hepatitis. *Clinical
#' Chemistry*, **26**, 1897-1899.
#'
"enzymes"

#' Multiple features data
#'
#' Digits (0-9) extracted from a collection of maps from a Dutch
#' public utility.  Two hundred `30 × 48` binary images per digit are
#' available, which have then been used to extract feature sets; see
#' Jain et al. (2000), for details, where that dataset is used for
#' assessing the performance of various classifiers for digit
#' recognition.
#'
#' @format A data frame with 2000 rows and 382 columns:
#'
#' * `digit`. The digits to which the feature sets `fou.*`, `kar.*` and `pix.*` correspond to.
#' * `fou.*`. 76 Fourier coefficients of the character shapes, which are computed to be rotation invariant.
#' * `kar.*`. 64 Karhunen-Lo\`eve coefficients of the character shapes.
#' * `pix.*`. 240 pixel averages in 2 x 3 windows of each character shape.
#' * `training`. `TRUE` if the digit is part of the training set
#'      and `FALSE` if the digit is allocated to the test set.
#'
#' @source
#'
#' The data provides the `fou`, `kar` and `pix` features of the
#' Multiple Features data set from the UCI Machine Learning Repository
#' (Duin, 1998).
#'
#' @seealso
#'
#' [mdypl_fit()]
#'
#' @references
#'
#' Duin, R. (1998). Multiple Features Dataset. UCI Machine Learning
#' Repository. \doi{10.24432/C5HC70}.
#'
#' Jain A, Duin R, Mao J (2000). Statistical pattern recognition: a
#' review. IEEE Transactions on Pattern Analysis and Machine
#' Intelligence, **22**, 4–37. \doi{10.1109/34.824819}.
#'
#' @examples
#'
#' \donttest{
#' data("MultipleFeatures", package = "brglm2")
#'
#' par(mfrow = c(10, 20), mar = numeric(4) + 0.1)
#' for (c_digit in 0:9) {
#'    df <- subset(MultipleFeatures, digit == c_digit)
#'    df <- as.matrix(df[, paste("pix", 1:240, sep = ".")])
#'    for (inst in 1:20) {
#'        m <- matrix(df[inst, ], 15, 16)[, 16:1]
#'        image(m, col = grey.colors(7, 1, 0), xaxt = "n", yaxt = "n")
#'   }
#' }
#'
#' }
"MultipleFeatures"

# Copyright (C) 2020- Euloge Clovis Kenne Pagui, Ioannis Kosmidis

#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 or 3 of the License
#  (at your option).
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  A copy of the GNU General Public License is available at
#  http://www.r-project.org/Licenses/

#' Bias reduction for negative binomial regression models
#'
#' [brnb()] is a function that fits negative binomial regression
#' models using implicit and explicit bias reduction methods.
#'
#' @inheritParams stats::glm
#' @param link The link function. Currently must be one of `"log"`,
#'     `"sqrt"` or `"identity"`.
#' @param control a list of parameters for controlling the fitting
#'     process. See [brglmControl()] for details.
#' @return A fitted model object of class [`"brnb"`][brnb] inheriting
#'     from [`"negbin"`][MASS::glm.nb()] and
#'     [`"brglmFit"`][brglmFit()]. The object is similar to the output
#'     of [brglmFit()] but contains four additional components:
#'     `theta` for the estimate of the dispersion parameter,
#'     `vcov.mean` for the estimated variance-covariance matrix of the
#'     regression coefficients, `vcov.dispersion` for the estimated
#'     variance of the dispersion parameter in the chosen
#'     parameterization (using the expected information), and
#'     `twologlik` for twice the log-likelihood function.
#'
#' @details
#'
#' A detailed description of the fitting procedure is given in the
#' iteration vignette (see, `vignette("iteration", "brglm2")` and
#' Kosmidis et al, 2020). The number of iterations when estimating
#' parameters are controlled by the `maxit` argument of
#' [brglmControl()].
#'
#' The type of score adjustment to be used is specified through the
#' `type` argument (see [brglmControl()] for details).
#'
#' The available options are:
#'
#' * `type = "AS_mixed"`: the mixed bias-reducing score
#' adjustments in Kosmidis et al (2020) that result in mean bias
#' reduction for the regression parameters and median bias reduction
#' for the dispersion parameter, if any; default.
#'
#' * `type = "AS_mean"`: the mean bias-reducing score
#' adjustments in Firth (1993) and Kosmidis & Firth (2009).
#'
#' * `type = "AS_median"`: the median bias-reducing score
#' adjustments in Kenne Pagui et al. (2017)
#'
#' * `type = "MPL_Jeffreys"`: maximum penalized likelihood
#' with powers of the Jeffreys prior as penalty.
#'
#' * `type = "ML"`: maximum likelihood.
#'
#' * `type = "correction"`: asymptotic bias correction, as in
#' Cordeiro & McCullagh (1991).
#'
#' The choice of the parameterization for the dispersion is controlled
#' by the `transformation` argument (see [brglmControl()] for
#' details).  The default is `"identity"`. Using `transformation =
#' "inverse"` uses the dispersion parameterization that
#' [MASS::glm.nb()] uses.
#'
#' @author Euloge Clovis Kenne Pagui `[aut]` \email{kenne@stat.unipd.it}, Ioannis Kosmidis `[aut, cre]` \email{ioannis.kosmidis@warwick.ac.uk}
#'
#' @references
#'
#' Cordeiro G M, McCullagh P (1991). Bias correction in generalized
#' linear models. *Journal of the Royal Statistical Society. Series B
#' (Methodological)*, **53**, 629-643. \doi{10.1111/j.2517-6161.1991.tb01852.x}.
#'
#' Firth D (1993). Bias reduction of maximum likelihood estimates.
#' *Biometrika*. **80**, 27-38. \doi{10.2307/2336755}.
#'
#' Kenne Pagui E C, Salvan A, Sartori N (2017). Median bias
#' reduction of maximum likelihood estimates. *Biometrika*, **104**,
#' 923–938. \doi{10.1093/biomet/asx046}.
#'
#' Kosmidis I, Kenne Pagui E C, Sartori N (2020). Mean and median bias
#' reduction in generalized linear models. *Statistics and Computing*,
#' **30**, 43-59. \doi{10.1007/s11222-019-09860-6}.
#'
#' Kosmidis I, Firth D (2009). Bias reduction in exponential family
#' nonlinear models. *Biometrika*, **96**, 793-804. \doi{10.1093/biomet/asp055}.
#'
#' @examples
#' ## Example in Saha, K., & Paul, S. (2005). Bias-corrected maximum
#' ## likelihood estimator of the negative binomial dispersion
#' ## parameter.  Biometrics, 61, 179--185.
#' #
#' # Number of revertant colonies of salmonella data
#' salmonella <- data.frame(freq = c(15, 16, 16, 27, 33, 20,
#'                                   21, 18, 26, 41, 38, 27,
#'                                   29, 21, 33, 60, 41, 42),
#'                          dose = rep(c(0, 10, 33, 100, 333, 1000), 3),
#'                          observation = rep(1:3, each = 6))
#'
#' # Maximum likelihood fit with glm.nb of MASS
#' salmonella_fm <- freq ~ dose + log(dose + 10)
#' fitML_glmnb <- MASS::glm.nb(salmonella_fm, data = salmonella)
#'
#' # Maximum likelihood fit with brnb
#' fitML <- brnb(salmonella_fm, data = salmonella,
#'               link = "log", transformation = "inverse", type = "ML")
#'
#' # Mean bias-reduced fit
#' fitBR_mean <- update(fitML, type = "AS_mean")
#'
#' # Median bias-reduced fit
#' fitBR_median <- update(fitML, type = "AS_median")
#'
#' # Mixed bias-reduced fit
#' fitBR_mixed <- update(fitML, type = "AS_mixed")
#'
#' # Mean bias-corrected fit
#' fitBC_mean <- update(fitML, type = "correction")
#'
#' # Penalized likelihood with Jeffreys-prior penalty
#' fit_Jeffreys <- update(fitML, type = "MPL_Jeffreys")
#'
#' # The parameter estimates from glm.nb and brnb with type = "ML" are
#' # numerically the same
#' all.equal(c(coef(fitML_glmnb), fitML_glmnb$theta),
#'             coef(fitML, model = "full"), check.attributes = FALSE)
#'
#' # Because of the invariance properties of the maximum likelihood,
#' # median reduced-bias, and mixed reduced-bias estimators the
#' # estimate of a monotone function of the dispersion should be
#' # (numerically) the same as the function of the estimate of the
#' # dispersion:
#'
#' # ML
#' coef(fitML, model = "dispersion")
#' 1 / coef(update(fitML, transformation = "identity"), model = "dispersion")
#'
#' # Median BR
#' coef(fitBR_median, model = "dispersion")
#' 1 / coef(update(fitBR_median, transformation = "identity"), model = "dispersion")
#'
#' # Mixed BR
#' coef(fitBR_mixed, model = "dispersion")
#' 1 / coef(update(fitBR_mixed, transformation = "identity"), model = "dispersion")
#'
#' ## The same is not true for mean BR
#' coef(fitBR_mean, model = "dispersion")
#' 1 / coef(update(fitBR_mean, transformation = "identity"), model = "dispersion")
#'
#' \donttest{
#' ## An example  from Venables & Ripley (2002, p.169).
#' data("quine", package = "MASS")
#' quineML <- brnb(Days ~ Sex/(Age + Eth*Lrn), link = "sqrt", transformation="inverse",
#'                 data = quine, type="ML")
#' quineBR_mean <- update(quineML, type = "AS_mean")
#' quineBR_median <- update(quineML, type = "AS_median")
#' quineBR_mixed <- update(quineML, type = "AS_mixed")
#' quine_Jeffreys <- update(quineML, type = "MPL_Jeffreys")
#'
#' fits <- list(ML = quineML,
#'              AS_mean = quineBR_mean,
#'              AS_median = quineBR_median,
#'              AS_mixed = quineBR_mixed,
#'              MPL_Jeffreys = quine_Jeffreys)
#' sapply(fits, coef, model = "full")
#' }
#'
#' @export
brnb <- function(formula, data, subset, weights = NULL, offset = NULL,
                 link = "log", start = NULL, etastart = NULL,
                 mustart = NULL,  control = list(...), na.action,
                 model = TRUE, x = FALSE, y = TRUE, contrasts = NULL,
                 intercept = TRUE, singular.ok = TRUE ,...) {

    loglik <- function(n, th, mu, y, w) {
        sum(w * (lgamma(th + y) - lgamma(th) - lgamma(y + 1) +
                 th * log(th) + y * log(mu + (y == 0)) - (th + y) * log(th + mu)))
    }

    trace_iteration <- function() {
        if (control$trace) {
            st <- max(abs(step_beta), na.rm = TRUE)
            gr <- max(abs(adjusted_grad_beta), na.rm = TRUE)
            cat("Coefficients update:\t", sprintf("%03f", betas), "\n", sep = " ")
            cat("Outer/Inner iteration:\t", sprintf("%03d", iter),
                "/", sprintf("%03d", step_factor), "\n", sep = "")
            cat("max |step|:", format(round(st, 6), nsmall = 6,
                                      scientific = FALSE), "\t", "max |gradient|:",
                format(round(gr, 6), nsmall = 6, scientific = FALSE),
                "\n")
            st <- abs(step_dispersion)
            gr <- abs(adjusted_grad_dispersion)
            cat("Dispersion update:\t",sprintf("%03f", dispersion), "\n", sep = "")
            cat("Outer iteration:\t", sprintf("%03d", iter),
                "\n")
            cat("max |step|:", format(round(st, 6), nsmall = 6,
                                      scientific = FALSE), "\t", "max |gradient|:",
                format(round(gr, 6), nsmall = 6, scientific = FALSE),
                "\n")
        }
    }

    mf <- Call <- match.call()
    m <- match(c("formula", "data", "subset", "weights", "na.action",
                 "etastart", "mustart", "offset"), names(mf), 0)
    mf <- mf[c(1, m)]
    mf$drop.unused.levels <- TRUE
    mf[[1L]] <- quote(stats::model.frame)
    mf <- eval.parent(mf)
    Terms <- attr(mf, "terms")
    return_x <- x
    return_y <- y

    y <- model.response(mf, "numeric")
    x <- if (!is.empty.model(Terms))
             model.matrix(Terms, mf, contrasts)
         else matrix(, NROW(y), 0)
    weights <- model.weights(mf)
    offset <- model.offset(mf)
    mustart <- model.extract(mf, "mustart")
    etastart <- model.extract(mf, "etastart")
    x <- as.matrix(x)
    nobs <- NROW(x)
    nvars <- NCOL(x)
    if (missing_offset <- is.null(offset)) {
        offset <- rep.int(0, nobs)
    }

    if (is.null(weights)) weights <- rep(1,nobs)
    if (any(weights < 0))
        stop("negative weights not allowed")

    control <- do.call("brglmControl", control)
    ok_links <- c("log", "sqrt", "identity")
    if (!isTRUE(link %in% ok_links))
        stop(link, " is not one of the appropriate link function")
    ok_transformations <- c("log", "sqrt", "identity", "inverse")
    if (!isTRUE(control$transformation %in% ok_transformations))
        stop(control$transformation, " is not one of the implemented dispersion transformations")

    linkobj <- enrichwith::enrich(make.link(link),with="d2mu.deta")
    linkfun <- linkobj$linkfun
    linkinv <- linkobj$linkinv
    dmu.deta <- linkobj$mu.eta
    d2mu.deta <- linkobj$d2mu.deta

    linkobj_disp <- enrichwith::enrich(make.link(control$transformation), with = "d2mu.deta")
    linkfun_disp <- linkobj_disp$linkfun
    linkinv_disp <- linkobj_disp$linkinv
    dkappa.dphi <- linkobj_disp$mu.eta
    d2kappa.dphi <- linkobj_disp$d2mu.deta

    fam0 <- do.call("poisson", list(link = link))

    is_ML <- control$type == "ML"
    is_AS_median <- control$type == "AS_median"
    is_AS_mixed <- control$type == "AS_mixed"
    is_correction <- control$type == "correction"

    ## computation of s1
    s1fun <- function(yy, k) {
        res <- 0
        if (yy > 0) {
            j <- 0:(yy - 1)
            res <- sum(j / (k * j + 1))
        }
        res
    }

    ## Computation of needed expectations almost exactly
    exp_quant <- function(mu, k) {
        n <- length(mu)
        E_s2 <-  E_s2y <- E_s1s2 <- E_s3 <- numeric(n)
        if (k < 0) {
            ## stop("Negative value of k was encountered during optimization")
            return(list(E_s2 = NA, E_s2y = NA, E_s1s2 = NA, E_s3 = NA))
        }
        ymax <- max(qnbinom(1 - 10 * .Machine$double.eps, mu = mu, size = 1 / k))
        ## if (ymax > 30000) stop("ymax too much large")  ## need control on largest value?
        if (ymax > 1) {
            out <- .C('expectedValues',
                      as.double(mu),
                      as.double(k),
                      as.integer(ymax),
                      as.integer(n),
                      E_s2 = as.double(E_s2),
                      E_s2y = as.double(E_s2y),
                      E_s1s2 = as.double(E_s1s2),
                      E_s3 = as.double(E_s3))
            E_s2 <- out$E_s2
            E_s2y <- out$E_s2y
            E_s1s2 <- out$E_s1s2
            E_s3 <- out$E_s3
        }
        list(E_s2 = E_s2, E_s2y = E_s2y, E_s1s2 = E_s1s2, E_s3 = E_s3)
    }

    ## ## ## IK, 20 Apr 2021: Optimized R version of exp_quant
    ## exp_quant <- function(mu, k) {
    ##     if (k < 0) {
    ##         stop("encountered negative value of k")
    ##     }
    ##     E_s2 <-  E_s2y <- E_s1s2 <- E_s3 <- numeric(length(mu))
    ##     ymax <- max(qnbinom(1 - 10 * .Machine$double.eps, mu = mu, size = 1 / k), na.rm = TRUE)
    ##     if (ymax > 1) {
    ##         yval <- 0:ymax
    ##         pmat <- sapply(mu, function(x) dnbinom(yval, mu = x, size = 1 / k))
    ##         j <- c(0, 0:(ymax - 1))
    ##         fra <- j / (k * j + 1)
    ##         s1 <- cumsum(fra)
    ##         s2 <- cumsum(fra^2)
    ##         s3 <- cumsum(fra^3)
    ##         temp <- pmat * s2
    ##         E_s2 <- .colSums(temp, ymax + 1, nobs, TRUE)
    ##         E_s2y <- .colSums(temp * yval, ymax + 1, nobs, TRUE)
    ##         E_s1s2 <- .colSums(temp * s1, ymax + 1, nobs, TRUE)
    ##         E_s3 <- .colSums(pmat * s3, ymax + 1, nobs, TRUE)
    ##     }
    ##     list(E_s2 = E_s2, E_s2y = E_s2y, E_s1s2 = E_s1s2, E_s3 = E_s3)
    ## }

    ## required quantities
    key_quantities <- function(pars) {
        betas <- pars[1:nvars]
        phi <- pars[nvars + 1]
        etas <- drop(x %*% betas + offset)
        mus <- linkinv(etas)
        d1 <- dmu.deta(etas)
        d2 <- d2mu.deta(etas)
        k <- linkinv_disp(phi)
        d1phi <- dkappa.dphi(phi)
        d2phi <- d2kappa.dphi(phi)
        varmus <- k * mus^2 + mus  ## variance
        d1varmus <- 2 * k * mus + 1  ## derivative of variance
        working_weights <- weights * d1^2 / varmus
        wx <- sqrt(working_weights) * x
        qr_decomposition <- qr(wx)
        s1 <- sapply(y, function(t) s1fun(t, k))
        expectations <- exp_quant(mus, k)
        out <- c(list(betas = betas,
                      k = k,
                      etas = etas,
                      phi = phi,
                      mus = mus,
                      d1 = d1,
                      d2 = d2,
                      d1phi = d1phi,
                      d2phi = d2phi,
                      working_weights = working_weights,
                      varmus = varmus,
                      d1varmus = d1varmus,
                      s1 = s1,
                      qr_decomposition = qr_decomposition),expectations)
        out
    }

    ## hat values
    hat_values <- function(pars, fit = NULL) {
        if (is.null(fit)) {
            fit <- key_quantities(pars)
        }
        with(fit, {
            Qmat <- qr.Q(qr_decomposition)
            .rowSums(Qmat * Qmat, nobs, nvars, TRUE)
        })
    }

    ## score function
    score <- function(pars, level = 0, fit = NULL) {
        if (is.null(fit)) {
            fit <- key_quantities(pars)
        }
        with(fit, {
            if (level == 0) {
                return(.colSums(x * working_weights * (y - mus) / d1, nobs, nvars, TRUE))
            }
            if (level == 1) {
                return(sum((s1 - mus * y / (k * mus + 1) + ((k * mus + 1) * log(k * mus + 1) - k * mus) / (k^3 * mus + k^2)) * d1phi * weights, na.rm = TRUE))
            }
        })
    }

    ## information matrix
    information <- function(pars, level = 0, inverse = FALSE, fit = NULL) {
        if (is.null(fit)) {
            fit <- key_quantities(pars)
        }
        with(fit, {
            if (level == 0) {
                R_matrix <- qr.R(qr_decomposition)
                if (inverse) {
                    return(chol2inv(R_matrix))
                } else {
                    return(crossprod(R_matrix))
                }
            }
            if (level == 1) {
                iphiphi <- sum( weights * (E_s2 + ((2 * k * mus + 2) * log((k * mus + 1)) - k^2 * mus^2 - 2 * k * mus) / (k^4 * mus + k^3)) * d1phi^2, na.rm = TRUE )
                if (inverse) {
                    return(1/iphiphi)
                } else {
                    return(iphiphi)
                }
            }
        })
    }

    ## Jeffreys adjustment
    AS_Jeffreys_adjustment <- function(pars, level = 0, fit = NULL) {
        if (is.null(fit)) {
            fit <-  key_quantities(pars)
        }
        with(fit, {
            iphiphi <- sum(weights * (E_s2 + ((2 * k * mus + 2) * log((k * mus + 1)) - k^2 * mus^2 - 2 * k * mus) / (k^4 * mus + k^3)) * d1phi^2, na.rm = TRUE )
            hatvalues <- hat_values(pars, fit = fit)
            if (level == 0) {
                ## Use only observations with keep = TRUE to ensure that no division with zero takes place
                return(control$a * ( .colSums(hatvalues * (2 * d2/d1 - d1varmus * d1 / varmus) * x, nobs, nvars, TRUE) +
                                     .colSums(d1phi^2 * x * working_weights * (E_s2y - mus * E_s2 - mus^3/(k * mus + 1)) / d1, nobs, nvars, TRUE) /  iphiphi))
            }
            if (level == 1) {
                return(control$a * (sum(-d1phi * hatvalues * mus^2 / varmus, na.rm = TRUE) +
                                    sum(weights * (-2 * E_s3  + (2 * k^3 * mus^3 + 9 * k^2 * mus^2 + 6 * k * mus - 6 * (k * mus + 1)^2 * log(k * mus + 1)) / (k^4 * (k * mus + 1)^2) + E_s1s2 -
                                                   mus * E_s2y / (k * mus + 1) -  (k * mus-(k * mus + 1) * log(k * mus + 1)) * E_s2 / (k^2 * (k * mus + 1))), na.rm = TRUE) * d1phi^3 +
                                    2 * sum(weights * (E_s2 + ((2 * k * mus + 2) * log((k * mus + 1)) - k^2 * mus^2 - 2 * k * mus)/(k^4 * mus + k^3)), na.rm = TRUE) * d1phi * d2phi / iphiphi))
            }
        })
    }

    ## mean adjustment
    AS_mean_adjustment <- function(pars, level = 0, fit = NULL) {
        if (is.null(fit)) {
            fit <- key_quantities(pars)
        }
        with(fit, {
            hatvalues <- hat_values(pars, fit = fit)
            if (level == 0) {
                return(.colSums(0.5 * hatvalues * d2 / d1 * x, nobs, nvars, TRUE))
            }
            if (level == 1) {
                iphiphi <- sum(weights * (E_s2 + ((2 * k * mus + 2) * log((k * mus + 1)) - k^2 * mus^2 - 2 * k * mus) / (k^4 * mus + k^3)) * d1phi^2, na.rm = TRUE)
                pqphiphi<- sum(weights * (-2 * E_s3 + (2 * k^3 * mus^3 + 9 * k^2 * mus^2 + 6 * k * mus - 6 * (k * mus + 1)^2 * log(k * mus + 1)) / (k^4 * (k * mus + 1)^2) +
                                          2 * E_s1s2 - 2 * mus * E_s2y / (k * mus + 1) - 2 * (k * mus - (k * mus + 1) * log(k * mus + 1)) * E_s2 / (k^2 * (k * mus + 1))), na.rm = TRUE) * d1phi^3 +
                           sum(weights * (E_s2 + ((2 * k * mus + 2) * log((k * mus + 1)) - k^2 * mus^2 - 2 * k * mus) / (k^4 * mus + k^3)) , na.rm = TRUE) * d1phi * d2phi
                return(0.5 * d1phi * sum(weights * hatvalues * d1^2 * mus^2 / (working_weights * varmus^2), na.rm = TRUE) + 0.5 * pqphiphi / iphiphi)
            }
        })
    }

    ## median adjustment
    AS_median_adjustment <- function(pars, level = 0, fit = NULL) {
        if (is.null(fit)) {
            fit <- key_quantities(pars)
        }
        with(fit, {
            hatvalues <- hat_values(pars, fit = fit)
            if (level == 0) {
                invInfo <- information(pars = pars, level = 0, inverse = TRUE, fit = fit)
                info <- information(pars = pars, level = 0, inverse = FALSE, fit = fit)
                F2_tilde <- numeric(nvars)
                for (j in seq.int(nvars)) {
                    invInfo_j <- invInfo[j,]
                    vcov_j <- tcrossprod(invInfo_j) / invInfo_j[j]
                    hats_j <- .rowSums((x %*% vcov_j) * x, nobs, nvars, TRUE) * working_weights
                    F2_tilde[j] <- invInfo_j %*% .colSums(x * (hats_j * (d1 * d1varmus / (6 * varmus) - 0.5 * d2 / d1)), nobs, nvars,  TRUE)
                }
                return( .colSums(0.5 * hatvalues * d2 / d1 * x, nobs, nvars, TRUE) + info %*% F2_tilde)
            }
            if (level == 1) {
                iphiphi <- sum(weights * (E_s2 + ((2 * k * mus + 2) * log((k * mus + 1)) - k^2 * mus^2 - 2 * k * mus) / (k^4 * mus + k^3)) * d1phi^2, na.rm = TRUE)
                pqphiphi <- sum(weights * (-2 * E_s3 + (2 * k^3 * mus^3 + 9 * k^2 * mus^2 + 6 * k * mus - 6 * (k * mus + 1)^2 * log(k * mus + 1)) / (k^4 * (k * mus + 1)^2) +
                                           2 * E_s1s2 - 2 * mus * E_s2y / (k * mus + 1) - 2 * (k * mus - (k * mus + 1) * log(k * mus + 1)) * E_s2 / (k^2 * (k * mus + 1))), na.rm = TRUE) * d1phi^3 +
                            sum(weights *(E_s2 + ((2 * k * mus + 2) * log((k * mus + 1)) - k^2 * mus^2 - 2 * k * mus) / (k^4 * mus + k^3)), na.rm = TRUE) * d1phi * d2phi
                pq2phiphi <- (sum(weights * (-2 * E_s3 / 3 + (2 * k^3 * mus^3 + 9 * k^2 * mus^2 + 6 * k * mus - 6 * (k * mus + 1)^2 * log(k * mus + 1)) / (3 * k^4 * (k * mus + 1)^2) +
                                             E_s1s2 / 2 - 0.5 * mus * E_s2y / (k * mus + 1) - 0.5 * (k * mus-(k * mus + 1) * log(k * mus + 1)) * E_s2 / (k^2 * (k * mus + 1))), na.rm = TRUE) * d1phi^3 +
                              0.5 * sum(weights * (E_s2 + ((2 * k * mus + 2) * log((k * mus + 1)) - k^2 * mus^2 - 2 * k * mus)/(k^4 * mus + k^3)), na.rm = TRUE) * d1phi * d2phi)
                return((0.5 * d1phi * sum(weights * hatvalues * d1^2 * mus^2 / (working_weights * varmus^2), na.rm = TRUE) + 0.5 * pqphiphi / iphiphi - pq2phiphi / iphiphi))
            }
        })
    }

    ## mixed adjustment
    AS_mixed_adjustment <- function(pars, level = 0, fit = NULL) {
        if (is.null(fit)) {
            fit <- key_quantities(pars)
        }
        with(fit, {
            hatvalues <- hat_values(pars, fit = fit)
            if (level == 0) {
                return(.colSums(0.5 * hatvalues * d2 / d1 * x, nobs, nvars, TRUE))
            }
            if (level == 1)
            {
                iphiphi <- sum(weights * (E_s2 + ((2 * k * mus + 2) * log((k * mus + 1)) - k^2 * mus^2 - 2 * k * mus) / (k^4 * mus + k^3)) * d1phi^2, na.rm = TRUE)
                pqphiphi <- (sum(weights *( -2 * E_s3 + (2 * k^3 * mus^3 + 9 * k^2 * mus^2 + 6 * k * mus - 6 * (k * mus + 1)^2 * log(k * mus + 1)) / (k^4 * (k * mus + 1)^2) +
                                            2 * E_s1s2 - 2 * mus * E_s2y / (k * mus + 1) - 2 * (k * mus - (k * mus + 1) * log(k * mus + 1)) * E_s2 / (k^2 * (k * mus + 1))), na.rm = TRUE) * d1phi^3 +
                             sum(weights * (E_s2 + ((2 * k * mus + 2) * log((k * mus + 1)) - k^2 * mus^2 - 2 * k * mus) / (k^4 * mus + k^3)), na.rm = TRUE) * d1phi * d2phi)
                pq2phiphi <- (sum(weights *(-2 * E_s3 / 3 + (2 * k^3 * mus^3 + 9 * k^2 * mus^2 + 6 * k * mus - 6 * (k * mus + 1)^2 * log(k * mus + 1)) / (3 * k^4 * (k * mus + 1)^2) +
                                            E_s1s2 / 2 - 0.5 * mus * E_s2y / (k * mus + 1) - 0.5 * (k * mus - (k * mus + 1) * log(k * mus + 1)) * E_s2 / (k^2 * (k * mus + 1))), na.rm = TRUE) * d1phi^3 +
                              0.5 * sum(weights * (E_s2 + ((2 * k * mus + 2) * log((k * mus + 1)) - k^2 * mus^2 - 2 * k * mus) / (k^4 * mus + k^3)), na.rm = TRUE) * d1phi * d2phi)
                return((0.5 * d1phi * sum(weights * hatvalues * d1^2 * mus^2 / (working_weights * varmus^2), na.rm = TRUE) + 0.5 * pqphiphi / iphiphi - pq2phiphi / iphiphi))
            }
        })
    }

    ## adjustment term function
    adjustment_function <- switch(control$type,
                                  AS_mean = AS_mean_adjustment,
                                  AS_median = AS_median_adjustment,
                                  AS_mixed = AS_mixed_adjustment,
                                  MPL_Jeffreys = AS_Jeffreys_adjustment,
                                  correction = function(pars, ...) 0,
                                  ML = function(pars, ...) 0)

    ## required components for computing the adjusted scores
    compute_step_components <- function(pars, level = 0, fit = NULL) {
        if (is.null(fit)) {
            fit <- key_quantities(pars)
        }
        if (level == 0) {
            grad <- score(pars, level = 0, fit = fit)
            inverse_info <- try(information(pars, level = 0, inverse = TRUE,fit = fit), silent = TRUE)
            failed_inversion <- inherits(inverse_info, "try-error")
            adjustment <- adjustment_function(pars, level = 0, fit = fit)
            failed_adjustment <- any(is.na(adjustment))
        }
        if (level == 1) {
            grad <- score(pars, level = 1, fit = fit)
            inverse_info <- try(information(pars, level = 1, inverse = TRUE,fit = fit), silent = TRUE)
            failed_inversion <- inherits(inverse_info, "try-error")
            adjustment <- adjustment_function(pars, level = 1, fit = fit)
            failed_adjustment <- any(is.na(adjustment))
        }
        out <- list(grad = grad,
                    inverse_info = inverse_info,
                    adjustment = adjustment,
                    failed_inversion = failed_inversion,
                    failed_adjustment = failed_adjustment)
        out
    }

    betas_names <- dimnames(x)[[2L]]
    ynames <- if (is.matrix(y))
                  rownames(y)
              else names(y)
    EMPTY <- nvars == 0

    valid_eta <- fam0$valideta
    valid_mu <- fam0$validmu
    eval(fam0$initialize)
    ## to be completed
    if (EMPTY)
        stop("invalid linear predictor values in empty model")

    qrx <- qr(x)
    rank <- qrx$rank
    is_full_rank <- rank == nvars
    if (!singular.ok && !is_full_rank) {
        stop("singular fit encountered")
    }

    if (!isTRUE(is_full_rank)) {
        aliased <- qrx$pivot[seq.int(qrx$rank + 1,nvars)]
        X_all <- x
        x <- x[, -aliased]
        nvars_all <- nvars
        nvars <- ncol(x)
        betas_names_all <- betas_names
        betas_names <- betas_names[-aliased]
    } else {
        nvars_all <- nvars
        betas_names_all <- betas_names
    }
    betas_all <- structure(rep(NA_real_, nvars_all), .Names = betas_names_all)
    keep <- weights > 0
    nkeep <- sum(keep)
    df_residual <- nkeep - rank
    if (is.null(start)) {
        start_fit <- glm.fit(x = x, y = y, weights = weights,
                             start = start, offset = offset, family =fam0,
                             control = list(epsilon = control$epsilon, maxit = 100, trace = FALSE),
                             intercept =  intercept)
        start <- coef(start_fit)
        sd <- sqrt(diag(solve(crossprod(start_fit$R))))
        inds <- sd > 100
        start[inds] <- sign(start[inds]) * 2
        ## dispersion <- linkfun_disp(1 / as.vector(MASS::theta.ml(y = y, mu = linkinv(x %*% start),
        ##                                                         n = nobs, weights = weights,
        ##                                                         trace = control$trace > 2)))
        dispersion <- linkfun_disp(1)
        betas <- start
        names(betas) <- betas_names
    } else {
        if (is.numeric(start) && (length(start) == nvars_all + 1)) {
            betas_all <- start[seq.int(nvars_all)]
            names(betas_all) <- betas_names_all
            if (!isTRUE(is_full_rank)) {
                betas_all[aliased] <- NA_real_
                betas <- betas_all[-aliased]
            } else {
                betas <- betas_all
            }
            dispersion <- start[nvars_all + 1]
        } else {
            stop(gettextf("length of 'start' should be equal to %d with initial values for the coefficients of %s and a starting value for the transformed dispersion parameter",
                          nvars_all + 1, paste(betas_names_all, collapse = ", ")))
        }
    }

    adjusted_grad_all <- rep(NA_real_, nvars_all + 1)
    names(adjusted_grad_all) <- c(betas_names_all, "dispersion")
    names(dispersion) <- "dispersion"
    par <- c(betas, dispersion)
    quantities <- key_quantities(par)
    step_components_beta <- compute_step_components(par, level = 0, fit = quantities)
    step_components_dispersion <- compute_step_components(par, level = 1, fit = quantities)
    if (step_components_beta$failed_inversion) {
        warning("failed to invert the information matrix")
    }
    if (step_components_beta$failed_adjustment) {
        warning("failed to calculate score adjustment")
    }
    adjusted_grad_beta <- with(step_components_beta, {
        grad + adjustment
    })
    step_beta <- drop(step_components_beta$inverse_info %*% adjusted_grad_beta)
    if (step_components_dispersion$failed_inversion)  {
        warning("failed to invert the information matrix")
    }
    if (step_components_dispersion$failed_adjustment) {
        warning("failed to calculate score adjustment")
    }
    adjusted_grad_dispersion <- with(step_components_dispersion, {
        grad + adjustment
    })
    step_dispersion <- as.vector(adjusted_grad_dispersion * step_components_dispersion$inverse_info)

    if (control$maxit == 0) {
        iter <- 0
        failed <- FALSE
    } else {
        for (iter in seq.int(control$maxit)) {
            step_factor <- 0
            testhalf <- TRUE
            while (testhalf & step_factor < control$max_step_factor) {
                step_beta_previous <- step_beta
                step_dispersion_previous <- step_dispersion
                betas <- betas + control$slowit * 2^(-step_factor) * step_beta
                dispersion <- dispersion + 2^(-step_factor) * step_dispersion
                ## If negative dispersion reset to 1
                if (linkinv_disp(dispersion) < 0) {
                    dispersion <- linkfun_disp(1)
                }
                par <- c(betas, dispersion)
                quantities <- key_quantities(par)
                step_components_beta <- compute_step_components(par, level = 0, fit = quantities)
                step_components_dispersion <- compute_step_components(par,  level = 1, fit = quantities)
                if (failed_inversion_beta <- step_components_beta$failed_inversion) {
                    warning("failed to invert the information matrix")
                    break
                }
                if (failed_adjustment_beta <- step_components_beta$failed_adjustment) {
                    warning("failed to calculate score adjustment")
                    break
                }
                adjusted_grad_beta <- with(step_components_beta, {
                    grad + adjustment
                })
                step_beta <- drop(step_components_beta$inverse_info %*%  adjusted_grad_beta)
                if (failed_inversion_dispersion <- step_components_dispersion$failed_inversion) {
                    warning("failed to invert the information matrix")
                    break
                }
                if (failed_adjustment_dispersion<- step_components_dispersion$failed_adjustment) {
                    warning("failed to calculate score adjustment")
                    break
                }
                adjusted_grad_dispersion <- with(step_components_dispersion, {
                    grad + adjustment
                })
                step_dispersion <- as.vector(adjusted_grad_dispersion * step_components_dispersion$inverse_info)
                if (step_factor == 0 & iter == 1) {
                    testhalf <- TRUE
                } else {
                    s2 <- c(abs(step_beta), abs(step_dispersion))
                    s1 <- c(abs(step_beta_previous), abs(step_dispersion_previous))
                    testhalf <- sum(s2, na.rm = TRUE) > sum(s1,  na.rm = TRUE)
                }
                step_factor <- step_factor + 1
                if (control$trace) {
                    trace_iteration()
                }
            }
            failed <- failed_adjustment_beta || failed_inversion_beta ||
                failed_adjustment_dispersion || failed_inversion_dispersion
            if (failed || max(abs(c(step_beta, step_dispersion)), na.rm = TRUE) < control$epsilon) {
                break
            }
        }
    }
    adjusted_grad_all[betas_names] <- adjusted_grad_beta
    adjusted_grad_all["dispersion"] <- adjusted_grad_dispersion
    betas_all[betas_names] <- betas

    if(iter >= control$maxit || failed) {
        convergence <- FALSE
        warning("optimization failed to converge")
    } else {
        convergence <- TRUE
    }
    if (!isTRUE(is_full_rank)) {
        x <- X_all
        betas <- betas_all
        betas[is.na(betas)] <- 0
        nvars <- nvars_all
    }

    par <- c(betas, dispersion)
    if (is_correction) {
        bias <- c(-step_components_beta$inverse_info %*% AS_mean_adjustment(par, level = 0, fit = quantities),
                  -step_components_dispersion$inverse_info %*% AS_mean_adjustment(par, level = 1, fit = quantities))
        par <- par - bias
        betas <- par[-(nvars + 1)]
        dispersion <-par[(nvars + 1)]
    }
    quantities <- key_quantities(par)
    step_components_beta <- compute_step_components(par, level = 0, fit = quantities)
    step_components_dispersion <- compute_step_components(par, level = 1, fit = quantities)

    qr.Wx <- quantities$qr_decomposition
    mus <- quantities$mus
    etas <- quantities$etas
    residuals <- with(quantities, (y - mus) / d1)
    working_weights <- quantities$working_weights
    wt <- rep.int(0, nobs)
    wt[keep] <- working_weights[keep]
    names(wt) <- names(residuals) <- names(mus) <- names(etas) <- names(weights) <- names(y) <- ynames

    fam <- do.call("negative.binomial", list(theta = 1 / linkinv_disp(dispersion),
                                             link = link))

    if (attr(Terms, "intercept")) {
        if (missing_offset) {
            nullFit <- try(glm.fit(x[, "(Intercept)", drop = FALSE], y, weights,
                                   offset = rep(0, nobs), family = fam,
                                   control = list(maxit = control$maxit,
                                                  epsilon = control$epsilon, trace = control$trace),
                               intercept = TRUE), silent = TRUE)
        } else {
            nullFit <- try(glm.fit(x[, "(Intercept)", drop = FALSE], y, weights,
                                   offset = offset, family = fam,
                                   control = list(maxit = control$maxit,
                                                  epsilon = control$epsilon, trace = control$trace > 1),
                                   intercept = TRUE), silent = TRUE)
        }
        if (inherits(nullFit, "try-error")) {
            stop("Computation of null deviance failed")
        }
        nullmus <- nullFit$fitted
    } else {
        nullmus <- linkinv(offset)
    }

    th <- 1 / linkinv_disp(dispersion)
    Lm <- loglik(nobs, th, mus, y, weights)
    dev.resids <- fam$dev.resids
    aic <- fam$aic
    nulldev <- sum(dev.resids(y, nullmus, weights))
    nulldf <- nkeep - as.integer(attr(Terms, "intercept"))
    deviance <- sum(dev.resids(y, mus, weights))
    aic.model <- aic(y, nobs, mus, weights, deviance) + 2 * (rank + 1)

    vcov.mean <-step_components_beta$inverse_info
    rownames(vcov.mean) <- colnames(vcov.mean)  <-  betas_names

    vcov.dispersion <- step_components_dispersion$inverse_info
    names(vcov.dispersion) <-  paste0(control$transformation, "(dispersion)")
    out <- list(coefficients = betas,
                vcov.mean = vcov.mean,
                vcov.dispersion =  vcov.dispersion,
                residuals = residuals,
                fitted.values = mus,
                R = if (!EMPTY) qr.R(qr.Wx),
                rank = rank,
                qr = if (!EMPTY) structure(qr.Wx[c("qr", "rank", "qraux", "pivot", "tol")], class = "qr"),
                family = fam,
                twologlik = as.vector(2 * Lm),
                linear.predictors = etas,
                deviance = deviance,
                aic = aic.model,
                link=link,
                null.deviance = nulldev,
                iter = iter,
                weights = wt,
                prior.weights = weights,
                df.residual = df_residual,
                df.null = nulldf,
                y = y,
                converged = convergence,
                dispersion = dispersion,
                grad = adjusted_grad_all,
                transformation = control$transformation,
                theta = as.vector(th), ## dispersion as in glm.nb
                type = control$type)
    out$terms <- Terms
    out$formula <- as.vector(attr(Terms,"formula"))
    Call$link <- link
    out$call <- Call
    if (model) out$model <- mf
    out$na.action <- attr(mf,"na.action")
    if (return_x) out$x <- x
    if (!return_y) out$y <- NULL
    out$contrasts <- attr(x,"contrasts")
    out$xlevels <- .getXlevels(Terms,mf)
    out$control <- control
    out$offset <- offset
    class(out)<-c("brnb", "negbin", "glm")
    out
}


#' Extract model coefficients from [`"brnb"`][brnb] objects
#'
#' @inheritParams stats::coef
#' @param model one of `"mean"` (default), `"full"`, `"dispersion"`,
#'     to return the estimates of the parameters in the linear
#'     prediction only, or both, the estimate of the dispersion
#'     parameter only, respectively.
#'
#' @details
#'
#' See [coef()] for more details.
#'
#' @method coef brnb
#' @export
coef.brnb <- function(object, model = c("mean", "full", "dispersion"), ...) {
    object$transformed_dispersion <- object$dispersion
    coef.brglmFit(object, model, ...)
}

#' Extract model variance-covariance matrix from [`"brnb"`][brnb] objects
#'
#'
#' @inheritParams stats::vcov.glm
#' @param object an object of class [`"brnb"`][brnb], typically, a result of a call to [brnb()].
#' @param model character specifying for which component of the model variance-covariance matrix should be extracted.
#'
#' @details
#'
#' The options for `model` are `"mean"` for mean regression only
#'  (default), `"dispersion"` for the dispersion parameter (in a
#'  chosen transformation; see [brglmControl()], and `"full"` for both
#'  the mean regression and the (transformed) dispersion parameters.
#'  See [vcov()] for more details.
#'
#' @seealso [vcov()]
#'
#' @method vcov brnb
#' @export
vcov.brnb <- function(object, model = c("mean", "full", "dispersion"), complete = TRUE, ...) {
    object$info_transformed_dispersion <- 1/object$vcov.dispersion
    object$dispersion <- 1
    vcov.brglmFit(object, model , complete , ...)
}

#' [summary()] method for [`"brnb"`][brnb] objects
#'
#' @inheritParams stats::summary.glm
#' @param object an object of class [`"brnb"`][brnb], typically, a
#'     result of a call to [brnb()].
#' @param x an object of class [`"summary.brnb"`][summary.brnb],
#'     usually, a result of a call to [summary.brnb].
#' @details The interface of the summary method for [`"brnb"`][brnb]
#'     objects is similar to that of [`"brglmFit"`][brglmFit] objects
#'     with additional information.
#'
#'     p-values of the individual Wald statistics are based on the
#'     standard normal distribution.
#'
#' @seealso [summary.brglmFit()] and [glm()]
#'
#' @examples
#' # For examples see examples(brnb)
#'
#' @method summary brnb
#' @export
summary.brnb <- function(object, ...) {
    ## coefficient table
    cf <- object$coefficients
    cf.dispersion <- object$dispersion
    se.mean <- sqrt(diag(object$vcov.mean))
    se.dispersion <- sqrt(object$vcov.dispersion)
    cf <- cbind(cf, se.mean, cf/se.mean, 2 * pnorm(-abs(cf/se.mean)))
    colnames(cf) <- c("Estimate", "Std. Error", "z value", "Pr(>|z|)")
    rownames(cf) <- names(object$coefficients)
    object$coefficients <- cf
    object$se.betas <- se.mean
    object$se.dispersion <- se.dispersion
    coef.dispersion <- cbind(cf.dispersion, se.dispersion)
    colnames(coef.dispersion) <- c("Estimate", "Std. Error")
    rownames(coef.dispersion) <- c("dispersion")
    object$coef.dispersion <- coef.dispersion
    ## number of iterations
    object$iterations <- object$iter
    ## delete some slots
    object$terms <- object$model <- object$y <-NULL
    object$x <- object$levels <- object$contrasts <-  NULL
    ## return
    class(object) <- "summary.brnb"
    object
}

#' @rdname summary.brnb
#' @method print summary.brnb
#' @export
print.summary.brnb <- function(x, digits = max(3, getOption("digits") - 3), ...) {
    cat("\nCall:", deparse(x$call, width.cutoff = floor(getOption("width") * 0.85)), "", sep = "\n")
    if (!x$converged) {
        cat("Estimation procedure did not converge\n")
    }
    if (NROW(x$coefficients)) {
        cat(paste("\nCoefficients (mean model with ", x$link, " link):\n", sep = ""))
        printCoefmat(x$coefficients, digits = digits, signif.legend = FALSE)
    } else {
        cat("\nNo coefficients (in mean model)\n")
    }
    if (getOption("show.signif.stars") & any( x$coefficients[, 4L] < 0.1)) {
        cat("---\nSignif. codes: ", "0 '***' 0.001 '**' 0.01 '*' 0.05 '.' 0.1 ' ' 1", "\n")
    }
    if (NROW(x$coef.dispersion)) {
        cat(paste("\n(Dispersion parameter  with ", x$transformation," transformation function: ",
                  round(x$coef.dispersion[,1], 7),")\n", sep = ""))
    } else {
        cat("\nNo coefficients (in precision model)\n")
    }
    cat("\n", apply(cbind(paste(format(c("Null",
                                         "Residual"), justify = "right"), "deviance:"),
                          format(unlist(x[c("null.deviance",
                                            "deviance")]), digits = max(5, digits + 1)), " on",
                          format(unlist(x[c("df.null", "df.residual")])), " degrees of freedom\n"),
                    1, paste, collapse = " "), sep = "")
    cat(paste("AIC:", round(x$aic,digits)))
    cat("\n\nType of estimator:", x$type, get_type_description(x$type))
    cat("\n", "Number of quasi-Fisher scoring iterations:", x$iter, "\n")
    invisible(x)
}


#' @method print brnb
#' @export
print.brnb <- function(x, digits = max(3, getOption("digits") - 3), ...) {
    cat("\nCall:", deparse(x$call, width.cutoff = floor(getOption("width") * 0.85)), "", sep = "\n")
    if (!x$converged) {
        cat("Estimation procedure did not converge\n")
    } else {
        if (length(x$coefficients)) {
            cat(paste("Coefficients (mean model with ", x$link, " link):\n", sep = ""))
            print.default(format(x$coefficients, digits = digits), print.gap = 2, quote = FALSE)
            cat("\n")
        } else {
            cat("No coefficients (in mean model)\n\n")
        }
        cat("\nDegrees of Freedom:", x$df.null, "Total (i.e. Null); ",
            x$df.residual, "Residual\n")
        if (nchar(mess <- naprint(x$na.action)))
            cat("  (", mess, ")\n", sep = "")
        cat("Null Deviance:\t", format(round(x$null.deviance, digits)), "\n")
        cat("Residual Deviance:\t", format(round(x$deviance, digits)),
            "\tAIC:", format(round(x$aic, digits)), "\n")
    }
    cat("\nType of estimator:", x$type, get_type_description(x$type), "\n")
}

#' Method for computing Wald confidence intervals for one or more
#' regression parameters in a [`"brnb"`][brnb] object
#'
#' @inheritParams stats::confint
#'
#' @method confint brnb
#' @export
confint.brnb <- function(object, parm, level = 0.95, ...) {
    confint.default(object, parm, level, ...)
}

#' Simulate Responses
#'
#' Simulate one or more responses from the distribution corresponding
#' to a fitted model [`"brnb"`][brnb] object.
#' @param object an object representing a fitted model.
#' @param nsim number of response vectors to simulate. Defaults to 1.
#' @param seed an object specifying if and how the random number
#'     generator should be initialized; see [set.seed()] for details.
#' @param ... extra arguments to be passed to methods. Not currently
#'     used.

#' @examples
#' # Example in Saha, K., & Paul, S. (2005). Bias-corrected maximum
#' # likelihood estimator of the negative binomial dispersion
#' # parameter.  Biometrics, 61, 179--185.
#' #
#' # Frequency distribution of red mites on apple leaves.
#' nomites <- 0:8
#' noleaves <- c(70, 38, 17, 10, 9, 3, 2, 1, 0)
#' fit_glmnb <- MASS::glm.nb(nomites~1,link="identity",weights = noleaves)
#' fit_brnb <- brnb(nomites ~ 1, link = "identity", transformation = "inverse",
#'                  type = "ML", weights = noleaves)
#' ## Let us simulate 10 response vectors
#' sim_glmnb <- simulate(fit_glmnb, nsim = 10, seed = 123)
#' sim_brnb <-  simulate(fit_brnb, nsim = 10, seed = 123)
#' # The results from glm.nb and brnb with type = "ML" are
#' # exactly the same
#' all.equal(sim_glmnb, sim_brnb, check.attributes = FALSE)


#' @method simulate brnb
#' @export
simulate.brnb <- function(object, nsim = 1, seed = NULL, ...) {
    if (!exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE))
        runif(1)
    if (is.null(seed))
        RNGstate <- get(".Random.seed", envir = .GlobalEnv)
    else {
        R.seed <- get(".Random.seed", envir = .GlobalEnv)
        set.seed(seed)
        RNGstate <- structure(seed, kind = as.list(RNGkind()))
        on.exit(assign(".Random.seed", R.seed, envir = .GlobalEnv))
    }
    theta <- object$theta
    mus <- fitted(object)
    n <- length(mus)
    val <- rnegbin(n*nsim, mu = mus, theta = theta)
    dim(val) <- c(n, nsim)
    val <- as.data.frame(val)
    names(val) <- paste0("sim_", seq_len(nsim))
    attr(val, "seed") <- RNGstate
    val
}

library("MASS")

salmonella <- data.frame(freq = c(15, 16, 16, 27, 33, 20,
                                  21, 18, 26, 41, 38, 27,
                                  29, 21, 33, 60, 41, 42),
                         dose = rep(c(0, 10, 33, 100, 333, 1000), 3),
                         observation = rep(1:3, each = 6))


salmonella_fm <- freq ~ dose + log(dose+10)

fitML_glmnb <- glm.nb(salmonella_fm, data = salmonella)
fitML <- brnb(salmonella_fm, data = salmonella, link = "log", transformation = "identity", type = "ML")
fitBR_mean <- update(fitML, type = "AS_mean")
fitBR_median <- update(fitML, type = "AS_median")
fitBR_mixed <- update(fitML, type = "AS_mixed")
fitBC_mean <- update(fitML, type = "correction")

## brnp returns the same estimates as MASS::glm.nb
expect_equal(coef(fitML), coef(fitML_glmnb), tol = 1e-06)

## brnp returns mean BR and median BR estimates as in Kenne Pagui et al (2021+)
expect_equal(coef(fitBR_mean, model = "full"),
             c(2.216, -0.001, 0.309, 0.065),
             check.attributes = FALSE, tol = 1e-03)
expect_equal(coef(fitBR_median, model = "full"),
             c(2.211, -0.001, 0.309, 0.069),
             check.attributes = FALSE, tol = 1e-03)
expect_equal(coef(fitBC_mean, model = "full"),
             c(2.210, -0.001, 0.311, 0.063),
             check.attributes = FALSE, tol = 1e-03)

## fit using weights

duptimes <- c(2,1,3,5,rep(1,14))
idx <- rep(1:nrow(salmonella), duptimes)
dupsalmonella <- salmonella[idx,]

fitMLe <- update(fitML, data = dupsalmonella)
fitMLw <- update(fitML, weights = duptimes, data = salmonella)

fitBR_mediane <- update(fitML, data = dupsalmonella, type = "AS_median")
fitBR_medianw <- update(fitML, weights = duptimes, data = salmonella, type = "AS_median")

fitBR_meane <- update(fitML, data = dupsalmonella, type = "AS_mean")
fitBR_meanw <- update(fitML, weights = duptimes, data = salmonella, type = "AS_mean")

fitBR_mixede <- update(fitML, data = dupsalmonella, type = "AS_mixed")
fitBR_mixedw <- update(fitML, weights = duptimes, data = salmonella, type = "AS_mixed")

fitBCe <- update(fitML, data = dupsalmonella, type = "AS_mixed")
fitBCw <- update(fitML, weights = duptimes, data = salmonella, type = "AS_mixed")

fitJe <- update(fitML, data = dupsalmonella, type = "MPL_Jeffreys",
                transformation = "inverse")
fitJw <- update(fitML, weights = duptimes, data = salmonella, type = "MPL_Jeffreys",
                transformation = "inverse")

## prior weights work as expected
## all numerical results are the same
expect_equal(coef(fitMLe, "full"), coef(fitMLw, "full"), tolerance = 1e-10)
expect_equal(coef(fitBR_mediane, "full"), coef(fitBR_medianw, "full"), tolerance = 1e-10)
expect_equal(coef(fitBR_mixede, "full"), coef(fitBR_mixedw, "full"), tolerance = 1e-10)
expect_equal(coef(fitBR_meane, "full"), coef(fitBR_meanw, "full"), tolerance = 1e-10)
expect_equal(coef(fitJe, "full"), coef(fitJw, "full"), tolerance = 1e-10)

## Dispersion transformations
## dispersion transformations work as expected for ML/mixed BR/median BR
for (f0 in list(fitML, fitBR_median, fitBR_mixed)) {
    fsqrt <- update(f0, transformation = "sqrt")
    flog <- update(f0, transformation = "log")
    expect_equal(coef(fsqrt, model = "dispersion"),
                 sqrt(coef(f0, model = "dispersion")),
                 tol = 1e-05, check.attributes = FALSE)
    expect_equal(coef(flog, model = "dispersion"),
                 log(coef(f0, model = "dispersion")),
                 tol = 1e-05, check.attributes = FALSE)
}


## dispersion transformations work as expected for mean BR
f0 <- fitBR_mean
fsqrt <- update(f0, transformation = "sqrt")
flog <- update(f0, transformation = "log")
expect_false(isTRUE(all.equal(coef(fsqrt, model = "dispersion"),
                              sqrt(1 / coef(f0, model = "dispersion")),
                              tol = 1e-04, check.attributes = FALSE)))
expect_false(isTRUE(all.equal(coef(flog, model = "dispersion"),
                              -log(coef(f0, model = "dispersion")),
                              tol = 1e-04, check.attributes = FALSE)))

## error is produced for not implemented transformations
expect_error(update(fitBR_mean, transformation = "asd"))



## From https://github.com/ikosmidis/brglm2/issues/31
## Infinite estimates
## df1 <- data.frame(production=c(15, 12, 10, 9, 6, 8, 9, 5, 3, 3, 2, 1, 0, 0, 0, 0),
##                   Treatment_Num=c(1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 4, 4, 4, 4),
##                   Treatment=c("Fresh", "Fresh", "Fresh", "Fresh", "low", "low", "low", "low", "med", "med", "med", "med", "high", "high", "high", "high"),
##                   Genotype=c(1, 1, 2, 2, 1, 1, 2, 2, 1, 1, 2, 2, 1, 1, 2, 2),
##                   Source_Salinity=c("Fresh", "Fresh", "Brackish", "Brackish", "Fresh", "Fresh", "Brackish", "Brackish", "Fresh", "Fresh", "Brackish", "Brackish", "Fresh", "Fresh", "Brackish", "Brackish"),
##                   Days_to_death <- c(500,  500,  500,  500,  400,  350,  300,  500,  200,  202,  260,  280,  150, 150, 160, 140),
##                   censored <- c(1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0))


## salmonella_fm <- production ~ Treatment
## mbr_inv <- brglm2::brnb(salmonella_fm, data = df1, link = "log", transformation = "inverse", type = "AS_mean")
## mbr_log <- brglm2::brnb(salmonella_fm, data = df1, link = "log", transformation = "log", type = "AS_mean")
## transformation = "identity" fails with default starting values
## mbr_sqrt <- brglm2::brnb(salmonella_fm, data = df1, link = "log", transformation = "identity", type = "AS_mean")

# Copyright (C) 2025- Ioannis Kosmidis

#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 or 3 of the License
#  (at your option).
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  A copy of the GNU General Public License is available at
#  http://www.r-project.org/Licenses/

#' Penalized likelihood ratio test for [`"mdyplFit"`][mdyplFit()] objects
#'
#' Computes the Diaconis-Ylvisaker prior penalized likelihood ratio
#' test statistic or its adjusted version using high-dimensionality
#' correction under proportional asymptotics. Associated p-values are
#' also computed using a chi squared distribution.
#'
#' @aliases plrtest
#' @inheritParams summary.mdyplFit
#' @param object1 a [`"mdyplFit"`][mdyplFit()] object
#' @param object2 a [`"mdyplFit"`][mdyplFit()] object
#' @param ... further arguments to be passed to [summary.mdyplFit()].
#'
#' @details
#'
#' Both `object1` and `object2` should have been fitted using the
#' [mdyplFit()] method for [glm()], and the same shrinkage parameter
#' `alpha`; see [mdyplFit()] and [mdyplControl()] for setting `alpha`.
#'
#' If `hd_correction = TRUE` then the deviance and the associated
#' p-value are adjusted using a high-dimensionality correction under
#' proportional asymptotics as in Sterzinger & Kosmidis (2024); see
#' [summary.mdyplFit()].
#'
#' @author Ioannis Kosmidis `[aut, cre]` \email{ioannis.kosmidis@warwick.ac.uk}
#'
#' @seealso [mdyplFit()], [summary.mdyplFit()], [mdypl_control()]
#'
#' @references
#'
#' Sterzinger P, Kosmidis I (2024). Diaconis-Ylvisaker prior
#' penalized likelihood for \eqn{p/n \to \kappa \in (0,1)} logistic
#' regression. *arXiv*:2311.07419v2, \url{https://arxiv.org/abs/2311.07419}.
#'
#' @method plrtest mdyplFit
#' @export
plrtest.mdyplFit <- function(object1, object2, hd_correction = FALSE, ...) {
    if (!inherits(object1, "mdyplFit") || !inherits(object2, "mdyplFit")) {
        stop("At least one of `object1` and `object2` is not of class `mdyplFit`")
    }
    if (!identical(object1$alpha, object2$alpha)) {
        stop("`object1` and `object2` have not been fitted using the same `alpha`. See ?mdypl_control.")
    }
    anv <- anova(object1, object2, test = "LRT")
    if (isTRUE(hd_correction)) {
        npars1 <- object1$rank
        npars2 <- object2$rank
        if (npars2 > npars1) {
            summ <- summary(object2, hd_correction = TRUE, ...)
        } else {
            summ <- summary(object1, hd_correction = TRUE, ...)
        }
        se_pars <- summ$se_parameters
        anv$Deviance <- anv$Deviance * se_pars[2] / (summ$kappa * se_pars[3]^2)
        anv[, "Pr(>Chi)"] <- pchisq(anv$Deviance, abs(anv$Df), lower.tail = FALSE)
        attr(anv, "hd_correction") <- TRUE
        attr(anv, "kappa") <- summ$kappa
        attr(anv, "signal_strength") <- summ$signal_strength
        attr(anv, "se_parameters") <- se_pars
    }
    class(anv) <- c("anova_mdyplFit", class(anv))
    anv
}

#' @method print anova_mdyplFit
#' @export
print.anova_mdyplFit <- function(x, ...) {
    print_anova(x, ...)
    if (isTRUE(attr(x, "hd_correction"))) {
        se_pars <- attr(x, "se_parameters")
        cat("\nHigh-dimensionality correction applied with")
        cat("\nDimentionality parameter (kappa) =", round(attr(x, "kappa"), 2))
        cat("\nEstimated signal strength (gamma) =", round(attr(x, "signal_strength"), 2))
        cat("\nState evolution parameters (mu, b, sigma) =", paste0("(", paste(round(se_pars[1:3], 2), collapse = ", "), ")"), "with max(|funcs|) =", max(abs(attr(se_pars, "funcs"))), "\n")
    }
}

print_anova <- utils::getFromNamespace("print.anova", "stats")

library("mbrglm")

data("lizards", package = "brglm2")
data("endometrial",package = "brglm2")

links <- lapply(c("logit", "probit", "cloglog", "cauchit"), make.link)

tol <- 1e-10
mbrglmControl <- mbrglm.control(mbr.epsilon = 1e-10, mbr.maxit = 1000)

for (l in seq_along(links)) {
    ## Lizards
    lizardsFormula <- cbind(grahami, opalinus) ~ height + diameter + light + time
    lizardsMBRlegacy <- mbrglm(lizardsFormula, family = binomial(links[[l]]), data = lizards,
                                       method = "mbrglm.fit",
                                       control.mbrglm = mbrglmControl)
    lizardsMBR <- glm(lizardsFormula, family = binomial(links[[l]]), data = lizards,
                      method = "brglmFit", type= "AS_median", epsilon = 1e-10, maxit = 1000)

    ## Endometrial
    endoFormula <- HG ~ NV + PI + EH
    endoMBRlegacy <- mbrglm(endoFormula, family = binomial(links[[l]]), data = endometrial,
                            method = "mbrglm.fit",
                            control.mbrglm = mbrglmControl)
    endoMBR <- glm(endoFormula, family = binomial(links[[l]]), data = endometrial,
                   method = "brglmFit", type = "AS_median", epsilon = 1e-10, maxit = 1000)

    c1 <- coef(summary(endoMBRlegacy))
    c2 <- coef(summary(endoMBR))

    ## glm with brglmFit method and mbrglm return the same coefficients when link is" links[[l]]$name
    expect_equal(coef(lizardsMBR), coef(lizardsMBRlegacy), tolerance = tol)
    expect_equal(c1, c2, tolerance = tol)
}

library("brglm")
data("lizards", package = "brglm2")
links <- lapply(c("logit", "probit", "cloglog", "cauchit"), make.link)

tol <- 1e-08
for (l in seq_along(links)) {
    lizardsBRlegacy <- brglm(cbind(grahami, opalinus) ~ height + diameter +
                                 light + time, family = binomial(links[[l]]),
                             data = lizards, method = "brglm.fit",
                             br.epsilon = 1e-10, br.maxit = 1000)

    lizardsBR <- glm(cbind(grahami, opalinus) ~ height + diameter +
                         light + time, family = binomial(links[[l]]), data=lizards,
                     method = "brglmFit", epsilon = 1e-10, maxit = 1000)
    ## glm with brglm.fit method and brglm_0 return the same coefficients for the lizards when link is links[[l]]$name)
    expect_equal(coef(lizardsBR), coef(lizardsBRlegacy), tolerance = tol)
}


## Performance comparisons BR versus ML
## link1 <- "cauchit"
## system.time(replicate(100, {lizardsBR <- glm(cbind(grahami, opalinus) ~ height + diameter + light + time, family = binomial(link1), data=lizards, method = "brglmFit", epsilon = 1e-10, maxit = 1000, type = "ML")}))

## system.time(replicate(100, {lizardsBR <- glm(cbind(grahami, opalinus) ~ height + diameter + light + time, family = binomial(link1), data=lizards, method = "glm.fit", epsilon = 1e-10, maxit = 1000)}))

## link1 <- "cauchit"
## system.time(replicate(100, {lizardsBR <- glm(cbind(grahami, opalinus) ~ height + diameter + light + time, family = binomial(link1), data=lizards, method = "brglmFit", epsilon = 1e-10, maxit = 1000)}))

## system.time(replicate(100, {lizardsBR <- brglm(cbind(grahami, opalinus) ~ height + diameter + light + time, family = binomial(link1), data=lizards, br.epsilon = 1e-10, br.maxit = 1000)}))

## A Gamma example, from McCullagh & Nelder (1989, pp. 300-2)
clotting <- data.frame(
    u = c(5,10,15,20,30,40,60,80,100, 5,10,15,20,30,40,60,80,100),
    conc = c(118,58,42,35,27,25,21,19,18,69,35,26,21,18,16,13,12,12),
    lot = factor(c(rep(1, 9), rep(2, 9))))
mod_identity <- glm(conc ~ lot*log(u), data = clotting, family = Gamma,
                    method = "brglm_fit", type = "ML",
                    transformation = "identity")
mod_log <- glm(conc ~ lot*log(u), data = clotting, family = Gamma,
                    method = "brglm_fit", type = "ML",
                    transformation = "log")
mod_sqrt <- glm(conc ~ lot*log(u), data = clotting, family = Gamma,
                    method = "brglm_fit", type = "ML",
                    transformation = "sqrt")
mod_inverse <- glm(conc ~ lot*log(u), data = clotting, family = Gamma,
                    method = "brglm_fit", type = "ML",
                    transformation = "inverse")

c_identity <- coef(mod_identity, model = "full")
c_log <- coef(mod_log, model = "full")
c_log[5] <- exp(c_log[5])
c_sqrt <- coef(mod_sqrt, model = "full")
c_sqrt[5] <- c_sqrt[5]^2
c_inverse <- coef(mod_inverse, model = "full")
c_inverse[5] <- 1/c_inverse[5]

tol <- 1e-08
## ML estimate of gamma dispersion from brglmFit is invariant to trasnformation
expect_equal(c_identity, c_log, tolerance = tol, check.attributes = FALSE)
expect_equal(c_identity, c_sqrt, tolerance = tol, check.attributes = FALSE)
expect_equal(c_identity, c_inverse, tolerance = tol, check.attributes = FALSE)

library("numDeriv")
library("brglm")

## A Gamma example, from McCullagh & Nelder (1989, pp. 300-2)
clotting <- data.frame(
    u = c(5,10,15,20,30,40,60,80,100, 5,10,15,20,30,40,60,80,100),
    conc = c(118,58,42,35,27,25,21,19,18,69,35,26,21,18,16,13,12,12),
    lot = factor(c(rep(1, 9), rep(2, 9))))
mod <- enrich(glm(conc ~ lot*log(u), data = clotting, family = Gamma()))

ifun <- get_information_function(mod)

penloglik <- function(pars, X, y) {
    p <- ncol(X)
    beta <- pars[1:p]
    dispersion <- pars[p + 1]
    eta <- X %*% beta
    mu <- 1/eta
    sum(dgamma(y, shape = 1/dispersion, scale = mu * dispersion, log = TRUE)) + 0.5 * log(det(ifun(beta, dispersion)))
}

modjef <- update(mod, method = "brglmFit", type = "MPL_Jeffreys", epsilon = 1e-15)

## the numerical gradient of the penalized log-likelihood is almost zero when evaluated at the estimates from type = 'MPL_Jeffreys
expect_equal(grad(penloglik,
                  x = coef(modjef, model = "full"),
                  X = model.matrix(mod),
                  y = model.response(mod$model)), rep(0, 5), tolerance = 1.5 * 1e-05)

## the numerical gradient of the penalized log-likelihood matches that from type = 'MPL_Jeffreys'
expect_warning(g1 <- update(mod, method = "brglmFit", type = "MPL_Jeffreys", maxit = 0, start = coef(mod, model = "full"))$grad)
g2 <- grad(penloglik,
           x = coef(mod, model = "full"),
           X = model.matrix(mod),
           y = model.response(mod$model))
expect_equal(unname(g1), g2, tolerance = 1e-05)

## source(system.file("inst", "brglm0/brglm0.R", package = "brglm2"))
data("lizards", package = "brglm2")

links <- lapply(c("logit", "probit", "cloglog", "cauchit"), make.link)


tol <- 1e-08
for (l in seq_along(links)) {
    expect_warning(
        lizardsBRlegacy <- brglm(cbind(grahami, opalinus) ~ height + diameter +
                                     light + time, family = binomial(links[[l]]), data=lizards,
                                 method = "brglm.fit", br.epsilon = 1e-10, br.maxit = 1000, pl = TRUE)
    )

    expect_warning(
        lizardsBR <- glm(cbind(grahami, opalinus) ~ height + diameter +
                             light + time, family = binomial(links[[l]]), data=lizards,
                         method = "brglmFit", epsilon = 1e-10, maxit = 1000, type = "MPL_Jeffreys")
    )
    ## glm with brglm.fit method and brglm_0 return the same coefficients for the lizards when link is links[[l]]$name)
    expect_equal(coef(lizardsBR), coef(lizardsBRlegacy), tolerance = tol)
}

library("MASS")

## A Gamma example, from McCullagh & Nelder (1989, pp. 300-2)
clotting <- data.frame(
    u = c(5,10,15,20,30,40,60,80,100, 5,10,15,20,30,40,60,80,100),
    conc = c(118,58,42,35,27,25,21,19,18,69,35,26,21,18,16,13,12,12),
    lot = factor(c(rep(1, 9), rep(2, 9))))
mod <- glm(conc ~ lot*log(u), data = clotting, family = Gamma)

tol <- sqrt(.Machine$double.eps)
## ML estimate of gamma dispersion from brglmFit and from MASS::gamma.dispersion are the same
expect_equal(update(mod, method = "brglmFit", epsilon = 1e-10)$dispersion_ML, MASS::gamma.dispersion(mod), tolerance = tol)

## ML estimate of gamma shape from brglmFit and from MASS::gamma.dispersion are the same
expect_equal(1/update(mod, method = "brglmFit", epsilon = 1e-10, transformation = "inverse")$dispersion_ML, gamma.shape(mod)$alpha, tolerance = tol)

library("enrichwith")

## A Gamma example, from McCullagh & Nelder (1989, pp. 300-2)
clotting <- data.frame(
    u = c(5,10,15,20,30,40,60,80,100, 5,10,15,20,30,40,60,80,100),
    conc = c(118,58,42,35,27,25,21,19,18,69,35,26,21,18,16,13,12,12),
    lot = factor(c(rep(1, 9), rep(2, 9))))

mod <- glm(conc ~ lot*log(u), data = clotting, family = Gamma)
emod <- enrich(mod, with = "all")

coefs <- coef(emod, model = "mean")
disp <- coef(emod, model = "dispersion")
coefs_bc <- c(coefs, disp) - get_bias_function(emod)(coefs, disp)
attributes(coefs_bc) <- NULL

mod_bc <- glm(conc ~ lot*log(u), data = clotting, family = Gamma, method = "brglmFit", type = "correction")

## bias corrected estimates computed using enrichwith are the same as those when having type = 'correction' in brglmFit
expect_equal(unname(coefs_bc), unname(coef(mod_bc, model = "full")), check.attributes = FALSE, tolerance = 1e-06)

data("anorexia", package = "MASS")

anorexML <- glm(Postwt ~ Prewt + Treat+ offset(Prewt),
                family = gaussian, data = anorexia)
anorexBR <- update(anorexML, method = "brglmFit", type = "AS_mean")

tol <- sqrt(.Machine$double.eps)
## dispersion_ML is the usual biased estimate for the residual variance
expect_equal(anorexBR$dispersion_ML, sum((anorexia$Postwt - fitted(anorexML))^2)/nrow(anorexia),
             tolerance = 1e-06)

## dispersion is the usual bias-corrected estimate for the residual variance
expect_equal(anorexBR$dispersion, sum((anorexia$Postwt - fitted(anorexML))^2)/(nrow(anorexia) - length(coef(anorexML))), tolerance = 1e-06)


## context("dispersion parameter estimation")

## set.seed(123)
## N <- 20
## x <- matrix(rnorm(2*N), N, 2)
## y <- 0.3 + drop(x %*% c(-3, 1)) + rnorm(N, 0.4)

## fitML <- glm(y ~ x, family = gaussian())
## fitBR <- update(fitML, method = "brglmFit")

## tol <- sqrt(.Machine$double.eps)
## test_that("dispersionML is the usual biased estimate for the residual variance",
##           expect_equal(fitBR$dispersionML, sum((fitML$y - fitted(fitML))^2)/N))

## test_that("dispersion is the usual bias-corrected estimate for the residual variance",
##           expect_equal(fitBR$dispersion, sum((fitML$y - fitted(fitML))^2)/(N - 3)))

## A Gamma example, from McCullagh & Nelder (1989, pp. 300-2)
clotting <- data.frame(
    u = c(5,10,15,20,30,40,60,80,100, 5,10,15,20,30,40,60,80,100),
    conc = c(118,58,42,35,27,25,21,19,18,69,35,26,21,18,16,13,12,12),
    lot = factor(c(rep(1, 9), rep(2, 9))))
mod <- glm(conc ~ lot*log(u), data = clotting, family = Gamma, epsilon = 1e-10, maxit = 1000)

mod1 <- update(mod, method = "brglmFit", start = c(coef(mod)*0.9, 1), epsilon = 1e-10, maxit = 1000)
mod2 <- update(mod, method = "brglmFit", start = c(coef(mod)*0.9), epsilon = 1e-10, maxit = 1000)

tol <- 1e-03
## start argument is passed correctly in brglmFit"
expect_equal(coef(mod), coef(mod1), tolerance = tol)
expect_equal(coef(mod), coef(mod2), tolerance = tol)

data("anorexia", package = "MASS")

anorexML <- glm(Postwt ~ Prewt + Treat+ offset(Prewt),
                family = gaussian, data = anorexia)
anorexBR <- update(anorexML, method = "brglmFit")
anorexMBR <- update(anorexML, method = "brglmFit", control = list(type="AS_median"))

tol <- sqrt(.Machine$double.eps)

## dispersion is RSS over residual degrees of freedom minus 2/3"
expect_equal(anorexMBR$dispersion, sum((anorexia$Postwt - fitted(anorexML))^2)/(nrow(anorexia) - length(coef(anorexML)) - 2/3))

## A Gamma example, from McCullagh & Nelder (1989, pp. 300-2)
clotting <- data.frame(
    u = c(5,10,15,20,30,40,60,80,100, 5,10,15,20,30,40,60,80,100),
    conc = c(118,58,42,35,27,25,21,19,18,69,35,26,21,18,16,13,12,12),
    lot = factor(c(rep(1, 9), rep(2, 9))))
mod <- glm(conc ~ lot*log(u) + I(2*log(u)), data = clotting, family = Gamma)

X <- model.matrix(mod)
Y <- mod$y
## brglmFit returns an error if singular.ok = TRUE
expect_error(brglm_fit(X, Y, family = Gamma(), singular.ok = FALSE),
             pattern = "singular fit encountered")
expect_true(is.na(coef(brglm_fit(X, Y, family = Gamma(), singular.ok = TRUE))["I(2 * log(u))"]))

data("lizards", package = "brglm2")

data("endometrial",package = "brglm2")

links <- lapply(c("logit", "probit", "cloglog", "cauchit"), make.link)

tol <- 1e-10
mbrglmControl <- mbrglm::mbrglm.control(mbr.epsilon = 1e-10, mbr.maxit = 1000)

for (l in seq_along(links)) {

    ## Lizards
    lizardsFormula <- cbind(grahami, opalinus) ~ height + diameter + light + time
    lizardsMBRlegacy <- mbrglm::mbrglm(lizardsFormula, family = binomial(links[[l]]), data = lizards,
                                       method = "mbrglm.fit",
                                       control.mbrglm = mbrglmControl)
    expect_warning(lizardsMBR <- glm(lizardsFormula, family = binomial(links[[l]]), data = lizards,
                                     method = "brglmFit", type="AS_median", epsilon = 1e-10, maxit = 1000))

    ## Endometrial
    endoFormula <- HG ~ NV + PI + EH
    expect_warning({
        endoMBRlegacy <- mbrglm::mbrglm(endoFormula, family = binomial(links[[l]]), data = endometrial,
                                        method = "mbrglm.fit",
                                        control.mbrglm = mbrglmControl)
        endoMBR <- glm(endoFormula, family = binomial(links[[l]]), data = endometrial,
                       method = "brglmFit", type = "AS_median", epsilon = 1e-10, maxit = 1000)
    })

    c1 <- coef(summary(endoMBRlegacy))
    c2 <- coef(summary(endoMBR))

    ## glm with brglmFit method and mbrglm return the same coefficients when link is" links[[l]]$name
    expect_equal(coef(lizardsMBR), coef(lizardsMBRlegacy), tolerance = tol)
    expect_equal(c1, c2, tolerance = tol)
}

library("numDeriv")
library("brglm")

## A Gamma example, from McCullagh & Nelder (1989, pp. 300-2)
clotting <- data.frame(
    u = c(5,10,15,20,30,40,60,80,100, 5,10,15,20,30,40,60,80,100),
    conc = c(118,58,42,35,27,25,21,19,18,69,35,26,21,18,16,13,12,12),
    lot = factor(c(rep(1, 9), rep(2, 9))))
mod <- enrich(glm(conc ~ lot*log(u), data = clotting, family = Gamma()))

ifun <- get_information_function(mod)

penloglik <- function(pars, X, y) {
    p <- ncol(X)
    beta <- pars[1:p]
    dispersion <- pars[p + 1]
    eta <- X %*% beta
    mu <- 1/eta
    sum(dgamma(y, shape = 1/dispersion, scale = mu * dispersion, log = TRUE)) + 0.5 * log(det(ifun(beta, dispersion)))
}

modjef <- update(mod, method = "brglmFit", type = "MPL_Jeffreys", epsilon = 1e-15)

## the numerical gradient of the penalized log-likelihood is almost zero when evaluated at the estimates from type = 'MPL_Jeffreys
expect_equal(grad(penloglik,
                  x = coef(modjef, model = "full"),
                  X = model.matrix(mod),
                  y = model.response(mod$model)), rep(0, 5), tolerance = 1.5 * 1e-05)

## the numerical gradient of the penalized log-likelihood matches that from type = 'MPL_Jeffreys'
expect_warning(
    g1 <- update(mod, method = "brglmFit", type = "MPL_Jeffreys", maxit = 0, start = coef(mod, model = "full"))$grad
)
g2 <- grad(penloglik,
           x = coef(mod, model = "full"),
           X = model.matrix(mod),
           y = model.response(mod$model))
expect_equal(unname(g1), g2, tolerance = 1e-05)

## source(system.file("inst", "brglm0/brglm0.R", package = "brglm2"))
data("lizards", package = "brglm2")

links <- lapply(c("logit", "probit", "cloglog", "cauchit"), make.link)


tol <- 1e-08
for (l in seq_along(links)) {
    expect_warning(
        lizardsBRlegacy <- brglm(cbind(grahami, opalinus) ~ height + diameter +
                                     light + time, family = binomial(links[[l]]), data=lizards,
                                 method = "brglm.fit", br.epsilon = 1e-10, br.maxit = 1000, pl = TRUE)
    )

    lizardsBR <- glm(cbind(grahami, opalinus) ~ height + diameter +
                         light + time, family = binomial(links[[l]]), data=lizards,
                     method = "brglmFit", epsilon = 1e-10, maxit = 1000, type = "MPL_Jeffreys")
    ## glm with brglm.fit method and brglm_0 return the same coefficients for the lizards when link is links[[l]]$name)
    expect_equal(coef(lizardsBR), coef(lizardsBRlegacy), tolerance = tol)
}

---
title: "Adjacent category logit models using **brglm2**"
author: "[Ioannis Kosmidis](https://www.ikosmidis.com)"
date: "05 February 2019"
output: rmarkdown::html_vignette
bibliography: brglm2.bib
nocite: |
  @kosmidis:2019
vignette: >
  %\VignetteIndexEntry{Adjacent category logit models using **brglm2**}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6,
  fig.height = 6
)
```

# **bracl**
The [**brglm2**](https://github.com/ikosmidis/brglm2) R package provides `bracl()` which is a wrapper of `brglmFit()` for fitting adjacent category models for ordinal responses using either maximum likelihood or maximum penalized likelihood or any of the various bias reduction methods described in `brglmFit()`. There is a formal equivalence between adjacent category logit models for ordinal responses and multinomial logistic regression models (see, e.g. the [Multinomial logistic regression using brglm2](https://cran.r-project.org/package=brglm2/vignettes/multinomial.html) vignette and the `brmultinom()` function). `bracl()` utilizes that equivalence and fits the corresponding Poisson log-linear model, by appropriately re-scaling the Poisson means to match the multinomial totals (a.k.a. the "Poisson trick"). The mathematical details and algorithm on using the Poisson trick for mean-bias reduction are given in @kosmidis:11.



# Citation
If you found this vignette or **brglm2**, in general, useful, please consider citing **brglm2** and the associated paper. You can find information on how to do this by typing `citation("brglm2")`.

# Opinion on stem cell research and religious fundamentalism

The `stemcell` data set ships with **brglm2**. @agresti:15[, Section 4.1] provides a detailed description of the variables recorded in this data set (see also `?stemcell`).
```{r, echo = TRUE}
library("brglm2")
data("stemcell", package = "brglm2")
stem <- within(stemcell, religion <- as.numeric(religion))
```
## Maximum likelihood estimation
The following chunk of code fits an adjacent category logit model with proportional odds and reproduces @agresti:10[, Table 4.2]. Note that the intercept parameters are different because @agresti:10[, Table 4.2] uses different contrasts for the intercept parameters.
```{r, echo = TRUE}
stem_formula <- research ~ religion + gender
stemcells_ml <- bracl(stem_formula, weights = frequency, data = stem,
                      parallel = TRUE, type = "ML")
summary(stemcells_ml)
```
`stemcells_ml` is an object inheriting from
```{r, echo = TRUE}
class(stemcells_ml)
```
**brglm2** implements `print`, `coef`, `fitted`, `predict`, `summary`, `vcov` and `logLik` methods for


We can check if a model with non-proportional odds fits the data equally well by fitting it and carrying out a likelihood ration test.
```{r, echo = TRUE}
stemcells_ml_full <- bracl(stem_formula, weights = frequency, data = stemcell,
                           parallel = FALSE, type = "ML")
summary(stemcells_ml_full)
```
The value of the log likelihood ratio statistic here is
```{r, echo = TRUE}
(lrt <- deviance(stemcells_ml) - deviance(stemcells_ml_full))
```
and has an asymptotic chi-squared distribution with
```{r, echo = TRUE}
(df1 <- df.residual(stemcells_ml) - df.residual(stemcells_ml_full))
```
The p-value from testing the hypothesis that `stemcells_ml_full` is an as good fit as `stemcells_ml` is
```{r, echo = TRUE}
pchisq(lrt, df1, lower.tail = FALSE)
```
hence, the simpler model is found to be as adequate as the full model is.

# Mean and median bias reduction
We can use `bracl()` to fit the adjacent category model using estimators with smaller mean or median bias. For mean bias reduction we do
```{r, echo = TRUE}
summary(update(stemcells_ml, type = "AS_mean"))
```
and for median
```{r, echo = TRUE}
summary(update(stemcells_ml, type = "AS_median"))
```
The estimates from mean and median bias reduction are similar to the maximum likelihood ones, indicating that estimation bias is not a major issue here.

# Prediction
We can predict the category probabilities using the `predict()` method
```{r, echo = TRUE}
predict(stemcells_ml, type = "probs")
```

# Relevant resources
`?brglmFit` and `?brglm_control` provide descriptions of the various bias reduction methods supported in **brglm2**. The [`iteration`](https://cran.r-project.org/package=brglm2/brglm2.pdf) vignette describes the iteration and gives the mathematical details for the bias-reducing adjustments to the score functions for generalized linear models.


# Citation
If you found this vignette or **brglm2**, in general, useful, please consider citing **brglm2** and the associated paper. You can find information on how to do this by typing `citation("brglm2")`.


# References

---
title: "Multinomial logistic regression using **brglm2**"
author: "[Ioannis Kosmidis](https://www.ikosmidis.com)"
date: "01 July 2017"
output: rmarkdown::html_vignette
bibliography: brglm2.bib
vignette: >
  %\VignetteIndexEntry{Multinomial logistic regression using **brglm2**}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6,
  fig.height = 6
)
```


# **brmultinom**
The [**brglm2**](https://github.com/ikosmidis/brglm2) R package provides `brmultinom()` which is a wrapper of `brglmFit` for fitting multinomial logistic regression models (a.k.a. baseline category logit models) using either maximum likelihood or maximum penalized likelihood or any of the various bias reduction methods described in `brglmFit()`. `brmultinom()` uses the equivalent Poisson log-linear model, by appropriately re-scaling the Poisson means to match the multinomial totals (a.k.a. the "Poisson trick"). The mathematical details and algorithm on using the Poisson trick for mean-bias reduction are given in @kosmidis:11.

This vignettes illustrates the use of `brmultinom()` and of the associated methods, using the alligator food choice example in @agresti:02[, Section 7.1]

# Alligator data

The alligator data set ships with **brglm2**. @agresti:02[, Section 7.1] provides a detailed description of the variables recorded in the data set.
```{r, echo = TRUE}
library("brglm2")
data("alligators", package = "brglm2")
```

## Maximum likelihood estimation

The following chunk of code reproduces @agresti:02[, Table 7.4]. Note that in order to get the estimates and standard errors reported in the latter table, we have to explicitly specify the contrasts that @agresti:02 uses.
```{r, echo = TRUE}
agresti_contrasts <- list(lake = contr.treatment(levels(alligators$lake), base = 4),
                          size = contr.treatment(levels(alligators$size), base = 2))
all_ml <- brmultinom(foodchoice ~ size + lake , weights = freq,
                     data = alligators,
                     contrasts = agresti_contrasts,
                     ref = 1,
                     type = "ML")
all_ml_summary <- summary(all_ml)
## Estimated regression parameters
round(all_ml_summary$coefficients, 2)
## Estimated standard errors
round(all_ml_summary$standard.errors, 2)
```

## Mean and median bias reduction

Fitting the model using mean-bias reducing adjusted score equations gives
```{r, echo = TRUE}
all_mean <- update(all_ml, type = "AS_mean")
summary(all_mean)
```
The corresponding fit using median-bias reducing adjusted score equations is
```{r, echo = TRUE}
all_median <- update(all_ml, type = "AS_median")
summary(all_median)
```

The estimates and the estimated standard errors from bias reduction
are close to those for maximum likelihood. As a result, it is unlikely
that either mean or median bias is of any real consequence for this
particular model and data combination.

# Infinite estimates and multinomial logistic regression

Let's scale the frequencies in `alligators` by 3 in order to get a sparser data set. The differences between maximum likelihood and mean and median bias reduction should be more apparent on the resulting data set. Here we have to "slow-down" the Fisher scoring iteration (by scaling the step-size), because otherwise the Fisher information matrix quickly gets numerically rank-deficient. The reason is data separation [@albert:84].
```{r, echo = TRUE, error = TRUE}
all_ml_sparse <- update(all_ml, weights = round(freq/3), slowit = 0.1)
summary(all_ml_sparse)
```
Specifically, judging from the estimated standard errors, the estimates for `(Intercept)`, `lakeHancock`, `lakeOklawaha` and `lakeTrafford` for `Reptile` and `lakeHancock` for `Bird` seem to be infinite.

To quickly check if that's indeed the case we can use the `check_infinite_estimates()` method of the [**detectseparation**](https://cran.r-project.org/package=detectseparation) R package.
```{r, echo = TRUE}
library("detectseparation")
se_ratios <- check_infinite_estimates(all_ml_sparse)
plot(se_ratios)
```

Some of the estimated standard errors diverge as the number of Fisher scoring iterations increases, which is evidence of complete or quasi-complete separation [@lesaffre:89].

In contrast, both mean and median bias reduction result in finite
estimates
```{r, echo = TRUE}
all_mean_sparse <- update(all_ml_sparse, type = "AS_mean")
summary(all_mean_sparse)

all_median_sparse <- update(all_ml_sparse, type = "AS_median")
summary(all_median_sparse)
```



# Relevant resources
`?brglmFit` and `?brglm_control` contain quick descriptions of the various bias reduction methods supported in **brglm2**. The [`iteration`](https://cran.r-project.org/package=brglm2/brglm2.pdf) vignette describes the iteration and gives the mathematical details for the bias-reducing adjustments to the score functions for generalized linear models.

# Citation
If you found this vignette or **brglm2**, in general, useful, please consider citing **brglm2** and the associated paper. You can find information on how to do this by typing `citation("brglm2")`.

# References

---
title: "Negative binomial regression using **brglm2**"
author: "Euloge Clovis Kenne Pagui, [Ioannis Kosmidis](https://www.ikosmidis.com)"
date: "12 June 2021"
output: rmarkdown::html_vignette
bibliography: brglm2.bib
vignette: >
  %\VignetteIndexEntry{Negative binomial regression using **brglm2**}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6,
  fig.height = 6
)
```


# **brnb**

The [**brglm2**](https://github.com/ikosmidis/brglm2) R package provides the `brnb()` function for fitting negative binomial regression models (see @agresti:15, Section 7.3, for a recent account on negative binomial regression models) using either maximum likelihood or any of the various bias reduction and adjusted estimating functions methods provided by `brglmFit()` (see `?brglmFit` for resources).

This vignette demonstrates the use of `brnb()` and of the associated methods, using the case studies in @kenne:20.

# Ames salmonella data

@magolin:89 provide data from an Ames salmonella reverse mutagenicity assay. The response variable corresponds to the number of revertant colonies observed (`freq`) on each of three replicate plates (`plate`), and the covariate (`dose`) is the dose level of quinoline on the plate in micro-grams. The code chunk below sets up a data frame with the data from replicate 1 in @magolin:89[, Table 1].

```{r, echo = TRUE}
freq <- c(15, 16, 16, 27, 33, 20,
          21, 18, 26, 41, 38, 27,
          29, 21, 33, 60, 41, 42)
dose <- rep(c(0, 10, 33, 100, 333, 1000), 3)
plate <- rep(1:3, each = 6)
(salmonella <- data.frame(freq, dose, plate))
```

The following code chunks reproduces @kenne:20[, Table 2] by estimating the negative binomial regression model with log link and model formula

```{r, echo = TRUE}
ames_f <- freq ~ dose + log(dose + 10)
```

using the various estimation methods that `brnb()` supports.

## Maximum likelihood estimation


```{r, echo = TRUE}
library("brglm2")
ames_ML <- brnb(ames_f, link = "log", data = salmonella,
                transformation = "identity",  type = "ML")
## Estimated regression and dispersion parameters
est <- coef(ames_ML, model = "full")
## Estimated standard errors for the regression parameters
sds <- sqrt(c(diag(ames_ML$vcov.mean), ames_ML$vcov.dispersion))
round(cbind(est, sds), 4)
```

## Bias reduction 

### Asymptotic mean-bias correction

The following code chunks updates the model fit using asymptotic mean-bias correction for estimating the model parameters

```{r, echo = TRUE}
ames_BC <- update(ames_ML, type = "correction")
## Estimated regression and dispersion parameters
est <- coef(ames_BC, model = "full")
## Estimated standard errors for the regression parameters
sds <- sqrt(c(diag(ames_BC$vcov.mean), ames_BC$vcov.dispersion))
round(cbind(est, sds), 4)
```

### Mean-bias reducing adjusted score equations

The corresponding fit using mean-bias reducing adjusted score equations is

```{r, echo = TRUE}
ames_BRmean <- update(ames_ML, type = "AS_mean")
## Estimated regression and dispersion parameters
est <- coef(ames_BRmean, model = "full")
## Estimated standard errors for the regression parameters
sds <- sqrt(c(diag(ames_BRmean$vcov.mean), ames_BRmean$vcov.dispersion))
round(cbind(est, sds), 4)
```

## Median-bias reducing adjusted score equations

The corresponding fit using median-bias reducing adjusted score equations is

```{r, echo = TRUE}
ames_BRmedian <- update(ames_ML, type = "AS_median")
## Estimated regression and dispersion parameters
est <- coef(ames_BRmedian, model = "full")
## Estimated standard errors for the regression parameters
sds <- sqrt(c(diag(ames_BRmedian$vcov.mean), ames_BRmedian$vcov.dispersion))
round(cbind(est, sds), 4)
```

## Mixed bias reducing adjusted score equations

As is done in @kosmidis:2019[, Section 4] for generalized linear models,
we can exploit the Fisher orthogonality of the regression parameters
and the dispersion parameter and use a composite bias reduction
adjustment to the score functions. Such an adjustment delivers
mean-bias reduced estimates for the regression parameters and a
median-bias reduced estimate for the dispersion parameter. The
resulting estimates of the regression parameters are invariant in terms
of their mean bias properties under arbitrary contrasts, and that of
the dispersion parameter is invariant in terms of its median bias
properties under monotone transformations.

Fitting the model using mixed-bias reducing adjusted score equations gives

```{r, echo = TRUE}
ames_BRmixed <- update(ames_ML, type = "AS_mixed")
## Estimated regression and dispersion parameters
est <- coef(ames_BRmixed, model = "full")
## Estimated standard errors for the regression parameters
sds <- sqrt(c(diag(ames_BRmixed$vcov.mean), ames_BRmixed$vcov.dispersion))
round(cbind(est, sds), 4)
```


<!-- # Epileptic seizures data -->

<!--  @kenne:20[, Section 5.2] provides a detailed description of the variables recorded in the data set. -->
<!-- ```{r, echo = TRUE} -->
<!-- library(MASS) -->
<!-- epil2 <- epil[epil$period == 1, ] -->
<!-- epil2["period"] <- rep(0, 59); epil2["y"] <- epil2["base"]; epil["time"] <- 1;  -->
<!-- epil2["time"] <- 4 -->
<!-- epil2 <- rbind(epil, epil2) -->
<!-- epil2$pred <- unclass(epil2$trt) * (epil2$period > 0); epil2$subject <- factor(epil2$subject) -->
<!-- epil3 <- aggregate(epil2, list(epil2$subject, epil2$period > 0), -->
<!--                    function(x) if(is.numeric(x)) sum(x) else x[1])  -->
<!-- epil3$pred <- factor(epil3$pred, -->
<!--                      labels = c("base", "placebo", "drug")) -->
<!-- contrasts(epil3$pred) <- structure(contr.sdif(3), -->
<!--                                    dimnames = list(NULL, c("placebo-base", "drug-placebo"))) -->
<!-- ``` -->

<!-- ## Maximum likelihood estimation -->

<!-- The following chunk of code reproduces @kenne:20[, Figure 4]. The model uses the log link on the mean and identity transformation for dispersion parameter. -->
<!-- ```{r, echo = TRUE} -->
<!-- epil3_ML <- brnb(y ~ -1+ factor(subject) + factor(pred),  data = epil3, type = "ML") -->
<!-- ## Estimated of interest regression coefficients and dispersion parameters -->
<!-- round(c(epil3_ML$coefficients[-c(1:59)], epil3_ML$dispersion), 2) -->
<!-- ## Estimated standard errors -->
<!-- round(c(summary(epil3_ML)$coefficients[-c(1:59), 2], sqrt(epil3_ML$vcov.dispersion)), 2) -->
<!-- ``` -->

<!-- ## Mean and median bias reduction -->

<!-- Fitting the model using mean-bias reducing adjusted score equations gives -->
<!-- ```{r, echo = TRUE} -->
<!-- epil3_BRmean <- brnb(y ~ -1+ factor(subject) + factor(pred),  data = epil3, type = "AS_mean") -->
<!-- ## Estimated  of interest regression coefficients and dispersion parameters -->
<!-- round(c(epil3_BRmean$coefficients[-c(1:59)], epil3_BRmean$dispersion), 2) -->
<!-- ## Estimated standard errors -->
<!-- round(c(summary(epil3_BRmean)$coefficients[-c(1:59), 2], sqrt(epil3_BRmean$vcov.dispersion)), 2) -->
<!-- ``` -->

<!-- The corresponding fit using median-bias reducing adjusted score equations is -->
<!-- ```{r, echo = TRUE} -->
<!-- epil3_BRmedian <- brnb(y ~ -1+ factor(subject) + factor(pred),  data = epil3, type = "AS_median") -->
<!-- ## Estimated  of interest regression coefficients and dispersion parameters -->
<!-- round(c(epil3_BRmedian$coefficients[-c(1:59)], epil3_BRmedian$dispersion), 2) -->
<!-- ## Estimated standard errors -->
<!-- round(c(summary(epil3_BRmedian)$coefficients[-c(1:59), 2], sqrt(epil3_BRmedian$vcov.dispersion)), 2) -->
<!-- ``` -->

<!-- ## Mean bias correction -->
<!-- ```{r, echo = TRUE} -->
<!-- epil3_BCmean <- brnb(y ~ -1+ factor(subject) + factor(pred),  data = epil3, type = "correction") -->
<!-- ## Estimated  of interest regression coefficients and dispersion parameters -->
<!-- round(c(epil3_BCmean$coefficients[-c(1:59)], epil3_BCmean$dispersion), 2) -->
<!-- ## Estimated standard errors -->
<!-- round(c(summary(epil3_BCmean)$coefficients[-c(1:59), 2], sqrt(epil3_BCmean$vcov.dispersion)), 2) -->
<!-- ``` -->

The differences between reduced-bias estimation and maximum likelihood are particularly pronounced for the dispersion parameter. Improved estimation of the dispersion parameter results to larger estimated standard errors than maximum likelihood. Hence, the estimated standard errors based on the maximum likelihood estimates appear to be smaller than they should be, which is also supported by the simulation results in @kenne:20[, Section 5].

# Relevant resources
`?brglmFit` and `?brglm_control` contain quick descriptions of the various bias reduction methods supported in **brglm2**. The [`iteration`](https://cran.r-project.org/package=brglm2/brglm2.pdf) vignette describes the iteration and gives the mathematical details for the bias-reducing adjustments to the score functions for generalized linear models.

# Citation
If you found this vignette or **brglm2**, in general, useful, please consider citing **brglm2** and the associated paper. You can find information on how to do this by typing `citation("brglm2")`.

# References

---
title: "Bias reduction in generalized linear models"
author: "[Ioannis Kosmidis](https://www.ikosmidis.com), Euloge Clovis Kenne Pagui"
date: "29 April 2017"
output: rmarkdown::html_vignette
bibliography: brglm2.bib
vignette: >
  %\VignetteIndexEntry{Bias reduction in generalized linear models}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6,
  fig.height = 6
)
```

# The **brglm2** package

[**brglm2**](https://github.com/ikosmidis/brglm2) provides tools for the estimation and inference from [generalized linear models](https://en.wikipedia.org/wiki/Generalized_linear_model) using various methods for bias reduction or maximum penalized likelihood with powers of the Jeffreys prior as penalty. Reduction of estimation bias is achieved either through the mean-bias reducing adjusted score equations in @firth:93 and @kosmidis:09, or through the direct subtraction of an estimate of the bias of the maximum likelihood estimator from the maximum likelihood estimates as prescribed in @cordeiro:91, or through the median-bias reducing adjusted score equations in @kenne:17.

In the special case of generalized linear models for binomial, Poisson and multinomial responses (both nominal and ordinal), mean and median bias reduction and maximum penalized likelihood return estimates with improved frequentist properties, that are also always finite, even in cases where the maximum likelihood estimates are infinite, like in complete and quasi-complete separation as defined in @albert:84.

The workhorse function is [`brglmFit()`](https://cran.r-project.org/package=brglm2/brglm2.pdf), which can be passed directly to the `method` argument of the `glm` function. `brglmFit` implements a quasi-[Fisher scoring](https://en.wikipedia.org/wiki/Scoring_algorithm) procedure, whose special cases result in various explicit and implicit bias reduction methods for generalized linear models [the classification of bias reduction methods into explicit and implicit is given in @kosmidis:14].

# This vignette

This vignette

+ presents the supported bias-reducing adjustments to the score functions for generalized linear models
+ describes the fitting algorithm at the core of **brglm2**

# Other resources

The bias-reducing quasi Fisher scoring iteration is also described in detail in the [bias vignette](https://cran.r-project.org/package=enrichwith/vignettes/bias.html) of the [**enrichwith**](https://cran.r-project.org/package=enrichwith) R package. @kosmidis:10 describe a parallel quasi [Newton-Raphson](https://en.wikipedia.org/wiki/Newton%27s_method) procedure.

Most of the material in this vignette comes from a presentation by [Ioannis Kosmidis](https://www.ikosmidis.com) at the useR! 2016 international conference at University of Stanford on 16 June 2016. The presentation was titled "Reduced-bias inference in generalized linear models".

# Generalized linear models

### Model

Suppose that $y_1, \ldots, y_n$ are observations on independent random variables $Y_1, \ldots, Y_n$, each with probability density/mass function of the form
$$
f_{Y_i}(y) = \exp\left\{\frac{y \theta_i - b(\theta_i) - c_1(y)}{\phi/m_i} - \frac{1}{2}a\left(-\frac{m_i}{\phi}\right) + c_2(y) \right\}
$$
for some sufficiently smooth functions $b(.)$, $c_1(.)$, $a(.)$ and $c_2(.)$, and fixed observation weights $m_1, \ldots, m_n$. The expected value and the variance of $Y_i$ are then
\begin{align*}
      E(Y_i) & = \mu_i =  b'(\theta_i) \\
      Var(Y_i) & = \frac{\phi}{m_i}b''(\theta_i) = \frac{\phi}{m_i}V(\mu_i)
\end{align*}
Hence, in this parameterization, $\phi$ is a dispersion parameter.

A generalized linear model links the mean $\mu_i$ to a linear predictor $\eta_i$ as
$$
g(\mu_i) = \eta_i = \sum_{t=1}^p \beta_t x_{it}
$$
where $g(.)$ is a monotone, sufficiently smooth link function, taking values on $\Re$, $x_{it}$ is the $(i,t)th$ component of a model matrix $X$, and $\beta = (\beta_1, \ldots, \beta_p)^\top$.

### Score functions and information matrix

Suppressing the dependence of the various quantities on the model parameters and the data, the derivatives of the log-likelihood about $\beta$ and $\phi$ (score functions) are
\begin{align*}
    s_\beta & = \frac{1}{\phi}X^TWD^{-1}(y - \mu) \\
	s_\phi & = \frac{1}{2\phi^2}\sum_{i = 1}^n (q_i - \rho_i)
\end{align*}
with $y = (y_1, \ldots, y_n)^\top$, $\mu = (\mu_1, \ldots, \mu_n)^\top$, $W = {\rm diag}\left\{w_1, \ldots, w_n\right\}$ and $D = {\rm diag}\left\{d_1, \ldots, d_n\right\}$, where $w_i = m_i d_i^2/v_i$ is the $i$th working weight, with  $d_i = d\mu_i/d\eta_i$ and $v_i = V(\mu_i)$. Furthermore, $q_i = -2 m_i \{y_i\theta_i - b(\theta_i) - c_1(y_i)\}$ and $\rho_i = m_i a'_i$ with $a'_i = a'(-m_i/\phi)$.
<!-- are the $i$th deviance residual (e.g. as is implemented in the `dev.resids` component of most `family` objects) and its expectation, respectively, with $a'_i = a'(-m_i/\phi)$. The only `family` object deviating from the above description is `Gamma` where `Gamma()$dev.resids` implements $q_i - 2m_i$ instead of $q_i$. For convenience in implementation, and just for Gamma we define $\rho_i = m_i a'_1 - 2m_i = -2\psi(-m_i/\phi) + 2\log(-m_i/\phi)$, where $\psi$ is the `digamma` function. This change affects none of the estimation methods discussed in this vignette. -->
The expected information matrix about $\beta$ and $\phi$ is
$$
i =
\left[
\begin{array}{cc}
i_{\beta\beta} & 0_p \\
0_p^\top & i_{\phi\phi}
\end{array}
\right]
=
\left[
\begin{array}{cc}
\frac{1}{\phi} X^\top W X & 0_p \\
0_p^\top & \frac{1}{2\phi^4}\sum_{i = 1}^n m_i^2 a''_i
\end{array}
\right]\,,
$$
where $0_p$ is a $p$-vector of zeros, and $a''_i = a''(-m_i/\phi)$.


### Maximum likelihood estimation

The maximum likelihood estimators $\hat\beta$ and $\hat\phi$ of $\beta$ and $\phi$, respectively, can be found by the solution of the score equations $s_\beta = 0_p$ and $s_\phi = 0$.

### Mean bias-reducing adjusted score functions

Let $A_\beta = -i_{\beta\beta} b_\beta$ and $A_\phi = -i_{\phi\phi} b_\phi$, where $b_\beta$ and $b_\phi$ are the first terms in the expansion of the mean bias of the maximum likelihood estimator of the regression parameters $\beta$ and dispersion $\phi$, respectively. The results in @firth:93 can be used to show that the solution of the adjusted score equations
\begin{align*}
s_\beta + A_\beta & = 0_p \\
s_\phi + A_\phi & = 0
\end{align*}
results in estimators $\tilde\beta$ and $\tilde\phi$ with bias of smaller asymptotic order than the maximum likelihood estimator.

The results in either @kosmidis:09 or @cordeiro:91 can then be used to re-express the adjustments in forms that are convenient for implementation. In particular, and after some algebra the bias-reducing adjustments for generalized linear models are
\begin{align*}
 A_\beta & = X^\top W \xi \,, \\
 A_\phi & = \frac{(p - 2)}{2\phi} + \frac{\sum_{i = 1}^n m_i^3
                   a'''_i}{2\phi^2\sum_{i = 1}^n m_i^2
                   a''_i}
%A_\phi & = \frac{(p - 2)\phi\sum_{i = 1}^n m_i^2
%                   a''(-m_i/\phi) + \sum_{i = 1}^n m_i^3
%                   a'''(-m_i/\phi))}{2\phi^2\sum_{i = 1}^n m_i^2
%                   a''(-m_i/\phi)}
\end{align*}
where  $\xi = (\xi_1, \ldots, \xi_n)^T$ with $\xi_i = h_id_i'/(2d_iw_i)$, $d_i' = d^2\mu_i/d\eta_i^2$, $a''_i = a''(-m_i/\phi)$, $a'''_i = a'''(-m_i/\phi)$, and $h_i$ is the "hat" value for the $i$th observation (see, e.g. `?hatvalues`).

### Median bias-reducing adjusted score functions
The results in @kenne:17 can be used to show that if
\begin{align*}
 A_\beta & =  X^\top W (\xi + X u) \\
 A_\phi & = \frac{p}{2\phi}+\frac{ \sum_{i = 1}^n m_i^3
                   a'''_i}{6\phi^2\sum_{i = 1}^n m_i^2 a''_i} \, ,
\end{align*}
then the solution of the adjusted score equations $s_\beta + A_\beta = 0_p$ and $s_\phi + A_\phi = 0$ results in estimators $\tilde\beta$ and $\tilde\phi$ with median bias of smaller asymptotic order than the maximum likelihood estimator. In the above expression, $u = (u_1, \ldots, u_p)^\top$ with
\begin{align*}
	u_j = [(X^\top W X)^{-1}]_{j}^\top X^\top \left[
	\begin{array}{c}
	\tilde{h}_{j,1} \left\{d_1 v'_1 / (6 v_1) - d'_1/(2 d_1)\right\} \\
	\vdots \\
	\tilde{h}_{j,n} \left\{d_n v'_n / (6 v_n) - d'_n/(2 d_n)\right\}
	\end{array}
	\right]
\end{align*}
where $[A]_j$ denotes the $j$th row of matrix $A$ as a column vector, $v'_i = V'(\mu_i)$, and $\tilde{h}_{j,i}$ is the $i$th diagonal element of $X K_j X^T W$, with $K_j = [(X^\top W X)^{-1}]_{j} [(X^\top W X)^{-1}]_{j}^\top / [(X^\top W X)^{-1}]_{jj}$\.

### Mixed adjustments
The results in @kosmidis:2019 can be used to show that if
\begin{align*}
 A_\beta & = X^\top W \xi \,, \\
 A_\phi & = \frac{p}{2\phi}+\frac{ \sum_{i = 1}^n m_i^3
                   a'''_i}{6\phi^2\sum_{i = 1}^n m_i^2 a''_i} \, ,
\end{align*}
then the solution of the adjusted score equations $s_\beta + A_\beta = 0_p$ and $s_\phi + A_\phi = 0$ results in estimators $\tilde\beta$ with mean bias of small asymptotic order than the maximum likelihood estimator and $\tilde\phi$ with median bias of smaller asymptotic order than the maximum likelihood estimator.

### Maximum penalized likelihood with powers of Jeffreys prior as penalty
The likelihood penalized by a power of the Jeffreys prior
\[
|i_{\beta\beta}|^a |i_{\phi\phi}|^a \quad a > 0
\]
can be maximized by solving the adjusted score equations $s_\beta + A_\beta = 0_p$ and $s_\phi + A_\phi = 0$  with
\begin{align*}
 A_\beta & = X^\top W \rho \,, \\
 A_\phi & = -\frac{p + 4}{2\phi}+\frac{ \sum_{i = 1}^n m_i^3
                   a'''_i}{6\phi^2\sum_{i = 1}^n m_i^2 a''_i} \, ,
\end{align*}
where $\rho = (\rho_1, \ldots, \rho_n)^T$ with $\rho_i = h_i \{2 d_i'/(d_i w_i) - v_i' d_i/(v_i w_i)\}$.

<!-- The results in @kenne:17 can be used to show that if -->
<!-- \begin{align*} -->
<!--  A_\beta & = i_\beta \left\{\frac{u_1}{2} + \frac{u_2}{6} \right\} \\ -->
<!--  A_\phi & = \frac{p}{2\phi}+\frac{ \sum_{i = 1}^n m_i^3 -->
<!--                    a'''_i}{6\phi^2\sum_{i = 1}^n m_i^2 a''_i} \, , -->
<!-- \end{align*} -->
<!-- then the solution of the adjusted score equations $s_\beta + A_\beta = 0_p$ and $s_\phi + A_\phi = 0$ results in estimators $\tilde\beta$ and $\tilde\phi$ with median bias of smaller asymptotic order than the maximum likelihood estimator.  -->
<!-- In the above expressions, -->
<!-- \begin{align*} -->
<!-- u_{1r} &  =  \sum_{s=1}^p i^{rs}_\beta tr\{\nu_\beta (L_s+K_s)\} \\ -->
<!-- u_{2r}  & =  \sum_{s=1}^p i^{rs}_\beta \{i^{r}_\beta\}^TL_si^{r}_\beta/i^{rr}_\beta -->
<!-- \end{align*} -->
<!-- where $i^{rs}_\beta$ is the $(r,s)$ element of $i^{-1}_\beta,\,$ $i^{r}_\beta$ is the $r$th column of $i^{-1}_\beta$, -->
<!-- $\nu_{\beta, r} = i^{-1}_\beta - i^{r}_\beta \left\{i^{r}_\beta\right\}^\top$, $L_s = X^TWQ_{1s}X/\phi,\,$  $K_s = -X^TWQ_{2s}X/\phi$, -->
<!-- <\!--$Q_{1s} = diag\{V'(\mu_1)d_1x_{1s}/V(\mu_1),\ldots,V'(\mu_n)d_1x_{ns}/V(\mu_n)\}$ and -\-> -->
<!-- <\!--$Q_{2s} = diag\{d_1x_{1s}(\frac{V'(\mu_1)}{V(\mu_1)}-\frac{d'_1}{d^2_1}),\ldots, -\-> -->
<!-- <\!--d_nx_{ns}(\frac{V'(\mu_n)}{V(\mu_n)}-\frac{d'_n}{d^2_n})\}$. -\-> -->
<!-- $Q_{1s} = {\rm diag}\{v'_1 d_1x_{1s}/v_1, \ldots, v'_n d_n x_{ns}/v_n\}$ and $Q_{2s} = Q_{1s} - {\rm diag}\{d_1x_{1s}/d'_1, \ldots, d_n x_{ns}/d'_n\}$, with $v'_i = V'(\mu_i)$. -->

# Fitting algorithm in `brglmFit`

`brglmFit()` implements a quasi Fisher scoring procedure for solving the adjusted score equations $s_\beta + A_\beta = 0_p$ and $s_\phi + A_\phi = 0$. The iteration consists of an outer loop and an inner loop that implements step-halving. The algorithm is as follows:

### Input

+ $s_\beta$, $i_{\beta\beta}$, $A_\beta$
+ $s_\phi$, $i_{\phi\phi}$, $A_\phi$
+ Starting values $\beta^{(0)}$ and $\phi^{(0)}$
+ $\epsilon > 0$: tolerance for the $L^\infty$ norm of the search direction before reporting convergence
+ $M$: maximum number of halving steps that can be taken

### Output

+ $\tilde\beta$, $\tilde\phi$

### Iteration

*Initialize outer loop*

1. $k \leftarrow 0$

2. $\upsilon_\beta^{(0)} \leftarrow \left\{i_{\beta\beta}\left(\beta^{(0)}, \phi^{(0)}\right)\right\}^{-1} \left\{s_\beta\left(\beta^{(0)}, \phi^{(0)}\right) + A_\beta\left(\beta^{(0)}, \phi^{(0)}\right)\right\}$

3. $\upsilon_\phi^{(0)} \leftarrow \left\{i_{\phi\phi}\left(\beta^{(0)}, \phi^{(0)}\right)\right\}^{-1} \left\{s_\phi\left(\beta^{(0)}, \phi^{(0)}\right) + A_\phi\left(\beta^{(0)}, \phi^{(0)}\right)\right\}$

*Initialize inner loop*

4. $m \leftarrow 0$

5. $b^{(m)} \leftarrow \beta^{(k)}$

6. $f^{(m)} \leftarrow \phi^{(k)}$

7. $v_\beta^{(m)} \leftarrow \upsilon_\beta^{(k)}$

8. $v_\phi^{(m)} \leftarrow \upsilon_\phi^{(k)}$

9. $d \leftarrow \left|\left|(v_\beta^{(m)}, v_\phi^{(m)})\right|\right|_\infty$

*Update parameters*

10. $b^{(m + 1)} \leftarrow b^{(m)} + 2^{-m} v_\beta^{(m)}$

11. $f^{(m + 1)} \leftarrow f^{(m)} + 2^{-m} v_\phi^{(m)}$

*Update direction*

12. $v_\beta^{(m + 1)} \leftarrow \left\{i_{\beta\beta}\left(b^{(m + 1)}, f^{(m + 1)}\right)\right\}^{-1} \left\{s_\beta\left(b^{(m + 1)}, f^{(m + 1)}\right) + A_\beta\left(b^{(m + 1)}, f^{(m + 1)}\right)\right\}$

13. $v_\phi^{(m + 1)} \leftarrow \left\{i_{\phi\phi}\left(b^{(m + 1)}, f^{(m + 1)}\right)\right\}^{-1} \left\{s_\phi\left(b^{(m + 1)}, f^{(m + 1)}\right) + A_\phi\left(b^{(m + 1)}, f^{(m + 1)}\right)\right\}$

*Continue or break halving within inner loop*

14. if $m + 1 < M$ and $\left|\left|(v_\beta^{(m + 1)}, v_\phi^{(m + 1)})\right|\right|_\infty > d$

	14.1. $m \leftarrow m + 1$

    14.2. GO TO 10

15. else

    15.1. $\beta^{(k + 1)} \leftarrow b^{(m + 1)}$

	15.2. $\phi^{(k + 1)} \leftarrow f^{(m + 1)}$

	15.3. $\upsilon_\beta^{(k + 1)} \leftarrow v_b^{(m + 1)}$

	15.4. $\upsilon_\phi^{(k + 1)} \leftarrow v_f^{(m + 1)}$

*Continue or break outer loop*

16. if $k + 1 < K$ and $\left|\left|(\upsilon_\beta^{(k + 1)}, \upsilon_\phi^{(k + 1)})\right|\right|_\infty > \epsilon$

	16.1 $k \leftarrow k + 1$

	16.2. GO TO 4

17. else

	17.1. $\tilde\beta \leftarrow \beta^{(k + 1)}$

    17.2. $\tilde\phi \leftarrow \phi^{(k + 1)}$

	17.3. STOP

# Notes

+ For $K = M = 1$, $\beta^{(0)} = \hat\beta$ and $\phi^{(0)} = \hat\phi$, the above iteration computes the bias-corrected estimates proposed in @cordeiro:91. This is achieved when the `brglmFit()` function is called with `type = "correction"` (see `?brglmFit`).

+ The mean-bias reducing adjusted score functions are solved when the `brglmFit()` function is called with `type = "AS_mean"`, and the median-bias reducing adjusted score functions with `type = "AS_median"` (see `?brglmFit`). Estimation using mixed adjustments is through `type = "AS_mixed"`. `type = "MPL_Jeffreys"` does maximum penalized likelihood with a power of the Jeffreys prior as penalty.

+ The steps where $\phi$ and the $\phi$ direction are updated are ignored for generalized linear models with known dispersion parameter, like in models with binomial and Poisson responses. Also, in that case, $v_\phi^{(.)}$ and $\upsilon_\phi^{(.)}$ in steps 9, 14 and 16 are set to zero.

+ The implementation of the adjusted score functions requires ready implementations of $d^2\mu_i/d\eta_i^2$, $a'(.)$, $a''(.)$ and $a'''(.)$. The [**enrichwith**](https://cran.r-project.org/package=enrichwith) R package is used internally to enrich the base `family` and `link-glm` objects with implementations of those functions (see `?enrich.family` and `?enrich.link-glm`).

+ The above iteration can be used to implement a variety of additive adjustments to the score function, by supplying the algorithm with appropriate adjustment functions $A_\beta$ and $A_\phi$

# Contributions to this vignette

The first version of the vignette has been written by Ioannis Kosmidis. Eugene Clovis Kenne Pagui and Nicola Sartori contributed the first version of the section "Median bias-reducing adjusted score functions", and Ioannis Kosmidis brought the expressions for the median bias-reducing adjustments in the reduced form that is shown above and is implemented in `brglmFit()`.

@kosmidis:2019 provides more details about mean and median bias reduction in generalized linear models.


# Citation
If you found this vignette or **brglm2**, in general, useful, please consider citing **brglm2** and the associated paper. You can find information on how to do this by typing `citation("brglm2")`.

# References

---
title: "Estimating the exponential of regression parameters using **brglm2**"
author: "[Ioannis Kosmidis](https://www.ikosmidis.com)"
date: "03 February 2023"
output: rmarkdown::html_vignette
bibliography: brglm2.bib
vignette: >
  %\VignetteIndexEntry{Estimating the exponential of regression parameters using **brglm2**}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  fig.width = 6,
  fig.height = 6
)
```

# The `expo()` method

The [**brglm2**](https://github.com/ikosmidis/brglm2) R package provides the `expo()` method for estimating exponentiated parameters of generalized linear models using various methods. 

The `expo()` method uses a supplied `"brglmFit"` or `"glm"` object to estimate exponentiated parameters of generalized linear models with maximum likelihood or various mean and median bias reduction methods. `expo()` is useful for computing (corrected) estimates of the multiplicative impact of a unit increase on a covariate on the mean of a Poisson log-linear model (`family = poisson("log")` in `glm()`) while adjusting for other covariates, the odds ratio associated with a unit increase on a covariate in a logistic regression model (`family = binomial("logit")` in `glm()`) while adjusting for other covariates, the relative risk associated with a unit increase on a covariate in a relative risk regression model (`family = binomial("log")` in `glm()`) while adjusting for other covariates, among others.

The vignette demonstrates the use of `expo()` and the associated methods by reproducing part of the analyses in @agresti:02[, Section 5.4.2] on the effects of AZT in slowing the development of AIDS symptoms.

# AIDS and AZT use

The data analyzed in @agresti:02[, Section 5.4.2] is from a 3-year
study on the effects of AZT in slowing the development of AIDS
symptoms. 338 veterans whose immune systems were beginning to falter
after infection with the AIDS virus were randomly assigned either to
receive AZT immediately or to wait until their T cells showed severe
immune weakness. See `?aids` for more details. 

The `aids` data set cross-classifies the veterans' race (`race`), whether they received AZT immediately (`AZT`), and whether they developed AIDS symptoms during the 3-year study (`symptomatic` and `asymptomatic`).
```{r, echo = TRUE}
library("brglm2")
data("aids", package = "brglm2")
aids
```

We now use a logistic regression model to model the probability of developing symptoms in terms of `AZT` and `race`, and reproduce part of the compute output in @agresti:02[, Table 5.6].
```{r, echo = TRUE}
aids_mod <- glm(cbind(symptomatic, asymptomatic) ~ AZT + race, 
                  family = binomial(), data = aids)
summary(aids_mod)
```

The Wald test for the hypothesis of conditional independence of AZT treatment and development of AIDS symptoms, controlling for race, returns a p-value of `r round(coef(summary(aids_mod))["AZTYes", "Pr(>|z|)"], 3)`, showing evidence of association.

The predicted probabilities for each combination of levels


The maximum likelihood estimates of the odds ratio between immediate AZT use and development of AIDS symptoms can be inferred from `aids_mod` through the `expo()` method, which also estimates standard errors using the delta method, and returns approximate 95% confidence intervals (see `?expo` for details).
```{r, echo = TRUE}
expo(aids_mod, type = "ML")
```

As noted in @agresti:02[, Section 5.4.2], for each race, the estimated
odds of symptoms are half as high for those who took AZT immediately,
with value $0.49$ and a nominally 95\% Wald confidence interval
$(0.28, 0.84)$.

The `expo()` method can be used to estimate the odds ratios using three methods that return estimates of the odds ratios with asymptotically smaller mean bias than the maximum likelihood estimator 
```{r, echo = TRUE}
expo(aids_mod, type = "correction*")
expo(aids_mod, type = "Lylesetal2012")
expo(aids_mod, type = "correction+")
```
and one method that returns estimates of the odds ratios with asymptotically smaller median bias than the maximum likelihood estimator 
```{r, echo = TRUE}
expo(aids_mod, type = "AS_median")
```

The estimated odds ratios and associated inferences from the methods that correct for mean and median bias are similar to those from maximum likelihood.

# Infinite odds ratio estimates 

When `expo()` is called with `type = correction*`, `type = correction+`, `type = Lylesetal2012`, and `type = AS_median`, then the estimates of the odds ratios can be shown to be always finite and greater than zero. The reason is that the corresponding odds-ratio estimators depend on regression parameter estimates that are finite even if the maximum likelihood estimates are infinite. See, @kosmidis:2019 and @kosmidis+firth:21 for details. 

As an example, consider the estimated odds ratios from a logistic regression model fitted on the `endometrial` data set using maximum likelihood. 
```{r, echo = TRUE}
data("endometrial", package = "brglm2")
endometrialML <- glm(HG ~ NV + PI + EH, data = endometrial, family = binomial())
endometrialML
```

The estimate of the coefficient for `NV` is in reality infinite as it can be verified using the [**detectseparation**](https://cran.r-project.org/package=detectseparation) R package
```{r, echo = TRUE}
library("detectseparation")
update(endometrialML, method = detect_separation)
```
and a naive estimate of the associated odds ratio while controlling for `PI` and `EH` is `r exp(coef(endometrialML)["NV"])`, which is in reality infinite.

In contrast, `expo()` returns finite reduced-mean-bias estimates of the odds ratios 
```{r, echo = TRUE}
expo(endometrialML, type = "correction*")
expo(endometrialML, type = "correction+")
expo(endometrialML, type = "Lylesetal2012")
```

# `brglmFit` objects

The `expo()` method also works seamlessly with `brglmFit` objects, returning the same results as above. For example, 
```{r, echo = TRUE}
aids_mod_br <- update(aids_mod, method = "brglmFit")
expo(aids_mod_br, type = "correction*")
```


# References

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/brglm2-package.R
\docType{package}
\name{brglm2}
\alias{brglm2}
\alias{brglm2-package}
\title{brglm2: Bias Reduction in Generalized Linear Models}
\description{
Estimation and inference from generalized linear models using
implicit and explicit bias reduction methods (Kosmidis, 2014), and
other penalized maximum likelihood methods. Currently supported
methods include the mean bias-reducing adjusted scores approach in
Firth (1993) and Kosmidis & Firth (2009), the median bias-reduction
adjusted scores approach in Kenne Pagui et al. (2017), the
correction of the asymptotic bias in Cordeiro & McCullagh (1991),
the mixed bias-reduction adjusted scores approach in Kosmidis et al
(2020), maximum penalized likelihood with powers of the Jeffreys
prior as penalty, and maximum likelihood.
}
\details{
In the special case of generalized linear models for binomial,
Poisson and multinomial responses (both nominal and ordinal), mean
and median bias reduction and maximum penalized likelihood return
estimates with improved frequentist properties, that are also
always finite, even in cases where the maximum likelihood estimates
are infinite (e.g. complete and quasi-complete separation in
multinomial regression). Estimation in all cases takes place via a
modified Fisher scoring algorithm, and S3 methods for the
construction of confidence intervals for the reduced-bias estimates
are provided.

The core model fitters are implemented by the functions
\code{\link[=brglm_fit]{brglm_fit()}} (univariate generalized linear models),
\code{\link[=brmultinom]{brmultinom()}} (baseline category logit models for nominal
multinomial responses), \code{\link[=bracl]{bracl()}} (adjacent category logit models
for ordinal multinomial responses), and \code{\link[=brnb]{brnb()}} for negative
binomial regression.

The similarly named \strong{brglm} R package can only handle generalized
linear models with binomial responses. Special care has been taken
when developing \strong{brglm2} in order not to have conflicts when the
user loads \strong{brglm2} and \strong{brglm} simultaneously. The development
and maintenance of the two packages will continue in parallel,
until \strong{brglm2} incorporates all \strong{brglm} functionality and
provides an appropriate wrapper to the \code{\link[brglm:brglm]{brglm::brglm()}} function.
}
\references{
Kosmidis I, Firth D (2021). Jeffreys-prior penalty, finiteness
and shrinkage in binomial-response generalized linear
models. \emph{Biometrika}, \strong{108}, 71-82. \doi{10.1093/biomet/asaa052}.

Cordeiro G M, McCullagh P (1991). Bias correction in generalized
linear models. \emph{Journal of the Royal Statistical Society. Series B
(Methodological)}, \strong{53}, 629-643. \doi{10.1111/j.2517-6161.1991.tb01852.x}.

Firth D (1993). Bias reduction of maximum likelihood estimates,
Biometrika, \strong{80}, 27-38. \doi{10.2307/2336755}.

Kenne Pagui E C, Salvan A, Sartori N (2017). Median bias
reduction of maximum likelihood estimates. \emph{Biometrika}, \strong{104},
923–938. \doi{10.1093/biomet/asx046}.

Kosmidis I, Kenne Pagui E C, Sartori N (2020). Mean and median bias
reduction in generalized linear models. \emph{Statistics and Computing},
\strong{30}, 43-59. \doi{10.1007/s11222-019-09860-6}.

Kosmidis I, Firth D (2009). Bias reduction in exponential family
nonlinear models. \emph{Biometrika}, \strong{96}, 793-804. \doi{10.1093/biomet/asp055}.

Kosmidis I, Firth D (2010). A generic algorithm for reducing
bias in parametric estimation. \emph{Electronic Journal of Statistics},
\strong{4}, 1097-1112. \doi{10.1214/10-EJS579}.

Kosmidis I (2014). Bias in parametric estimation: reduction and
useful side-effects. \emph{WIRE Computational Statistics}, \strong{6},
185-196. \doi{10.1002/wics.1296}.
}
\seealso{
\code{\link[=brglm_fit]{brglm_fit()}}, \code{\link[=brmultinom]{brmultinom()}}, \code{\link[=bracl]{bracl()}}
}
\author{
Ioannis Kosmidis \verb{[aut, cre]} \email{ioannis.kosmidis@warwick.ac.uk}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{MultipleFeatures}
\alias{MultipleFeatures}
\title{Multiple features data}
\format{
A data frame with 2000 rows and 382 columns:
\itemize{
\item \code{digit}. The digits to which the feature sets \verb{fou.*}, \verb{kar.*} and \verb{pix.*} correspond to.
\item \verb{fou.*}. 76 Fourier coefficients of the character shapes, which are computed to be rotation invariant.
\item \verb{kar.*}. 64 Karhunen-Lo\`eve coefficients of the character shapes.
\item \verb{pix.*}. 240 pixel averages in 2 x 3 windows of each character shape.
\item \code{training}. \code{TRUE} if the digit is part of the training set
and \code{FALSE} if the digit is allocated to the test set.
}
}
\source{
The data provides the \code{fou}, \code{kar} and \code{pix} features of the
Multiple Features data set from the UCI Machine Learning Repository
(Duin, 1998).
}
\usage{
MultipleFeatures
}
\description{
Digits (0-9) extracted from a collection of maps from a Dutch
public utility.  Two hundred \verb{30 × 48} binary images per digit are
available, which have then been used to extract feature sets; see
Jain et al. (2000), for details, where that dataset is used for
assessing the performance of various classifiers for digit
recognition.
}
\examples{

\donttest{
data("MultipleFeatures", package = "brglm2")

par(mfrow = c(10, 20), mar = numeric(4) + 0.1)
for (c_digit in 0:9) {
   df <- subset(MultipleFeatures, digit == c_digit)
   df <- as.matrix(df[, paste("pix", 1:240, sep = ".")])
   for (inst in 1:20) {
       m <- matrix(df[inst, ], 15, 16)[, 16:1]
       image(m, col = grey.colors(7, 1, 0), xaxt = "n", yaxt = "n")
  }
}

}
}
\references{
Duin, R. (1998). Multiple Features Dataset. UCI Machine Learning
Repository. \doi{10.24432/C5HC70}.

Jain A, Duin R, Mao J (2000). Statistical pattern recognition: a
review. IEEE Transactions on Pattern Analysis and Machine
Intelligence, \strong{22}, 4–37. \doi{10.1109/34.824819}.
}
\seealso{
\code{\link[=mdypl_fit]{mdypl_fit()}}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/brnb.R
\name{simulate.brnb}
\alias{simulate.brnb}
\title{Simulate Responses}
\usage{
\method{simulate}{brnb}(object, nsim = 1, seed = NULL, ...)
}
\arguments{
\item{object}{an object representing a fitted model.}

\item{nsim}{number of response vectors to simulate. Defaults to 1.}

\item{seed}{an object specifying if and how the random number
generator should be initialized; see \code{\link[=set.seed]{set.seed()}} for details.}

\item{...}{extra arguments to be passed to methods. Not currently
used.}
}
\description{
Simulate one or more responses from the distribution corresponding
to a fitted model \code{\link[=brnb]{"brnb"}} object.
}
\examples{
# Example in Saha, K., & Paul, S. (2005). Bias-corrected maximum
# likelihood estimator of the negative binomial dispersion
# parameter.  Biometrics, 61, 179--185.
#
# Frequency distribution of red mites on apple leaves.
nomites <- 0:8
noleaves <- c(70, 38, 17, 10, 9, 3, 2, 1, 0)
fit_glmnb <- MASS::glm.nb(nomites~1,link="identity",weights = noleaves)
fit_brnb <- brnb(nomites ~ 1, link = "identity", transformation = "inverse",
                 type = "ML", weights = noleaves)
## Let us simulate 10 response vectors
sim_glmnb <- simulate(fit_glmnb, nsim = 10, seed = 123)
sim_brnb <-  simulate(fit_brnb, nsim = 10, seed = 123)
# The results from glm.nb and brnb with type = "ML" are
# exactly the same
all.equal(sim_glmnb, sim_brnb, check.attributes = FALSE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/se.R
\name{se0_ridge}
\alias{se0_ridge}
\title{Logistic ridge regression state evolution functions with no intercept}
\usage{
se0_ridge(mu, b, sigma, kappa, gamma, lambda, gh = NULL, prox_tol = 1e-10)
}
\arguments{
\item{mu}{aggregate bias parameter.}

\item{b}{parameter \code{b} in the state evolution functions.}

\item{sigma}{square root of the aggregate variance of the MDYPL
estimator.}

\item{kappa}{asymptotic ratio of columns/rows of the design
matrix. \code{kappa} should be in \verb{(0, 1)}.}

\item{gamma}{the square root of the limit of the variance of the
linear predictor.}

\item{lambda}{the shrinkage parameter of the logistic regression penalty
estimator. \code{lambda} should be in greater than zero.}

\item{gh}{A list with the Gauss-Hermite quadrature nodes and
weights, as returned from \code{statmod::gauss.quad()} with \code{kind = "hermite"}. Default is \code{NULL}, in which case \code{gh} is set to
\code{statmod::gauss.quad(200, kind = "hermite")} is used.}

\item{prox_tol}{tolerance for the computation of the proximal
operator; default is \code{1e-10}.}
}
\description{
Logistic ridge regression state evolution functions with no intercept
}
\details{
It is assumed that the ridge penalty to the logistic regression
log-likelihood is \code{n * lambda * sum(beta^2) / (2 * length(beta))},
where \code{n} is the sum of the binomial totals.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/brnb.R
\name{vcov.brnb}
\alias{vcov.brnb}
\title{Extract model variance-covariance matrix from \code{\link[=brnb]{"brnb"}} objects}
\usage{
\method{vcov}{brnb}(object, model = c("mean", "full", "dispersion"), complete = TRUE, ...)
}
\arguments{
\item{object}{an object of class \code{\link[=brnb]{"brnb"}}, typically, a result of a call to \code{\link[=brnb]{brnb()}}.}

\item{model}{character specifying for which component of the model variance-covariance matrix should be extracted.}

\item{complete}{for the \code{aov}, \code{lm}, \code{glm}, \code{mlm}, and where
    applicable \code{summary.lm} etc methods: logical indicating if the
    full variance-covariance matrix should be returned also in case of
    an over-determined system where some coefficients are undefined and
    \code{\link[stats]{coef}(.)} contains \code{NA}s correspondingly.   When
    \code{complete = TRUE},  \code{vcov()} is compatible with
    \code{coef()} also in this singular case.}

\item{...}{
    additional arguments for method functions.  For the
    \code{\link[stats]{glm}} method this can be used to pass a
    \code{dispersion} parameter.}
}
\description{
Extract model variance-covariance matrix from \code{\link[=brnb]{"brnb"}} objects
}
\details{
The options for \code{model} are \code{"mean"} for mean regression only
(default), \code{"dispersion"} for the dispersion parameter (in a
chosen transformation; see \code{\link[=brglmControl]{brglmControl()}}, and \code{"full"} for both
the mean regression and the (transformed) dispersion parameters.
See \code{\link[=vcov]{vcov()}} for more details.
}
\seealso{
\code{\link[=vcov]{vcov()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bracl.R
\name{bracl}
\alias{bracl}
\title{Bias reduction for adjacent category logit models for ordinal
responses using the Poisson trick.}
\usage{
bracl(
  formula,
  data,
  weights,
  subset,
  na.action,
  parallel = FALSE,
  contrasts = NULL,
  model = TRUE,
  x = TRUE,
  control = list(...),
  ...
)
}
\arguments{
\item{formula}{
    a formula expression as for regression models, of the form
    \code{response ~ predictors}. The response should be a factor
    (preferably an ordered factor), which will be interpreted as an
    ordinal response, with levels ordered as in the factor.  
    The model must have an intercept: attempts to remove one will
    lead to a warning and be ignored.  An offset may be used.  See the
    documentation of \code{\link{formula}} for other details.
  }

\item{data}{
    an optional data frame, list or environment in which to interpret
    the variables occurring in \code{formula}.
  }

\item{weights}{
    optional case weights in fitting.  Default to 1.
  }

\item{subset}{
    expression saying which subset of the rows of the data should  be used
    in the fit.  All observations are included by default.
  }

\item{na.action}{
    a function to filter missing data.
  }

\item{parallel}{if \code{FALSE} (default), then a non-proportional odds
adjacent category model is fit, assuming different effects per
category; if \code{TRUE} then a proportional odds adjacent category
model is fit. See Details.}

\item{contrasts}{
    a list of contrasts to be used for some or all of
    the factors appearing as variables in the model formula.
  }

\item{model}{
    logical for whether the model matrix should be returned.
  }

\item{x}{should the model matrix be included with in the result
(default is \code{TRUE}).}

\item{control}{a list of parameters for controlling the fitting
process. See \code{\link[=brglmControl]{brglmControl()}} for details.}

\item{...}{arguments to be used to form the default \code{control}
argument if it is not supplied directly.}
}
\description{
\code{\link[=bracl]{bracl()}} is a wrapper of \code{\link[=brglmFit]{brglmFit()}} that fits adjacent category
logit models with or without proportional odds using implicit and
explicit bias reduction methods. See Kosmidis & Firth (2011) for
details.
}
\details{
The \code{\link[=bracl]{bracl()}} function fits adjacent category models, which assume
multinomial observations with probabilities with proportional odds
of the form

\deqn{\log\frac{\pi_{ij}}{\pi_{ij + 1}} = \alpha_j + \beta^T x_i}{log(pi[i, j]/pi[i, j+1]) = alpha[j] + sum(beta * x[i, ])}

or with non-proportional odds of the form

\deqn{\log\frac{\pi_{ij}}{\pi_{ij + 1}} = \alpha_j + \beta_j^T x_i}{log(pi[i, j]/pi[i, j+1]) = alpha[j] + sum(beta[j, ] * x[i, ])}

where \eqn{x_i}{x[i, ]} is a vector of covariates and \eqn{\pi_{ij}}{pi[i, j]} is the
probability that category \eqn{j} is observed at the covariate setting \eqn{i}.
}
\examples{

data("stemcell", package = "brglm2")

# Adjacent category logit (non-proportional odds)
fit_bracl <- bracl(research ~ as.numeric(religion) + gender, weights = frequency,
                   data = stemcell, type = "ML")
# Adjacent category logit (proportional odds)
fit_bracl_p <- bracl(research ~ as.numeric(religion) + gender, weights = frequency,
                    data = stemcell, type = "ML", parallel = TRUE)


}
\references{
Kosmidis I, Kenne Pagui E C, Sartori N (2020). Mean and median bias
reduction in generalized linear models. \emph{Statistics and Computing},
\strong{30}, 43-59. \doi{10.1007/s11222-019-09860-6}.

Agresti, A (2010). \emph{Analysis of Ordinal Categorical Data} (2nd
edition).  Wiley Series in Probability and Statistics. Wiley.

Albert A, Anderson J A (1984). On the Existence of Maximum
Likelihood Estimates in Logistic Regression Models. \emph{Biometrika},
\strong{71}, 1-10. \doi{10.2307/2336390}.

Kosmidis I, Firth D (2011). Multinomial logit bias reduction
via the Poisson log-linear model. \emph{Biometrika}, \strong{98},
755-759. \doi{10.1093/biomet/asr026}.

Palmgren J (1981). The Fisher Information Matrix for Log Linear
Models Arguing Conditionally on Observed Explanatory
Variables. \emph{Biometrika}, \strong{68},
563-566. \doi{10.1093/biomet/68.2.563}.
}
\seealso{
\code{\link[nnet:multinom]{nnet::multinom()}}, \code{\link[=brmultinom]{brmultinom()}}
}
\author{
Ioannis Kosmidis \verb{[aut, cre]} \email{ioannis.kosmidis@warwick.ac.uk}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{endometrial}
\alias{endometrial}
\title{Histology grade and risk factors for 79 cases of endometrial cancer}
\format{
A data frame with 79 rows and 4 variables:
\itemize{
\item \code{NV}: neovasculization with coding 0 for absent and 1 for present
\item \code{PI}: pulsality index of arteria uterina
\item \code{EH}: endometrium height
\item \code{HG} histology grade with coding 0 for low grade and 1 for high grade
}
}
\source{
The packaged data set was downloaded in \code{.dat} format from
\url{https://users.stat.ufl.edu/~aa/glm/data/}. The latter link
provides the data sets used in Agresti (2015).

\if{html}{\out{<div class="sourceCode">}}\preformatted{The endometrial data set was first analyzed in Heinze and
Schemper (2002), and was originally provided by Dr
E. Asseryanis from the Medical University of Vienna.
}\if{html}{\out{</div>}}
}
\usage{
endometrial
}
\description{
Histology grade and risk factors for 79 cases of endometrial cancer
}
\references{
Agresti A (2015). \emph{Foundations of Linear and Generalized Linear
Models}.  Wiley Series in Probability and Statistics. Wiley.

Heinze G, Schemper M (2002). A Solution to the Problem of
Separation in Logistic Regression. \emph{Statistics in Medicine},
\strong{21}, 2409–2419. \doi{10.1002/sim.1047}.

Kosmidis I, Firth D (2021). Jeffreys-prior penalty, finiteness
and shrinkage in binomial-response generalized linear
models. \emph{Biometrika}, \strong{108}, 71-82. \doi{10.1093/biomet/asaa052}.
}
\seealso{
\code{\link[=brglm_fit]{brglm_fit()}}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bracl.R
\name{predict.bracl}
\alias{predict.bracl}
\title{Predict method for \link{bracl} fits}
\usage{
\method{predict}{bracl}(object, newdata, type = c("class", "probs"), ...)
}
\arguments{
\item{object}{a fitted object of class inheriting from \code{\link[=bracl]{"bracl"}}.}

\item{newdata}{optionally, a data frame in which to look for
variables with which to predict.  If omitted, the fitted linear
predictors are used.}

\item{type}{the type of prediction required. The default is
\code{"class"}, which produces predictions of the response category
at the covariate values supplied in \code{"newdata"}, selecting the
category with the largest probability; the alternative
\code{"probs"} returns all category probabilities at the covariate
values supplied in \code{newdata}.}

\item{...}{further arguments passed to or from other methods.}
}
\value{
If \code{type = "class"} a vector with the predicted response
categories; if \code{type = "probs"} a matrix of probabilities for all
response categories at \code{newdata}.
}
\description{
Obtain class and probability predictions from a fitted adjacent
category logits model.
}
\details{
If \code{newdata} is omitted the predictions are based on the data
used for the fit.
}
\examples{

data("stemcell", package = "brglm2")

# Adjacent category logit (non-proportional odds)
fit_bracl <- bracl(research ~ as.numeric(religion) + gender, weights = frequency,
                   data = stemcell, type = "ML")
# Adjacent category logit (proportional odds)
fit_bracl_p <- bracl(research ~ as.numeric(religion) + gender, weights = frequency,
                    data = stemcell, type = "ML", parallel = TRUE)

# New data
newdata <- expand.grid(gender = c("male", "female"),
                       religion = c("liberal", "moderate", "fundamentalist"))

# Predictions
sapply(c("class", "probs"), function(what) predict(fit_bracl, newdata, what))
sapply(c("class", "probs"), function(what) predict(fit_bracl_p, newdata, what))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{stemcell}
\alias{stemcell}
\title{Opinion on stem cell research and religious fundamentalism}
\format{
A data frame with 24 rows and 4 variables:
\itemize{
\item \code{research}: opinion about funding stem cell research with levels \code{definitely}, \code{probably}, \verb{probably not}, \verb{definitely not}
\item \code{gender}: the gender of the respondent with levels \code{female} and \code{male}
\item \code{religion}: the fundamentalism/liberalism of one’s religious beliefs with levels \code{fundamentalist}, \code{moderate},
\code{liberal}
\code{frequency}: the number of times a respondent fell in each of the combinations of levels for \code{research}, \code{religion} and \code{gender}
}
}
\source{
The \code{stemcell} data set is analyzed in Agresti (2010, Subsection 4.1.5).
}
\usage{
stemcell
}
\description{
A data set from the 2006 General Social Survey that shows the
relationship in the United States between opinion about funding
stem cell research and the fundamentalism/liberalism of one’s
religious beliefs, stratified by gender.
}
\references{
Agresti A (2010). \emph{Analysis of Ordinal Categorical Data} (2nd edition).  Wiley Series in
Probability and Statistics. Wiley.
}
\seealso{
\code{\link[=bracl]{bracl()}}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/brglm2-defunct.R
\name{brglm2-defunct}
\alias{brglm2-defunct}
\alias{check_infinite_estimates}
\alias{detect_separation}
\title{Defunct Functions in package \pkg{brglm2}}
\usage{
check_infinite_estimates(...)

detect_separation(...)
}
\arguments{
\item{...}{arguments to be passed to functions and methods.}
}
\description{
The functions or variables listed here are no longer part of
\pkg{brglm2}.
}
\details{
\itemize{
\item \code{\link[=detect_separation]{detect_separation()}}: This function is defunct from \pkg{brglm2}
since version 0.8.0. A new version of \code{\link[=detect_separation]{detect_separation()}} is now
maintained in the
\href{https://cran.r-project.org/package=detectseparation}{\pkg{detectseparation}}
R package.
\item \code{\link[=check_infinite_estimates]{check_infinite_estimates()}} is defunct from \pkg{brglm2} since
version 0.8.0. An new version of \code{\link[=check_infinite_estimates]{check_infinite_estimates()}} is
now maintained in the
\href{https://cran.r-project.org/package=detectseparation}{\pkg{detectseparation}}
R package.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plrtest.R
\name{plrtest.mdyplFit}
\alias{plrtest.mdyplFit}
\alias{plrtest}
\title{Penalized likelihood ratio test for \code{\link[=mdyplFit]{"mdyplFit"}} objects}
\usage{
\method{plrtest}{mdyplFit}(object1, object2, hd_correction = FALSE, ...)
}
\arguments{
\item{object1}{a \code{\link[=mdyplFit]{"mdyplFit"}} object}

\item{object2}{a \code{\link[=mdyplFit]{"mdyplFit"}} object}

\item{hd_correction}{if \code{FALSE} (default), then the corresponding
quantities are computed according to standard asymptotics. If
\code{TRUE} then the high-dimensionality corrections in Sterzinger &
Kosmidis (2024) are employed to updates estimates, estimated
standard errors, z-statistics, etc. See Details.}

\item{...}{further arguments to be passed to \code{\link[=summary.mdyplFit]{summary.mdyplFit()}}.}
}
\description{
Computes the Diaconis-Ylvisaker prior penalized likelihood ratio
test statistic or its adjusted version using high-dimensionality
correction under proportional asymptotics. Associated p-values are
also computed using a chi squared distribution.
}
\details{
Both \code{object1} and \code{object2} should have been fitted using the
\code{\link[=mdyplFit]{mdyplFit()}} method for \code{\link[=glm]{glm()}}, and the same shrinkage parameter
\code{alpha}; see \code{\link[=mdyplFit]{mdyplFit()}} and \code{\link[=mdyplControl]{mdyplControl()}} for setting \code{alpha}.

If \code{hd_correction = TRUE} then the deviance and the associated
p-value are adjusted using a high-dimensionality correction under
proportional asymptotics as in Sterzinger & Kosmidis (2024); see
\code{\link[=summary.mdyplFit]{summary.mdyplFit()}}.
}
\references{
Sterzinger P, Kosmidis I (2024). Diaconis-Ylvisaker prior
penalized likelihood for \eqn{p/n \to \kappa \in (0,1)} logistic
regression. \emph{arXiv}:2311.07419v2, \url{https://arxiv.org/abs/2311.07419}.
}
\seealso{
\code{\link[=mdyplFit]{mdyplFit()}}, \code{\link[=summary.mdyplFit]{summary.mdyplFit()}}, \code{\link[=mdypl_control]{mdypl_control()}}
}
\author{
Ioannis Kosmidis \verb{[aut, cre]} \email{ioannis.kosmidis@warwick.ac.uk}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/se.R
\name{se1}
\alias{se1}
\title{MDYPL state evolution functions with intercept}
\usage{
se1(
  mu,
  b,
  sigma,
  iota,
  kappa,
  gamma,
  alpha,
  intercept,
  gh = NULL,
  prox_tol = 1e-10
)
}
\arguments{
\item{mu}{aggregate bias parameter.}

\item{b}{parameter \code{b} in the state evolution functions.}

\item{sigma}{square root of the aggregate variance of the MDYPL
estimator.}

\item{iota}{limits of the MDYPL estimate for the intercept as the sample size goes to +Inf}

\item{kappa}{asymptotic ratio of columns/rows of the design
matrix. \code{kappa} should be in \verb{(0, 1)}.}

\item{gamma}{the square root of the limit of the variance of the
linear predictor.}

\item{alpha}{the shrinkage parameter of the MDYPL
estimator. \code{alpha} should be in \verb{(0, 1)}.}

\item{intercept}{intercept of the logistic regression model}

\item{gh}{A list with the Gauss-Hermite quadrature nodes and
weights, as returned from \code{statmod::gauss.quad()} with \code{kind = "hermite"}. Default is \code{NULL}, in which case \code{gh} is set to
\code{statmod::gauss.quad(200, kind = "hermite")} is used.}

\item{prox_tol}{tolerance for the computation of the proximal
operator; default is \code{1e-10}. fixed point problem solved via Newton-Raphson}
}
\description{
MDYPL state evolution functions with intercept
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/brnb.R
\name{coef.brnb}
\alias{coef.brnb}
\title{Extract model coefficients from \code{\link[=brnb]{"brnb"}} objects}
\usage{
\method{coef}{brnb}(object, model = c("mean", "full", "dispersion"), ...)
}
\arguments{
\item{object}{an object for which the extraction of model coefficients is
    meaningful.}

\item{model}{one of \code{"mean"} (default), \code{"full"}, \code{"dispersion"},
to return the estimates of the parameters in the linear
prediction only, or both, the estimate of the dispersion
parameter only, respectively.}

\item{...}{other arguments.}
}
\description{
Extract model coefficients from \code{\link[=brnb]{"brnb"}} objects
}
\details{
See \code{\link[=coef]{coef()}} for more details.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/brglmFit.R
\name{coef.brglmFit}
\alias{coef.brglmFit}
\title{Extract model coefficients from \code{\link[=brglmFit]{"brglmFit"}} objects}
\usage{
\method{coef}{brglmFit}(object, model = c("mean", "full", "dispersion"), ...)
}
\arguments{
\item{object}{an object for which the extraction of model coefficients is
    meaningful.}

\item{model}{one of \code{"mean"} (default), \code{"dispersion"}, `"full",
to return the estimates of the parameters in the linear
prediction only, the estimate of the dispersion parameter only,
or both, respectively.}

\item{...}{other arguments.}
}
\description{
Extract model coefficients from \code{\link[=brglmFit]{"brglmFit"}} objects
}
\details{
See \code{\link[=coef]{coef()}} for more details.
}
\seealso{
\code{\link[=coef]{coef()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/se.R
\name{se0}
\alias{se0}
\title{MDYPL state evolution functions with no intercept}
\usage{
se0(mu, b, sigma, kappa, gamma, alpha, gh = NULL, prox_tol = 1e-10)
}
\arguments{
\item{mu}{aggregate bias parameter.}

\item{b}{parameter \code{b} in the state evolution functions.}

\item{sigma}{square root of the aggregate variance of the MDYPL
estimator.}

\item{kappa}{asymptotic ratio of columns/rows of the design
matrix. \code{kappa} should be in \verb{(0, 1)}.}

\item{gamma}{the square root of the limit of the variance of the
linear predictor.}

\item{alpha}{the shrinkage parameter of the MDYPL
estimator. \code{alpha} should be in \verb{(0, 1)}.}

\item{gh}{A list with the Gauss-Hermite quadrature nodes and
weights, as returned from \code{statmod::gauss.quad()} with \code{kind = "hermite"}. Default is \code{NULL}, in which case \code{gh} is set to
\code{statmod::gauss.quad(200, kind = "hermite")} is used.}

\item{prox_tol}{tolerance for the computation of the proximal
operator; default is \code{1e-10}.}
}
\description{
MDYPL state evolution functions with no intercept
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expo.R
\name{expo.brglmFit}
\alias{expo.brglmFit}
\alias{brglmFit_expo}
\alias{expo}
\alias{expo.glm}
\title{Estimate the exponential of parameters of generalized linear models
using various methods}
\usage{
\method{expo}{brglmFit}(
  object,
  type = c("correction*", "correction+", "Lylesetal2012", "AS_median", "ML"),
  level = 0.95
)

\method{expo}{glm}(
  object,
  type = c("correction*", "correction+", "Lylesetal2012", "AS_median", "ML"),
  level = 0.95
)
}
\arguments{
\item{object}{an object of class \code{\link[=brglmFit]{"brglmFit"}} or
\code{\link[=glm]{"glm"}}.}

\item{type}{the type of correction to be used. The available
options are \code{"correction*"} (explicit mean bias correction with
a multiplicative adjustment), \code{"correction*"} (explicit mean
bias correction with an additive adjustment), \code{"Lylesetal2012"}
(explicit median bias correction using the multiplicative
adjustment in Lyles et al., 2012), \code{"AS_median"} (median bias
reduction), and \code{"ML"} (maximum likelihood). See Details.}

\item{level}{the confidence level required. Default is \code{0.95}.}
}
\value{
a list inheriting from class \code{\link[=brglmFit_expo]{"brglmFit_expo"}} with
components \code{coef} (the estimates of the exponentiated
regression parameters), \code{se} (the corresponding estimated
standard errors for the exponentiated parameters), \code{ci}
(confidence intervals of level \code{level} for the exponentiated
parameters), and \code{type} for the \code{type} of correction that has
been requested.
}
\description{
The \code{\link[=expo]{expo()}} method uses the supplied \code{\link[=brglmFit]{"brglmFit"}} or
\code{\link[=glm]{"glm"}} object to estimate the exponential of parameters of
generalized linear models with maximum likelihood or various mean
and median bias reduction methods. \code{\link[=expo]{expo()}} is useful for computing
(corrected) estimates of the multiplicative impact of a unit
increase on a covariate on the mean of a Poisson log-linear model
(\code{family = poisson("log")} in \code{\link[=glm]{glm()}}) while adjusting for other
covariates, the odds ratio associated with a unit increase on a
covariate in a logistic regression model (\code{family = binomial("logit")} in \code{\link[=glm]{glm()}}) while adjusting for other
covariates, the relative risk associated with a unit increase on a
covariate in a relative risk regression model (\code{family = binomial("log")} in \code{\link[=glm]{glm()}}) while adjusting for other covariates,
among others.
}
\details{
The supported methods through the \code{type} argument are:
\itemize{
\item \code{"ML"}: the estimates of the exponentiated parameters are
\eqn{\exp(\hat\theta_j)}, where \eqn{\theta_j} is the maximum
likelihood estimates for the \eqn{j}th regression parameter.
\item \code{"correction*"}: the estimates of the exponentiated parameters
are \eqn{\exp(\hat\theta_j) / (1 + \hat{v}_j / 2)}, where
\eqn{\hat\theta_j} is the estimate of the \eqn{j}th regression
parameter using \code{type = "AS_mixed"} in \code{\link[=brglmFit]{brglmFit()}}.
\item \code{"correction+"}: the estimates of the exponentiated parameters
are \eqn{\exp(\hat\theta_j) (1 - \hat{v}_j / 2)}, where
\eqn{\hat\theta_j} is the estimate of the \eqn{j}th regression
parameter using \code{type = "AS_mixed"} in \code{\link[=brglmFit]{brglmFit()}}.
\item \code{"Lylesetal2012"}: the estimates of the exponentiated parameters
are \eqn{\exp(\hat\theta_j) exp(- \hat{v}_j / 2)}, where
\eqn{\hat\theta_j} is the estimate of the \eqn{j}th regression
parameter using \code{type = "AS_mixed"} in \code{\link[=brglmFit]{brglmFit()}}. This estimator
has been proposed in Lyles et al. (2012).
\item \code{"AS_median"}: the estimates of the exponentiated parameters are
\eqn{\exp(\hat\theta_j)}, where \eqn{\hat\theta_j} is the estimate
of the \eqn{j}th regression parameter using \code{type = "AS_median"} in
\code{\link[=brglmFit]{brglmFit()}}.
}

\code{"correction*"} and \code{"correction+"} are based on multiplicative and
additive adjustments, respectively, of the exponential of a
reduced-bias estimator (like the ones coming from \code{\link[=brglmFit]{brglmFit()}} with
\code{type = "AS_mixed"}, \code{type = "AS_mean"}, and \code{type = "correction"}). The form of those adjustments results from the
expression of the first-term in the mean bias expansion of the
exponential of a reduced-bias estimator. See, for example, Di
Caterina & Kosmidis (2019, expression 12) for the general form of
the first-term of the mean bias of a smooth transformation of a
reduced-bias estimator.

The estimators from \code{"correction+"}, \code{"correction*"},
\code{"Lylesetal2012"} have asymptotic mean bias of order smaller than
than of the maximum likelihood estimator. The estimators from
\code{"AS_median"} are asymptotically closed to being median unbiased
than the maximum likelihood estimator is.

Estimated standard errors are computed using the delta method,
where both the Jacobin and the information matrix are evaluated at
the logarithm of the estimates of the exponentiated parameters.

Confidence intervals results by taking the exponential of the
limits of standard Wald-type intervals computed at the logarithm of
the estimates of the exponentiated parameters.
}
\examples{

## The lizards example from ?brglm::brglm
lizardsML <- glm(cbind(grahami, opalinus) ~ height + diameter +
                 light + time, family = binomial(logit), data = lizards,
                 method = "glm.fit")
# Get estimates, standard errors, and confidence intervals of odds
# ratios with various methods
expo(lizardsML, type = "ML")
expo(lizardsML, type = "correction*")
expo(lizardsML, type = "Lylesetal2012")
expo(lizardsML, type = "correction+")
expo(lizardsML, type = "AS_median")

## Example from ?glm
## Dobson (1990) Page 93: Randomized Controlled Trial :
df_D93 <- data.frame(counts = c(18,17,15,20,10,20,25,13,12),
                     outcome = gl(3,1,9),
                     treatment = gl(3,3))
glm.D93 <- glm(counts ~ outcome + treatment, family = poisson(), data = df_D93)
expo(glm.D93, type = "correction*")

}
\references{
Di Caterina C, Kosmidis I (2019). Location-Adjusted Wald Statistics for Scalar
Parameters. \emph{Computational Statistics & Data Analysis}, \strong{138},
126-142. \doi{10.1016/j.csda.2019.04.004}.

Kosmidis I, Kenne Pagui E C, Sartori N (2020). Mean and median bias
reduction in generalized linear models. \emph{Statistics and Computing},
\strong{30}, 43-59. \doi{10.1007/s11222-019-09860-6}.

Cordeiro G M, McCullagh P (1991). Bias correction in generalized
linear models. \emph{Journal of the Royal Statistical Society. Series B
(Methodological)}, \strong{53}, 629-643. \doi{10.1111/j.2517-6161.1991.tb01852.x}.

Lyles R H, Guo Y, Greenland S (2012). Reducing bias and mean
squared error associated with regression-based odds ratio
estimators. \emph{Journal of Statistical Planning and Inference},
\strong{142} 3235–3241. \doi{10.1016/j.jspi.2012.05.005}.
}
\seealso{
\code{\link[=brglm_fit]{brglm_fit()}} and and \code{\link[=brglm_control]{brglm_control()}}
}
\author{
Ioannis Kosmidis \verb{[aut, cre]} \email{ioannis.kosmidis@warwick.ac.uk}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/brnb.R
\name{brnb}
\alias{brnb}
\title{Bias reduction for negative binomial regression models}
\usage{
brnb(
  formula,
  data,
  subset,
  weights = NULL,
  offset = NULL,
  link = "log",
  start = NULL,
  etastart = NULL,
  mustart = NULL,
  control = list(...),
  na.action,
  model = TRUE,
  x = FALSE,
  y = TRUE,
  contrasts = NULL,
  intercept = TRUE,
  singular.ok = TRUE,
  ...
)
}
\arguments{
\item{formula}{an object of class \code{"\link[stats]{formula}"} (or one that
    can be coerced to that class): a symbolic description of the
    model to be fitted.  The details of model specification are given
    under \sQuote{Details}.}

\item{data}{an optional data frame, list or environment (or object
    coercible by \code{\link{as.data.frame}} to a data frame) containing
    the variables in the model.  If not found in \code{data}, the
    variables are taken from \code{environment(formula)},
    typically the environment from which \code{glm} is called.}

\item{subset}{an optional vector specifying a subset of observations
    to be used in the fitting process.  (See additional details about how
    this argument interacts with data-dependent bases in the
    \sQuote{Details} below.)}

\item{weights}{an optional vector of \sQuote{prior weights} to be used
    in the fitting process.  Should be \code{NULL} or a numeric vector.}

\item{offset}{this can be used to specify an \emph{a priori} known
    component to be included in the linear predictor during fitting.
    This should be \code{NULL} or a numeric vector of length equal to
    the number of cases.  One or more \code{\link[stats]{offset}} terms can be
    included in the formula instead or as well, and if more than one is
    specified their sum is used.  See \code{\link[stats]{model.offset}}.}

\item{link}{The link function. Currently must be one of \code{"log"},
\code{"sqrt"} or \code{"identity"}.}

\item{start}{starting values for the parameters in the linear predictor.}

\item{etastart}{starting values for the linear predictor.}

\item{mustart}{starting values for the vector of means.}

\item{control}{a list of parameters for controlling the fitting
process. See \code{\link[=brglmControl]{brglmControl()}} for details.}

\item{na.action}{a function which indicates what should happen
    when the data contain \code{NA}s.  The default is set by
    the \code{na.action} setting of \code{\link{options}}, and is
    \code{\link[stats]{na.fail}} if that is unset.  The \sQuote{factory-fresh}
    default is \code{\link[stats]{na.omit}}.  Another possible value is
    \code{NULL}, no action.  Value \code{\link[stats]{na.exclude}} can be useful.}

\item{model}{a logical value indicating whether \emph{model frame}
    should be included as a component of the returned value.}

\item{x, y}{For \code{glm}:
    logical values indicating whether the response vector and model
    matrix used in the fitting process should be returned as components
    of the returned value.

    For \code{glm.fit}: \code{x} is a design matrix of dimension
    \code{n * p}, and \code{y} is a vector of observations of length
    \code{n}.
  }

\item{contrasts}{an optional list. See the \code{contrasts.arg}
    of \code{model.matrix.default}.}

\item{intercept}{logical. Should an intercept be included in the
    \emph{null} model?}

\item{singular.ok}{logical; if \code{FALSE} a singular fit is an
    error.}

\item{...}{
    For \code{glm}: arguments to be used to form the default
    \code{control} argument if it is not supplied directly.

    For \code{weights}: further arguments passed to or from other methods.
  }
}
\value{
A fitted model object of class \code{\link[=brnb]{"brnb"}} inheriting
from \code{\link[MASS:glm.nb]{"negbin"}} and
\code{\link[=brglmFit]{"brglmFit"}}. The object is similar to the output
of \code{\link[=brglmFit]{brglmFit()}} but contains four additional components:
\code{theta} for the estimate of the dispersion parameter,
\code{vcov.mean} for the estimated variance-covariance matrix of the
regression coefficients, \code{vcov.dispersion} for the estimated
variance of the dispersion parameter in the chosen
parameterization (using the expected information), and
\code{twologlik} for twice the log-likelihood function.
}
\description{
\code{\link[=brnb]{brnb()}} is a function that fits negative binomial regression
models using implicit and explicit bias reduction methods.
}
\details{
A detailed description of the fitting procedure is given in the
iteration vignette (see, \code{vignette("iteration", "brglm2")} and
Kosmidis et al, 2020). The number of iterations when estimating
parameters are controlled by the \code{maxit} argument of
\code{\link[=brglmControl]{brglmControl()}}.

The type of score adjustment to be used is specified through the
\code{type} argument (see \code{\link[=brglmControl]{brglmControl()}} for details).

The available options are:
\itemize{
\item \code{type = "AS_mixed"}: the mixed bias-reducing score
adjustments in Kosmidis et al (2020) that result in mean bias
reduction for the regression parameters and median bias reduction
for the dispersion parameter, if any; default.
\item \code{type = "AS_mean"}: the mean bias-reducing score
adjustments in Firth (1993) and Kosmidis & Firth (2009).
\item \code{type = "AS_median"}: the median bias-reducing score
adjustments in Kenne Pagui et al. (2017)
\item \code{type = "MPL_Jeffreys"}: maximum penalized likelihood
with powers of the Jeffreys prior as penalty.
\item \code{type = "ML"}: maximum likelihood.
\item \code{type = "correction"}: asymptotic bias correction, as in
Cordeiro & McCullagh (1991).
}

The choice of the parameterization for the dispersion is controlled
by the \code{transformation} argument (see \code{\link[=brglmControl]{brglmControl()}} for
details).  The default is \code{"identity"}. Using \code{transformation = "inverse"} uses the dispersion parameterization that
\code{\link[MASS:glm.nb]{MASS::glm.nb()}} uses.
}
\examples{
## Example in Saha, K., & Paul, S. (2005). Bias-corrected maximum
## likelihood estimator of the negative binomial dispersion
## parameter.  Biometrics, 61, 179--185.
#
# Number of revertant colonies of salmonella data
salmonella <- data.frame(freq = c(15, 16, 16, 27, 33, 20,
                                  21, 18, 26, 41, 38, 27,
                                  29, 21, 33, 60, 41, 42),
                         dose = rep(c(0, 10, 33, 100, 333, 1000), 3),
                         observation = rep(1:3, each = 6))

# Maximum likelihood fit with glm.nb of MASS
salmonella_fm <- freq ~ dose + log(dose + 10)
fitML_glmnb <- MASS::glm.nb(salmonella_fm, data = salmonella)

# Maximum likelihood fit with brnb
fitML <- brnb(salmonella_fm, data = salmonella,
              link = "log", transformation = "inverse", type = "ML")

# Mean bias-reduced fit
fitBR_mean <- update(fitML, type = "AS_mean")

# Median bias-reduced fit
fitBR_median <- update(fitML, type = "AS_median")

# Mixed bias-reduced fit
fitBR_mixed <- update(fitML, type = "AS_mixed")

# Mean bias-corrected fit
fitBC_mean <- update(fitML, type = "correction")

# Penalized likelihood with Jeffreys-prior penalty
fit_Jeffreys <- update(fitML, type = "MPL_Jeffreys")

# The parameter estimates from glm.nb and brnb with type = "ML" are
# numerically the same
all.equal(c(coef(fitML_glmnb), fitML_glmnb$theta),
            coef(fitML, model = "full"), check.attributes = FALSE)

# Because of the invariance properties of the maximum likelihood,
# median reduced-bias, and mixed reduced-bias estimators the
# estimate of a monotone function of the dispersion should be
# (numerically) the same as the function of the estimate of the
# dispersion:

# ML
coef(fitML, model = "dispersion")
1 / coef(update(fitML, transformation = "identity"), model = "dispersion")

# Median BR
coef(fitBR_median, model = "dispersion")
1 / coef(update(fitBR_median, transformation = "identity"), model = "dispersion")

# Mixed BR
coef(fitBR_mixed, model = "dispersion")
1 / coef(update(fitBR_mixed, transformation = "identity"), model = "dispersion")

## The same is not true for mean BR
coef(fitBR_mean, model = "dispersion")
1 / coef(update(fitBR_mean, transformation = "identity"), model = "dispersion")

\donttest{
## An example  from Venables & Ripley (2002, p.169).
data("quine", package = "MASS")
quineML <- brnb(Days ~ Sex/(Age + Eth*Lrn), link = "sqrt", transformation="inverse",
                data = quine, type="ML")
quineBR_mean <- update(quineML, type = "AS_mean")
quineBR_median <- update(quineML, type = "AS_median")
quineBR_mixed <- update(quineML, type = "AS_mixed")
quine_Jeffreys <- update(quineML, type = "MPL_Jeffreys")

fits <- list(ML = quineML,
             AS_mean = quineBR_mean,
             AS_median = quineBR_median,
             AS_mixed = quineBR_mixed,
             MPL_Jeffreys = quine_Jeffreys)
sapply(fits, coef, model = "full")
}

}
\references{
Cordeiro G M, McCullagh P (1991). Bias correction in generalized
linear models. \emph{Journal of the Royal Statistical Society. Series B
(Methodological)}, \strong{53}, 629-643. \doi{10.1111/j.2517-6161.1991.tb01852.x}.

Firth D (1993). Bias reduction of maximum likelihood estimates.
\emph{Biometrika}. \strong{80}, 27-38. \doi{10.2307/2336755}.

Kenne Pagui E C, Salvan A, Sartori N (2017). Median bias
reduction of maximum likelihood estimates. \emph{Biometrika}, \strong{104},
923–938. \doi{10.1093/biomet/asx046}.

Kosmidis I, Kenne Pagui E C, Sartori N (2020). Mean and median bias
reduction in generalized linear models. \emph{Statistics and Computing},
\strong{30}, 43-59. \doi{10.1007/s11222-019-09860-6}.

Kosmidis I, Firth D (2009). Bias reduction in exponential family
nonlinear models. \emph{Biometrika}, \strong{96}, 793-804. \doi{10.1093/biomet/asp055}.
}
\author{
Euloge Clovis Kenne Pagui \verb{[aut]} \email{kenne@stat.unipd.it}, Ioannis Kosmidis \verb{[aut, cre]} \email{ioannis.kosmidis@warwick.ac.uk}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/brglmFit.R
\name{vcov.brglmFit}
\alias{vcov.brglmFit}
\title{Return the variance-covariance matrix for the regression parameters
in a \code{\link[=brglmFit]{brglmFit()}} object}
\usage{
\method{vcov}{brglmFit}(object, model = c("mean", "full", "dispersion"), complete = TRUE, ...)
}
\arguments{
\item{object}{
    a fitted model object, typically.  Sometimes also a
    \code{\link{summary}()} object of such a fitted model.
  }

\item{model}{character specifying for which component of the model coefficients should be extracted.}

\item{complete}{for the \code{aov}, \code{lm}, \code{glm}, \code{mlm}, and where
    applicable \code{summary.lm} etc methods: logical indicating if the
    full variance-covariance matrix should be returned also in case of
    an over-determined system where some coefficients are undefined and
    \code{\link[stats]{coef}(.)} contains \code{NA}s correspondingly.   When
    \code{complete = TRUE},  \code{vcov()} is compatible with
    \code{coef()} also in this singular case.}

\item{...}{
    additional arguments for method functions.  For the
    \code{\link[stats]{glm}} method this can be used to pass a
    \code{dispersion} parameter.}
}
\description{
Return the variance-covariance matrix for the regression parameters
in a \code{\link[=brglmFit]{brglmFit()}} object
}
\details{
The options for \code{model} are \code{"mean"} for mean regression parameters
only (default), \code{"dispersion"} for the dispersion parameter (or the
transformed dispersion; see \code{\link[=brglm_control]{brglm_control()}}), and \code{"full"} for
both the mean regression and the (transformed) dispersion
parameters.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{hepatitis}
\alias{hepatitis}
\title{Post-transfusion hepatitis: impact of non-A, non-B hepatitis
surrogate tests}
\format{
A data frame with 28 rows and the following 6 columns:
\itemize{
\item \code{city}: Subjects were recruited from 3 Canadian Red Cross Society
Blood Centres and 13 university-affiliated hospitals in 3 cities:
Toronto, Hamilton and Winnipeg.
\item \code{group}: Eligible subjects were assigned to one of two allogenic
blood recipient groups.  One group received products that had only
routine Canadian transfusion-transmissible disease marker screening
(no-withhold).  The other group received only products that were
not positive for NANB surrogate markers (withhold).
\item \code{time}: Hepatitis C (HCV) screening was introduced in Canada in
May, 1990.  Subjects were recruited into the study before (pre) and
after (post) the introduction of anti-HCV testing.
\item \code{HCV}: Post-transfusion HCV hepatitis present (1) or absent (0).
\item \code{nonABC}: Post-transfusion non-A, non-B, non-C hepatitis present (1) or absent (0)
\item \code{counts}: Number of subjects
}
}
\source{
Data is from Blajchman et al. (1995), also analyzed in Bull et
al. (2002), and is also provided by the
\href{https://cran.r-project.org/package=pmlr}{\pkg{pmlr}} R package.
}
\usage{
hepatitis
}
\description{
Data from a randomized double-blind trial to assess whether
withholding donor blood positive for the non-A, non-B (\code{"NANB"})
surrogate markers would reduce the frequency of post-transfusion
hepatitis.  The dataset contains \code{4588} subjects enrolled from 1988
to 1992 into two study groups that received allogenic blood from
which units positive for NANB surrogate markers were withheld (n =
\code{2311}) or not withheld (n = \code{2277}).  Subjects were followed up
for 6 months and assessed for the presence of post-transfusion
hepatitis.
}
\references{
Bull S B, Mak C, Greenwood C M T (2002). A modified score function
estimator for multinomial logistic regression in small
samples. \emph{Computational Statistics & Data Analysis}, \strong{39},
57-74. \doi{10.1016/S0167-9473(01)00048-2}

Blajchman M A, Bull S B and Feinman S V (1995). Post-transfusion
hepatitis: impact of non-A, non-B hepatitis surrogate tests. \emph{The
Lancet}, \strong{345}, 21--25. \doi{10.1016/S0140-6736(95)91153-7}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary-mdyplFit.R
\name{sloe}
\alias{sloe}
\title{Estimate the corrupted signal strength in a model with
(sub-)Gaussian covariates}
\usage{
sloe(object)
}
\arguments{
\item{object}{an \code{\link[=mdyplFit]{"mdyplFit"}} object.}
}
\value{
A scalar.
}
\description{
Estimate the corrupted signal strength in a model with
(sub-)Gaussian covariates
}
\details{
The Signal Strength Leave-One-Out Estimator (SLOE) is defined in
Yadlowsky et al. (2021) when the model is estimated using maximum
likelihood (i.e. when \code{object$alpha = 1}; see \code{\link[=mdyplControl]{mdyplControl()}} for
what \code{alpha} is). The SLOE adaptation when estimation is through
maximum Diaconis-Ylvisaker prior penalized likelihood
(\code{\link[=mdypl_fit]{mdypl_fit()}}) has been put forward in Sterzinger & Kosmidis
(2025).

In particular, \code{\link[=sloe]{sloe()}} computes an estimate of the corrupted
signal strength which is the limit \deqn{\nu^2} of \eqn{var(X
\hat\beta(\alpha))}, where \eqn{\hat\beta(\alpha)} is the maximum
Diaconis-Ylvisaker prior penalized likelihood (MDYPL) estimator as
computed by \code{\link[=mdyplFit]{mdyplFit()}} with shrinkage parameter \eqn{alpha}.
}
\references{
Sterzinger P, Kosmidis I (2024). Diaconis-Ylvisaker prior
penalized likelihood for \eqn{p/n \to \kappa \in (0,1)} logistic
regression. \emph{arXiv}:2311.07419v2, \url{https://arxiv.org/abs/2311.07419}.

Yadlowsky S, Yun T, McLean C Y, D' Amour A (2021). SLOE: A Faster
Method for Statistical Inference in High-Dimensional Logistic
Regression. In M Ranzato, A Beygelzimer, Y Dauphin, P Liang, J W
Vaughan (eds.), \emph{Advances in Neural Information Processing
Systems}, \strong{34}, 29517–29528. Curran Associates,
Inc. \url{https://proceedings.neurips.cc/paper_files/paper/2021/file/f6c2a0c4b566bc99d596e58638e342b0-Paper.pdf}.
}
\seealso{
\code{\link[=summary.mdyplFit]{summary.mdyplFit()}}
}
\author{
Ioannis Kosmidis \verb{[aut, cre]} \email{ioannis.kosmidis@warwick.ac.uk}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{enzymes}
\alias{enzymes}
\title{Liver enzyme data}
\format{
A data frame with 218 rows and the following 6 columns:
\itemize{
\item \code{Patient}: Patient ID
\item \code{Group}: Four diagnostic groups were considered: acute viral
hepatitis (1), persistent chronic hepatitis (2), aggressive chronic
hepatitis (3) and post-necrotic cirrhosis (4).
\item \code{AST}: Aspartate aminotransferase (in U/L)
\item \code{ALT}: Alanine aminotransferase (in U/L)
\item \code{GLDH}: Glutamate dehydrogenase (in U/L)
\item \code{OCT}: Ornithine carbamyltransferase (in U/L)
}
}
\source{
Data from Albert and Harris (1984, Chapter 5, Appendix I), and is
also provided by the
\href{https://cran.r-project.org/package=pmlr}{\pkg{pmlr}} R package.
}
\usage{
enzymes
}
\description{
Liver enzyme data collected from 218 patients with liver disease
(Plomteux, 1980). The laboratory profile consists of enzymatic
activity measured for four liver enzymes: aspartate
aminotransferase (\code{AST}), alanine aminotransferase (\code{ALT}),
glutamate dehydrogenase (\code{GLDH}) and ornithine carbamyltransferase
(\code{OCT}).
}
\references{
Albert A, Harris E K (1984). \emph{Multivariate Interpretation of
Clinical Laboratory Data}. Dekker: New York.

Plomteux G (1980). Multivariate analysis of an enzyme profile for
the differential diagnosis of viral hepatitis. \emph{Clinical
Chemistry}, \strong{26}, 1897-1899.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ordinal_superiority.R
\name{ordinal_superiority.bracl}
\alias{ordinal_superiority.bracl}
\alias{ordinal_superiority}
\title{Ordinal superiority scores of Agresti and Kateri (2017)}
\usage{
\method{ordinal_superiority}{bracl}(
  object,
  formula,
  data,
  measure = c("gamma", "Delta"),
  level = 0.95,
  bc = FALSE
)
}
\arguments{
\item{object}{a fitted object from an ordinal regression
model. Currently only models from class \code{\link[=bracl]{"bracl"}} are supported.}

\item{formula}{a RHS formula indicating the group variable to use.}

\item{data}{an optional data frame in which to look for variables
with which to compute ordinal superiority measures.  If
omitted, an attempt is made to use the data that produced
\code{object}.}

\item{measure}{either \code{"gamma"} (default) or \code{"Delta"}, specifying
the ordinal superiority measure to be returned.}

\item{level}{the confidence level required when computing
confidence intervals for the ordinal superiority measures.}

\item{bc}{logical. If \code{FALSE} (default) then the ordinal
superiority measures are computed using the estimates in
\code{object}. If \code{TRUE} then the ordinal superiority measure
estimates are corrected for mean bias.}
}
\description{
\code{\link[=ordinal_superiority]{ordinal_superiority()}} is a method for the estimation and
inference about model-based ordinal superiority scores introduced
in Agresti and Kateri (2017, Section 5) from fitted objects. The
mean bias of the estimates of the ordinal superiority scores can be
corrected.
}
\examples{

data("stemcell", package = "brglm2")

# Adjacent category logit (proportional odds)
stem <- within(stemcell, {nreligion = as.numeric(religion)})
fit_bracl_p <- bracl(research ~ nreligion + gender, weights = frequency,
                     data = stem, type = "ML", parallel = TRUE)

# Estimates and 95\% confidence intervals for the probabilities that the response
# category for gender "female" is higher than the response category for gender "male",
# while adjusting for religion.
ordinal_superiority(fit_bracl_p, ~ gender)

\donttest{
# And their (very-similar in value here) bias corrected versions
# with 99\% CIs
ordinal_superiority(fit_bracl_p, ~ gender, bc = TRUE, level = 0.99)
# Note that the object is refitted with type = "AS_mean"

}

}
\references{
Agresti, A., Kateri, M. (2017). Ordinal probability effect measures
for group comparisons in multinomial cumulative link models.
\emph{Biometrics}, \strong{73} 214-219. \doi{10.1111/biom.12565}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/brmultinom.R, R/zzz_conventions.R
\name{residuals.brmultinom}
\alias{residuals.brmultinom}
\alias{residuals.bracl}
\title{Residuals for multinomial logistic regression and adjacent category logit models}
\usage{
\method{residuals}{brmultinom}(object, type = c("pearson", "response", "deviance", "working"), ...)

\method{residuals}{bracl}(object, type = c("pearson", "response", "deviance", "working"), ...)
}
\arguments{
\item{object}{the object coming out of \code{\link[=bracl]{bracl()}} and
\code{\link[=brmultinom]{brmultinom()}}.}

\item{type}{the type of residuals which should be returned.  The
options are: \code{"pearson"} (default), \code{"response"}, \code{"deviance"},
\code{"working"}. See Details.}

\item{...}{Currently not used.}
}
\description{
Residuals for multinomial logistic regression and adjacent category logit models
}
\details{
The residuals computed are the residuals from the equivalent
Poisson log-linear model fit, organized in a form that matches the
output of \code{fitted(object, type = "probs")}. As a result, the output
is residuals defined in terms of the object and expected
multinomial counts.
}
\seealso{
brmultinom bracl
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mdyplFit.R, R/zzz_conventions.R
\name{mdyplControl}
\alias{mdyplControl}
\alias{mdypl_control}
\title{Auxiliary function for \code{\link[=glm]{glm()}} fitting using the \code{\link[=brglmFit]{brglmFit()}}
method.}
\usage{
mdyplControl(alpha = NULL, epsilon = 1e-08, maxit = 25, trace = FALSE)

mdypl_control(alpha = NULL, epsilon = 1e-08, maxit = 25, trace = FALSE)
}
\arguments{
\item{alpha}{the shrinkage parameter (in \verb{[0, 1]}) in the
Diaconis-Ylvisaker prior penalty. Default is \code{NULL}, which
results in \code{alpha = n / (n + p)}, where \code{n} is the sum of the
binomial totals and \code{p} is the number of model
parameters. Setting \code{alpha = 1} corresponds to using maximum
likelihood, i.e. no penalization. See Details.}

\item{epsilon}{positive convergence tolerance epsilon. Default is
\code{1e-08}.}

\item{maxit}{integer giving the maximal number of iterations
allowed. Default is \code{25}.}

\item{trace}{logical indicating if output should be produced for
each iteration. Default is \code{FALSE}.}
}
\value{
A list with components named as the arguments.
}
\description{
Typically only used internally by \code{\link[=brglmFit]{brglmFit()}}, but may be used to
construct a \code{control} argument.
}
\details{
Internally, \code{\link[=mdyplFit]{mdyplFit()}} uses \code{\link[stats:glm]{stats::glm.fit()}} to fit a logistic
regression model on responses \code{alpha * y + (1 - alpha) / 2}, where
\code{y} are the original binomial responses scaled by the binomial
totals. \code{epsilon}, \code{maxit} and \code{trace} control the
\code{\link[stats:glm]{stats::glm.fit()}} call; see \code{\link[stats:glm.control]{stats::glm.control()}}.
}
\seealso{
\code{\link[=mdyplFit]{mdyplFit()}}, \code{\link[=glm.control]{glm.control()}}
}
\author{
Ioannis Kosmidis \verb{[aut, cre]} \email{ioannis.kosmidis@warwick.ac.uk}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/brmultinom.R
\name{confint.brmultinom}
\alias{confint.brmultinom}
\title{Method for computing confidence intervals for one or more
regression parameters in a \code{\link[=brmultinom]{"brmultinom"}} object}
\usage{
\method{confint}{brmultinom}(object, parm, level = 0.95, ...)
}
\arguments{
\item{object}{a fitted model object.}

\item{parm}{a specification of which parameters are to be given
    confidence intervals, either a vector of numbers or a vector of
    names.  If missing, all parameters are considered.}

\item{level}{the confidence level required.}

\item{...}{additional argument(s) for methods.}
}
\description{
Method for computing confidence intervals for one or more
regression parameters in a \code{\link[=brmultinom]{"brmultinom"}} object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/summary-mdyplFit.R
\name{summary.mdyplFit}
\alias{summary.mdyplFit}
\alias{print.summary.mdyplFit}
\title{Summary method for \code{\link[=mdyplFit]{"mdyplFit"}} objects}
\usage{
\method{summary}{mdyplFit}(object, hd_correction = FALSE, solve_se_control = list(), ...)

\method{print}{summary.mdyplFit}(
  x,
  digits = max(3L, getOption("digits") - 3L),
  symbolic.cor = x$symbolic.cor,
  signif.stars = getOption("show.signif.stars"),
  ...
)
}
\arguments{
\item{object}{an object of class \code{"glm"}, usually, a result of a
    call to \code{\link[stats]{glm}}.}

\item{hd_correction}{if \code{FALSE} (default), then the corresponding
quantities are computed according to standard asymptotics. If
\code{TRUE} then the high-dimensionality corrections in Sterzinger &
Kosmidis (2024) are employed to updates estimates, estimated
standard errors, z-statistics, etc. See Details.}

\item{solve_se_control}{a list of further arguments to be passed to
\code{\link[=solve_se]{solve_se()}}. Even if explicitly specified, the arguments
\code{kappa}, \code{ss}, \code{alpha}, and \code{intercept} are always set
according to \code{object}, and \code{corrupted} is set to \code{TRUE}.}

\item{...}{further arguments to be passed to \code{\link[=summary.glm]{summary.glm()}}.}

\item{x}{an object of class \code{"summary.glm"}, usually, a result of a
    call to \code{summary.glm}.}

\item{digits}{the number of significant digits to use when printing.}

\item{symbolic.cor}{logical. If \code{TRUE}, print the correlations in
    a symbolic form (see \code{\link[stats]{symnum}}) rather than as numbers.}

\item{signif.stars}{logical. If \code{TRUE}, \sQuote{significance stars}
    are printed for each coefficient.}
}
\value{
A list with objects as in the result of \code{\link[stats:summary.glm]{stats::summary.glm()}},
with extra component \code{se_parameters}, which is the vector of the
solution to the state evolution equations with extra attributes
(see \code{\link[=solve_se]{solve_se()}}).
}
\description{
Summary method for \code{\link[=mdyplFit]{"mdyplFit"}} objects
}
\details{
If \code{hd_correction = TRUE}, the \code{\link[=sloe]{sloe()}} estimator of the square
root of the corrupted signal strength is estimated from \code{object},
as are the conditional variances of each covariate given the others
(excluding the intercept). The latter are estimated using residual
sums of squares from the linear regression of each covariate on all
the others, as proposed in Zhao et al (2021, Section 5.1). Then the
appropriate state evolution equations are solved using \code{\link[=solve_se]{solve_se()}}
with \code{corrupted = TRUE}, and the obtained constants are used to
rescale the estimates, and adjust estimated standard errors and
z-statistics as in Sterzinger & Kosmidis (2024).

The key assumptions under which the rescaled estimates and corrected
standard errors and z-statistics are asymptotically valid are that
the covariates have sub-Gaussian distributions, and that the signal
strength, which is the limit \eqn{\gamma^2} of \eqn{var(X \beta)}
is finite as \eqn{p / n \to \kappa \in (0, 1)}, with \eqn{\kappa \in
(0, 1)}. See Sterzinger & Kosmidis (2024).

If \code{hd_correction = TRUE}, and the model has an intercept, then the
result provides only a corrected estimate of the intercept with no
accompanying standard error, z-statistic, and p-value. Also,
\code{vcov(summary(object, hd_correction = TRUE))} is always
\code{NULL}. Populating those objects with appropriate estimates is the
subject of current work.
}
\examples{

\donttest{

set.seed(123)
n <- 2000
p <- 400
set.seed(123)
betas <- c(rnorm(p / 2, mean = 7, sd = 1), rep(0, p / 2))
X <- matrix(rnorm(n * p, 0, 1/sqrt(n)), nrow = n, ncol = p)
probs <- plogis(drop(X \%*\% betas))
y <- rbinom(n, 1, probs)
fit_mdypl <- glm(y ~ -1 + X, family = binomial(), method = "mdyplFit")

st_summary <- summary(fit_mdypl)
hd_summary <- summary(fit_mdypl, hd_correction = TRUE)

cols <- hcl.colors(3, alpha = 0.2)
par(mfrow = c(1, 2))
plot(betas, type = "l", ylim = c(-3, 14),
     main = "MDYPL estimates",
     xlab = "Parameter index", ylab = NA)
points(coef(st_summary)[, "Estimate"], col = NA, bg = cols[1], pch = 21)

plot(betas, type = "l", ylim = c(-3, 14),
     main = "rescaled MDYPL estimates",
     xlab = "Parameter index", ylab = NA)
points(coef(hd_summary)[, "Estimate"], col = NA, bg = cols[2], pch = 21)

## z-statistics
z_mdypl <- coef(st_summary)[betas == 0, "z value"]
qqnorm(z_mdypl, col = NA, bg = cols[1], pch = 21, main = "z value")
abline(0, 1, lty = 2)
z_c_mdypl <- coef(hd_summary)[betas == 0, "z value"]
qqnorm(z_c_mdypl, col = NA, bg = cols[2], pch = 21, main = "corrected z value")
abline(0, 1, lty = 2)

}

}
\references{
Zhao Q, Sur P, Cand\`es E J (2022). The asymptotic distribution of
the MLE in high-dimensional logistic models: Arbitrary
covariance. \emph{Bernoulli}, \strong{28},
1835–1861. \doi{10.3150/21-BEJ1401}.

Sterzinger P, Kosmidis I (2024). Diaconis-Ylvisaker prior
penalized likelihood for \eqn{p/n \to \kappa \in (0,1)} logistic
regression. \emph{arXiv}:2311.07419v2, \url{https://arxiv.org/abs/2311.07419}.
}
\seealso{
\code{\link[=mdyplFit]{mdyplFit()}}, \code{\link[=solve_se]{solve_se()}}
}
\author{
Ioannis Kosmidis \verb{[aut, cre]} \email{ioannis.kosmidis@warwick.ac.uk}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expo.R
\name{coef.brglmFit_expo}
\alias{coef.brglmFit_expo}
\title{Extract estimates from \code{\link[=brglmFit_expo]{"brglmFit_expo"}} objects}
\usage{
\method{coef}{brglmFit_expo}(object, ...)
}
\arguments{
\item{object}{an object for which the extraction of model coefficients is
    meaningful.}

\item{...}{other arguments.}
}
\description{
Extract estimates from \code{\link[=brglmFit_expo]{"brglmFit_expo"}} objects
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/brmultinom.R
\name{brmultinom}
\alias{brmultinom}
\title{Bias reduction for multinomial response models using the
Poisson trick.}
\usage{
brmultinom(
  formula,
  data,
  weights,
  subset,
  na.action,
  contrasts = NULL,
  ref = 1,
  model = TRUE,
  x = TRUE,
  control = list(...),
  ...
)
}
\arguments{
\item{formula}{
a formula expression as for regression models, of the form
\code{response ~ predictors}. The response should be a factor or a
matrix with K columns, which will be interpreted as counts for each of
K classes.
A log-linear model is fitted, with coefficients zero for the first
class. An offset can be included: it should be a numeric matrix with K columns
if the response is either a matrix with K columns or a factor with K >= 2
classes, or a numeric vector for a response factor with 2 levels.
See the documentation of \code{\link{formula}()} for other details.
}

\item{data}{
an optional data frame in which to interpret the variables occurring
in \code{formula}.
}

\item{weights}{
optional case weights in fitting.
}

\item{subset}{
expression saying which subset of the rows of the data should  be used
in the fit. All observations are included by default.
}

\item{na.action}{
a function to filter missing data.
}

\item{contrasts}{
a list of contrasts to be used for some or all of
the factors appearing as variables in the model formula.
}

\item{ref}{the reference category to use for multinomial
regression. Either an integer, in which case
\code{levels(response)[ref]} is used as a baseline, or a character
string. Default is 1.}

\item{model}{
  logical. If true, the model frame is saved as component \code{model}
  of the returned object.
}

\item{x}{should the model matrix be included with in the result
(default is \code{TRUE}).}

\item{control}{a list of parameters for controlling the fitting
process. See \code{\link[=brglmControl]{brglmControl()}} for details.}

\item{...}{arguments to be used to form the default \code{control}
argument if it is not supplied directly.}
}
\description{
\code{\link[=brmultinom]{brmultinom()}} is a wrapper of \code{\link[=brglmFit]{brglmFit()}} that fits multinomial
regression models using implicit and explicit bias reduction
methods. See Kosmidis & Firth (2011) for details.
}
\details{
The models \code{\link[=brmultinom]{brmultinom()}} handles are also known as
baseline-category logit models (see, Agresti, 2002, Section 7.1),
because they model the log-odds of every category against a
baseline category. The user can control which baseline (or
reference) category is used via the \code{ref}. By default
\code{\link[=brmultinom]{brmultinom()}} uses the first category as reference.

The maximum likelihood estimates for the parameters of
baseline-category logit models have infinite components with
positive probability, which can result in problems in their
estimation and the use of inferential procedures (e.g. Wad
tests). Albert and Andreson (1984) have categorized the possible
data patterns for such models into the exclusive and exhaustive
categories of complete separation, quasi-complete separation and
overlap, and showed that infinite maximum likelihood estimates
result when complete or quasi-complete separation occurs.

The adjusted score approaches to bias reduction that \code{\link[=brmultinom]{brmultinom()}}
implements for \code{type = "AS_mean"} and \code{type = "AS_median"} are
alternatives to maximum likelihood that result in estimates with
smaller asymptotic mean and median bias, respectively, that are
also \emph{always} finite, even in cases of complete or quasi-complete
separation.

\code{\link[=brmultinom]{brmultinom()}} is a wrapper of \code{\link[=brglmFit]{brglmFit()}} that fits multinomial
logit regression models through the 'Poisson trick' (see, for
example, Palmgren, 1981; Kosmidis & Firth, 2011).

The implementation relies on the construction of an extended model
matrix for the log-linear model and constraints on the sums of the
Poisson means. Specifically, a log-linear model is fitted on a
\href{https://en.wikipedia.org/wiki/Kronecker_product}{Kronecker product} of the
original model matrix \code{X} implied by the formula, augmented by
\code{nrow(X)} dummy variables.

The extended model matrix is sparse, and the
\href{https://cran.r-project.org/package=Matrix}{\pkg{Matrix}} package is
used for its effective storage.

While \code{\link[=brmultinom]{brmultinom()}} can be used for analyses using multinomial
regression models, the current implementation is more of a proof of
concept and is not expected to scale well with either of \code{nrow(X)},
\code{ncol(X)} or the number of levels in the categorical response.
}
\examples{
## The housing data analysis from ?MASS::housing

data("housing", package = "MASS")

# Maximum likelihood using nnet::multinom
houseML1nnet <- nnet::multinom(Sat ~ Infl + Type + Cont, weights = Freq,
                               data = housing)
# Maximum likelihood using brmultinom with baseline category 'Low'
houseML1 <- brmultinom(Sat ~ Infl + Type + Cont, weights = Freq,
                       data = housing, type = "ML", ref = 1)
# The estimates are numerically the same as houseML0
all.equal(coef(houseML1nnet), coef(houseML1), tolerance = 1e-04)

# Maximum likelihood using brmultinom with 'High' as baseline
houseML3 <- brmultinom(Sat ~ Infl + Type + Cont, weights = Freq,
                      data = housing, type = "ML", ref = 3)
# The fitted values are the same as houseML1
all.equal(fitted(houseML3), fitted(houseML1), tolerance = 1e-10)

# Bias reduction
houseBR3 <- update(houseML3, type = "AS_mean")
# Bias correction
houseBC3 <- update(houseML3, type = "correction")

## Reproducing Bull et al. (2002, Table 3)
data("hepatitis", package = "brglm2")

# Construct a variable with the multinomial categories according to
# the HCV and nonABC columns
hepat <- hepatitis
hepat$type <- with(hepat, factor(1 - HCV * nonABC + HCV + 2 * nonABC))
hepat$type <- factor(hepat$type, labels = c("noDisease", "C", "nonABC"))
contrasts(hepat$type) <- contr.treatment(3, base = 1)

# Maximum likelihood estimation fails to converge because some estimates are infinite
hepML <- brmultinom(type ~ group * time, data = hepat, weights = counts, type = "ML", slowit = 0.1)

# Mean bias reduction returns finite estimates
hep_meanBR <- brmultinom(type ~ group * time, data = hepat, weights = counts, type = "AS_mean")
# The estimates in Bull et al. (2002, Table 3, DOI: 10.1016/S0167-9473(01)00048-2)
coef(hep_meanBR)

# Median bias reduction also returns finite estimates, which are a bit larger in absolute value
hep_medianBR <- brmultinom(type ~ group * time, data = hepat, weights = counts, type = "AS_median")
coef(hep_medianBR)

}
\references{
Kosmidis I, Kenne Pagui E C, Sartori N (2020). Mean and median bias
reduction in generalized linear models. \emph{Statistics and Computing},
\strong{30}, 43-59. \doi{10.1007/s11222-019-09860-6}.

Agresti A (2002). \emph{Categorical data analysis} (2nd edition). Wiley
Series in Probability and Statistics. Wiley.

Albert A, Anderson J A (1984). On the Existence of Maximum
Likelihood Estimates in Logistic Regression Models. \emph{Biometrika},
\strong{71} 1--10. \doi{10.2307/2336390}.

Kosmidis I, Firth D (2011). Multinomial logit bias reduction
via the Poisson log-linear model. \emph{Biometrika}, \strong{98}, 755-759.
\doi{10.1093/biomet/asr026}.

Palmgren, J (1981). The Fisher Information Matrix for Log Linear
Models Arguing Conditionally on Observed Explanatory
Variables. \emph{Biometrika}, \strong{68}, 563-566.
\doi{10.1093/biomet/68.2.563}.
}
\seealso{
\code{\link[nnet:multinom]{nnet::multinom()}}, \code{\link[=bracl]{bracl()}} for adjacent category logit models for ordinal responses
}
\author{
Ioannis Kosmidis \verb{[aut, cre]} \email{ioannis.kosmidis@warwick.ac.uk}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/solve_se.R
\name{solve_se}
\alias{solve_se}
\title{Solve the MDYPL state evolution equations with or without
intercept, with signal strength or contaminated signal strength}
\usage{
solve_se(
  kappa,
  ss,
  alpha,
  intercept = NULL,
  start,
  corrupted = FALSE,
  gh = NULL,
  prox_tol = 1e-10,
  transform = TRUE,
  init_method = "Nelder-Mead",
  init_iter = 50,
  ...
)
}
\arguments{
\item{kappa}{asymptotic ratio of columns/rows of the design
matrix. \code{kappa} should be in \verb{(0, 1)}.}

\item{ss}{square root of signal strength or of corrupted signal
strength, depending on whether \code{corrupted = TRUE} or not. See
Details.}

\item{alpha}{the shrinkage parameter of the MDYPL
estimator. \code{alpha} should be in \verb{(0, 1)}.}

\item{intercept}{if \code{NULL} (default) then the MDYPL state evolution
equations for the model with no intercept parameter are
solved. If a real then the equations for the models with
intercept parameter equal to \code{intercept} are solved. See
Details.}

\item{start}{a vector with starting values for \code{mu}, \code{b},\code{sigma}
(and \code{iota} if \code{intercept} is numeric).}

\item{corrupted}{if \code{FALSE} (default) then \code{ss} is the square root
of the signal strength and \code{intercept}, if numeric, is the
oracle intercept value. If \code{TRUE}, then \code{ss} is the square root
of the corrupted signal strength, and \code{intercept}, if numeric,
is the limit of the estimator computed by \code{\link[=mdyplFit]{mdyplFit()}} with
shrinkage parameter \code{alpha}. See Details.}

\item{gh}{A list with the Gauss-Hermite quadrature nodes and
weights, as returned from \code{statmod::gauss.quad()} with \code{kind = "hermite"}. Default is \code{NULL}, in which case \code{gh} is set to
\code{statmod::gauss.quad(200, kind = "hermite")}.}

\item{prox_tol}{tolerance for the computation of the proximal
operator; default is \code{1e-10}.}

\item{transform}{if \code{TRUE} (default), the optimization is with
respect to \code{log(mu)}, \code{log(b)},\code{log(sigma)}, (and \code{iota} if
\code{intercept} is numeric). If \code{FALSE}, then it is over \code{mu}, \code{b},
\code{sigma} (and \code{iota} if \code{intercept} is numeric). The solution is
returned in terms of the latter set, regardless of how
optimization took place.}

\item{init_method}{The method to be passed to \code{\link[=optim]{optim()}}. Default is
\code{"Nelder-Mead"}.}

\item{init_iter}{how many iterations of \code{\link[=optim]{optim()}} should we make to
get starting values for \code{\link[nleqslv:nleqslv]{nleqslv::nleqslv()}}? Default is \code{50},
but can also be \code{0} in which case \code{start} is directly passed to
\code{nleqslv:nleqslv()}. \code{init_iter = "only"} results in only
\code{\link[=optim]{optim()}} being used. See Details.}

\item{...}{further arguments to be passed to \code{\link[nleqslv:nleqslv]{nleqslv::nleqslv()}},
unless \code{init_iter = "only"}, in which case \code{...} is further
arguments to be passed to \code{\link[=optim]{optim()}}.}
}
\value{
If \code{intercept = NULL}, a vector with the values of \code{mu},
\code{b},\code{sigma}. Otherwise, a vector with the values of \code{mu},
\code{b},\code{sigma}, and \code{iota}, if \code{corrupted = FALSE}, or the value of
the intercept otherwise. The vector has attributes the state
evolution functions at the solution (\code{"funcs"}), the number of
iterations used by the last optimization method (\code{"iter"}), any
messages from the last optimization method (\code{"message"}), and
information on the optimization methods used
(\code{"optimization-chain"}).
}
\description{
Solve the MDYPL state evolution equations with or without
intercept, with signal strength or contaminated signal strength
}
\details{
\code{init_iter} iterations of \code{\link[=optim]{optim()}} with \code{method = init_method} are
used towards minimizing \code{sum(se)^2}, where \code{se} is a vector of the
state evolution functions. The solution is then passed to
\code{nleqslv::nleqslv()} for a more aggressive iteration. The state
evolution equations are given in expressions (8) (model without
intercept) and expression (15) (model with intercept) in Sterzinger
& Kosmidis (2024).

If \code{corrupted = FALSE} (default), then \code{ss} is the square root of
the signal strength, which is the limit \eqn{\gamma^2} of
\eqn{var(X \beta)}. If \code{corrupted = TRUE}, then \code{ss} is the square
root of the corrupted signal strength which is the limit
\eqn{\nu^2} of \eqn{var(X hat(beta)(\alpha))}, where
\eqn{hat(\beta)(\alpha)} is the maximum Diaconis-Ylvisaker prior
penalized likelihood (MDYPL) estimator as computed by \code{\link[=mdyplFit]{mdyplFit()}}
with shrinkage parameter \eqn{alpha}.

If \code{intercept = NULL}, then the state evolution equations are
solved for the model without intercept. If \code{intercept} is a real
number, then the state evolution equations for the model with
intercept are solved (i.e. with predictor \eqn{\eta_i = \theta +
x_i^T \beta}). In that case, what \code{intercept} represents depends on
the value of \code{corrupted}. If \code{corrupted = FALSE}, \code{intercept}
represents the oracle value of \eqn{\theta}, otherwise it represents
the limit \code{iota} of the MDYPL estimator of \eqn{\theta} as computed by
\code{\link[=mdyplFit]{mdyplFit()}} with shrinkage parameter \code{alpha}.

Note that \code{start} is always for \code{mu}, \code{b},\code{sigma}, as is the
result, regardless whether \code{transform = TRUE} or
not. Transformations during optimization are done internally.
}
\examples{

## Reproducing Table 13 of Zhao et al. (2022, DOI: 10.3150/21-BEJ1401)
\donttest{

thetas <- c(0, 0.5, 1, 2, 2.5)
gamma0 <- 5
pars3 <- matrix(NA, length(thetas), 3)
pars4 <- matrix(NA, length(thetas), 4)
colnames(pars4) <- c("I_mu", "I_b", "I_sigma", "I_iota")
colnames(pars3) <- c("II_mu", "II_b", "II_sigma")
for (i in seq_along(thetas)) {
    start3 <- c(0.5, 1, 1)
    pars3[i, ] <- solve_se(kappa = 0.2, ss = sqrt(5 + thetas[i]^2),
                           alpha = 1, start = start3, init_iter = 0)
    start4 <- c(pars3[i, ], thetas[i])
    pars4[i, ] <- solve_se(kappa = 0.2, ss = sqrt(5), intercept = thetas[i],
                           alpha = 1, start = start4, init_iter = 0)
}

cbind(pars3, pars4)

}
}
\references{
Zhao Q, Sur P, Cand\`es E J (2022). The asymptotic distribution of
the MLE in high-dimensional logistic models: Arbitrary
covariance. \emph{Bernoulli}, \strong{28}, 1835–1861. \doi{10.3150/21-BEJ1401}.

Sterzinger P, Kosmidis I (2024). Diaconis-Ylvisaker prior
penalized likelihood for \eqn{p/n \to \kappa \in (0,1)} logistic
regression. \emph{arXiv}:2311.07419v2, \url{https://arxiv.org/abs/2311.07419}.
}
\author{
Ioannis Kosmidis \verb{[aut, cre]} \email{ioannis.kosmidis@warwick.ac.uk}, Federico Boiocchi \verb{[ctb]} \email{federico.boiocchi@gmail.com}, Philipp Sterzinger \verb{[ctb, earlier Julia code by]} \email{P.Sterzinger@lse.ac.uk}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mdyplFit.R, R/zzz_conventions.R
\name{mdyplFit}
\alias{mdyplFit}
\alias{mdypl_fit}
\title{Fitting function for \code{\link[=glm]{glm()}} for maximum Diaconis-Ylvisaker prior
penalized likelihood estimation of logistic regression models}
\usage{
mdyplFit(
  x,
  y,
  weights = rep(1, nobs),
  start = NULL,
  etastart = NULL,
  mustart = NULL,
  offset = rep(0, nobs),
  family = binomial(),
  control = list(),
  intercept = TRUE,
  singular.ok = TRUE
)

mdypl_fit(
  x,
  y,
  weights = rep(1, nobs),
  start = NULL,
  etastart = NULL,
  mustart = NULL,
  offset = rep(0, nobs),
  family = binomial(),
  control = list(),
  intercept = TRUE,
  singular.ok = TRUE
)
}
\arguments{
\item{x}{a design matrix of dimension \code{n * p}.}

\item{y}{a vector of observations of length \code{n}.}

\item{weights}{an optional vector of \sQuote{prior weights} to be used
    in the fitting process.  Should be \code{NULL} or a numeric vector.}

\item{start}{starting values for the parameters in the linear predictor.}

\item{etastart}{starting values for the linear predictor.}

\item{mustart}{starting values for the vector of means.}

\item{offset}{this can be used to specify an \emph{a priori} known
    component to be included in the linear predictor during fitting.
    This should be \code{NULL} or a numeric vector of length equal to
    the number of cases.  One or more \code{\link[stats]{offset}} terms can be
    included in the formula instead or as well, and if more than one is
    specified their sum is used.  See \code{\link[stats]{model.offset}}.}

\item{family}{a description of the error distribution and link
    function to be used in the model.  For \code{glm} this can be a
    character string naming a family function, a family function or the
    result of a call to a family function.  For \code{glm.fit} only the
    third option is supported.  (See \code{\link[stats]{family}} for details of
    family functions.)}

\item{control}{a list of parameters controlling the fitting
process. See \code{\link[=mdyplControl]{mdyplControl()}} for details.}

\item{intercept}{logical. Should an intercept be included in the
    \emph{null} model?}

\item{singular.ok}{logical; if \code{FALSE} a singular fit is an
    error.}
}
\value{
An object inheriting from \code{\link[=mdyplFit]{"mdyplFit"}} object, which
is a list having the same elements to the list that
\code{\link[stats:glm]{stats::glm.fit()}} returns, with a few extra arguments.
}
\description{
\code{\link[=mdyplFit]{mdyplFit()}} is a fitting method for \code{\link[=glm]{glm()}} that fits logistic
regression models using maximum Diaconis-Ylvisaker prior penalized
likelihood estimation.
}
\details{
\code{\link[=mdyplFit]{mdyplFit()}} uses \code{\link[stats:glm]{stats::glm.fit()}} to fit a logistic regression
model on responses \code{alpha * y + (1 - alpha) / 2}, where \code{y} are the
original binomial responses scaled by the binomial totals. This is
equivalent to penalizing the likelihood by the Diaconis-Ylvisaker
prior with shrinkage parameter \eqn{\alpha} and regression parameters
set to zero. See Rigon & Aliverti (2023) and Sterzinger & Kosmidis
(2024).

By default, \code{alpha = n / (p + n)} is used, where \code{n} is the sum of
the binomial totals. Alternative values of \code{alpha} can be passed to
the \code{control} argument; see \code{\link[=mdyplControl]{mdyplControl()}} for setting up the
list passed to \code{control}. If \code{alpha = 1} then \code{\link[=mdyplFit]{mdyplFit()}} will
simply do maximum likelihood estimation.

Note that \code{null.deviance}, \code{deviance} and \code{aic} in the resulting
object are computed at the adjusted responses. Hence, methods such
as \link[stats:logLik]{logLik()} and \link[stats:AIC]{AIC()} use the
penalized log-likelihood. With the default \code{alpha}, the inferential
procedures based on penalized likelihood are asymptotically
equivalent to the ones that use the unpenalized likelihood when
\code{p/n} is vanishing asymptotically.

For high-dimensionality corrected estimates, standard errors and z
statistics, use the \code{\link[=summary.mdyplFit]{summary}} method for
\code{\link[=mdyplFit]{"mdyplFit"}} objects with \code{hd_correction = TRUE}.

\code{\link[=mdypl_fit]{mdypl_fit()}} is an alias to \code{\link[=mdyplFit]{mdyplFit()}}.
}
\examples{

data("lizards", package = "brglm2")
liz_fm <- cbind(grahami, opalinus) ~ height + diameter + light + time
## ML fit = MDYPL fit with `alpha = 1`
liz_ml <- glm(liz_fm, family = binomial(), data = lizards,
              method = "mdyplFit", alpha = 1)
liz_ml0 <- glm(liz_fm, family = binomial(), data = lizards)

## liz_ml is the same fit as liz_ml0
summ_liz_ml <- summary(liz_ml)
summ_liz_ml0 <- summary(liz_ml0)
all.equal(coef(summ_liz_ml), coef(summ_liz_ml0))

## MDYPL fit with default `alpha` (see `?mdyplControl`)
liz_fm <- cbind(grahami, opalinus) ~ height + diameter + light + time
liz_mdypl <- glm(liz_ml, family = binomial(), data = lizards,
                 method = "mdyplFit")

## Comparing outputs from ML and MDYPL, with and without
## high-dimensionality corrections.
summary(liz_mdypl)
summary(liz_mdypl, hd_correction = TRUE)
summ_liz_ml
summary(liz_ml, hd_correction = TRUE)
## Not much difference in fits here as this is a low dimensional
## problem with dimensionality constant
(liz_ml$rank - 1) / sum(weights(liz_ml))



## The case study in Section 8 of Sterzinger and
## Kosmidis (2024)
data("MultipleFeatures", package = "brglm2")

## Center the fou.* and kar.* features
vars <- grep("fou|kar", names(MultipleFeatures), value = TRUE)
train_id <- which(MultipleFeatures$training)
MultipleFeatures[train_id, vars] <- scale(MultipleFeatures[train_id, vars], scale = FALSE)
## Compute the MDYPL fits
kappa <- length(vars) / sum(MultipleFeatures$training)
full_fm <- formula(paste("I(digit == 7) ~", paste(vars, collapse = " + ")))
nest_vars <- grep("fou", vars, value = TRUE)
nest_fm <- formula(paste("I(digit == 7) ~", paste(nest_vars, collapse = " + ")))
full_m <- glm(full_fm, data = MultipleFeatures, family = binomial(),
              method = mdyplFit, alpha = 1 / (1 + kappa), subset = training)
nest_m <- update(full_m, nest_fm)

## With a naive penalized likelihood ratio test we get no evidence
## against the hypothesis that the model with only `fou` features
## is an as good descrition of `7` as the model with both `fou` and
## `kar` features.
plrtest(nest_m, full_m)

## With a high-dimensionality correction theres is strong evidence
## against the model with only `fou` features
plrtest(nest_m, full_m, hd_correction = TRUE)


\donttest{
## A simulated data set as in Rigon & Aliverti (2023, Section 4.3)

set.seed(123)
n <- 1000
p <- 500
gamma <- sqrt(5)
X <- matrix(rnorm(n * p, 0, 1), nrow = n, ncol = p)
betas0 <- rep(c(-1, -1/2, 0, 2, 3), each = p / 5)
betas <- gamma * betas0 / sqrt(sum(betas0^2))
probs <- plogis(drop(X \%*\% betas))
y <- rbinom(n, 1, probs)
fit_mdypl <- glm(y ~ -1 + X, family = binomial(), method = "mdyplFit")

## The default value of `alpha` is `n / (n + p)` here
identical(n / (n + p), fit_mdypl$alpha)

## Aggregate bias of MDYPL and rescaled MDYPL estimators
ag_bias <- function(estimates, beta) mean(estimates - beta)
ag_bias(coef(summary(fit_mdypl))[, "Estimate"], betas)
ag_bias(coef(summary(fit_mdypl, hd_correction = TRUE))[, "Estimate"], betas)

}
}
\references{
Sterzinger P, Kosmidis I (2024). Diaconis-Ylvisaker prior
penalized likelihood for \eqn{p/n \to \kappa \in (0,1)} logistic
regression. \emph{arXiv}:2311.07419v2, \url{https://arxiv.org/abs/2311.07419}.

Rigon T, Aliverti E (2023). Conjugate priors and bias reduction for
logistic regression models. \emph{Statistics & Probability Letters},
\strong{202}, 109901. \doi{10.1016/j.spl.2023.109901}.
}
\seealso{
\code{\link[=mdyplControl]{mdyplControl()}}, \code{\link[=summary.mdyplFit]{summary.mdyplFit()}}, \code{\link[=plrtest.mdyplFit]{plrtest.mdyplFit()}}, \code{\link[=glm]{glm()}}
}
\author{
Ioannis Kosmidis \verb{[aut, cre]} \email{ioannis.kosmidis@warwick.ac.uk}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/brmultinom.R
\name{simulate.brmultinom}
\alias{simulate.brmultinom}
\title{Method for simulating a data set from \code{\link[=brmultinom]{"brmultinom"}} and \code{\link[=bracl]{"bracl"}}
objects}
\usage{
\method{simulate}{brmultinom}(object, ...)
}
\arguments{
\item{object}{an object of class \code{\link[=brmultinom]{"brmultinom"}} or \code{\link[=bracl]{"bracl"}}.}

\item{...}{currently not used.}
}
\value{
A \code{\link[=data.frame]{"data.frame"}} with \code{object$ncat} times the rows that
\code{model.frame(object)} have and the same variables. If \code{weights} has
been specified in the call that generated \code{object}, then the
simulate frequencies will populate the weights variable. Otherwise,
the resulting \link{data.frame} will have a \code{".weights"} variable with
the simulated multinomial counts.
}
\description{
Method for simulating a data set from \code{\link[=brmultinom]{"brmultinom"}} and \code{\link[=bracl]{"bracl"}}
objects
}
\examples{

## Multinomial logistic regression
data("housing", package = "MASS")
houseML1 <- brmultinom(Sat ~ Infl + Type + Cont, weights = Freq,
                       data = housing, type = "ML", ref = 1)
simulate(houseML1)

## Adjacent-category logits
data("stemcell", package = "brglm2")
stemML1 <- bracl(research ~ religion + gender, weights = frequency,
                data = stemcell, type = "ML")

simulate(stemML1)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/brglmControl.R, R/zzz_conventions.R
\name{brglmControl}
\alias{brglmControl}
\alias{brglm_control}
\title{Auxiliary function for \code{\link[=glm]{glm()}} fitting using the \code{\link[=brglmFit]{brglmFit()}}
method.}
\usage{
brglmControl(
  epsilon = 1e-06,
  maxit = 100,
  check_aliasing = TRUE,
  trace = FALSE,
  type = c("AS_mixed", "AS_mean", "AS_median", "correction", "MPL_Jeffreys", "ML"),
  transformation = "identity",
  slowit = 1,
  response_adjustment = NULL,
  max_step_factor = 12,
  a = 1/2,
  ...
)

brglm_control(
  epsilon = 1e-06,
  maxit = 100,
  check_aliasing = TRUE,
  trace = FALSE,
  type = c("AS_mixed", "AS_mean", "AS_median", "correction", "MPL_Jeffreys", "ML"),
  transformation = "identity",
  slowit = 1,
  response_adjustment = NULL,
  max_step_factor = 12,
  a = 1/2,
  ...
)
}
\arguments{
\item{epsilon}{positive convergence tolerance epsilon. Default is
\code{1e-06}.}

\item{maxit}{integer giving the maximal number of iterations
allowed. Default is \code{100}.}

\item{check_aliasing}{logical indicating where a QR decomposition
of the model matrix should be used to check for
aliasing. Default is \code{TRUE}. See Details.}

\item{trace}{logical indicating if output should be produced for
each iteration. Default is \code{FALSE}.}

\item{type}{the type of fitting method to be used. The options are
\code{"AS_mean"} (mean-bias reducing adjusted scores), \code{"AS_median"}
(median-bias reducing adjusted scores), \code{"AS_mixed"} (bias
reduction using mixed score adjustments; default),
\code{"correction"} (asymptotic bias correction), \code{"MPL_Jeffreys"}
(maximum penalized likelihood with powers of the Jeffreys prior
as penalty) and \code{"ML"} (maximum likelihood).}

\item{transformation}{the transformation of the dispersion to be
estimated. Default is \code{"identity"}. See Details.}

\item{slowit}{a positive real used as a multiplier for the
stepsize. The smaller it is the smaller the steps are. Default
is \code{1}.}

\item{response_adjustment}{a (small) positive constant or a vector
of such. Default is \code{NULL}. See Details.}

\item{max_step_factor}{the maximum number of step halving steps to
consider. Default is \code{12}.}

\item{a}{power of the Jeffreys prior penalty. See Details.}

\item{...}{further arguments passed to \code{\link[=brglmControl]{brglmControl()}}. Currently
ignored in the output.}
}
\value{
a list with components named as the arguments, including
symbolic expressions for the dispersion transformation
(\code{Trans}) and its inverse (\code{inverseTrans})
}
\description{
Typically only used internally by \code{\link[=brglmFit]{brglmFit()}}, but may be used to
construct a \code{control} argument.
}
\details{
\code{\link[=brglmControl]{brglmControl()}} provides default values and sanity checking for
the various constants that control the iteration and generally the
behaviour of \code{\link[=brglmFit]{brglmFit()}}.

When \code{trace = TRUE}, calls to \code{\link[=cat]{cat()}} produce the output for each
iteration.  Hence, \verb{options(digits = *)} can be used to increase
the precision.

When \code{check_aliasing = TRUE} (default), a QR decomposition of the
model matrix is computed to check for aliasing. If the model matrix
is known to be of full rank, then \code{check_aliasing = FALSE} avoids
the extra computational overhead of an additional QR decomposition,
which can be substantial for large model matrices. However, setting
\code{check_aliasing = FALSE} tells \code{\link[=brglmFit]{brglmFit()}} that the model matrix
is full rank, and hard to trace back errors will result if it is
rank deficient.

\code{transformation} sets the transformation of the dispersion
parameter for which the bias reduced estimates are computed. Can be
one of \code{"identity"}, \code{"sqrt"}, \code{"inverse"}, \code{"log"} and
\code{"inverseSqrt"}. Custom transformations are accommodated by
supplying a list of two expressions (transformation and inverse
transformation). See the examples for more details.

The value of \code{response_adjustment} is only relevant if \code{\link[=brglmFit]{brglmFit()}}
is called with \code{start = NULL}, and \code{family} is \code{\link[=binomial]{binomial()}} or
\code{\link[=poisson]{poisson()}}. For those models, an initial maximum likelihood fit is
obtained on adjusted data to provide starting values for the
iteration in \code{\link[=brglmFit]{brglmFit()}}. The value of \code{response_adjustment}
governs how the data is adjusted. Specifically, if \code{family} is
\code{\link[=binomial]{binomial()}}, then the responses and totals are adjusted by
\code{response_adjustment} and \code{2 * response_adjustment}, respectively;
if \code{family} is \code{\link[=poisson]{poisson()}}, then the responses are adjusted by and
\code{response_adjustment}. \code{response_adjustment = NULL} (default)
is equivalent to setting it to
\code{"number of parameters" / "number of observations"}.

When \code{type = "AS_mixed"} (default), mean bias reduction is used for
the regression parameters, and median bias reduction for the
dispersion parameter, if that is not fixed. This adjustment has
been developed based on equivariance arguments (see, Kosmidis et
al, 2020, Section 4) in order to produce regression parameter
estimates that are invariant to arbitrary contrasts, and estimates
for the dispersion parameter that are invariant to arbitrary
non-linear transformations. \code{type = "AS_mixed"} and \code{type = "AS_mean"} return the same results if \code{\link[=brglmFit]{brglmFit()}} is called with
\code{family} \code{\link[=binomial]{binomial()}} or \code{\link[=poisson]{poisson()}} (i.e. families with fixed
dispersion).

When \code{type = "MPL_Jeffreys"}, \code{\link[=brglmFit]{brglmFit()}} will maximize the
penalized log-likelihood \deqn{l(\beta, \phi) + a\log \det i(\beta,
\phi)}{l(beta, phi) + a log det i(beta, phi)} where \eqn{i(\beta,
\phi)}{i(beta, phi)} is the expected information matrix about the
regression parameters \eqn{\beta} and the dispersion parameter
\eqn{\phi}. See, \code{vignette("iteration", "brglm2")} for more
information. The argument \code{a} controls the amount of penalization
and its default value is \code{a = 1/2}, corresponding to maximum
penalized likelihood using a Jeffreys-prior penalty. See, Kosmidis
& Firth (2021) for proofs and discussion about the finiteness and
shrinkage properties of the maximum penalized likelihood estimators
for binomial-response generalized linear models.

The estimates from \code{type = "AS_mean"} and \code{type = "MPL_Jeffreys"} with \code{a = 1/2} (default) are identical for
Poisson log-linear models and logistic regression models, i.e. for
binomial and Poisson regression models with canonical links. See,
Firth (1993) for details.

\code{\link[=brglm_control]{brglm_control()}} is an alias to \code{\link[=brglmControl]{brglmControl()}}.
}
\examples{

data("coalition", package = "brglm2")
## The maximum likelihood fit with log link
coalitionML <- glm(duration ~ fract + numst2, family = Gamma, data = coalition)

## Bias reduced estimation of the dispersion parameter
coalitionBRi <- glm(duration ~ fract + numst2, family = Gamma, data = coalition,
                    method = "brglmFit")
coef(coalitionBRi, model = "dispersion")

## Bias reduced estimation of log(dispersion)
coalitionBRl <- glm(duration ~ fract + numst2, family = Gamma, data = coalition,
                    method = "brglmFit", transformation = "log")
coef(coalitionBRl, model = "dispersion")

## Just for illustration: Bias reduced estimation of dispersion^0.25
my_transformation <- list(expression(dispersion^0.25), expression(transformed_dispersion^4))
coalitionBRc <- update(coalitionBRi, transformation = my_transformation)
coef(coalitionBRc, model = "dispersion")

}
\references{
Kosmidis I, Firth D (2021). Jeffreys-prior penalty, finiteness
and shrinkage in binomial-response generalized linear
models. \emph{Biometrika}, \strong{108}, 71-82. \doi{10.1093/biomet/asaa052}.

Kosmidis I, Kenne Pagui E C, Sartori N (2020). Mean and median bias
reduction in generalized linear models. \emph{Statistics and Computing},
\strong{30}, 43-59. \doi{10.1007/s11222-019-09860-6}.

Firth D (1993). Bias reduction of maximum likelihood estimates.
Biometrika, \strong{80}, 27-38. \doi{10.2307/2336755}.
}
\seealso{
\code{\link[=brglm_fit]{brglm_fit()}} and \code{\link[=glm.fit]{glm.fit()}}
}
\author{
Ioannis Kosmidis \verb{[aut, cre]} \email{ioannis.kosmidis@warwick.ac.uk}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{lizards}
\alias{lizards}
\title{Habitat preferences of lizards}
\format{
A data frame with 23 rows and 6 columns:
\itemize{
\item \code{grahami}. count of grahami lizards
\item \code{opalinus}. count of opalinus lizards
\item \code{height}. a factor with levels \verb{<5ft}, \verb{>=5ft}
\item \code{diameter}. a factor with levels \verb{<=2in}, \verb{>2in}
\item \code{light}. a factor with levels \code{sunny}, \code{shady}
\item \code{time}. a factor with levels \code{early}, \code{midday}, \code{late}
}

The variables \code{grahami} and \code{opalinus} are counts of two lizard
species at two different perch heights, two different perch
diameters, in sun and in shade, at three times of day.
}
\source{
McCullagh P, Nelder J A (1989) \emph{Generalized Linear
Models} (2nd Edition).  London: Chapman and Hall.

Originally from

\if{html}{\out{<div class="sourceCode">}}\preformatted{Schoener T W (1970) Nonsynchronous spatial overlap of lizards
in patchy habitats.  _Ecology_ *51*, 408-418.
}\if{html}{\out{</div>}}
}
\usage{
lizards
}
\description{
Habitat preferences of lizards
}
\seealso{
\code{\link[=brglm_fit]{brglm_fit()}}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{aids}
\alias{aids}
\title{Effects of AZT in slowing the development of AIDS symptoms}
\format{
A data frame with 4 rows and 4 variables:
\itemize{
\item \code{symptomatic}: counts of veterans showing AIDS symptoms during the 3-year study
\item \code{asymptomatic}: counts of veterans not showing AIDS symptoms during the 3-year study
\item \code{race}: the race of the veterans with levels \code{"White"} and \code{"Black"}
\item \code{AZT_use}: whether the veterans received AZT immediately (\code{"Yes"})
or waited until their T cells showed severe immune weakness (\code{"No"})
}
}
\source{
The data set is analyzed in Agresti (2002, Subsection 5.4.2). Its
original source is New York Times, Feb. 15, 1991.
}
\usage{
aids
}
\description{
The data is from a 3-year study on the effects of AZT in slowing the
development of AIDS symptoms. 338 veterans whose immune systems
were beginning to falter after infection with the AIDS virus were
randomly assigned either to receive AZT immediately or to wait
until their T cells showed severe immune weakness.
}
\references{
Agresti A (2002). \emph{Categorical Data Analysis}.  Wiley Series in
Probability and Statistics. Wiley.
}
\seealso{
\code{\link[=brmultinom]{brmultinom()}}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mdyplFit.R
\name{confint.mdyplFit}
\alias{confint.mdyplFit}
\title{Method for computing confidence intervals for one or more
regression parameters in a \code{\link[=mdyplFit]{"mdyplFit"}} object}
\usage{
\method{confint}{mdyplFit}(object, parm, level = 0.95, hd_correction = FALSE, ...)
}
\arguments{
\item{object}{a fitted model object.}

\item{parm}{a specification of which parameters are to be given
    confidence intervals, either a vector of numbers or a vector of
    names.  If missing, all parameters are considered.}

\item{level}{the confidence level required.}

\item{hd_correction}{if \code{FALSE} (default), then the corresponding
quantities are computed according to standard asymptotics. If
\code{TRUE} then the high-dimensionality corrections in Sterzinger &
Kosmidis (2024) are employed to updates estimates, estimated
standard errors, z-statistics, etc. See Details.}

\item{...}{additional argument(s) for methods.}
}
\description{
Method for computing confidence intervals for one or more
regression parameters in a \code{\link[=mdyplFit]{"mdyplFit"}} object
}
\examples{

\donttest{

set.seed(123)
n <- 2000
p <- 800
set.seed(123)
betas <- c(rnorm(p / 4, mean = 7, sd = 1), rep(0, 3 * p / 4))
X <- matrix(rnorm(n * p, 0, 1/sqrt(n)), nrow = n, ncol = p)
probs <- plogis(drop(X \%*\% betas))
y <- rbinom(n, 1, probs)
fit_mdypl <- glm(y ~ -1 + X, family = binomial(), method = "mdyplFit")

wald_ci <- confint(fit_mdypl)
adj_wald_ci <- confint(fit_mdypl, hd_correction = TRUE)
ag_coverage <- function(cis, beta) mean((cis[, 1] < beta) & (cis[, 2] > beta))
ag_coverage(wald_ci, betas)
ag_coverage(adj_wald_ci, betas)

}

}
\seealso{
\code{\link[=mdyplFit]{mdyplFit()}}, \code{\link[=summary.mdyplFit]{summary.mdyplFit()}}
}
\author{
Ioannis Kosmidis \verb{[aut, cre]} \email{ioannis.kosmidis@warwick.ac.uk}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mis-link.R
\name{mis}
\alias{mis}
\title{A \code{\link[=make.link]{"link-glm"}} object for misclassified responses in binomial regression models}
\usage{
mis(link = "logit", sensitivity = 1, specificity = 1)
}
\arguments{
\item{link}{the baseline link to be used.}

\item{sensitivity}{the probability of observing a success given that a success actually took place given any covariate values.}

\item{specificity}{the probability of observing a failure given that a failure actually took place given any covariate values.}
}
\description{
\code{\link[=mis]{mis()}} is a \code{\link[=make.link]{"link-glm"}} object that specifies the link function in Neuhaus (1999, expression (8)) for handling misclassified responses in binomial regression models using maximum likelihood. A prior specification of the sensitivity and specificity is required.
}
\details{
\code{sensitivity + specificity} should be greater or equal to 1,
otherwise it is implied that the procedure producing the responses
performs worse than chance in terms of misclassification.
}
\examples{

## Define a few links with some misclassification
logit_mis <- mis(link = "logit", sensitivity = 0.9, specificity = 0.9)

lizards_f <- cbind(grahami, opalinus) ~ height + diameter + light + time

lizardsML <- glm(lizards_f, family = binomial(logit), data = lizards)

lizardsML_mis <- update(lizardsML, family = binomial(logit_mis),
                        start = coef(lizardsML))

## A notable change is coefficients is noted here compared to when
## specificity and sensitity are 1
coef(lizardsML)
coef(lizardsML_mis)

## Bias reduction is also possible
update(lizardsML_mis, method = "brglmFit", type = "AS_mean",
       start = coef(lizardsML))

update(lizardsML_mis, method = "brglmFit", type = "AS_median",
       start = coef(lizardsML))

}
\references{
Neuhaus J M (1999). Bias and efficiency loss due to misclassified
responses in binary regression. Biometrika, \strong{86}, 843-855.
\url{https://www.jstor.org/stable/2673589}.
}
\seealso{
\code{\link[=glm]{glm()}}, \code{\link[=brglm_fit]{brglm_fit()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/brmultinom.R
\name{predict.brmultinom}
\alias{predict.brmultinom}
\title{Predict method for \link{brmultinom} fits}
\usage{
\method{predict}{brmultinom}(object, newdata, type = c("class", "probs"), ...)
}
\arguments{
\item{object}{a fitted object of class inheriting from
\code{\link[=brmultinom]{"brmultinom"}}.}

\item{newdata}{optionally, a data frame in which to look for
variables with which to predict.  If omitted, the fitted linear
predictors are used.}

\item{type}{the type of prediction required. The default is
\code{"class"}, which produces predictions of the response category
at the covariate values supplied in \code{"newdata"}, selecting the
category with the largest probability; the alternative
\code{"probs"} returns all category probabilities at the covariate
values supplied in \code{newdata}.}

\item{...}{further arguments passed to or from other methods.}
}
\value{
If \code{type = "class"} a vector with the predicted response
categories; if \code{type = "probs"} a matrix of probabilities for all
response categories at \code{newdata}.
}
\description{
Obtain class and probability predictions from a fitted baseline
category logits model.
}
\details{
If \code{newdata} is omitted the predictions are based on the data used
for the fit.
}
\examples{

data("housing", package = "MASS")

# Maximum likelihood using brmultinom with baseline category 'Low'
houseML1 <- brmultinom(Sat ~ Infl + Type + Cont, weights = Freq,
                       data = housing, type = "ML", ref = 1)

# New data
newdata <- expand.grid(Infl = c("Low", "Medium"),
                       Type = c("Tower", "Atrium", "Terrace"),
                       Cont = c("Low", NA, "High"))

## Predictions
sapply(c("class", "probs"), function(what) predict(houseML1, newdata, what))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/brglmFit.R
\name{confint.brglmFit}
\alias{confint.brglmFit}
\title{Method for computing confidence intervals for one or more
regression parameters in a \code{\link[=brglmFit]{"brglmFit"}} object}
\usage{
\method{confint}{brglmFit}(object, parm, level = 0.95, ...)
}
\arguments{
\item{object}{a fitted model object.}

\item{parm}{a specification of which parameters are to be given
    confidence intervals, either a vector of numbers or a vector of
    names.  If missing, all parameters are considered.}

\item{level}{the confidence level required.}

\item{...}{additional argument(s) for methods.}
}
\description{
Method for computing confidence intervals for one or more
regression parameters in a \code{\link[=brglmFit]{"brglmFit"}} object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{coalition}
\alias{coalition}
\title{Coalition data}
\format{
A data frame with 314 rows and the 7 variables "duration",
"ciep12", "invest", "fract", "polar", "numst2", and "crisis".  For
variable descriptions, please refer to King et al (1990).
}
\usage{
coalition
}
\description{
This data set contains survival data on government coalitions in
parliamentary democracies (Belgium, Canada, Denmark, Finland,
France, Iceland, Ireland, Israel, Italy, Netherlands, Norway,
Portugal, Spain, Sweden, and the United Kingdom) for the period
1945-1987.  For parsimony, country indicator variables are omitted
in the sample data.
}
\note{
Data is as it is provided by the
\href{https://cran.r-project.org/package=Zelig}{\pkg{Zeilig}} R package.
}
\references{
King G, Alt J E, Burns N E, Laver M. (1990). A Unified
Model of Cabinet Dissolution in Parliamentary
Democracies. \emph{American Journal of Political Science}, \strong{34},
846-870. \doi{10.2307/2111401}.

King G, Alt J E, Burns N E, Laver M. ICPSR
Publication Related Archive, 1115.
}
\seealso{
\code{\link[=brglm_fit]{brglm_fit()}}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/brglmFit.R, R/zzz_conventions.R
\name{brglmFit}
\alias{brglmFit}
\alias{brglm_fit}
\title{Fitting function for \code{\link[=glm]{glm()}} for reduced-bias estimation and
inference}
\usage{
brglmFit(
  x,
  y,
  weights = rep(1, nobs),
  start = NULL,
  etastart = NULL,
  mustart = NULL,
  offset = rep(0, nobs),
  family = gaussian(),
  control = list(),
  intercept = TRUE,
  fixed_totals = NULL,
  singular.ok = TRUE
)

brglm_fit(
  x,
  y,
  weights = rep(1, nobs),
  start = NULL,
  etastart = NULL,
  mustart = NULL,
  offset = rep(0, nobs),
  family = gaussian(),
  control = list(),
  intercept = TRUE,
  fixed_totals = NULL,
  singular.ok = TRUE
)
}
\arguments{
\item{x}{a design matrix of dimension \code{n * p}.}

\item{y}{a vector of observations of length \code{n}.}

\item{weights}{an optional vector of \sQuote{prior weights} to be used
    in the fitting process.  Should be \code{NULL} or a numeric vector.}

\item{start}{starting values for the parameters in the linear
predictor. If \code{NULL} (default) then the maximum likelihood
estimates are calculated and used as starting values.}

\item{etastart}{applied only when start is not \code{NULL}. Starting
values for the linear predictor to be passed to
\code{\link[=glm.fit]{glm.fit()}} when computing starting values using maximum
likelihood.}

\item{mustart}{applied only when start is not \code{NULL}. Starting
values for the vector of means to be passed to
\code{\link[=glm.fit]{glm.fit()}} when computing starting values using maximum
likelihood.}

\item{offset}{this can be used to specify an \emph{a priori} known
    component to be included in the linear predictor during fitting.
    This should be \code{NULL} or a numeric vector of length equal to
    the number of cases.  One or more \code{\link[stats]{offset}} terms can be
    included in the formula instead or as well, and if more than one is
    specified their sum is used.  See \code{\link[stats]{model.offset}}.}

\item{family}{a description of the error distribution and link
    function to be used in the model.  For \code{glm} this can be a
    character string naming a family function, a family function or the
    result of a call to a family function.  For \code{glm.fit} only the
    third option is supported.  (See \code{\link[stats]{family}} for details of
    family functions.)}

\item{control}{a list of parameters controlling the fitting
process. See \code{\link[=brglmControl]{brglmControl()}} for details.}

\item{intercept}{logical. Should an intercept be included in the
    \emph{null} model?}

\item{fixed_totals}{effective only when \code{family} is
\code{\link[=poisson]{poisson()}}. Either \code{NULL} (no effect) or a vector that
indicates which counts must be treated as a group. See Details
for more information and \code{\link[=brmultinom]{brmultinom()}}.}

\item{singular.ok}{logical. If \code{FALSE}, a singular model is an
error.}
}
\value{
An object inheriting from \code{\link[=brglmFit]{"brglmFit"}} object, which
is a list having the same elements to the list that
\code{\link[stats:glm]{stats::glm.fit()}} returns, with a few extra arguments.
}
\description{
\code{\link[=brglmFit]{brglmFit()}} is a fitting method for \code{\link[=glm]{glm()}} that fits generalized
linear models using implicit and explicit bias reduction methods
(Kosmidis, 2014), and other penalized maximum likelihood
methods. Currently supported methods include the mean bias-reducing
adjusted scores approach in Firth (1993) and Kosmidis & Firth
(2009), the median bias-reduction adjusted scores approach in Kenne
Pagui et al. (2017), the correction of the asymptotic bias in
Cordeiro & McCullagh (1991), the mixed bias-reduction adjusted
scores approach in Kosmidis et al (2020), maximum penalized
likelihood with powers of the Jeffreys prior as penalty, and
maximum likelihood. Estimation is performed using a quasi Fisher
scoring iteration (see \code{vignette("iteration", "brglm2")}, which, in
the case of mean-bias reduction, resembles an iterative correction
of the asymptotic bias of the Fisher scoring iterates.
}
\details{
A detailed description of the supported adjustments and the quasi
Fisher scoring iteration is given in the iteration vignette (see,
\code{vignette("iteration", "brglm2")} or Kosmidis et al, 2020).  A
shorter description of the quasi Fisher scoring iteration is also
given in one of the vignettes of the \emph{enrichwith} R package (see,
\url{https://cran.r-project.org/package=enrichwith/vignettes/bias.html}).
Kosmidis and Firth (2010) describe a parallel quasi Newton-Raphson
iteration with the same stationary point.

In the special case of generalized linear models for binomial,
Poisson and multinomial responses, the adjusted score equation
approaches for \code{type = "AS_mixed"}, \code{type = "AS_mean"}, and \code{type = "AS_median"} (see below for what methods each \code{type} corresponds)
return estimates with improved frequentist properties, that are
also always finite, even in cases where the maximum likelihood
estimates are infinite (e.g. complete and quasi-complete separation
in multinomial regression). See, Kosmidis and Firth (2021) for a
proof for binomial-response GLMs with Jeffreys-prior penalties to
the log-likelihood, which is equivalent to mean bias reduction for
logistic regression. See, also,
\code{\link[detectseparation:detect_separation]{detectseparation::detect_separation()}} and
\code{\link[detectseparation:check_infinite_estimates]{detectseparation::check_infinite_estimates()}} for pre-fit and
post-fit methods for the detection of infinite estimates in
binomial response generalized linear models.

The type of score adjustment to be used is specified through the
\code{type} argument (see \code{\link[=brglmControl]{brglmControl()}} for details). The available
options are
\itemize{
\item \code{type = "AS_mixed"}: the mixed bias-reducing score adjustments in
Kosmidis et al (2020) that result in mean bias reduction for the
regression parameters and median bias reduction for the dispersion
parameter, if any; default.
\item \code{type = "AS_mean"}: the mean bias-reducing score adjustments in
Firth, 1993 and Kosmidis & Firth, 2009. \code{type = "AS_mixed"} and
\code{type = "AS_mean"} will return the same results when \code{family} is
\code{\link[=binomial]{binomial()}} or \code{\link[=poisson]{poisson()}}, i.e. when the dispersion is fixed
\item \code{type = "AS_median"}: the median bias-reducing score
adjustments in Kenne Pagui et al. (2017)
\item \code{type = "MPL_Jeffreys"}: maximum penalized likelihood
with powers of the Jeffreys prior as penalty.
\item \code{type = "ML"}: maximum likelihood.
\item \code{type = "correction"}: asymptotic bias correction, as in
Cordeiro & McCullagh (1991).
}

The null deviance is evaluated based on the fitted values using the
method specified by the \code{type} argument (see \code{\link[=brglmControl]{brglmControl()}}).

The \code{family} argument of the current version of \code{\link[=brglmFit]{brglmFit()}} can
accept any combination of \code{\link[=family]{"family"}} objects and link functions,
including families with user-specified link functions, \code{\link[=mis]{mis()}}
links, and \code{\link[=power]{power()}} links, but excluding \code{\link[=quasi]{quasi()}},
\code{\link[=quasipoisson]{quasipoisson()}} and \code{\link[=quasibinomial]{quasibinomial()}} families.

The description of \code{method} argument and the \verb{Fitting functions}
section in \code{\link[=glm]{glm()}} gives information on supplying fitting
methods to \code{\link[=glm]{glm()}}.

\code{fixed_totals} specifies groups of observations for which the sum
of the means of a Poisson model will be held fixed to the observed
count for each group. This argument is used internally in
\code{\link[=brmultinom]{brmultinom()}} and \code{\link[=bracl]{bracl()}} for baseline-category logit models and
adjacent category logit models, respectively.

\code{\link[=brglm_fit]{brglm_fit()}} is an alias to \code{\link[=brglmFit]{brglmFit()}}.
}
\examples{
## The lizards example from ?brglm::brglm
data("lizards", package = "brglm2")
# Fit the model using maximum likelihood
lizardsML <- glm(cbind(grahami, opalinus) ~ height + diameter +
                 light + time, family = binomial(logit), data = lizards,
                 method = "glm.fit")
# Mean bias-reduced fit:
lizardsBR_mean <- glm(cbind(grahami, opalinus) ~ height + diameter +
                      light + time, family = binomial(logit), data = lizards,
                      method = "brglmFit")
# Median bias-reduced fit:
lizardsBR_median <- glm(cbind(grahami, opalinus) ~ height + diameter +
                        light + time, family = binomial(logit), data = lizards,
                        method = "brglmFit", type = "AS_median")
summary(lizardsML)
summary(lizardsBR_median)
summary(lizardsBR_mean)

# Maximum penalized likelihood with Jeffreys prior penatly
lizards_Jeffreys <- glm(cbind(grahami, opalinus) ~ height + diameter +
                        light + time, family = binomial(logit), data = lizards,
                        method = "brglmFit", type = "MPL_Jeffreys")
# lizards_Jeffreys is the same fit as lizardsBR_mean (see Firth, 1993)
all.equal(coef(lizardsBR_mean), coef(lizards_Jeffreys))

# Maximum penalized likelihood with powers of the Jeffreys prior as
# penalty. See Kosmidis & Firth (2021) for the finiteness and
# shrinkage properties of the maximum penalized likelihood
# estimators in binomial response models
\donttest{
a <- seq(0, 20, 0.5)
coefs <- sapply(a, function(a) {
      out <- glm(cbind(grahami, opalinus) ~ height + diameter +
             light + time, family = binomial(logit), data = lizards,
             method = "brglmFit", type = "MPL_Jeffreys", a = a)
      coef(out)
})
# Illustration of shrinkage as a grows
matplot(a, t(coefs), type = "l", col = 1, lty = 1)
abline(0, 0, col = "grey")
}

\donttest{
## Another example from
## King, Gary, James E. Alt, Nancy Elizabeth Burns and Michael Laver
## (1990).  "A Unified Model of Cabinet Dissolution in Parliamentary
## Democracies", _American Journal of Political Science_, **34**, 846-870

data("coalition", package = "brglm2")
# The maximum likelihood fit with log link
coalitionML <- glm(duration ~ fract + numst2, family = Gamma, data = coalition)
# The mean bias-reduced fit
coalitionBR_mean <- update(coalitionML, method = "brglmFit")
# The bias-corrected fit
coalitionBC <- update(coalitionML, method = "brglmFit", type = "correction")
# The median bias-corrected fit
coalitionBR_median <- update(coalitionML, method = "brglmFit", type = "AS_median")
}

\donttest{
## An example with offsets from Venables & Ripley (2002, p.189)
data("anorexia", package = "MASS")

anorexML <- glm(Postwt ~ Prewt + Treat + offset(Prewt),
                family = gaussian, data = anorexia)
anorexBC <- update(anorexML, method = "brglmFit", type = "correction")
anorexBR_mean <- update(anorexML, method = "brglmFit")
anorexBR_median <- update(anorexML, method = "brglmFit", type = "AS_median")

# All methods return the same estimates for the regression
# parameters because the maximum likelihood estimator is normally
# distributed around the `true` value under the model (hence, both
# mean and component-wise median unbiased). The Wald tests for
# anorexBC and anorexBR_mean differ from anorexML because the
# bias-reduced estimator of the dispersion is the unbiased, by
# degree of freedom adjustment (divide by n - p), estimator of the
# residual variance. The Wald tests from anorexBR_median are based
# on the median bias-reduced estimator of the dispersion that
# results from a different adjustment of the degrees of freedom
# (divide by n - p - 2/3)
summary(anorexML)
summary(anorexBC)
summary(anorexBR_mean)
summary(anorexBR_median)
}

## endometrial data from Heinze & Schemper (2002) (see ?endometrial)
data("endometrial", package = "brglm2")
endometrialML <- glm(HG ~ NV + PI + EH, data = endometrial,
                     family = binomial("probit"))
endometrialBR_mean <- update(endometrialML, method = "brglmFit",
                             type = "AS_mean")
endometrialBC <- update(endometrialML, method = "brglmFit",
                        type = "correction")
endometrialBR_median <- update(endometrialML, method = "brglmFit",
                               type = "AS_median")
summary(endometrialML)
summary(endometrialBC)
summary(endometrialBR_mean)
summary(endometrialBR_median)

}
\references{
Kosmidis I, Firth D (2021). Jeffreys-prior penalty, finiteness
and shrinkage in binomial-response generalized linear
models. \emph{Biometrika}, \strong{108}, 71-82. \doi{10.1093/biomet/asaa052}.

Kosmidis I, Kenne Pagui E C, Sartori N (2020). Mean and median bias
reduction in generalized linear models. \emph{Statistics and Computing},
\strong{30}, 43-59. \doi{10.1007/s11222-019-09860-6}.

Cordeiro G M, McCullagh P (1991). Bias correction in generalized
linear models. \emph{Journal of the Royal Statistical Society. Series B
(Methodological)}, \strong{53}, 629-643. \doi{10.1111/j.2517-6161.1991.tb01852.x}.

Firth D (1993). Bias reduction of maximum likelihood estimates.
\emph{Biometrika}. \strong{80}, 27-38. \doi{10.2307/2336755}.

Kenne Pagui E C, Salvan A, Sartori N (2017). Median bias
reduction of maximum likelihood estimates. \emph{Biometrika}, \strong{104},
923–938. \doi{10.1093/biomet/asx046}.

Kosmidis I, Firth D (2009). Bias reduction in exponential family
nonlinear models. \emph{Biometrika}, \strong{96}, 793-804. \doi{10.1093/biomet/asp055}.

Kosmidis I, Firth D (2010). A generic algorithm for reducing
bias in parametric estimation. \emph{Electronic Journal of Statistics},
\strong{4}, 1097-1112. \doi{10.1214/10-EJS579}.

Kosmidis I (2014). Bias in parametric estimation: reduction and
useful side-effects. \emph{WIRE Computational Statistics}, \strong{6},
185-196. \doi{10.1002/wics.1296}.
}
\seealso{
\code{\link[=brglmControl]{brglmControl()}}, \code{\link[=glm.fit]{glm.fit()}}, \code{\link[=glm]{glm()}}
}
\author{
Ioannis Kosmidis \verb{[aut, cre]} \email{ioannis.kosmidis@warwick.ac.uk}, Euloge Clovis Kenne Pagui \verb{[ctb]} \email{kenne@stat.unipd.it}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/brglmFit.R
\name{summary.brglmFit}
\alias{summary.brglmFit}
\alias{print.summary.brglmFit}
\title{\code{\link[=summary]{summary()}} method for \code{\link[=brglmFit]{"brglmFit"}} objects}
\usage{
\method{summary}{brglmFit}(
  object,
  dispersion = NULL,
  correlation = FALSE,
  symbolic.cor = FALSE,
  ...
)

\method{print}{summary.brglmFit}(
  x,
  digits = max(3L, getOption("digits") - 3L),
  symbolic.cor = x$symbolic.cor,
  signif.stars = getOption("show.signif.stars"),
  ...
)
}
\arguments{
\item{object}{an object of class \code{"glm"}, usually, a result of a
    call to \code{\link[stats]{glm}}.}

\item{dispersion}{the dispersion parameter for the family used.
    Either a single numerical value or \code{NULL} (the default), when
    it is inferred from \code{object} (see \sQuote{Details}).}

\item{correlation}{logical; if \code{TRUE}, the correlation matrix of
    the estimated parameters is returned and printed.}

\item{symbolic.cor}{logical. If \code{TRUE}, print the correlations in
    a symbolic form (see \code{\link[stats]{symnum}}) rather than as numbers.}

\item{...}{further arguments passed to or from other methods.}

\item{x}{an object of class \code{"summary.glm"}, usually, a result of a
    call to \code{summary.glm}.}

\item{digits}{the number of significant digits to use when printing.}

\item{signif.stars}{logical. If \code{TRUE}, \sQuote{significance stars}
    are printed for each coefficient.}
}
\description{
\code{\link[=summary]{summary()}} method for \code{\link[=brglmFit]{"brglmFit"}} objects
}
\details{
The interface of the summary method for \code{\link[=brglmFit]{"brglmFit"}}
objects is identical to that of \code{\link[=glm]{"glm"}} objects. The summary
method for \code{\link[=brglmFit]{"brglmFit"}} objects computes the p-values of the
individual Wald statistics based on the standard normal
distribution, unless the family is Gaussian, in which case a t
distribution with appropriate degrees of freedom is used.
}
\examples{
## For examples see `examples(brglmFit)`

}
\seealso{
\code{\link[=summary.glm]{summary.glm()}} and \code{\link[=glm]{glm()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{alligators}
\alias{alligators}
\title{Alligator food choice data}
\format{
A data frame with 80 rows and 5 variables:
\itemize{
\item \code{foodchoice}: primary food type, in volume, found in an alligator’s stomach, with levels \code{fish}, \code{invertebrate},\code{reptile}, \code{bird}, \code{other}
\item \code{lake}: lake of capture with levels \code{Hancock}, \code{Oklawaha}, \code{Trafford}, \code{George}.
\item \code{gender}: gender of the alligator with levels \code{Male} and \code{Female}
\item \code{size}: size of the alligator with levels \verb{<=2.3} meters long and \verb{>2.3} meters long
\item \code{freq}: number of alligators for each foodchoice, lake, gender and size combination
}
}
\source{
The alligators data set is analyzed in Agresti (2002, Subsection 7.1.2).
}
\usage{
alligators
}
\description{
Alligator food choice data
}
\references{
Agresti A (2002). \emph{Categorical Data Analysis}.  Wiley Series in
Probability and Statistics. Wiley.
}
\seealso{
\code{\link[=brmultinom]{brmultinom()}}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/brnb.R
\name{summary.brnb}
\alias{summary.brnb}
\alias{print.summary.brnb}
\title{\code{\link[=summary]{summary()}} method for \code{\link[=brnb]{"brnb"}} objects}
\usage{
\method{summary}{brnb}(object, ...)

\method{print}{summary.brnb}(x, digits = max(3, getOption("digits") - 3), ...)
}
\arguments{
\item{object}{an object of class \code{\link[=brnb]{"brnb"}}, typically, a
result of a call to \code{\link[=brnb]{brnb()}}.}

\item{...}{further arguments passed to or from other methods.}

\item{x}{an object of class \code{\link[=summary.brnb]{"summary.brnb"}},
usually, a result of a call to \link{summary.brnb}.}

\item{digits}{the number of significant digits to use when printing.}
}
\description{
\code{\link[=summary]{summary()}} method for \code{\link[=brnb]{"brnb"}} objects
}
\details{
The interface of the summary method for \code{\link[=brnb]{"brnb"}}
objects is similar to that of \code{\link[=brglmFit]{"brglmFit"}} objects
with additional information.

\if{html}{\out{<div class="sourceCode">}}\preformatted{p-values of the individual Wald statistics are based on the
standard normal distribution.
}\if{html}{\out{</div>}}
}
\examples{
# For examples see examples(brnb)

}
\seealso{
\code{\link[=summary.brglmFit]{summary.brglmFit()}} and \code{\link[=glm]{glm()}}
}
